/*

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                Arlips compiler v. 3.1           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

File tds.c, table of symbols and stuff
April 2006

Copyright (C) 2006 Emilio Vivancos Rubio

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                              Contact                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Author:              Emilio Vivancos Rubio  vivancos@dsic.upv.es
Developer Arlips v2: Mario Gonzlez Gimnez mgonzale@dsic.upv.es
Developer Arlips v3: Carlos Garca Montoro  cargarcia@dsic.upv.es

Departamento de Sistemas Informticos y Computacin
Universidad Politcnica de Valencia
Camino de Vera s/n
46022 Valencia
Spain

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

*/


#include "tds.h"
#include "alex.h"

/******  tablas de simbolos  ******/

EHTIPO   * tipos[HASH_SIZE];
EHTDS    * tds[HASH_SIZE];
EHREGLA  * reglas[HASH_SIZE];
EHFUNC   * funciones[HASH_SIZE];

REGTDS   * elemtds;    /* elemento de la tabla tds al que accedo */
REGTIPO  * elemtip;    /* elemento de la tabla tipo al que accedo */

/* estructura de clasificacion de las condiciones */
/* hay una entrada por cada operador relacional, mas una por las exp bool */
LISTA_LISTA_CONDIC * clas_cond[7];


int ihash;      /*indice hash del identificador */

int is_kdm_level; //Indicar si una instancia es del kdm
/****************************************************************************/
/*                      GESTION DE LAS TABLAS HASH                          */
/****************************************************************************/

int indice_hash(ident)
  char * ident;
{
  /* esta funcion recibe el nombre de un identificador       */
  /* y devuelve la posicion donde debe estar almacenada      */
  /* dentro del array hash que le corresponda (tds,tipos...).*/

  /*  Robert Sedgewick */
  /* 'Algorithms in C' */
  /* Addison-Wesley Publishing Company, 1990 */

  int h=0;
  while(*ident!='\0')
  h=(64+h+*ident++)%HASH_SIZE;
  return(h);
}


/****************************************************************************/
/***************        GESTION  DE  LAS  TABLAS        *********************/
/****************************************************************************/

char busca_tipo_en_tds(ident)
     char * ident;
{
  /* esta funcion busca por todas las tablas hash para determinar si existe */
  /* ese identificador y si es asi, de que tipo es */
  if((elemtds=busca_en_tds(ident))!=NULL) {
    /* encontrado en tds */
    /* sera var, pattern_var,inputvar o constante */
    return (elemtds->clvar);
  }
  else
    if((elemtip=busca_en_tipos(ident))!=NULL) {
      return (clase);
    }
  /*     else
          if((elemregla=busca_en_reglas(ident))!=NULL)
    	  {
		 return(regla);
  	  } */    /* lo comento, por que las reglas pueden llamarse */
                  /* igual que los identificadores del tds, ya que estos */
                  /* nombres no pueden utilizarse en sitio alguno (mas que */
                  /* a efectos de resultado) */
    else return(FALSE);
}

REGTDS * busca_en_tds(ident)
  char * ident;
{
  char encontrado;
  EHTDS * aux;
  LISTA_PARAM *lp;

  encontrado=FALSE;
  if (declarando_funcion) {  /* primero se busca en los parametros */
     lp = busca_param(entfuncion->lp,ident);
     if (lp!=NULL) return (lp->ptro);
  }
  ihash=indice_hash(ident); /* doy el indice hash del identificador */
  aux=tds[ihash];
  while(aux!=NULL && !encontrado){
    if(strcmp(aux->etds.nombre,ident)!=0) aux=aux->sig;
    else encontrado=TRUE;
  }
  if(encontrado)
     /* lo ha encontrado */
     return(&(aux->etds));
  else return(NULL);
}

REGTIPO * busca_en_tipos(ident)
  char * ident;
{
  char encontrado;
  EHTIPO * aux;

  encontrado=FALSE;
  ihash=indice_hash(ident);  /* doy el indice hash del identificador */
  aux=tipos[ihash];
  while(aux!=NULL && !encontrado){
    if(strcmp(aux->etipo.nombre,ident)!=0) aux=aux->sig;
    else encontrado=TRUE;
  }
  if(encontrado)
     /* lo ha encontrado */
     return(&(aux->etipo));
  else return(NULL);
}

REGREGLA * busca_en_reglas(ident)
  char * ident;
{
  char encontrado;
  EHREGLA * aux;

  encontrado=FALSE;
  ihash=indice_hash(ident);  /* doy el indice hash del identificador */
  aux=reglas[ihash];
  while(aux!=NULL && !encontrado){
    if(strcmp(aux->eregla.nombre,ident)!=0) aux=aux->sig;
    else encontrado=TRUE;
  }
  if(encontrado)
     /* lo ha encontrado */
     return(&(aux->eregla));
  else return(NULL);
}

REGFUNC * busca_en_funciones(ident)
  char * ident;
{
  char encontrado;
  EHFUNC * aux;

  encontrado=FALSE;
  ihash=indice_hash(ident);  /* doy el indice hash del identificador */
  aux=funciones[ihash];
  while(aux!=NULL && !encontrado){
    if(strcmp(aux->efunc.nombre,ident)!=0) aux=aux->sig;
    else encontrado=TRUE;
  }
  if(encontrado)
     /* lo ha encontrado */
     return(&(aux->efunc));
  else return(NULL);
}

REGREGLA * busca_regla(numregla)
  int numregla;
{
  char encontrado;
  EHREGLA * aux;
  int i;

  for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
           if (aux->eregla.num_regla==numregla) return (&(aux->eregla));
                                           else aux=aux->sig;
         }
     }
  return(NULL);
}



/****************************************************************************/

/* GESTION DE LA LISTA DE IDENTIFICADORES,LA LISTA_DE_CAMPOS DE REGISTROS   */
/*                           Y LA LISTA DE NUMEROS                          */

/****************************************************************************/

LISTA * cr_lista_id(list,id)
 LISTA * list;
 char * id;
{
 /* funcion que crea el primer elemento de la lista de identificadores */

 list=(LISTA *)malloc(sizeof(LISTA));
 strcpy(list->nombre,id);
 list->sig=NULL;
 return (list);
}

LISTA * insertar(list,id)
 LISTA * list;
 char * id;
{
/*funcion que inserta un nuevo identificador en la lista de identificadores*/

 LISTA * aux;

 aux=list;
 while(aux->sig!=NULL) aux=aux->sig;
 aux->sig=(LISTA *)malloc(sizeof(LISTA));
 aux=aux->sig;
 strcpy(aux->nombre,id);
 aux->sig=NULL;
 return(list);
}


LISTA * insertar_sindup(list,id)
 LISTA * list;
 char * id;
{
/* funcion que inserta un nuevo identificador en la lista de identificadores */
/* siempre que este identificador no este previamente en la lista         */
/* devuelve NULL si no se inserta (por estar ya en la lista) o el puntero */
/* a la lista si se inserta */

 LISTA * aux;
 int encontrado;

 encontrado = FALSE;
 aux = list;
 while (aux!=NULL && !encontrado)
   if (strcmp(aux->nombre,id)!=0) aux=aux->sig;
                        else encontrado = TRUE;
 if (!encontrado) {
        if (list==NULL) list=cr_lista_id(list,id);
                  else  list=insertar(list,id);
        return(list);
 }
 return (NULL);
}

LISTA * borrar(list)
LISTA * list;
{
 /* funcion que borra una lista de identificadores */

 LISTA * aux;

 while(list!=NULL)
 {
  aux=list;
  list=list->sig;
  free(aux);
 }
 return(list);
}

int busca_lista(list,id)
 LISTA * list;
 char * id;
{
/* funcion que busca un identificador en la lista de identificadores */
/* devuelve FALSE si no lo encuentra o TRUE si lo encuentra          */

 LISTA * aux;
 int encontrado;

 encontrado = FALSE;
 aux = list;
 while (aux!=NULL && !encontrado)
   if (strcmp(aux->nombre,id)!=0) aux=aux->sig;
                        else encontrado = TRUE;
 return (encontrado);
}


/* EXPRESIONES */

EXPRESION * crea_lista_exp(le,esoper,escamp,tipop,tipoper,clasoper,ident)
   EXPRESION *le;
   int esoper;         /* booleano que indica si es un operando */
   int escamp;         /* booleano que indica si es un campo */
   char tipop;         /* indica el tipo de operador */
   char tipoper;       /* indica el tipo de operando */
   char clasoper;      /* indica la clase del operando */
   char *ident;        /* identificador del operando */
{
 le=(EXPRESION *)malloc(sizeof(EXPRESION));
 le->esoperando=esoper;
 le->escampo=escamp;
 le->tipooperador=tipop;
 le->tipooperando=tipoper;
 le->claseoperando=clasoper;
 strcpy(le->ident,ident);
 le->sig=NULL;
 return(le);
}

EXPRESION * inserta_lista_exp(EXPRESION *le,
			      int esoper,         /* booleano que indica si es un operando */
			      int escamp,         /* booleano que indica si es un campo */
			      char tipop,         /* indica el tipo de operador */
			      char tipoper,       /* indica el tipo de operando */
			      char clasoper,      /* indica la clase del operando */
			      char *ident,        /* identificador del operando */
			      ENTREAL val) {      /* valor si es una constante */
  
  EXPRESION *aux;
  
  aux=le;
  if (aux!=NULL) {
    while (aux->sig!=NULL)
      aux=aux->sig;
    aux->sig=(EXPRESION *)malloc(sizeof(EXPRESION));
    aux=aux->sig;
  } 
  else {
    aux=(EXPRESION *)malloc(sizeof(EXPRESION));
    le=aux;
  }
  aux->indice_array = NULL;
  aux->esoperando=esoper;
  aux->escampo=escamp;
  aux->tipooperador=tipop;
  aux->tipooperando=tipoper;
  aux->claseoperando=clasoper;
  if (aux->tipooperando==cadena) {
    aux->cad = (char *) malloc(255);
    strcpy(aux->cad,ident);
  } 
  else 
    if (ident!=NULL)
      strcpy(aux->ident,ident);
  aux->valor=val; 
  aux->sig=NULL;
  return(le);
}

EXPRESION *busca_expresion(expre, numexp)
     EXPRESION *expre;
     int numexp;
/* esta funcion devuelve de una expresion multiple (separada por comas) */
/* la expresion determinada por el numero pasado como segundo parametro */
{
  EXPRESION *expi;
  int contexp;

  expi = expre;
  contexp = numexp;
  while (expi!=NULL && contexp>0) {
     if (!expi->esoperando && expi->tipooperador==coma) contexp--;
     expi=expi->sig;
  }
  return (expi);
}


int es_tipo_simple(char simb) {
  switch (simb) {
    
  case arraysim:
    return(array);

  case realsim :
    return(real);
  
  case intesim : 
    return (entero);
  
 case booleansim: 
   return(boolean);
  
  case intlistsim:
    return (lista_enteros);
  
  case realistsim:
    return (lista_reales);
  
  default: 
    return (FALSE);
  }
}


int es_obj_simple(exp)
  EXPRESION *exp;
/* esta funcion se encarga de comprobar si la expresion pasada como parametro */
/* es una expresion con constantes y operadores, o se trata de un objeto      */
/* (instancia, variable) simple */
{
  EXPRESION *expaux;

  expaux = exp;
  while (expaux!=NULL) {
    /* si se encuentra algun operador distinto del punto */
    if (!expaux->esoperando && expaux->tipooperador!=punto
        && expaux->tipooperador!=existe) return (FALSE);
    if (expaux->claseoperando==constante) return (FALSE);
    expaux=expaux->sig;
  }
  return(TRUE);
}

int es_constante(exp)
  EXPRESION *exp;
/* esta funcion se encarga de comprobar si la expresion pasada como parametro */
/* es una expresion con una sola constante                                   */
{

    if (exp->esoperando && exp->claseoperando==constante
        && exp->tipooperando==entero && exp->sig==NULL)
        return(TRUE);
    else return(FALSE);
}


/* CONDICIONES */

LISTA_CONDIC * crea_lista_condic(lc,oprel,li,ld,neg,lobj_cond,lparam,numvpe,numvpu,nc)
   LISTA_CONDIC *lc;
   char oprel;        /* operador relacional */
   struct EXPRESION * li;
   struct EXPRESION * ld;
   int neg;
   LISTA_OBJ *lobj_cond;
   LISTA *lparam;
   int numvpe;          /* numero de variables patron existencialmente cuantificadas */
   int numvpu;          /* numero de variables patron universalmente cuantificadas */
   int nc;              /* numero de condicion, dentro la regla */
{
 /* funcion que crea el primer elemento de la lista de condiciones */

 lc=(LISTA_CONDIC *)malloc(sizeof(LISTA_CONDIC));
 lc->oprel=oprel;
 lc->lado_izq=li;
 lc->lado_der=ld;
 lc->esta_negada =neg;
 lc->lobj_cond=lobj_cond;
 lc->param=lparam;
 lc->num_vpe=numvpe;
 lc->num_vpu=numvpu;
 lc->num_condic=nc;
 lc->es_orig = TRUE;
 lc->regla_orig[0]='\x0';
 lc->cond_orig=0;

 lc->sig=NULL;
 return(lc);
}

LISTA_CONDIC * inserta_lista_condic(lc,oprel,li,ld,neg,lobj_cond,lparam,numvpe,numvpu,nc)
   LISTA_CONDIC *lc;
   char oprel;        /* operador relacional */
   struct EXPRESION * li;
   struct EXPRESION * ld;
   int neg;
   LISTA_OBJ *lobj_cond;
   LISTA *lparam;
   int numvpe;          /* numero de variables patron existencialmente cuantificadas */
   int numvpu;          /* numero de variables patron universalmente cuantificadas */
   int nc;              /* numero de condicion, dentro la regla */
{
 /* funcion que inserta una nueva condicion en la lista de condiciones */

 LISTA_CONDIC *aux;

 aux=lc;
 if (aux!=NULL) {
    while(aux->sig!=NULL) aux=aux->sig;
    aux->sig=(LISTA_CONDIC *)malloc(sizeof(LISTA_CONDIC));
    aux=aux->sig;
 } else {
    aux=(LISTA_CONDIC *)malloc(sizeof(LISTA_CONDIC));
    lc=aux;
 }
 aux->oprel=oprel;
 aux->esta_negada= neg;
 aux->lado_izq=li;
 aux->lado_der=ld;
 aux->lobj_cond=lobj_cond;
 aux->param=lparam;
 aux->num_vpe=numvpe;
 aux->num_vpu=numvpu;
 aux->num_condic=nc;
 aux->es_orig = TRUE;
 aux->regla_orig[0]='\x0';
 aux->cond_orig=0;
 aux->sig=NULL;
 return(lc);
}

LISTA_CONDIC * busca_en_cond(lc,numc)
  LISTA_CONDIC *lc;
  int numc;
{
  while (numc>0 && lc!=NULL) {
     lc = lc->sig;
     numc--;
  }
  return(lc);
}

ARB_COND * crea_nodo_condic(oprel,li,ld,neg)
   char oprel;        /* operador relacional */
   struct EXPRESION * li;
   struct EXPRESION * ld;
   int neg;
{
 /* funcion que crea un hoja condicion del arbol de condiciones */

 ARB_COND *aux;

 aux=(ARB_COND *)malloc(sizeof(ARB_COND));
 aux->oprel=oprel;
 aux->esta_negada= neg;
 aux->tipo_nodo = nodo_condicion;
 aux->lado_izq=li;
 aux->lado_der=ld;
 aux->subarb_and =NULL;
 aux->subarb_or =NULL;
 aux->sig=NULL;
 return(aux);
}

ARB_COND * inserta_arb_condic_and(arbc,subarb)
   ARB_COND *arbc;
   ARB_COND *subarb;
{
 /* funcion que inserta una nueva condicion en el arbol de condiciones */

 ARB_COND *aux;

 aux=arbc;
 if (aux!=NULL) {
    if (subarb->tipo_nodo!=nodo_or) {
       aux = aux->subarb_and;
       if (aux!=NULL) {
          while(aux->sig!=NULL) aux=aux->sig;
          aux->sig=subarb;
       } else arbc->subarb_and=subarb;
    } else {
      aux = aux->subarb_or;
      if (aux!=NULL) {
          while(aux->sig!=NULL) aux=aux->sig;
          aux->sig=subarb;
       } else arbc->subarb_or=subarb;
    }
 } else { /* si es el primero que se aade se crea el nodo and padre */
    aux = (ARB_COND *)malloc(sizeof(ARB_COND));
    if (subarb->tipo_nodo!=nodo_or) {
          aux->subarb_and = subarb;
          aux->subarb_or=NULL;
    } else {
          aux->subarb_or = subarb;
          aux->subarb_and= NULL;
    }
    aux->esta_negada= FALSE;
    aux->tipo_nodo = nodo_and;
    aux->oprel = nulo;
    aux->lado_izq=NULL;
    aux->lado_der=NULL;

    aux->sig=NULL;
    arbc=aux;
 }
 return(arbc);
}

ARB_COND * inserta_arb_condic_or(arbc,subarb1,subarb2)
   ARB_COND *arbc;
   ARB_COND *subarb1;
   ARB_COND *subarb2;
{
 /* funcion que inserta una nueva condicion OR en el arbol de condiciones */

 ARB_COND *aux;
 ARB_COND *nod_or;

 aux=arbc;
 /* se crea el nodo OR padre */
 nod_or = (ARB_COND *)malloc(sizeof(ARB_COND));
 nod_or->subarb_and = NULL;
 nod_or->esta_negada= FALSE;
 nod_or->tipo_nodo = nodo_or;
 nod_or->oprel = nulo;
 nod_or->lado_izq=NULL;
 nod_or->lado_der=NULL;
 nod_or->subarb_or = subarb1;
 subarb1->sig= subarb2;
 nod_or->sig=NULL;

 if (aux!=NULL) {
    aux = aux->subarb_or;
    if (aux!=NULL) {
        while(aux->sig!=NULL) aux=aux->sig;
        aux->sig=nod_or;
    } else arbc->subarb_or=nod_or;
 } else
   arbc=nod_or;
 return (arbc);
}

/* CONDICIONES ACUMULADAS */

void inserta_condac(cond_ac,num_cond,num_regla,param,num_param,num_asig)
  CONDACUM **cond_ac;
  int num_cond;
  int num_regla;
  int *param;
  int num_param;
  int num_asig;
{
  int i;
  int p;
  int esta_incluido;
  CONDACUM *ptro, *ptro_aux;
  ENTREAL temp;

  /* se comprueba si esta condicion ya esta insertada o esta incluida en otra */
  esta_incluido = FALSE;
  ptro = *cond_ac;
  while (ptro!=NULL && !esta_incluido)
  {
    if (ptro->num_regla==num_regla) {
       if (ptro->num_cond==num_cond) {
           /* el numero de parametros debe ser el mismo */
           p=0;
           while (p<num_param) {
              if ((ptro->param[p]==param[p]) ||
                  (ptro->param[p]==-1)) p++;
                  else break;
           }
           if (p==num_param) {
              esta_incluido=TRUE;
              /* ahora aadimos el numero de la asignacion a la lista */
              /* de asignaciones que producen la llamada a esta condicion */
              temp.entero = num_asig;
              ptro->asig = inserta_ln(ptro->asig,temp);
           }
       }
    }
    ptro = ptro->sig;
  }
  if (!esta_incluido) {
     /* tambien se comprueba si esta condicion incluye a otra que ya esta previamente */
     /* insertada */
    esta_incluido = FALSE;
    ptro = *cond_ac;
    ptro_aux = NULL;
    while (ptro!=NULL)
    {
      if (ptro->num_regla==num_regla) {
         if (ptro->num_cond==num_cond) {
             /* el numero de parametros debe ser el mismo */
             p=0;
             while (p<num_param) {
                if ((ptro->param[p]==param[p]) ||
                    (param[p]==-1)) p++;
                    else break;
             }
             if (p==num_param) { /* se borra el elemento i */
                free(ptro->param);
                if (ptro_aux == NULL) {
                   *cond_ac = ptro->sig;
                } else {
                   ptro_aux->sig = ptro->sig;
                }
                free(ptro);
             }
         }
      }
      ptro_aux = ptro;
      ptro = ptro->sig;
    }
    /* se aade */
    ptro_aux = (CONDACUM *)malloc(sizeof(CONDACUM));
    ptro_aux->num_regla = num_regla;
    ptro_aux->num_cond = num_cond;
    ptro_aux->num_param = num_param;
    ptro_aux->param = (int *) malloc(sizeof(int)*num_param);
    ptro_aux->asig = (LIS_NUM *)malloc(sizeof(LIS_NUM));
    ptro_aux->asig->num.entero = num_asig;
    ptro_aux->asig->sig = NULL;
    ptro_aux->sig =NULL;
    i = 0;
    while (i<num_param) {
        ptro_aux->param[i]=param[i];
        i++;
    }
    if (*cond_ac==NULL) {
        *cond_ac = ptro_aux;
    } else {
        ptro = *cond_ac;
        while (ptro->sig!=NULL){
          ptro=ptro->sig;
        }
        ptro->sig = ptro_aux;
    }
 }
}
/* ACCIONES */

LISTA_ACCIONES * inserta_lista_acciones(lac,tipoac,ident,cam,cam_exp,exp,tip_exp,tip_lis_exp,tip_izq)
   LISTA_ACCIONES *lac;
   char tipoac;
   char *ident;
   char *cam;
   EXPRESION *cam_exp;
   EXPRESION *exp;
   char tip_exp;
   char *tip_lis_exp;
   char tip_izq;
{
 LISTA_ACCIONES *aux;
 int i;

 aux=lac;
 if (aux!=NULL) {
    while(aux->sig!=NULL) aux=aux->sig;
    aux->sig=(LISTA_ACCIONES *)malloc(sizeof(LISTA_ACCIONES));
    aux=aux->sig;
 } else {
    aux=(LISTA_ACCIONES *)malloc(sizeof(LISTA_ACCIONES));
    lac=aux;
 }
 aux->tipo_accion=tipoac;
 aux->expr=exp;
 aux->campos=cam_exp;
 aux->tip_expr = tip_exp;
 aux->tip_izq = tip_izq;
 if (ident!=NULL) strcpy(aux->ident,ident);
 if (cam!=NULL) strcpy(aux->campo,cam);
 for (i=0;i<30;i++) aux->tip_lista_expr[i]= tip_lis_exp[i];
 aux->sig=NULL;
 return(lac);
}

/* LISTA DE OBJETOS */

LISTA_OBJ * inserta_lista_obj(//lo,claseobj,tipoobj,cuanti,superclase,nom,numero_obj,campo_obj,lcampos)
			      LISTA_OBJ *lo,
			      char claseobj,                 /* instancia o patternvar */
			      char tipoobj,                 /* tipo del objeto: entero, real, definido */
			      char cuanti,                   /* indica como esta cuantificada */
			      char *superclase,
			      char *nom,
			      int numero_obj,
			      char *campo_obj,
			      LISTA *lcampos) {

  /* funcion que inserta un nuevo objeto en la lista de objetos */

  LISTA_OBJ *aux, *ant;
  char encontrado=FALSE;
  char error_;
  aux = lo;
  if (aux!=NULL) {
    /* en la lista de objetos se considera que esta repetido si ya existe un */
    /* objeto con un campo asi */
    while(aux != NULL) {
      if (strcmp(aux->nombre,nom) == 0 && strcmp(aux->campo,campo_obj) == 0) {
	if (claseobj == paternvar && aux->cuantificador != cuanti) {
	  menerror(128, &error_);
	}
	encontrado=TRUE;
	break;
      }
      ant = aux;
      aux = aux->sig;
    }
	
    if (!encontrado) {
      aux = ant;
      aux->sig=(LISTA_OBJ *)malloc(sizeof(LISTA_OBJ));
      aux=aux->sig;
    }
  } 
  else {
    aux=(LISTA_OBJ *)malloc(sizeof(LISTA_OBJ));
    lo=aux;
  }
  if (!encontrado) {
    aux->clvar=claseobj;
    aux->tipo = tipoobj;
    aux->cuantificador = cuanti;
    aux->campos=lcampos;
    strcpy(aux->instancia_de, superclase);
    strcpy(aux->nombre, nom);
    strcpy(aux->campo, campo_obj);
    switch (claseobj)  {
    case instancia :
      aux->cod.num_instancia=numero_obj;
      break;
    case paternvar :
      aux->cod.num_paternvar=numero_obj;
      break;
    case paramt :
      aux->por_ref=cuanti;
      /* FALTA: puesto que los parametros no iran nunca cuantificados, uso este */
      /* parametro para indicar en la lista de objetos si el parametro es por valor */
      /* o por referencia, pero esto lo deberia cambiar por un nuevo parametro */
      /* no lo he hecho por que habria que cambiar todas las llamadas */
    }
    aux->sig=NULL;
  }
  return(lo);
}

LISTA_OBJ * busca_obj(lo,identif)
    LISTA_OBJ *lo;
    char * identif;
{
  LISTA_OBJ *aux;
  char encontrado=FALSE;

 aux=lo;
 if (aux!=NULL) {
    while(aux!=NULL && !encontrado)
     if (strcmp(aux->nombre,identif)!=0) {
       aux=aux->sig;
     } else encontrado=TRUE;
 }
 return(aux);   /* si no lo encuentra devuelve NULL */
}

/* PARAMETROS */

LISTA_PARAM * cr_lista_param(list_p,id,entds,porref)
  LISTA_PARAM * list_p;
  char id[long_ident+1];
  REGTDS * entds;
  int porref;
{
  /* funcion que crea el primer elemento de la lista de parametros */

 list_p=(LISTA_PARAM *)malloc(sizeof(LISTA_PARAM));
 strcpy(list_p->nombre,id);
 list_p->ptro = entds;
 list_p->por_ref = porref;
 list_p->sig=NULL;
 return (list_p);
}

LISTA_PARAM * insertar_lista_param(list_p,id,entds,porref)
  LISTA_PARAM * list_p;
  char id[long_ident+1];
  REGTDS * entds;
  int porref;
{
/*funcion que inserta un nuevo parametro en la lista de parametros*/

 LISTA_PARAM * aux;

 aux=list_p;
 while(aux->sig!=NULL) aux=aux->sig;

 aux->sig=(LISTA_PARAM *)malloc(sizeof(LISTA_PARAM));
 aux=aux->sig;
 strcpy(aux->nombre,id);
 aux->ptro = entds;
 aux->por_ref = porref;
 aux->sig=NULL;
 return(list_p);
}

LISTA_PARAM * busca_param(lp,param) /* busca el lugar que ocupa el */
                                /* parametro param entre la lista */
                                /* de parametros de la funcion func */
 LISTA_PARAM *lp;
 char *param;
{
  LISTA_PARAM *aux1;
  char encontrado;

  aux1=lp;
  encontrado = FALSE;
  while (aux1!=NULL && !encontrado) {
     if (strcmp(aux1->nombre,param)!=0) aux1=aux1->sig;
             else encontrado=TRUE;

  }
  if(encontrado) return(aux1);
  else return(NULL);
}

/* LISTA DE NUMEROS */

LIS_NUM * crea_ln(ln,num)
   LIS_NUM *ln;
   ENTREAL num;
{
 /* funcion que crea el primer elemento de la lista de numeros */

 ln=(LIS_NUM *)malloc(sizeof(LIS_NUM));
 ln->num=num;
 ln->sig=NULL;
 return(ln);
}

LIS_NUM * inserta_ln(ln,num)
  LIS_NUM *ln;
  ENTREAL num;
{
 /* funcion que inserta un nuevo numero en la lista de numeros */

 LIS_NUM *aux;

 aux=ln;
 while(aux->sig!=NULL) aux=aux->sig;
 aux->sig=(LIS_NUM *)malloc(sizeof(LIS_NUM));
 aux=aux->sig;
 aux->num=num;
 aux->sig=NULL;
 return(ln);
}


LIS_NUM * borrar_ln(ln)
  LIS_NUM * ln;
{
 /* borra la lista de numeros */

 LIS_NUM * aux;

 while(ln!=NULL)
 {
  aux=ln;
  ln=ln->sig;
  free(aux);
 }
 return(ln);
}


LISTA_NUM * crea_lista_num(ln)
   LISTA_NUM *ln;
{
 /* funcion que crea el primer elemento del tipo de datos lista de numeros */

 ln=(LISTA_NUM *)malloc(sizeof(LISTA_NUM));
 ln->prim=NULL;
 ln->ult=NULL;
 return(ln);
}

LISTA_NUM * borrar_lista_num(l_num)
   LISTA_NUM *l_num;
{
 /* borrar la lista de numeros */

 LIS_NUM * ln;
 LIS_NUM * aux;

 ln = l_num->prim;
 while(ln!=NULL)
 {
  aux=ln;
  ln=ln->sig;
  free(aux);
 }
 return(NULL);
}

void ver_lista_campos(l)
 LISTA_CAMPOS * l;
{
 while(l!=NULL){
             printf("%s",l->nombre);
             switch (l->tipo) {
                          case entero  :  printf(", entero(%d)",l->valor.entero);
                                          break;
                          case subrango:  printf(", subrango[%d..%d](%d)",l->lim_inf,l->lim_sup,l->valor.entero);
                                          break;
                          case definido:  printf(", definido ");
                                          break;
                          case real    :  printf(", real(%f)",l->valor.real);
                                          break;
                          case caracter:  printf(", caracter ");
                                          break;
                          case boolean :  printf(", boolean  ");
                                          break;
                          case lista_enteros :  printf(", listaent[%d]",l->tam_lista);
                                          break;
                          case lista_reales  :  printf(", listareal[%d]",l->tam_lista);
                                                break;
                          case err           :  printf(", tipoerror");
                                                break;
                          default:              printf(", tipo no inicializado");
             }

/*           if(l->instancia_de[0])
                          printf(", %s",l->instancia_de); */
             printf("\n");
             l=l->siguiente;
 }
}

void ver_lista_exp(expre)
 EXPRESION * expre;
{
 EXPRESION * aux;

  aux=expre;
  while (aux!=NULL) {
        if (!aux->esoperando)
           switch (aux->tipooperador) {
                  case mas: printf("+");
                            break;
                  case menos: printf("-");
                            break;
                  case opmult: printf("*");
                            break;
                  case opdiv: printf("/");
                            break;
                  case existe: printf("?");
                               break;
                  case abrepar: printf("(");
                                break;
                  case cierrapar: printf(")");
                                  break;
                  case coma: printf(", ");
                                  break;
                  case punto: printf(".");
           }
        else {
           if (aux->claseoperando==constante) {
                  if (aux->tipooperando==real) printf("%f",aux->valor.real);
                    else if (aux->tipooperando==entero) printf("%d",aux->valor.entero);
                    else if (aux->tipooperando==cadena) printf("""%s""",aux->cad);
           } else
              printf("%s",aux->ident);
        }
        aux=aux->sig;
  }
}

void ver_condic(list_con)
   LISTA_CONDIC* list_con;
{
   LISTA_CONDIC* aux;

   aux=list_con;
   while (aux!=NULL) {
            ver_lista_exp(aux->lado_izq);
            switch (aux->oprel) {
                  case igual : printf("=");
                               break;
                  case distinto: printf("<>");
                                 break;
                  case menor: printf("<");
                              break;
                  case mayor: printf(">");
                              break;
                  case mayorig: printf(">=");
                              break;
                  case menorig: printf("<=");
                              break;
            }
            ver_lista_exp(aux->lado_der);
            aux=aux->sig;
            printf("\n");
   }
}

void ver_condiciones(cond)
   ARB_COND * cond;
{
   ARB_COND* aux;

   aux=cond;
   while (aux!=NULL) {
            if (aux->subarb_and!=NULL) ver_condiciones(aux->subarb_and);
            if (aux->tipo_nodo==nodo_condicion) {
               if (aux->esta_negada) printf("!");
               printf("(");
               ver_lista_exp(aux->lado_izq);
               switch (aux->oprel) {
                  case igual : printf("=");
                               break;
                  case distinto: printf("<>");
                                 break;
                  case menor: printf("<");
                              break;
                  case mayor: printf(">");
                              break;
                  case mayorig: printf(">=");
                              break;
                  case menorig: printf("<=");
                              break;
               }
               ver_lista_exp(aux->lado_der);
               printf(")");
            }
            aux=aux->sig;
            printf("\n");
   }
}

void ver_acciones(list_acc)
   LISTA_ACCIONES* list_acc;
{
   LISTA_ACCIONES* aux;

   aux=list_acc;
   while (aux!=NULL) {
            switch (aux->tipo_accion) {
                  case asig_patter : printf("?%s:=",aux->ident);
                                     ver_lista_exp(aux->expr);
                                     break;
                  case asig_instan: printf("!FALTA!");
                                    break;
                  case imprimir: printf("Print(");
                                 ver_lista_exp(aux->expr);
                                 printf(")");
                                 break;
                  case llamada_subprg: printf("!FALTA!");
                              break;
                  case oplista_addhead: printf("add_head(");
                                        ver_lista_exp(aux->expr);
                                        printf(" %s)", aux->ident);
                                        break;
                  case oplista_addtail: printf("add_tail(");
                                        ver_lista_exp(aux->expr);
                                        printf(" %s)", aux->ident);
                                        break;
                  case retorno_subprg: printf("return(");
                                        ver_lista_exp(aux->expr);
                                        printf(")");
                              break;
                  case parada: printf("!FALTA!");
                              break;
            }
            aux=aux->sig;
            printf("\n");
   }
}

void ver_lista(list)
LISTA * list;
{
 LISTA * aux;

aux=list;
printf("\n(");
while(aux!=NULL)
{
 printf("nombre :%s |",aux->nombre);
 aux=aux->sig;
}
printf(")\n");
}

void ver_lista_num(ln)
LIS_NUM * ln;
{
 LIS_NUM * aux;

aux=ln;
printf("\n(");
while(aux!=NULL)
{
 printf("%f,",aux->num);
 aux=aux->sig;
}
printf(")\n");
}

LISTA_CAMPOS * crea_lista_campos(//l,nom,tip,linf,lsup,def,lnum,tam_lista,ini,instancia_de)
				 LISTA_CAMPOS * l,
				 char * nom,
				 char tip,
				 int linf,
				 int lsup,
				 ENTREAL def,
				 LISTA_NUM *lnum,
				 int tam_lista,
				 char ini,     /* campo que indica si esta inicializado con un valor */
				 char * instancia_de,
				 int t_array) {

  REGTIPO * auxptro;
  LISTA_CAMPOS * auxlc_orig;
  LISTA * lis_id;
  
  if(tip!=err){
    /* si no ha ocurrido un error en el tipo del campo, lo creo */
    l=(LISTA_CAMPOS *)malloc(sizeof(LISTA_CAMPOS));
    strcpy(l->nombre,nom);
    strcpy(l->instancia_de, instancia_de);
    l->tipo=tip;
    l->tbase_array = t_array;
    l->valor=def;
    l->inicializado=ini;
    l->ld = NULL;
    l->lc = NULL;
    l->tam_lista=0;
    l->reglas=NULL;
    l->num_conds=NULL;
    
    switch(tip){
    case subrango:  
      l->lim_inf=linf;
      l->lim_sup=lsup;
      break;
      
    case definido:   /* ahora se copian los campos de la clase en */
      /* el campo lc del campo */
      auxptro = busca_en_tipos(instancia_de);
      auxlc_orig = auxptro->lc;
      while (auxlc_orig!=NULL) {
	if (l->lc==NULL)
	  l->lc = crea_lista_campos(l->lc,auxlc_orig->nombre,auxlc_orig->tipo,
				    auxlc_orig->lim_inf,auxlc_orig->lim_sup,auxlc_orig->valor,
				    auxlc_orig->ld,auxlc_orig->tam_lista,auxlc_orig->inicializado,
				    auxlc_orig->instancia_de, auxlc_orig->tbase_array);
	else {
	  lis_id=NULL;
	  lis_id=cr_lista_id(lis_id,auxlc_orig->nombre);
	  l->lc = inserta_lista_campos(l->lc,lis_id,auxlc_orig->tipo,
				       auxlc_orig->lim_inf,auxlc_orig->lim_sup,auxlc_orig->valor,
				       auxlc_orig->ld,auxlc_orig->tam_lista,auxlc_orig->inicializado,
				       auxlc_orig->instancia_de, auxlc_orig->tbase_array);
	  lis_id = borrar(lis_id);
	}
	auxlc_orig=auxlc_orig->siguiente;
      }
      break;
      
    case entero:    
      break;
  
    case real:
      break;
      
    case lista_enteros :
      l->ld = lnum;
      l->tam_lista=tam_lista;
      break;
    case lista_reales:
      l->ld = lnum;
      l->tam_lista=tam_lista;
      break;
    case array:
      l->ld = lnum;
      l->tam_lista = tam_lista;
    }
    l->campo_buscado = FALSE;
    l->siguiente=NULL;
  }
  return(l);
}

LISTA_CAMPOS * inserta_lista_campos(//l1,l2,tip,linf,lsup,def,lnum,tam_lista,ini,instancia_de)
				    LISTA_CAMPOS * l1,  /*lista de campos del registro (se va construyendo)*/
				    LISTA * l2,
				    char tip,
				    int linf,
				    int lsup,
				    ENTREAL def,
				    LISTA_NUM *lnum,
				    int tam_lista,
				    char ini,        /* campo que indica si esta inicializado con un valor */
				    char * instancia_de,
				    int t_array) {
  
  char repetido;
  int desp;   /* para calcular el desplazamiento */
  REGTIPO * auxptro;
  LISTA_CAMPOS * aux;
  LISTA_CAMPOS * rlc; /* rlc me sirve para recorrer la lista_campos y */
		      /* comprobar que no se repite el nombre del campo */
  LISTA_CAMPOS * auxlc_orig;
  LISTA *l;
  
  if(tip!=err){
    /* si no ha ocurrido un error en el tipo del campo, lo creo */
    aux=l1;
     while(aux->siguiente!=NULL) aux=aux->siguiente;
     while(l2!=NULL){
       repetido=FALSE;
       rlc=l1;
       while(rlc!=NULL && !repetido){
	 if(strcmp(rlc->nombre,l2->nombre)!=0) rlc=rlc->siguiente;
	 else repetido=TRUE;
       }
       if(!repetido){
	 aux->siguiente=(LISTA_CAMPOS *) malloc (sizeof(LISTA_CAMPOS));
	 aux=aux->siguiente;
	 strcpy(aux->nombre,l2->nombre);
         strcpy(aux->instancia_de, instancia_de);
	 aux->tipo=tip;
         aux->valor=def;
         aux->ld = NULL;
         aux->lc = NULL;
         aux->tam_lista=0;
	 aux->siguiente=NULL;
         aux->inicializado=ini;
         aux->reglas=NULL;
         aux->num_conds=NULL;
	 aux->campo_buscado = FALSE;
	 switch(tip){
	 case subrango:  aux->lim_inf=linf;
	   aux->lim_sup=lsup;
	   break;
	   
         case definido:   /* ahora se copian los campos de la clase en */
	   /* el campo lc del campo */
	   auxptro = busca_en_tipos(instancia_de);
	   auxlc_orig = auxptro->lc;
	   while (auxlc_orig!=NULL) {
	     if (aux->lc==NULL)
	       aux->lc = crea_lista_campos(aux->lc,auxlc_orig->nombre,auxlc_orig->tipo,
					   auxlc_orig->lim_inf,auxlc_orig->lim_sup,auxlc_orig->valor,
					   auxlc_orig->ld,auxlc_orig->tam_lista,auxlc_orig->inicializado,
					   auxlc_orig->instancia_de, auxlc_orig->tbase_array);
	     else {
                             l=NULL;
                             l=cr_lista_id(l,auxlc_orig->nombre);
                             aux->lc = inserta_lista_campos(aux->lc,l,auxlc_orig->tipo,
							    auxlc_orig->lim_inf,auxlc_orig->lim_sup,auxlc_orig->valor,
							    auxlc_orig->ld,auxlc_orig->tam_lista,auxlc_orig->inicializado,
							    auxlc_orig->instancia_de, auxlc_orig->tbase_array);
                             l = borrar(l);
	     }
	     auxlc_orig=auxlc_orig->siguiente;
	   }
	   break;
	   
	 case entero:    break;
	   
	 case real:      break;
	   
         case lista_enteros :    
	   aux->ld = lnum;
	   aux->tam_lista=tam_lista;
	   break;
         case lista_reales:  
	   aux->ld = lnum;
	   aux->tam_lista=tam_lista;
	   break;
	 case array:
	   aux->ld = lnum;
	   aux->tbase_array = t_array;
	   aux->tam_lista = tam_lista;
	   break;
	   
	 } /* fin del case */
       }/* fin del if(!repetido) */
       else{
	 /* el nombre del campo esta duplicado */
	 menerror_1(76);
       }
       l2=l2->sig;
     }/* fin del while */
     aux->siguiente=NULL;
  }
  return(l1);
}

LISTA_CAMPOS * busca_campo(ide, list)
     char * ide;
     LISTA_CAMPOS * list;
{
  while (list!=NULL) {
    if (!strcmp(list->nombre,ide)) {
      list->campo_buscado = TRUE;
      return (list);
    }
    else  list=list->siguiente;
  }
  return (NULL);
}


//Marca como buscado el campo correspondiente de todas las instancias de la misma clase.
//Resulta til al encontrar la utilizacin de campos por parte de variables patrn,
//pues estas afectarn a todas las instacias.
void marcar_campo_iclase_buscado(char *campo, char *clase) {
  int i;
  EHTDS *reg_tds;

  for (i=0; i<HASH_SIZE; i++) {
    if (tds[i] != NULL) {
      reg_tds = tds[i];
      while (reg_tds != NULL) {
	if ((reg_tds->etds.clvar == instancia) &&
	    (!strcmp(reg_tds->etds.instancia_de, clase))) {
	  busca_campo(campo, reg_tds->etds.lc);
	} //if es uan instancia de la clase a marcar
	reg_tds = reg_tds->sig;
      } //while haya registros en un elemento de la hash de la tds
    } //if es un registro/elemento de la tds
  } //for cada indice de la hash de la tds
} //marca_campo_iclases_buscado
  


/***********************************************************/
/*******  Insercion de la lista de identificadores  ********/
/******        en la lista correspondiente           *******/
/**********************************************************/

void insertar_en_const(ls,n,esreal)
  LISTA * ls;
  ENTREAL n;
  char esreal;
{
 EHTDS * aux1, * aux2;
 int i;

 /* recorro la lista de elementos e inserto la nueva constante al final */

 while(ls!=NULL){
    i=indice_hash(ls->nombre); /* obtengo el indice hash del identificador */
    if(tds[i]==NULL){
	tds[i]=(EHTDS *)malloc(sizeof(EHTDS));
	aux2=tds[i];           /* apunto aqui para luego introducir los datos */
    }
    else{
       aux1=tds[i];
       aux2=tds[i]->sig;
       while(aux2!=NULL){
	 aux1=aux1->sig;
	 aux2=aux2->sig;
       }
       aux2=(EHTDS *)malloc(sizeof(EHTDS));
       /*en aux2, tengo el lugar donde he de introducir los datos */
       aux1->sig=aux2; /*uno el nuevo elemento al final de la lista */
    }
    strcpy(aux2->etds.nombre,ls->nombre);
    aux2->etds.clvar=constante;
    aux2->etds.valor=n;
    if(esreal)
       aux2->etds.tipo=real; /* es una constante real */
    else
       aux2->etds.tipo=entero; /* es una constante entera */
    aux2->sig=NULL;
    aux2->etds.ld=NULL;
    aux2->etds.lc=NULL;
    aux2->etds.instancia_de[0]='\0';
    ls=ls->sig;
 }
}

REGTDS *crea_regtds_simple(clv,tip,nombre)
     char clv; /*dice de que tipo es la variable (var, inputvar, patternvar)*/
     char tip;
     char nombre[long_ident+1]; /* nombre de la variable de tipo simple */
{
  REGTDS * aux1;
  
  aux1=(REGTDS *)malloc(sizeof(REGTDS));
  
  strcpy(aux1->nombre,nombre);
  aux1->clvar = clv;
  aux1->tipo=tip;
  aux1->ld=NULL;
  aux1->tam_lista = 0;
  aux1->lc=NULL;
  aux1->instancia_de[0]='\0';
  aux1->reglas =NULL;
  aux1->inicializado = FALSE;
  aux1->num_conds=NULL;
  return(aux1);
}

REGTDS *crea_regtds_tipdef(clv,ident,nombre)
     char clv; /*dice de que tipo es la variable (var, inputvar, patternvar)*/
     char * ident;  /* nombre de la clase a la que pertenece */
     char nombre[long_ident+1]; /* nombre de la variable de tipo simple */
{
  REGTDS * aux1;
  REGTIPO * auxptro;
  LISTA_CAMPOS * lcaux2, * lcaux;
  
  aux1=(REGTDS *)malloc(sizeof(REGTDS));
  strcpy(aux1->nombre,nombre);
  strcpy(aux1->instancia_de,ident);
  auxptro = busca_en_tipos(ident);
  lcaux2 = NULL;
  if (auxptro!=NULL) {   /* se copian los campos de la clase padre */
    lcaux = auxptro->lc;
    while (lcaux!=NULL) {
      if (lcaux2==NULL) {
	aux1->lc=(LISTA_CAMPOS *) malloc (sizeof(LISTA_CAMPOS));
	lcaux2 = aux1->lc;
	} 
      else {
	lcaux2->siguiente=(LISTA_CAMPOS *) malloc (sizeof(LISTA_CAMPOS));
	lcaux2 = lcaux2->siguiente;
      }
      strcpy(lcaux2->nombre,lcaux->nombre);
      strcpy(lcaux2->instancia_de,lcaux->instancia_de);
      lcaux2->tipo=	lcaux->tipo;
      lcaux2->lim_inf = lcaux->lim_inf;
      lcaux2->lim_sup = lcaux->lim_sup;
      lcaux2->valor = lcaux->valor;
      lcaux2->tam_lista = lcaux->tam_lista;
      lcaux2->ld = lcaux->ld;   /* las listas son las mismas, ya que aunque utilicen */
                                /* la misma lista esto solo sirve a efectos de inicializacion */
      lcaux2->lc = lcaux->lc; /* no importa que apunten todos a la misma */
                              /* lista de campos, ya que de hecho, todas */
                              /* las instancias tendran los mismos valores */
                                /* iniciales (no hay forma de dar valores a */
                                /* campos de tipo definido) */
      lcaux2->inicializado = lcaux->inicializado;
      lcaux2->reglas = NULL;
      lcaux2->num_conds=NULL;
      lcaux = lcaux->siguiente;
    }
  }
  if (lcaux2==NULL) aux1->lc = NULL;
  else lcaux2->siguiente = NULL;
  aux1->tipo=definido;
  aux1->clvar=clv;         /*indico de que clase es la variable */
  aux1->reglas =NULL;
  aux1->num_conds =NULL;
  aux1->inicializado = FALSE;
  aux1->ld=NULL;
  
  return(aux1);
}

void instipsub_entds(clv,linf,lsup,ls)
 char clv; /*indico de que clase es la variable (var, inputvar, patternvar)*/
 int linf;
 int lsup;
 LISTA * ls;
{
  EHTDS * aux1, * aux2;
  int i;
  char error_=FALSE;
  char encontrado=FALSE;

  while(ls!=NULL)
  {
      i=indice_hash(ls->nombre);    /*obtengo el indice hash del identificador */
      if(tds[i]==NULL){
	tds[i]=(EHTDS *)malloc(sizeof(EHTDS));
	aux2=tds[i]; /*apunto aqui para luego introducir los datos */
      }
      else{
	aux1=tds[i];
	aux2=tds[i]->sig;
	if(strcmp(aux1->etds.nombre,ls->nombre)==0) encontrado=TRUE;
	while(aux2!=NULL && !encontrado){
	  if(strcmp(aux2->etds.nombre,ls->nombre)!=0){
	    aux1=aux1->sig;
	    aux2=aux2->sig;
	  }
	  else encontrado=TRUE;
	}
	if(encontrado){
	  menerror(70,&error_);
	}
	else{
	  aux2=(EHTDS *)malloc(sizeof(EHTDS));
	  /*en aux2, tengo el lugar donde he de introducir los datos */
	  aux1->sig=aux2; /*uno el nuevo elemento al final de la lista */
	}
      }
      if(!encontrado){
	  strcpy(aux2->etds.nombre,ls->nombre);
	  aux2->etds.tipo=subrango;
	  aux2->etds.lim_inf=linf;
	  aux2->etds.lim_sup=lsup;
	  aux2->etds.ld=NULL;
          aux2->etds.lc=NULL;
          aux2->etds.reglas = NULL;
          aux2->etds.num_conds=NULL;
          aux2->etds.inicializado = FALSE;
	  aux2->sig=NULL;
	  aux2->etds.clvar=clv; /* introduzco de que clase es la variable */
      }
      else{
	encontrado=FALSE;
      }
      ls=ls->sig;
  }
}

REGTDS * instipdef_entds(clv,ident,ls)
     char clv;      /*indica de que clase son las variables (var, inputvar, patternvar*/
     char * ident;  /* nombre de la clase a la que pertenece */
     LISTA * ls;    /* lista con los identificadores */
{
  /* inserta en la tabla 'tds' los identificadores de la lista pasada,   */
  /* el tipo que se introduce es el definido (tipo definido por el       */
  /* usuario) adems se introduce el nombre de la clase a la que pertenece */
  /* y que se encuentra en la tabla clases. */
  
  EHTDS * aux1, * aux2;
  REGTIPO * auxptro;
  LISTA_CAMPOS * lcaux, * lcaux2;
  int i,j;
  char tipo;
  char error_;
  char encontrado=FALSE;
  int kdm = FALSE;

  while(ls!=NULL) {
    i=indice_hash(ls->nombre);    /*obtengo el indice hash del identificador */
    if(tds[i]==NULL){
      tds[i]=(EHTDS *)malloc(sizeof(EHTDS));
      aux2=tds[i]; /*apunto aqui para luego introducir los datos */
    }
    else{
      aux1=tds[i];
      aux2=tds[i]->sig;
      if(strcmp(aux1->etds.nombre,ls->nombre)==0) encontrado=TRUE;
      while(aux2!=NULL && !encontrado){
	if(strcmp(aux2->etds.nombre,ls->nombre)!=0){
	  aux1=aux1->sig;
	  aux2=aux2->sig;
	}
	else encontrado=TRUE;
      }
      if(encontrado) {
	menerror(70,&error_);
      }
      else{
	aux2=(EHTDS *)malloc(sizeof(EHTDS));
	/*en aux2, tengo el lugar donde he de introducir los datos */
	aux1->sig=aux2; /*uno el nuevo elemento al final de la lista */
      }
    }
    if(!encontrado) {
      strcpy(aux2->etds.nombre,ls->nombre);
      strcpy(aux2->etds.instancia_de,ident);
      aux2->etds.from_kdm = is_kdm_level; //indica si es instancia importada de KDM
      /* aadimos en el campo lc de la TDS la lista de campos de la clase */
      /* a la que pertenece la var, instancia, o patternvar, y tambien todos */
      /* los campos de sus clases ascendentes */
      auxptro = busca_en_tipos(ident);
      /* se aumenta el numero de instancias o variables patron de la clase */
      switch (clv)  {
      case instancia : aux2->etds.cod.num_instancia=auxptro->num_instancias++;
	break;
      case paternvar : aux2->etds.cod.num_paternvar=auxptro->num_paternvars++;
	break;
      case var: aux2->etds.cod.num_instancia=auxptro->num_vars++;
      }
      lcaux2 = NULL;
      if (auxptro!=NULL) {   /* se copian los campos de la clase padre */
	lcaux = auxptro->lc;
	while (lcaux!=NULL) {
          if (lcaux2==NULL) {
	    aux2->etds.lc=(LISTA_CAMPOS *) malloc (sizeof(LISTA_CAMPOS));
	    lcaux2 = aux2->etds.lc;
          } else {
	    lcaux2->siguiente=(LISTA_CAMPOS *) malloc (sizeof(LISTA_CAMPOS));
	    lcaux2 = lcaux2->siguiente;
          }
          strcpy(lcaux2->nombre,lcaux->nombre);
          strcpy(lcaux2->instancia_de,lcaux->instancia_de);
          lcaux2->tipo=	lcaux->tipo;
	  lcaux2->tbase_array = lcaux->tbase_array;
          lcaux2->lim_inf = lcaux->lim_inf;
          lcaux2->lim_sup = lcaux->lim_sup;
          lcaux2->valor = lcaux->valor;
          lcaux2->tam_lista = lcaux->tam_lista;
          lcaux2->ld = lcaux->ld;   /* las listas son las mismas, ya que aunque utilicen */
                                    /* la misma lista esto solo sirve a efectos de inicializacion */
          lcaux2->lc = lcaux->lc; /* no importa que apunten todos a la misma */
                                  /* lista de campos, ya que de hecho, todas */
                                  /* las instancias tendran los mismos valores */
                                  /* iniciales (no hay forma de dar valores a */
                                  /* campos de tipo definido) */
          lcaux2->inicializado = lcaux->inicializado;
          lcaux2->reglas = NULL;
          lcaux2->num_conds=NULL;
          lcaux = lcaux->siguiente;
	}
      }
      if (lcaux2==NULL) aux2->etds.lc = NULL;
      else lcaux2->siguiente = NULL;
      aux2->etds.tipo=definido;
      aux2->etds.clvar=clv;      /*indico de que clase es la variable */
      aux2->etds.reglas =NULL;
      aux2->etds.num_conds =NULL;
      aux2->etds.inicializado = FALSE;
      aux2->etds.ld=NULL;
      aux2->sig=NULL;
    }
    else{
      encontrado=FALSE;
    }
    ls=ls->sig;
  }
  return(&aux2->etds);
}


void instipdef_entipos(ls,superclase,lc)
 LISTA * ls;
 REGTIPO * superclase;
 LISTA_CAMPOS * lc;
{
  /* inserta en la tabla de tipos, los identificadores de la lista */
  /* que se pasa, , adems se introduce un puntero */
  /* que apunta al tipo definido que estara en la tabla 'tipos' */

  EHTIPO * aux1, * aux2; /*para recorrer la lista de elementos hash de tipos*/
  REGTIPO * auxptro;
  LISTA_CAMPOS * lcaux, * lcaux2;
  int i;
  LISTA_CAMPOS * auxlc_orig;
  LISTA *l;

 while(ls!=NULL){
    i=indice_hash(ls->nombre); /*obtengo el indice hash del identificador */
    if(tipos[i]==NULL){
	tipos[i]=(EHTIPO *)malloc(sizeof(EHTIPO));
	aux2=tipos[i]; /*apunto aqui para luego introducir los datos */
    }
    else{
       aux1=tipos[i];
       aux2=tipos[i]->sig;
       while(aux2!=NULL){
	 aux1=aux1->sig;
	 aux2=aux2->sig;
       }
       aux2=(EHTIPO *)malloc(sizeof(EHTIPO));
       /*en aux2, tengo el lugar donde he de introducir los datos */
       aux1->sig=aux2; /*uno el nuevo elemento al final de la lista */
    }
   /* introducimos los datos en el lugar correspondiente */
   strcpy(aux2->etipo.nombre,ls->nombre);
   aux2->etipo.lc=lc;
   aux2->etipo.superclase=superclase;
   aux2->etipo.num_instancias = 0;
   aux2->etipo.num_paternvars = 0;
   aux2->etipo.num_vars = 0;
   aux2->etipo.ya_declarada = FALSE;
   aux2->etipo.declarando_actualmente = FALSE;
   aux2->sig=NULL;

   /* se aade a esta clase todos los campos heredados de sus superclases */

   auxptro=superclase;
   lcaux2=aux2->etipo.lc;
   if (lcaux2!=NULL) while (lcaux2->siguiente!=NULL) lcaux2=lcaux2->siguiente;

   /* no es necesario recorrer toda la jerarquia de clases, ya que en los tipos */
   /* ya hemos cogido los campos de todas sus clases padres */
   if (auxptro!=NULL) {
      lcaux = auxptro->lc;
      while (lcaux!=NULL) {
          if (lcaux2==NULL) {
             aux2->etipo.lc=(LISTA_CAMPOS *) malloc (sizeof(LISTA_CAMPOS));
             lcaux2 = aux2->etipo.lc;
          } else {
             lcaux2->siguiente=(LISTA_CAMPOS *) malloc (sizeof(LISTA_CAMPOS));
             lcaux2 = lcaux2->siguiente;
          }
          strcpy(lcaux2->nombre,lcaux->nombre);
          strcpy(lcaux2->instancia_de,lcaux->instancia_de);
          lcaux2->tipo=	lcaux->tipo;
	  lcaux2->tbase_array = lcaux->tbase_array;
          lcaux2->lim_inf = lcaux->lim_inf;
          lcaux2->lim_sup = lcaux->lim_sup;
          lcaux2->valor = lcaux->valor;
          lcaux2->tam_lista = lcaux->tam_lista;
          lcaux2->ld = lcaux->ld;
          lcaux2->lc = NULL;
          if (lcaux->tipo==definido) { /* tambien tenemos que copiar los campos  */
                                       /* de los tipos estructurados */
                       auxptro = busca_en_tipos(lcaux->instancia_de);
                       auxlc_orig = auxptro->lc;
                       while (auxlc_orig!=NULL) {
                          if (lcaux2->lc==NULL)
                             lcaux2->lc = crea_lista_campos(lcaux2->lc,auxlc_orig->nombre,auxlc_orig->tipo,
                                     auxlc_orig->lim_inf,auxlc_orig->lim_sup,auxlc_orig->valor,
                                     auxlc_orig->ld,auxlc_orig->tam_lista,auxlc_orig->inicializado,
                                     auxlc_orig->instancia_de, auxlc_orig->tbase_array);
                          else {
                             l=NULL;
                             l=cr_lista_id(l,auxlc_orig->nombre);
                             lcaux2->lc = inserta_lista_campos(lcaux2->lc,l,auxlc_orig->tipo,
                                     auxlc_orig->lim_inf,auxlc_orig->lim_sup,auxlc_orig->valor,
                                     auxlc_orig->ld,auxlc_orig->tam_lista,auxlc_orig->inicializado,
                                     auxlc_orig->instancia_de, auxlc_orig->tbase_array);
                             l = borrar(l);
                          }
                          auxlc_orig=auxlc_orig->siguiente;
                       }
          }
          lcaux2->inicializado = lcaux->inicializado;

	  //CGM: Creo que lo siguiente est mal, pues no se trata de la misma
	  //clase, sino de una nueva que se est declarando.
	  //Por tanto no ha podido haber regla alguna que use la clase.
	  //Creo que en general no pasa nada por el hecho de que 
	  //por lo general primero se hacen las declaraciones de clases 
	  //y objetos, y cuando estn todos, se hacen las de reglas.
	  //Con lo que las listas  aun no se han inicializado, y no hay problema.
	  //Pero creo que lo correcto sera inicializar ambas listas a NULL.

          lcaux2->reglas = lcaux->reglas;
          lcaux2->num_conds = lcaux->num_conds;
	  /////////////////////////////////////

	  lcaux2->campo_buscado = 0;

          lcaux = lcaux->siguiente;
      }
   }
   if (lcaux2==NULL) aux2->etipo.lc = NULL;
             else lcaux2->siguiente = NULL;
   ls=ls->sig;
 }
}

REGTDS * instipsimple_entds(clv,tip,nombre)
 char clv; /*dice de que tipo es la variable (var, inputvar, patternvar)*/
 char tip;
 char nombre[long_ident+1]; /* nombre de la variable de tipo simple */
{
  EHTDS * aux1, * aux2;
  int i;
  char error_=FALSE;
  char encontrado=FALSE;
  REGTDS *elemtdsdevuelto;

  i=indice_hash(nombre);    /*obtengo el indice hash del identificador */
  if(tds[i]==NULL){
     tds[i]=(EHTDS *)malloc(sizeof(EHTDS));
     aux2=tds[i]; /*apunto aqui para luego introducir los datos */
  }
  else{
     aux1=tds[i];
     aux2=tds[i]->sig;
     if(strcmp(aux1->etds.nombre,nombre)==0)  encontrado=TRUE;
     while(aux2!=NULL && !encontrado){
	 if(strcmp(aux2->etds.nombre,nombre)!=0){
	    aux1=aux1->sig;
	    aux2=aux2->sig;
	 }
	 else encontrado=TRUE;
     }
     if(encontrado){
	 menerror(70,&error_);
     }
     else{
	aux2=(EHTDS *)malloc(sizeof(EHTDS));
	/*en aux2, tengo el lugar donde he de introducir los datos */
	aux1->sig=aux2; /*uno el nuevo elemento al final de la lista */
     }
  }
  if(!encontrado){
     strcpy(aux2->etds.nombre,nombre);
     aux2->etds.clvar = clv;
     aux2->etds.tipo=tip;
     aux2->etds.ld=NULL;
     aux2->etds.tam_lista = 0;
     aux2->etds.lc=NULL;
     aux2->etds.instancia_de[0]='\0';
     aux2->etds.reglas =NULL;
     aux2->etds.inicializado = FALSE;
     aux2->etds.num_conds=NULL;
     aux2->sig=NULL;
     elemtdsdevuelto=&(aux2->etds);
  }
  else{
     encontrado=FALSE;
  }
  return(elemtdsdevuelto);
}

REGREGLA * insregla_enreglas(id,num)
  char * id;
  char num; /* numero de regla */
{
  EHREGLA * aux1, * aux2;
  REGREGLA * elemregladevuelto; 
  int i;
  char error_=FALSE;
  char encontrado=FALSE;

  i=indice_hash(id);    /*obtengo el indice hash del identificador */
  if(reglas[i]==NULL){
    reglas[i]=(EHREGLA *)malloc(sizeof(EHREGLA));
    aux2=reglas[i]; /*apunto aqui para luego introducir los datos */
  }
  else{
    aux1=reglas[i];
    aux2=reglas[i]->sig;
    if(strcmp(aux1->eregla.nombre,id)==0) encontrado=TRUE;
    while(aux2!=NULL && !encontrado){
      if(strcmp(aux2->eregla.nombre,id)!=0){
	aux1=aux1->sig;
	aux2=aux2->sig;
      }
      else encontrado=TRUE;
    }
    if(encontrado){
      menerror(69,&error_);
    }
    else{
      aux2=(EHREGLA *)malloc(sizeof(EHREGLA));
      /*en aux2, tengo el lugar donde he de introducir los datos */
      aux1->sig=aux2; /*uno el nuevo elemento al final de la lista */
    }
  }
  if(!encontrado){
      strcpy(aux2->eregla.nombre,id);
      aux2->eregla.expr_prio=NULL;
      aux2->eregla.num_regla=num;
      aux2->eregla.lcond=NULL;
      aux2->eregla.lobj_cond= NULL;
      aux2->eregla.lacc=NULL;
      aux2->eregla.lobj_acc = NULL;
      aux2->eregla.condiciones = NULL;
      aux2->eregla.tiene_copias = FALSE;
      aux2->eregla.num_copias = 0;
      aux2->eregla.copia_de = NULL;
      aux2->eregla.ya_generado = FALSE;
      aux2->sig=NULL;
      elemregladevuelto=&(aux2->eregla); /* anoto la direccion a devolver */
  }
  else{
    encontrado=FALSE;
  }
 return(elemregladevuelto); 
 /* devuelvo la direccin, donde se encuentra la informacion  */
 /* correspondiente a la regla */
}

REGFUNC * insertar_en_func(//nom,l_param,tipo,tipo_dev,cad_param)
			   char nom[long_ident+1],
			   LISTA_PARAM *l_param,
			   /* los siguientes parametros solo tendran valor en llamadas externas */
			   char tipo,
			   char tipo_dev,
			   char * cad_param)  /* si es una funcion externa */
{
  EHFUNC * aux1, * aux2;
  REGFUNC * devuelve;
  //char nom_temp[long_ident+1]; Ya no hace falta. Ahora Arlips es sensible a maysculas/minsculas
  //para todos los identificadores
  int i;

  /* recorro la lista de elementos e inserto la nueva funcion al final */

  /*Antes el nombre se intrtoduca en maysculas. Ya no...
  if (tipo==externa) { // el nombre de la funcion se introduce en mayusculas
    strcpy(nom_temp,nom);
    i=0;
    while (nom[i]!='\x0') {
      if((nom[i]>=97)&&(nom[i]<=122)) nom[i]=nom[i]-32; // paso de minusculas a mayusculas
      i++;
    }
  }
  */

  i=indice_hash(nom);    /*obtengo el indice hash del identificador */
  if(funciones[i]==NULL){
      funciones[i]=(EHFUNC *)malloc(sizeof(EHFUNC));
      aux2=funciones[i]; /*apunto aqui para luego introducir los datos */
  }
  else{
     aux1=funciones[i];
     aux2=funciones[i]->sig;
     while(aux2!=NULL){
       aux1=aux1->sig;
       aux2=aux2->sig;
     }
     aux2=(EHFUNC *)malloc(sizeof(EHFUNC));
     /*en aux2, tengo el lugar donde he de introducir los datos */
     aux1->sig=aux2; /*uno el nuevo elemento al final de la lista */
  }
  strcpy(aux2->efunc.nombre,nom); /* se introduce el nombre de la funcion */
  if (tipo==externa) strcpy(aux2->efunc.nombre_ext, nom);  //nom_temp);<-Antes, en lugar de nom
  aux2->efunc.lp=l_param; /* se introducen los parametros de la funcion */
  aux2->efunc.tipo_dev=tipo_dev; /* tipo devuelto por la funcion */
  aux2->efunc.tipo = tipo;       /* externa o interna */
  aux2->efunc.lacc=NULL;
  aux2->efunc.lobj_func = NULL;
  aux2->efunc.modif = NULL;
  /* se introducen los datos correspondientes a las llamadas externas */
  aux2->efunc.param = NULL;
  if (cad_param!=NULL) {
     if (cad_param[0]!='\0') {
        aux2->efunc.param = (char *) malloc(strlen(cad_param)+1);
        strcpy(aux2->efunc.param, cad_param);
     }
  }

  aux2->sig=NULL;
  devuelve = &(aux2->efunc);
  return(devuelve);
}


void ver_tabla_tds()
{
  EHTDS *aux;
  int i;
  
  for(i=0;i<HASH_SIZE;i++)
    if(tds[i]!=NULL) {
      aux=tds[i];
      while(aux!=NULL) {
	printf("%s",aux->etds.nombre);
	switch (aux->etds.tipo) {
	case entero  :  printf(", entero(%d)",aux->etds.valor.entero);
	  break;
	case subrango:  printf(", subrango[%d..%d](%d)",aux->etds.lim_inf,aux->etds.lim_sup,aux->etds.valor.entero);
	  break;
	case definido:  printf(", definido ");
	  break;
	case real    :  printf(", real(%f)",aux->etds.valor.real);
	  break;
	case caracter:  printf(", caracter ");
	  break;
	case boolean :  printf(", boolean  ");
	  break;
	case lista_enteros :  printf(", listaent ");
	  break;
	case lista_reales  :  printf(", listareal");
	  break;
	case err           :  printf(", tipoerror");
	  break;
	default:              printf(", tipo no inicializado");
	}
	
	switch (aux->etds.clvar) {
	case var  :  printf(", variable  ");
	  break;
	case paternvar:  printf(", paternvar (%d) ",aux->etds.cod.num_paternvar);
	  break;
	case inpvar:    printf(", inputvar  ");
	  break;
	case paramt    :  printf(", parametro ");
	  break;
	case constante:  printf(", constante ");
	  break;
	case instancia:  
	  printf(", la instancia (%d)",aux->etds.cod.num_instancia);
	  if (!aux->etds.from_kdm) printf(" no");
	  printf(" es del KDM");
	  break;
	default:  printf(", clase no inicializada");
	}
	if(aux->etds.instancia_de[0])
	  printf(", %s",aux->etds.instancia_de);
	if ((aux->etds.lc!=NULL)&&(aux->etds.clvar==instancia)) {
	  printf("\n----Campos:----\n");
	  ver_lista_campos(aux->etds.lc);
	  printf("---------------\n");
	} else printf("\n");
	aux=aux->sig;
      }
    }
}

void ver_tabla_tipos()
{
  EHTIPO *aux;
  int i;

  for(i=0;i<HASH_SIZE;i++)
     if(tipos[i]!=NULL) {
         aux=tipos[i];
         while(aux!=NULL) {
             printf("%s",aux->etipo.nombre);
             if (aux->etipo.superclase!=NULL)
                printf(", subclase de %s",aux->etipo.superclase->nombre);
             printf("\nNumero de instancias: %d\n",aux->etipo.num_instancias);
             printf("Numero de var. patron: %d",aux->etipo.num_paternvars);
             if (aux->etipo.lc!=NULL) {
                printf("\n----Campos:----\n");
                ver_lista_campos(aux->etipo.lc);
                printf("---------------\n");
             }
             printf("\n");
             aux=aux->sig;
         }
     }
}

void ver_tabla_reglas()
{
  EHREGLA *aux;
  int i;

  for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
             printf("REGLA: %d\n",aux->eregla.num_regla);
/*           printf("%s ,%d\n",aux->eregla.nombre,aux->eregla.prio); */
             printf("Condiciones: \n");
             ver_condic(aux->eregla.lcond);
             printf("Acciones: \n");
             ver_acciones(aux->eregla.lacc);
             aux=aux->sig;
         }
     }
}

/******************************************************************************/
/************************** PROCESAMIENTO DE REGLAS ***************************/
/******************************************************************************/

void inserta_reglacond(lr,lc,nomr,numc)
     LISTA  **lr;
     LIS_NUM **lc;
     char *nomr;
     int numc;
/* esta funcion inserta una referencia a una regla-condicion en una lista de
reglas y condiciones, siempre que no se halle ya en esta lista */
{
     LISTA *laux;
     LIS_NUM *lcaux;
     ENTREAL val;

     laux = *lr;
     lcaux = *lc;

     while (laux!=NULL)
     { 
       if (!strcmp(laux->nombre,nomr)) {
	  if (lcaux->num.entero == numc) return;
       }
       laux = laux->sig;
       lcaux = lcaux->sig;
     }
     if (*lr==NULL) *lr=cr_lista_id(*lr,nomr);
              else *lr=insertar(*lr,nomr);
     
     val.entero = numc;
     if (*lc==NULL) *lc=crea_ln(*lc,val);
              else *lc=inserta_ln(*lc,val);
}
/************************** PROCESA_CONDIC ****************************/

int Procesa_condic(REGREGLA* reglita) {
     /* este procecimiento se encarga de insertar en la lista de objetos (instancias
	y var.patron) de la regla, todos los objetos que se hallen en las condiciones.
	Tambien se encarga de crear la lista de condiciones, ya que cuando se procesan
	las condiciones de la regla, todas estan al mismo nivel y relacionadas mediante
	el operador AND. */
  
  /* esta es la pila necesaria para recorrer iterativamente los arboles */
  /* de las condiciones de las reglas */
  struct ARB_COND *pila_arb_cond[100];
  int    num_elem_pila;
  
  ARB_COND *arbc;
  EXPRESION *exp;
  LISTA_OBJ *lo;
  int num_vpe,num_vpu;
  LISTA *lvpe, *lvpeaux;
  int num_cond;
  LISTA *param_cond;
  LISTA_OBJ *lo_cond;
  int num_vpe_cond, num_vpu_cond;
  
  /*  if (reglita->tiene_copias) return; */
  arbc = reglita->condiciones;
  num_elem_pila=0;
  num_cond = 0;
  
  do {
    while(arbc!=NULL) {
      /* apilamos */
      if (num_elem_pila>99) {
	printf("Desbordamiento de pila...\n");
	exit(1);
      }
      pila_arb_cond[num_elem_pila]=arbc;
      num_elem_pila++;
      arbc = arbc->subarb_and;
    }
    if (num_elem_pila>0) {
      /* desapilamos */
      num_elem_pila--;
      arbc = pila_arb_cond[num_elem_pila];
      
      /* Procesamos */
      if (arbc->tipo_nodo == nodo_condicion) {
	lo_cond = NULL;
	param_cond = NULL;
	num_vpe_cond = 0;
	num_vpu_cond = 0;
	
	/* Despues se procesan los objetos que contiene las condiciones */
	
	exp = arbc->lado_izq;
	procesa_expr_en_condicion(reglita, exp, num_cond, &lo_cond, &num_vpe_cond, &num_vpu_cond, lvpeaux, &param_cond);
	
	
	exp =arbc->lado_der;
	procesa_expr_en_condicion(reglita, exp, num_cond, &lo_cond, &num_vpe_cond, &num_vpu_cond, lvpeaux, &param_cond); 


	/* Ahora se aade a la lista de condiciones */
	if (reglita->lcond==NULL) 
	  reglita->lcond = crea_lista_condic(reglita->lcond, arbc->oprel, arbc->lado_izq, 
					     arbc->lado_der, arbc->esta_negada, lo_cond,
					     param_cond, num_vpe_cond, num_vpu_cond, num_cond);
	else reglita->lcond = inserta_lista_condic(reglita->lcond, arbc->oprel, arbc->lado_izq,
						   arbc->lado_der, arbc->esta_negada, lo_cond,
						   param_cond, num_vpe_cond, num_vpu_cond, num_cond);
	num_cond++;
	
      }
      if (arbc->sig==NULL && arbc->subarb_or!=NULL)
	arbc=arbc->subarb_or;
      else
	arbc=arbc->sig;
    }
  } while (arbc!=NULL || num_elem_pila>0);
  
  num_vpe=0;
  num_vpu=0;
  /* construimos la lista de las variables patron */
  lvpe = NULL;
  lo = reglita->lobj_cond;
  while (lo!=NULL) {
    if (lo->clvar==paternvar) {
      lvpeaux=insertar_sindup(lvpe,lo->nombre);
      if (lvpeaux!=NULL) {    /* se ha insertado con exito */
	lvpe = lvpeaux;
	if (lo->cuantificador==existe) 
	  num_vpe++;
	else num_vpu++;
      }
    }
    lo=lo->sig;
  }
  if (num_vpe>num_maxvp)
    num_maxvp=num_vpe;
  reglita->param = lvpe;
  reglita->num_vpe = num_vpe;
  reglita->num_vpu = num_vpu;
  return(FALSE);
}


// PROCESAMIENTO DE LAS EXPRESIONES EN LAS CONDICIONES DE UNA REGLA //
/* Esta funcin procesa las expresiones que puedan aparecer en las condiciones
   de una regla, generando las listas de objetos implicados. Est pensada para 
   completar la labor de Procesa_condic, permitiendo que pueda ser llamada 
   recursivamente. Esto permite procesar los ndices de los arrays de forma
   flexible
*/

void procesa_expr_en_condicion(REGREGLA *regla,
			       EXPRESION *exp,
			       int num_cond, 
			       LISTA_OBJ **lo_cond,
			       int *num_vpe,
			       int *num_vpu,
			       LISTA *lvpeaux,
			       LISTA **param_cond) {

  EXPRESION *exp_ant = exp;
  REGTDS *entvp;
  REGTIPO *auxptro;
  LISTA_CAMPOS *lcamp, *lcamp_aux;
  LISTA *campos;

  int clase_elemento, tipo_elemento;

  while (exp != NULL) { 
    campos = NULL; 
    if (exp->esoperando && !exp->escampo &&
	(exp->claseoperando == var || exp->claseoperando == instancia || exp->claseoperando == paternvar)) {
      clase_elemento = exp->claseoperando;  //instancia, paternvar o var
      entvp= busca_en_tds(exp->ident);
      tipo_elemento = entvp->tipo; //definido, entero, real, ... o array
      if (tipo_elemento == array) {
	tipo_elemento = entvp->tbase_array;
	procesa_expr_en_condicion(regla, exp->indice_array, num_cond, lo_cond, 
				  num_vpe, num_vpu, lvpeaux, param_cond);
      }
      while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
	//Solo entrar cuando el tipo sea definido
	exp = exp->sig;
	exp = exp->sig;
	if (campos == NULL)
	  campos = cr_lista_id(campos, exp->ident);
	else
	  campos = insertar(campos, exp->ident);
	if (exp->indice_array != NULL) {
	  procesa_expr_en_condicion(regla, exp->indice_array, num_cond, lo_cond, num_vpe,
				    num_vpu, lvpeaux, param_cond);
	}
      }
      regla->lobj_cond = inserta_lista_obj(regla->lobj_cond, clase_elemento, tipo_elemento,
					   exp_ant->tipooperador, entvp->instancia_de, entvp->nombre,
					   entvp->cod.num_instancia, exp->ident, campos);
      *lo_cond = inserta_lista_obj(*lo_cond, clase_elemento, tipo_elemento, exp_ant->tipooperador,
				  entvp->instancia_de, entvp->nombre, entvp->cod.num_instancia, 
				  exp->ident, campos);
      if (tipo_elemento == definido && clase_elemento != var) {
	//Cuando es de tipo definido hay que aadir la condicin al campo afectado por la misma
	//de forma que la tds recuerde esta informacin. Esto se hace segn el tipo, y no es
	//necesario para variables puesto que no intervienen en el pattern matching.
	switch(clase_elemento) {
	case paternvar:
	  auxptro = busca_en_tipos(entvp->instancia_de);
	  lcamp_aux = auxptro->lc;
	  break;
	case instancia:
	  lcamp_aux = entvp->lc;
	  break;
	}
	while (campos != NULL) {
	  lcamp = busca_campo(campos->nombre, lcamp_aux);
	  lcamp_aux = lcamp->lc;
	  campos = campos->sig;
	}
	inserta_reglacond(&lcamp->reglas, &lcamp->num_conds, regla->nombre, num_cond);
	
	
	//Si es una variable patrn es aadida a la lista de parmetros de la condicin.
	if (clase_elemento == paternvar) {
	  lvpeaux = insertar_sindup(*param_cond, entvp->nombre);
	  if (lvpeaux!=NULL) {    /* se ha insertado con exito */
	    *param_cond = lvpeaux;
	    if (exp_ant->tipooperador==existe) 
	      (*num_vpe)++;
	    else 
	      (*num_vpu)++;
	  }
	}
      }
      else {
	//Cuando no se trata de un elemento de tipo definido solo las instancias pueden 
	//provocar fase de pattern matching. Si el elemento es de tipo definido
	//pero es una variable, tampoco debe entrar por aqu
	if (clase_elemento == instancia) {
	  inserta_reglacond(&entvp->reglas, &entvp->num_conds, regla->nombre, num_cond);
	}
      }
    }
    exp_ant = exp;
    exp = exp->sig;
  }
}

/************************** PROCESA_ACCION ****************************/

int Procesa_accion(REGREGLA* reglita) {
  /* este procecimiento se encarga de insertar en la lista de objetos (instancias
     y var.patron) de la regla, todos los objetos que se hallen en las
     acciones de la regla */
  
  LISTA_ACCIONES *la;
  EXPRESION *exp, *exp_ant;
  REGTDS *entvp;
  LISTA_OBJ *lo;
  LISTA * campos;
  
  
  la = reglita->lacc;
  while (la!=NULL) {
    if (la->tipo_accion==asig_patter) {
      /* si es una vp se busca su correspondiente referencia en la parte izquierda */
      /* de la regla */
      lo=busca_obj(reglita->lobj_cond,la->ident);
      if (lo==NULL) {
	printf("ERROR: La variable patron %s no aparece en la condicion de activacion \n",la->ident);
	printf("de la regla %s, y por tanto, no se instanciara\n",reglita->nombre);
	return(TRUE);  /* si no existe se devuelve el error */
      }
    }


    if (la->tipo_accion == asig_instan || la->tipo_accion == asig_patter ||
	la->tipo_accion == asig_var) {
      exp = la->campos;
      procesa_expr_en_accion(reglita, la->ident, exp);
    }

    exp =la->expr;
    while (exp!=NULL) {
           if (exp->esoperando && exp->claseoperando==paternvar && !exp->escampo) {
                /* si es una vp se busca su correspondiente referencia en la parte izquierda */
                /* de la regla */
                lo=busca_obj(reglita->lobj_cond,exp->ident);
                if (lo==NULL) {
                       printf("ERROR: La variable patron %s no aparece en la condicion de activacion \n",la->ident);
                       printf("de la regla %s, y por tanto, no se instanciara\n",reglita->nombre);
                       return(TRUE);  /* si no existe se devuelve el error */
                }
                entvp= busca_en_tds(exp->ident);
                campos = NULL;
                while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
                    exp=exp->sig;  /* ahora obtenemos el campo: las patern var siempre lo tienen */
                    exp=exp->sig;
                    if (campos==NULL) campos = cr_lista_id(campos,exp->ident);
                                 else campos = insertar(campos, exp->ident);
                }
                reglita->lobj_acc=inserta_lista_obj(reglita->lobj_acc, paternvar,definido,exp_ant->tipooperador,entvp->instancia_de,entvp->nombre,entvp->cod.num_paternvar,exp->ident,campos);
           } else if (exp->esoperando && exp->claseoperando==instancia && !exp->escampo) {
                entvp = busca_en_tds(exp->ident);
                campos = NULL;
                if (entvp->tipo==definido) {
                   while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
                         exp=exp->sig;
                         exp=exp->sig;
                         if (campos==NULL) campos = cr_lista_id(campos,exp->ident);
                                      else campos = insertar(campos, exp->ident);
                   }
                   reglita->lobj_acc=inserta_lista_obj(reglita->lobj_acc, instancia, definido,0,entvp->instancia_de,entvp->nombre,entvp->cod.num_instancia, exp->ident,campos);
                } else
                   reglita->lobj_acc=inserta_lista_obj(reglita->lobj_acc,instancia,entvp->tipo, 0, entvp->instancia_de, entvp->nombre,0, exp->ident,campos);
           } else if (exp->esoperando && exp->claseoperando==var && !exp->escampo) {
                entvp = busca_en_tds(exp->ident);
                campos = NULL;
                if (entvp->tipo==definido) {
                   while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
                         exp=exp->sig;
                         exp=exp->sig;
                         if (campos==NULL) campos = cr_lista_id(campos,exp->ident);
                                      else campos = insertar(campos, exp->ident);
                   }
                   reglita->lobj_acc=inserta_lista_obj(reglita->lobj_acc, var, definido,0,entvp->instancia_de,entvp->nombre,entvp->cod.num_instancia, exp->ident,campos);
                } else
                   reglita->lobj_acc=inserta_lista_obj(reglita->lobj_acc, var,entvp->tipo, 0, entvp->instancia_de, entvp->nombre,0, exp->ident,campos);
           }

           exp_ant = exp;
           exp=exp->sig;
     }
     /*  FALTA: Es realmente necesario incluir las expresiones?: solo lo utilizamos */
     /*  en Genera_Der_reglas para saber cual es la clase de la varible */
      la=la->sig;
  }
  return(FALSE);
}


void procesa_expr_en_accion(REGREGLA *reglita, char *ident, EXPRESION *exp) {
  REGTDS *entvp;
  LISTA *campos;
  char *ident_aux;
  int inserta_obj = 0;
  int cuantificador = 0;

  if (strcmp(ident, "")) {
    entvp = busca_en_tds(ident);
    inserta_obj = 1;
    campos = NULL;
    ident_aux = ident;
  }
  while (exp != NULL) {
    if (exp->esoperando) {
      if (!exp->escampo) {
	if (exp->claseoperando != constante) {
	  entvp = busca_en_tds(exp->ident);
	  inserta_obj = 1;
	  ident_aux  = exp->ident;
	  campos = NULL;
	}
      }
      else {
	if (campos == NULL) {
	  campos = cr_lista_id(campos, exp->ident);
	}
	else {
	  campos = insertar(campos, exp->ident);
	}
      }
      if (exp->indice_array != NULL) {
	procesa_expr_en_accion(reglita, exp->indice_array->ident, exp->indice_array);
      }
    }
    exp = exp->sig;
  }
  if (inserta_obj) {
    //Previene la insercin de constantes, que provocan una violacin de segmento

    //Si se trata de una variable patrn, su cantificador solo podr ser existencial
    //puesto que ya estamos procesando una accin y lo hemos comprobado antes.
    //De poner esto en orden se encarga la siguiente condicin...
    if (entvp->clvar == paternvar) {
      cuantificador = existe;
    }
    reglita->lobj_acc = inserta_lista_obj(reglita->lobj_acc, entvp->clvar, entvp->tipo, //Si tipo == array...???
					  cuantificador, entvp->instancia_de, ident_aux, entvp->cod.num_instancia, 
					  reglita->lacc->campo, campos);
  }
}


ARB_COND * copia_rama(orig)
   ARB_COND * orig;
{
  ARB_COND * arbc_aux;
  EXPRESION * exp_aux;
  ARB_COND * aux1, *aux2;
  ARB_COND * arb_ant;

  if (orig->tipo_nodo == nodo_condicion) {
           arbc_aux=(ARB_COND *) malloc(sizeof(ARB_COND));
           arbc_aux->tipo_nodo = orig->tipo_nodo;
           arbc_aux->oprel = orig->oprel;
           arbc_aux->esta_negada = orig->esta_negada;
           arbc_aux->lado_izq = NULL;
           exp_aux = orig->lado_izq;
           while (exp_aux != NULL) {
               arbc_aux->lado_izq = inserta_lista_exp(arbc_aux->lado_izq, exp_aux->esoperando,
                       exp_aux->escampo,exp_aux->tipooperador,exp_aux->tipooperando,
                       exp_aux->claseoperando,exp_aux->ident,exp_aux->valor);
               exp_aux=exp_aux->sig;
           }
           arbc_aux->lado_der = NULL;
           exp_aux = orig->lado_der;
           while (exp_aux != NULL) {
               arbc_aux->lado_der = inserta_lista_exp(arbc_aux->lado_der, exp_aux->esoperando,
                       exp_aux->escampo,exp_aux->tipooperador,exp_aux->tipooperando,
                       exp_aux->claseoperando,exp_aux->ident,exp_aux->valor);
               exp_aux=exp_aux->sig;
           }
           arbc_aux->sig = NULL;
           arbc_aux->subarb_and = NULL;
           arbc_aux->subarb_or = NULL;
  } else if (orig->tipo_nodo == nodo_and) {
           arbc_aux = (ARB_COND *) malloc(sizeof(ARB_COND));
           arbc_aux->sig = NULL;
           arbc_aux->tipo_nodo = nodo_and;
           arbc_aux->lado_izq = NULL;
           arbc_aux->lado_der = NULL;
           arbc_aux->subarb_and = NULL;
           arbc_aux->subarb_or = NULL;
           aux1 = arbc_aux->subarb_and;
           aux2 = orig->subarb_and;
           while (aux2!=NULL) {
                  if (aux1==NULL) {
                          arbc_aux->subarb_and = copia_rama(aux2);
                          aux1 = arbc_aux->subarb_and;
                  } else {
                          aux1->sig = copia_rama(aux2);
                          aux1 = aux1->sig;
                  }
                  aux2 = aux2->sig;
           }
           aux2 = orig->subarb_or;
           arb_ant = NULL;
           while (aux2!=NULL) {
              aux1 = (ARB_COND *) malloc(sizeof(ARB_COND));
              aux1->sig = NULL;
              aux1->tipo_nodo = nodo_or;
              aux1->lado_izq = NULL;
              aux1->lado_der = NULL;
              aux1->subarb_and = NULL;
              aux1->subarb_or = copia_rama(aux2->subarb_or);
              aux1->subarb_or->sig = copia_rama(aux2->subarb_or->sig);
              if (arbc_aux->subarb_or==NULL) arbc_aux->subarb_or = aux1;
                                        else arb_ant->sig = aux1;
              arb_ant = aux1;
              aux2=aux2->sig;
           }
  }
  return (arbc_aux);
}

void escinde_OR(c1, c2, orig, rama_OR)
   ARB_COND **c1;
   ARB_COND **c2;
   ARB_COND *orig;
   ARB_COND *rama_OR;
{
  ARB_COND * arbc_aux;
  EXPRESION * exp_aux;
  ARB_COND * copia1, *copia2;
  ARB_COND * aux1, *aux2;
  ARB_COND * aux3;
  ARB_COND * arb_ant1, *arb_ant2;
  ARB_COND * ult_and1, *ult_and2;

  if (orig->tipo_nodo == nodo_condicion) {
           arbc_aux=(ARB_COND *) malloc(sizeof(ARB_COND));
           arbc_aux->tipo_nodo = orig->tipo_nodo;
           arbc_aux->oprel = orig->oprel;
           arbc_aux->esta_negada = orig->esta_negada;
           arbc_aux->lado_izq = NULL;
           exp_aux = orig->lado_izq;
           while (exp_aux != NULL) {
               arbc_aux->lado_izq = inserta_lista_exp(arbc_aux->lado_izq, exp_aux->esoperando,
                       exp_aux->escampo,exp_aux->tipooperador,exp_aux->tipooperando,
                       exp_aux->claseoperando,exp_aux->ident,exp_aux->valor);
               exp_aux=exp_aux->sig;
           }
           arbc_aux->lado_der = NULL;
           exp_aux = orig->lado_der;
           while (exp_aux != NULL) {
               arbc_aux->lado_der = inserta_lista_exp(arbc_aux->lado_der, exp_aux->esoperando,
                       exp_aux->escampo,exp_aux->tipooperador,exp_aux->tipooperando,
                       exp_aux->claseoperando,exp_aux->ident,exp_aux->valor);
               exp_aux=exp_aux->sig;
           }
           *c1 = arbc_aux;
           arbc_aux=(ARB_COND *) malloc(sizeof(ARB_COND));
           arbc_aux->tipo_nodo = orig->tipo_nodo;
           arbc_aux->oprel = orig->oprel;
           arbc_aux->esta_negada = orig->esta_negada;
           arbc_aux->lado_izq = NULL;
           exp_aux = orig->lado_izq;
           while (exp_aux != NULL) {
               arbc_aux->lado_izq = inserta_lista_exp(arbc_aux->lado_izq, exp_aux->esoperando,
                       exp_aux->escampo,exp_aux->tipooperador,exp_aux->tipooperando,
                       exp_aux->claseoperando,exp_aux->ident,exp_aux->valor);
               exp_aux=exp_aux->sig;
           }
           arbc_aux->lado_der = NULL;
           exp_aux = orig->lado_der;
           while (exp_aux != NULL) {
               arbc_aux->lado_der = inserta_lista_exp(arbc_aux->lado_der, exp_aux->esoperando,
                       exp_aux->escampo,exp_aux->tipooperador,exp_aux->tipooperando,
                       exp_aux->claseoperando,exp_aux->ident,exp_aux->valor);
               exp_aux=exp_aux->sig;
           }
           *c2 = arbc_aux;
           (*c1)->sig = NULL;
           (*c1)->subarb_and = NULL;
           (*c1)->subarb_or = NULL;
           (*c2)->sig = NULL;
           (*c2)->subarb_and = NULL;
           (*c2)->subarb_or = NULL;
  } else
    if (orig->tipo_nodo == nodo_and) {
           (*c1) = (ARB_COND *) malloc(sizeof(ARB_COND));
           (*c2) = (ARB_COND *) malloc(sizeof(ARB_COND));
           (*c1)->sig = NULL;
           (*c2)->sig = NULL;
           (*c1)->tipo_nodo = nodo_and;
           (*c2)->tipo_nodo = nodo_and;
           (*c1)->lado_izq = NULL;
           (*c1)->lado_der = NULL;
           (*c2)->lado_izq = NULL;
           (*c2)->lado_der = NULL;
           (*c1)->subarb_and = NULL;
           (*c2)->subarb_and = NULL;
           (*c1)->subarb_or = NULL;
           (*c2)->subarb_or = NULL;

           aux1 = (*c1)->subarb_and;
           aux2 = (*c2)->subarb_and;
           arbc_aux = orig->subarb_and;
           while (arbc_aux!=NULL) {
                  escinde_OR(&copia1, &copia2, arbc_aux, rama_OR);
                  if (aux1==NULL) {
                          (*c1)->subarb_and = copia1;
                          aux1 = copia1;
                  } else {
                          aux1->sig = copia1;
                          aux1 = copia1;
                  }
                  if (aux2==NULL) {
                          (*c2)->subarb_and = copia2;
                          aux2 = copia2;
                  } else {
                          aux2->sig = copia2;
                          aux2 = copia2;
                  }
                  arbc_aux = arbc_aux->sig;
           }
           ult_and1 = aux1;
           ult_and2 = aux2;
           aux3 = orig->subarb_or;
           arb_ant1 = NULL;
           arb_ant2 = NULL;
           while (aux3!=NULL) {
             if (aux3==rama_OR) { /* si es la rama que buscamos */
                 arbc_aux = aux3;
                 if ((*c1)->subarb_and==NULL) {
                       (*c1)->subarb_and=copia_rama(arbc_aux->subarb_or);
                       ult_and1 = (*c1)->subarb_and;
                 } else {
                       ult_and1->sig = copia_rama(arbc_aux->subarb_or);
                       ult_and1=ult_and1->sig;
                 }
                 arbc_aux = arbc_aux->subarb_or;
                 if ((*c2)->subarb_and==NULL) {
                       (*c2)->subarb_and=copia_rama(arbc_aux->sig);
                       ult_and2 = (*c2)->subarb_and;
                 } else {
                       ult_and2->sig = copia_rama(arbc_aux->sig);
                       ult_and2=ult_and2->sig;
                 }
             } else {  /* si no es la rama que buscamos */
                  aux1 = (ARB_COND *) malloc(sizeof(ARB_COND));
                  aux2 = (ARB_COND *) malloc(sizeof(ARB_COND));
                  aux1->sig = NULL;
                  aux2->sig = NULL;
                  aux1->tipo_nodo = nodo_or;
                  aux2->tipo_nodo = nodo_or;
                  aux1->lado_izq = NULL;
                  aux1->lado_der = NULL;
                  aux2->lado_izq = NULL;
                  aux2->lado_der = NULL;
                  aux1->subarb_and = NULL;
                  aux2->subarb_and = NULL;
                  aux1->subarb_or = NULL;
                  aux2->subarb_or = NULL;
                  if (arb_ant1!=NULL) arb_ant1->sig = aux1;
                             else (*c1)->subarb_or = aux1;
                  if (arb_ant2!=NULL) arb_ant2->sig = aux2;
                             else (*c2)->subarb_or = aux2;
                  arb_ant1 = aux1;
                  arb_ant2 = aux2;
                  arbc_aux = aux3;

                  /* primera rama del OR */
                  arbc_aux = arbc_aux->subarb_or;
                  escinde_OR(&copia1, &copia2, arbc_aux, rama_OR);
                  aux1->subarb_or = copia1;
                  aux2->subarb_or = copia2;

                  /* segunda rama del OR */
                  aux1 = aux1->subarb_or;
                  aux2 = aux2->subarb_or;
                  arbc_aux = arbc_aux->sig;
                  escinde_OR(&copia1, &copia2, arbc_aux, rama_OR);
                  aux1->sig = copia1;
                  aux2->sig = copia2;
             }
             aux3=aux3->sig;
           }
  } else
      if (orig->tipo_nodo == nodo_or) {
           aux3 = orig;
           arb_ant1 = NULL;
           arb_ant2 = NULL;
           while (aux3!=NULL) {
             if (aux3==rama_OR) { /* si es la rama que buscamos */
                 arbc_aux = aux3;
                 (*c1)=copia_rama(arbc_aux->subarb_or);
                 arbc_aux = arbc_aux->subarb_or;
                 (*c2)=copia_rama(arbc_aux->sig);
             } else {  /* si no es la rama que buscamos */
                  aux1 = (ARB_COND *) malloc(sizeof(ARB_COND));
                  aux2 = (ARB_COND *) malloc(sizeof(ARB_COND));
                  aux1->sig = NULL;
                  aux2->sig = NULL;
                  aux1->tipo_nodo = nodo_or;
                  aux2->tipo_nodo = nodo_or;
                  aux1->lado_izq = NULL;
                  aux1->lado_der = NULL;
                  aux2->lado_izq = NULL;
                  aux2->lado_der = NULL;
                  aux1->subarb_and = NULL;
                  aux2->subarb_and = NULL;
                  aux1->subarb_or = NULL;
                  aux2->subarb_or = NULL;
                  if (arb_ant1!=NULL) arb_ant1->sig = aux1;
                             else (*c1) = aux1;
                  if (arb_ant2!=NULL) arb_ant2->sig = aux2;
                             else (*c2) = aux2;
                  arb_ant1 = aux1;
                  arb_ant2 = aux2;
                  arbc_aux = aux3;

                  /* primera rama del OR */
                  arbc_aux = arbc_aux->subarb_or;
                  escinde_OR(&copia1, &copia2, arbc_aux, rama_OR);
                  aux1->subarb_or = copia1;
                  aux2->subarb_or = copia2;

                  /* segunda rama del OR */
                  aux1 = aux1->subarb_or;
                  aux2 = aux2->subarb_or;
                  arbc_aux = arbc_aux->sig;
                  escinde_OR(&copia1, &copia2, arbc_aux, rama_OR);
                  aux1->sig = copia1;
                  aux2->sig = copia2;
             }
             aux3=aux3->sig;
           }
      } else {
        printf("ERROR FATAL: nodo inesperado en escinde_OR()");
        exit(1);
        }
}

void duplica_regla(regla, pto_rup)
   REGREGLA *regla;        /* regla a duplicar */
   ARB_COND *pto_rup;      /* condicion OR que se debe romper en dos AND */
{
  REGREGLA *copia1, *copia2;
  char nombre1[long_ident+1];
  char nombre2[long_ident+1];
  char cad1[3];
  char cad2[3];
  char err1, err2;

  ARB_COND *arbc1,*arbc2;
  EXPRESION *exp_aux;
  REGREGLA *regla_aux;


  regla->tiene_copias=TRUE;

  /* ahora buscamos al padre de todas las reglas escindidas, para */
  /* controlar el nombre y el codigo numerico */
  regla_aux = regla;
  while (regla_aux->copia_de !=NULL) regla_aux=regla_aux->copia_de;
  regla_aux->num_copias++;
  sprintf(cad1,"%02d",regla_aux->num_copias);
  regla_aux->num_copias++;
  sprintf(cad2,"%02d",regla_aux->num_copias);

  /* ahora se modifica el nombre */
  strcpy(nombre1, regla_aux->nombre);
  strcat(nombre1,cad1);
  strcpy(nombre2, regla_aux->nombre);
  strcat(nombre2,cad2);

  copia1=insregla_enreglas(nombre1,numregla++);
  copia2=insregla_enreglas(nombre2,numregla++);
  copia1->copia_de=regla;
  copia2->copia_de=regla;

  /* ahora hay que copiar las condiciones modificandolas en el OR */
  /* de forma que se convierta en AND */

  escinde_OR(&arbc1,&arbc2,regla->condiciones,pto_rup);
  copia1->condiciones = arbc1;
  copia2->condiciones = arbc2;
  /* FALTA: puede ser necesario volver a escindir estas reglas para */
  /* otros OR's, es decir, que se deberia llamar a Procesa_ORs con estas */
  /* dos nuevas reglas */
  Procesa_regla(copia1,&err1);
  Procesa_regla(copia2,&err1);
}

void Procesa_ORs(regla)
   REGREGLA *regla;
{
  /* esta es la pila necesaria para recorrer iterativamente los arboles */
  /* de las condiciones de las reglas */
  struct ARB_COND *pila_arb_cond[100];
  int    num_elem_pila;

  ARB_COND *arbc;

  arbc = regla->condiciones;
  num_elem_pila=0;
  do {
      while(arbc!=NULL) {
           /* apilamos */
           if (num_elem_pila>99) {
              printf("Desbordamiento de pila...\n");
              exit(1);
           }
           pila_arb_cond[num_elem_pila]=arbc;
           num_elem_pila++;
           arbc = arbc->subarb_and;
      }
      if (num_elem_pila>0) {
           /* desapilamos */
           num_elem_pila--;
           arbc = pila_arb_cond[num_elem_pila];

           /* Procesamos la rama OR */
           if (arbc->tipo_nodo!=nodo_condicion) {
               if (arbc->tipo_nodo==nodo_or) {
                  duplica_regla(regla,arbc);
                   /* solo procesamos el primer OR, y despues las reglas */
                   /* copias creadas ya procesaran sus ORs */
                  return;
               } else if (arbc->subarb_or!=NULL) {
                      duplica_regla(regla,arbc->subarb_or);
                      /* solo procesamos el primer OR, y despues las reglas */
                      /* copias creadas ya procesaran sus ORs */
                      return;
               }
           }
           arbc=arbc->sig;
      }
  } while (arbc!=NULL || num_elem_pila>0);

}
/************************** PROCESA_REGLAS ****************************/

void Procesa_regla(regla,error)
  REGREGLA *regla;
  char *error;
{
  int err_Procesa_condic;
  int err_Procesa_accion;

  Procesa_ORs(regla);
  err_Procesa_condic=Procesa_condic(regla);
  err_Procesa_accion=Procesa_accion(regla);

  if (err_Procesa_condic || err_Procesa_accion) *error = TRUE;
                                       else     *error = FALSE;

}

/************************** PROCESA_FUNCION ****************************/
void Procesa_funcion(funcion,error)
  REGFUNC *funcion;
  char *error;
{
  LISTA_ACCIONES *la;
  EXPRESION *exp, *exp_ant;
  REGTDS *entvp;
  LISTA_OBJ *lo;
  LISTA_PARAM *lparam;
  LISTA * campos;
  MODIFICACIONES * mod_aux, *mod_aux2;

  la = funcion->lacc;
  while (la!=NULL) {
      if (la->tipo_accion==asig_paramt) {
         /* se busca el identificador en la TDS */
         entvp = busca_en_tds(la->ident);
         exp = la->campos;
         campos = NULL;
         while (exp!=NULL) {
                 exp=exp->sig;  /* ahora obtenemos el campo: las patern var siempre lo tienen */
                 if (campos==NULL) campos = cr_lista_id(campos,exp->ident);
                              else campos = insertar(campos, exp->ident);
                 exp=exp->sig;
         }
         lparam = busca_param(funcion->lp,la->ident);
         funcion->lobj_func=inserta_lista_obj(funcion->lobj_func, paramt,entvp->tipo, lparam->por_ref,entvp->instancia_de,la->ident,0,la->campo,campos);
         if (lparam->por_ref) {
            /* si es una asignacion a un parametro por referencia */
            /* se anota una posible modificacion de la BH */
            mod_aux = (MODIFICACIONES *) malloc (sizeof(MODIFICACIONES));
            strcpy(mod_aux->nombre,la->ident);
            mod_aux->campos = la->campos;  /* no es necesario copiarlos */
            mod_aux->clase_obj = paramt;
            mod_aux->tip_izq = la->tip_izq;
            mod_aux->sig = NULL;
            if (funcion->modif==NULL) funcion->modif = mod_aux;
            else {
              mod_aux2= funcion->modif;
              while(mod_aux2->sig!=NULL) mod_aux2=mod_aux2->sig;
              mod_aux2->sig = mod_aux;
            }
         }
      }
      if (la->tipo_accion==asig_instan) {
         entvp = busca_en_tds(la->ident);
         exp = la->campos;
         campos = NULL;
         while (exp!=NULL) {
                 exp=exp->sig;
                 if (campos==NULL) campos = cr_lista_id(campos,exp->ident);
                              else campos = insertar(campos, exp->ident);
                 exp=exp->sig;
         }
         funcion->lobj_func=inserta_lista_obj(funcion->lobj_func, instancia,entvp->tipo, 0 ,entvp->instancia_de,la->ident,entvp->cod.num_instancia,la->campo,campos);
         /* se anota una posible modificacion de la BH */
         mod_aux = (MODIFICACIONES *) malloc (sizeof(MODIFICACIONES));
         strcpy(mod_aux->nombre,la->ident);
         mod_aux->campos = la->campos;  /* no es necesario copiarlos */
         mod_aux->clase_obj = instancia;
         mod_aux->tip_izq = la->tip_izq;
         mod_aux->sig = NULL;
         if (funcion->modif==NULL) funcion->modif = mod_aux;
         else {
              mod_aux2= funcion->modif;
              while(mod_aux2->sig!=NULL) mod_aux2=mod_aux2->sig;
              mod_aux2->sig = mod_aux;
         }
      }

      if (la->tipo_accion==asig_var) {
         entvp = busca_en_tds(la->ident);
         exp = la->campos;
         campos = NULL;
         while (exp!=NULL) {
                 exp=exp->sig;
                 if (campos==NULL) campos = cr_lista_id(campos,exp->ident);
                              else campos = insertar(campos, exp->ident);
                 exp=exp->sig;
         }
         funcion->lobj_func=inserta_lista_obj(funcion->lobj_func, var,entvp->tipo, 0 ,entvp->instancia_de,la->ident,entvp->cod.num_instancia,la->campo,campos);
      }
 
        exp =la->expr;
        while (exp!=NULL) {
           if (exp->esoperando && exp->claseoperando==paramt && !exp->escampo) {
                entvp= busca_en_tds(exp->ident);
                lparam = busca_param(funcion->lp,exp->ident);
                campos = NULL;
                while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
                    exp=exp->sig;
                    exp=exp->sig;
                    if (campos==NULL) campos = cr_lista_id(campos,exp->ident);
                                 else campos = insertar(campos, exp->ident);
                }
                funcion->lobj_func=inserta_lista_obj(funcion->lobj_func, paramt,entvp->tipo,lparam->por_ref,entvp->instancia_de,entvp->nombre,0,exp->ident,campos);
           } else if (exp->esoperando && exp->claseoperando==instancia && !exp->escampo) {
                entvp = busca_en_tds(exp->ident);
                campos = NULL;
                if (entvp->tipo==definido) {
                   while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
                         exp=exp->sig;
                         exp=exp->sig;
                         if (campos==NULL) campos = cr_lista_id(campos,exp->ident);
                                      else campos = insertar(campos, exp->ident);
                   }
                   funcion->lobj_func=inserta_lista_obj(funcion->lobj_func, instancia, definido,0,entvp->instancia_de,entvp->nombre,entvp->cod.num_instancia, exp->ident,campos);
                } else
                   funcion->lobj_func=inserta_lista_obj(funcion->lobj_func,instancia,entvp->tipo, 0, entvp->instancia_de, entvp->nombre,0, exp->ident,campos);
           } else if (exp->esoperando && exp->claseoperando==var && !exp->escampo) {
                entvp = busca_en_tds(exp->ident);
                campos = NULL;
                if (entvp->tipo==definido) {
                   while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
                         exp=exp->sig;
                         exp=exp->sig;
                         if (campos==NULL) campos = cr_lista_id(campos,exp->ident);
                                      else campos = insertar(campos, exp->ident);
                   }
                   funcion->lobj_func=inserta_lista_obj(funcion->lobj_func, var, definido,0,entvp->instancia_de,entvp->nombre,entvp->cod.num_instancia, exp->ident,campos);
                } else
                   funcion->lobj_func=inserta_lista_obj(funcion->lobj_func,var,entvp->tipo, 0, entvp->instancia_de, entvp->nombre,0, exp->ident,campos);
           }
           exp_ant = exp;
           exp=exp->sig;
        }
     /*  FALTA: Es realmente necesario incluir las expresiones?: solo lo utilizamos */
     /*  en Genera_Der_reglas para saber cual es la clase de la varible */
      la=la->sig;
  }
}

/* DETECCION DE CONDICIONES REPETIDAS */

LISTA_LISTA_CONDIC * inserta_condic_regla(llc,cond,reg)
     LISTA_LISTA_CONDIC *llc;
     LISTA_CONDIC *cond;
     REGREGLA *reg;
{

 LISTA_LISTA_CONDIC *aux;

 aux=llc;
 if (aux!=NULL) {
    while(aux->sig!=NULL) aux=aux->sig;
    aux->sig=(LISTA_LISTA_CONDIC *)malloc(sizeof(LISTA_LISTA_CONDIC));
    aux=aux->sig;
 } else {
    aux=(LISTA_LISTA_CONDIC *)malloc(sizeof(LISTA_LISTA_CONDIC));
    llc=aux;
 }
 aux->condicion = cond;
 aux->regla = reg;
 aux->sig=NULL;
 return(llc);
}

int compara_expr(exp1,exp2)
  EXPRESION * exp1;
  EXPRESION * exp2;
{
 if (exp1->esoperando==exp2->esoperando)
  if (exp1->esoperando) {
    if (exp1->claseoperando== exp2->claseoperando)
     switch (exp1->claseoperando) {
       case constante: if (exp1->tipooperando==exp2->tipooperando) {
                          if (exp1->tipooperando == real) {
                            if (exp1->valor.real == exp2->valor.real)
                              return(TRUE);
                          } else if (exp1->valor.entero == exp2->valor.entero)
                              return(TRUE);
                       }
                       return(FALSE);
                       break;
       case instancia:
       case paternvar: if (exp1->tipooperando==exp2->tipooperando) {
                         if (exp1->escampo==exp2->escampo)
                           if (!strcmp(exp1->ident,exp2->ident))
                              return(TRUE);
                       }
                       return(FALSE);
                       break;

     }
  } else { /* es operador */
    if (exp1->tipooperador == exp2->tipooperador) return (TRUE);
                                         else    return (FALSE);
  }
  return(FALSE);
}

int compara_lista_expr(exp1,exp2)
  EXPRESION * exp1;
  EXPRESION * exp2;
{
    EXPRESION * aux1;
    EXPRESION * aux2;

    aux1=exp1;
    aux2=exp2;
    while (aux1!=NULL && aux2!=NULL) {
       if (compara_expr(aux1,aux2) && compara_lista_expr(aux1->indice_array, aux2->indice_array)) {
              aux1=aux1->sig;
              aux2=aux2->sig;
       } else return(FALSE);
    }
    if (aux1!=NULL || aux2!=NULL) return (FALSE);
    return(TRUE);
}

int compara_condic(cond1, cond2)
  LISTA_CONDIC * cond1;
  LISTA_CONDIC * cond2;
{
/* esta funcion devuelve un booleano que indica si las dos condiciones */
/* pasadas como parametros son iguales */
  if (cond1->num_vpe==cond2->num_vpe)
    if (cond1->num_vpu==cond2->num_vpu)
      if (cond1->esta_negada == cond2->esta_negada)
        if (compara_lista_expr(cond1->lado_izq,cond2->lado_izq))
           if (compara_lista_expr(cond1->lado_der,cond2->lado_der))
               return(TRUE);
   return(FALSE);
}

void Detecta_cond_repetidas()
{
/* esta funcion detecta las condiciones repetidas */

  EHREGLA * aux;
  int i;
  LISTA_CONDIC *lcaux;
  LISTA_LISTA_CONDIC *llcaux;
  LISTA_LISTA_CONDIC *llc, *llcant;
  int cont;

  /* primero se inicializa la estructura empleada como base */
  for(i=0;i<7; i++) clas_cond[i]=NULL;


  for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
          if (!aux->eregla.tiene_copias) {
            /* primero se clasifican las condiciones segun su op. relacional */
            lcaux = aux->eregla.lcond;
            while (lcaux!=NULL) {
              switch (lcaux->oprel) {
                   case menorig : clas_cond[0]=inserta_condic_regla(clas_cond[0],lcaux,&(aux->eregla));
                                  break;
                   case mayorig : clas_cond[1]=inserta_condic_regla(clas_cond[1],lcaux,&(aux->eregla));
                                  break;
                   case mayor:    clas_cond[2]=inserta_condic_regla(clas_cond[2],lcaux,&(aux->eregla));
                                  break;
                   case menor:    clas_cond[3]=inserta_condic_regla(clas_cond[3],lcaux,&(aux->eregla));
                                  break;
                   case igual:    clas_cond[4]=inserta_condic_regla(clas_cond[4],lcaux,&(aux->eregla));
                                  break;
                   case distinto: clas_cond[5]=inserta_condic_regla(clas_cond[5],lcaux,&(aux->eregla));
                                  break;
                   case nulo:     clas_cond[6]=inserta_condic_regla(clas_cond[6],lcaux,&(aux->eregla));
              }
              lcaux=lcaux->sig;
            }
          }
          aux=aux->sig;
         }
     }
     /* ahora se procesa cada op.relacional buscando condiciones iguales */
     /* para ello se compara cada condicion con la siguiente */
     for(cont=0;cont<7;cont++)
     {
         if (clas_cond[cont]!=NULL) {
             llc = clas_cond[cont];
             while (llc!=NULL) {
                 llcaux = llc->sig;
                 llcant = llc;
                 while (llcaux!=NULL) {
                    if (compara_condic(llcaux->condicion,llc->condicion)) {
                       printf("Encontradas condiciones iguales: \n");
                       printf("1: REGLA: %s Numero de condicion: %d\n",llcaux->regla->nombre,llcaux->condicion->num_condic);
                       printf("2: REGLA: %s Numero de condicion: %d\n",llc->regla->nombre,llc->condicion->num_condic);
                       printf("\n");

                       /* ahora se modifica la condicion para que sea una copia de la original */
                       llcaux->condicion->es_orig = FALSE;
                       strcpy(llcaux->condicion->regla_orig,llc->regla->nombre);
                       llcaux->condicion->cond_orig = llc->condicion->num_condic;

                    /* aqui se podria eliminar la segunda condicion igual de la lista */
                    /* ya que si cualquier otra fuera igual a esta ya lo habria */
                    /* sido de la primera: asi se optimiza un poco la busqueda */
                       llcant->sig = llcaux->sig;
                       free(lcaux);

                    }
                    llcant = llcaux;
                    llcaux=llcaux->sig;
                 }
                 llc = llc->sig;
             }
         }
     }

}

