/*

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                Arlips compiler v. 3.1           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

File genera.c, code generation
April 2006

Copyright (C) 2006 Emilio Vivancos Rubio

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                              Contact                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Author:              Emilio Vivancos Rubio  vivancos@dsic.upv.es
Developer Arlips v2: Mario Gonzlez Gimnez mgonzale@dsic.upv.es
Developer Arlips v3: Carlos Garca Montoro  cargarcia@dsic.upv.es

Departamento de Sistemas Informticos y Computacin
Universidad Politcnica de Valencia
Camino de Vera s/n
46022 Valencia
Spain

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

*/




#include "genera.h"
#include "tds.h"
#include "alex.h"
#include "genera.h"
#include "tds.h"
#include "alex.h"
#include "es.h"


extern int verbose;
FILE * fich_gen;        /* contiene el fichero donde se escribe */

char fich_make[80]="Makefile";

int num_maxvp=0;
int debug=FALSE;
int nomain=FALSE;
int nomake=FALSE;
int noprints = FALSE;
int verbos = FALSE;
int artis_mode = FALSE;
int log_ = FALSE;
char prefijo[9];              /* contiene un prefijo para todos los identificadores, para cuando se compila el */
                              /* sistema de produccion sin la funcion main */
enum fich_gener{cab_agenda,fich_agenda,cab_tipos,fich_tipos, 
		cab_log, fich_log, fich_bh};
char fich_generados[7][80]={"agenda_.h", "agenda_.c","tipos_.h","tipos_.c", 
			   "log_.c", "log_.h", "fich_bh.h"};


/* esta es la pila necesaria para recorrer iterativamente los arboles */
/* de las condiciones de las reglas */
struct ARB_COND *pila_arb_cond[100];
int    num_elem_pila;

/* Estas dos listas mantienen las condiciones-reglas que se ven afectadas */
/* por las inicializaciones */
LISTA * reglas_reset=NULL;
LIS_NUM * condiciones_reset=NULL;

/* nombre del sistema de produccion */
char nom_sp[20];

//Nombres de las funciones ofertadas por el KDM para obtener o escribir informacin
char f_GetKDM[] = "KDM_Get_Value";
char f_SetKDM[] = "KDM_Set_Value";


/****************************** GENERA_EXP *******************************/
//NOTA IMPORTANTE:
//Es importante que Genera_exp solo pueda finalizar su ejecucin o bien por
//haber acabado de generar la expresin, o bien por un break tras encontrar una coma
//Dicho break se usa para determinadas funciones internas, pero la generacin de cdigo
//para la llamada a funciones externas depende de esas dos condiciones.
EXPRESION * Genera_exp(
		       EXPRESION *exp,     /* expresion que se quiere imprimir */
		       char * cadena_expr,  /* cadena generada por la expresion */
		       LISTA_OBJ * lobj) { /* lista de objetos que pueden aparecer en la expresion: */
                                           /* variables patron, etc... */


  LISTA_OBJ * lo;
  LISTA * campos;
  LISTA * campaux;
  char cad_aux[255];
  
  //Variables necesarias si hay que imprimir un array:
  EXPRESION *exp_ant;
  REGTDS *elemento;
  LISTA_CAMPOS *l_campos;
  int tbase_array;
  int tam_array;

  //Variables para imprimir la lista de parmetros de una funcin externa
  int comas_en_lista_param_funcion = 0; //Servir para saber cuantas comas han de ser impresas.
  REGFUNC *funcion_en_tds; //Registro de la TDS que contiene los datos de la funcin.
  int i; //Contador
  
  cad_aux[0]='\x0';
  cadena_expr[0]='\x0';
  while (exp!=NULL) {
    
    if (!exp->esoperando && exp->tipooperador==coma) {
      break;
    }
    
    if ((!exp->esoperando) && ((exp->tipooperador==existe) || (exp->tipooperador==paratodo)) ) {
      exp=exp->sig;
      lo = busca_obj(lobj,exp->ident);
      sprintf(cad_aux,"%si_%s[%s]",prefijo,lo->instancia_de,lo->nombre);
      strcat(cadena_expr,cad_aux);
    } 
    else
      if (!exp->esoperando)
	switch (exp->tipooperador) {
	case mas: 
	  sprintf(cad_aux,"+");
	  strcat(cadena_expr,cad_aux);
	  break;
	case menos:
	  sprintf(cad_aux,"-");
	  strcat(cadena_expr,cad_aux);
	  break;
	case oposim: 
	  sprintf(cad_aux," || ");
	  strcat(cadena_expr,cad_aux);
	  break;
	case opysim:
	  sprintf(cad_aux," && ");
	  strcat(cadena_expr,cad_aux);
	  break;
	case opmult:
	  sprintf(cad_aux,"*");
	  strcat(cadena_expr,cad_aux);
	  break;
	case opdiv:
	  sprintf(cad_aux,"/");
	  strcat(cadena_expr,cad_aux);
	  break;
	case abrepar:
	  sprintf(cad_aux,"(");
	  strcat(cadena_expr,cad_aux);
	  break;
	case cierrapar:
	  sprintf(cad_aux,")");
	  strcat(cadena_expr,cad_aux);
	  break;
	case punto:
	  sprintf(cad_aux,".");
	  strcat(cadena_expr,cad_aux);
	  break;
	case sizesim :
	  exp=exp->sig;
	  exp=exp->sig; /* pasamos el parentesis */
	  if (exp->tipooperador==existe) { /* si es una variable patron */
	    exp=exp->sig;
	    lo = busca_obj(lobj,exp->ident);
	    campos = NULL;
	    while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
	      exp=exp->sig;  /* ahora obtenemos el campo: las patern var siempre lo tienen */
	      exp=exp->sig;
	      if (campos==NULL)
		campos = cr_lista_id(campos,exp->ident);
	      else
		campos = insertar(campos, exp->ident);
	    }
	    switch (exp->tipooperando) {
	    case lista_enteros:
	      sprintf(cad_aux,"size_lent(");
	      strcat(cadena_expr,cad_aux);
	      break;
	    case lista_reales:
	      sprintf(cad_aux,"size_lreal(");
	      strcat(cadena_expr,cad_aux);
	      break;
	    }
	    sprintf(cad_aux,"&%si_%s[%s]", prefijo, lo->instancia_de, lo->nombre);
	    strcat(cadena_expr,cad_aux);
	    campaux = campos;
	    while ( campaux!=NULL) {
	      sprintf(cad_aux,".%s",campaux->nombre);
	      strcat(cadena_expr,cad_aux);
	      campaux = campaux->sig;
	    }
	    sprintf(cad_aux,")");
	    strcat(cadena_expr,cad_aux);
	    borrar(campos);
	  }
	  else { 
	    /* se trata de una instancia */
	    //CGM: Tambin puede tratarse de una variable global de tipo definido
	    lo = busca_obj(lobj,exp->ident);
	    if (exp->tipooperando==definido) {
	      campos = NULL;
	      while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
		exp=exp->sig;
		exp=exp->sig;
		if (campos==NULL)
		  campos = cr_lista_id(campos,exp->ident);
		else
		  campos = insertar(campos, exp->ident);
	      }
	      switch (exp->tipooperando) {
	      case lista_enteros: 
		sprintf(cad_aux,"size_lent(");
		strcat(cadena_expr,cad_aux);
		break;
	      case lista_reales:
		sprintf(cad_aux,"size_lreal(");
		strcat(cadena_expr,cad_aux);
		break;
	      }
	      switch (lo->clvar) {
	      case instancia:
		sprintf(cad_aux,"&%si_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
		break;
	      case var:
		sprintf(cad_aux,"&%sv_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
		break;
	      default:
		fprintf(stderr, "ERROR del compilador Arlips.\n");
		fprintf(stderr, "Se esperaba un objeto lo con lo->clvar instancia o var,\n");
		fprintf(stderr, "Pero se recibi %d. (Ver en tds.h la enum clasobj\n", lo->clvar);
		fprintf(stderr, "Revisa en genera.c la funcin Genera_exp() en su caso sizesim.\n");
		fprintf(stderr, "Revisa la insercin de elementos en la tds\n");
		break;
	      }
	      strcat(cadena_expr,cad_aux);
	      campaux = campos;
	      while ( campaux!=NULL) {
		sprintf(cad_aux,".%s",campaux->nombre);
		strcat(cadena_expr,cad_aux);
		campaux = campaux->sig;
	      }
	      sprintf(cad_aux,")");
	      strcat(cadena_expr,cad_aux);
	      borrar(campos);
	    } 
	    else {
	      switch (exp->tipooperando) {
	      case lista_enteros: 
		sprintf(cad_aux,"size_lent(");
		strcat(cadena_expr,cad_aux);
		break;
	      case lista_reales:
		sprintf(cad_aux,"size_lreal(");
		strcat(cadena_expr,cad_aux);
		break;
	      }
	      sprintf(cad_aux,"&%s%s)",prefijo, lo->nombre);
	      strcat(cadena_expr,cad_aux);
	    }
	  }
	  exp=exp->sig; /* pasamos el parentesis */
	  break;

	case headsim :  /* todavia no sabemos si es una lista */
	  /* de enteros o reales */
	  exp=exp->sig;
	  exp=exp->sig; /* pasamos el parentesis */
	  if (exp->tipooperador==existe) { /* si es una variable patron */
	    exp=exp->sig;
	    lo = busca_obj(lobj,exp->ident);
	    campos = NULL;
	    while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
	      exp=exp->sig;  /* ahora obtenemos el campo: las patern var siempre lo tienen */
	      exp=exp->sig;
	      if (campos==NULL)
		campos = cr_lista_id(campos,exp->ident);
	      else 
		campos = insertar(campos, exp->ident);
	    }
	    switch (exp->tipooperando) {
	    case lista_enteros: sprintf(cad_aux,"head_lent(");
	      strcat(cadena_expr,cad_aux);
	      break;
	    case lista_reales: sprintf(cad_aux,"head_lreal(");
	      strcat(cadena_expr,cad_aux);
	      break;
	    }
	    sprintf(cad_aux,"&%si_%s[%s]",prefijo,lo->instancia_de,lo->nombre);
	    strcat(cadena_expr,cad_aux);
	    campaux = campos;
	    while ( campaux!=NULL) {
	      sprintf(cad_aux,".%s",campaux->nombre);
	      strcat(cadena_expr,cad_aux);
	      campaux = campaux->sig;
	    }
	    sprintf(cad_aux,")");
	    strcat(cadena_expr,cad_aux);
	    borrar(campos);
	  } 
	  else { /* se trata de una instancia */
	    //o de una variable
	    lo = busca_obj(lobj,exp->ident);
	    if (exp->tipooperando==definido) {
	      campos = NULL;
	      while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
		exp=exp->sig;
		exp=exp->sig;
		if (campos==NULL)
		  campos = cr_lista_id(campos,exp->ident);
		else
		  campos = insertar(campos, exp->ident);
	      }
	      switch (exp->tipooperando) {
	      case lista_enteros:
		sprintf(cad_aux,"head_lent(");
		strcat(cadena_expr,cad_aux);
		break;
	      case lista_reales:
		sprintf(cad_aux,"head_lreal(");
		strcat(cadena_expr,cad_aux);
		break;
	      }
              switch (lo->clvar) {
	      case instancia:
		sprintf(cad_aux,"&%si_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
		break;
	      case var:
		sprintf(cad_aux,"&%sv_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
		break;
	      default:
		fprintf(stderr, "ERROR del compilador Arlips.\n");
		fprintf(stderr, "Se esperaba un objeto lo con lo->clvar instancia o var,\n");
		fprintf(stderr, "Pero se recibi %d. (Ver en tds.h la enum clasobj\n", lo->clvar);
		fprintf(stderr, "Revisa en genera.c la funcin Genera_exp() en su caso headsim.\n");
		fprintf(stderr, "Revisa la insercin de elementos en la tds\n");
		break;
	      }  
	      strcat(cadena_expr,cad_aux);
	      campaux = campos;
	      while ( campaux!=NULL) {
		sprintf(cad_aux,".%s",campaux->nombre);
		strcat(cadena_expr,cad_aux);
		campaux = campaux->sig;
	      }
	      sprintf(cad_aux,")");
	      strcat(cadena_expr,cad_aux);
	      borrar(campos);
	    } 
	    else {
	      switch (exp->tipooperando) {
	      case lista_enteros:
		sprintf(cad_aux,"head_lent(");
		strcat(cadena_expr,cad_aux);
		break;
	      case lista_reales:
		sprintf(cad_aux,"head_lreal(");
		strcat(cadena_expr,cad_aux);
		break;
	      }
	      sprintf(cad_aux,"&%s%s)",prefijo, lo->nombre);
	      strcat(cadena_expr,cad_aux);
	    }
	  }
	  exp=exp->sig; /* pasamos el parentesis */
	  break;
	  
	case tailsim :  /* todavia no sabemos si es una lista */
                                    /* de enteros o reales */
	  exp=exp->sig;
	  exp=exp->sig; /* pasamos el parentesis */
	  if (exp->tipooperador==existe) { /* si es una variable patron */
	    exp=exp->sig;
	    lo = busca_obj(lobj,exp->ident);
	    campos = NULL;
	    while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
	      exp=exp->sig;  /* ahora obtenemos el campo: las patern var siempre lo tienen */
	      exp=exp->sig;
	      if (campos==NULL)
		campos = cr_lista_id(campos,exp->ident);
	      else
		campos = insertar(campos, exp->ident);
	    }
	    switch (exp->tipooperando) {
	    case lista_enteros:
	      sprintf(cad_aux,"tail_lent(");
	      strcat(cadena_expr,cad_aux);
	      break;
	    case lista_reales:
	      sprintf(cad_aux,"tail_lreal(");
	      strcat(cadena_expr,cad_aux);
	      break;
	    }
	    sprintf(cad_aux,"&%si_%s[%s]",prefijo,lo->instancia_de,lo->nombre);
	    strcat(cadena_expr,cad_aux);
	    campaux = campos;
	    while ( campaux!=NULL) {
	      sprintf(cad_aux,".%s",campaux->nombre);
	      strcat(cadena_expr,cad_aux);
	      campaux = campaux->sig;
	    }
	    sprintf(cad_aux,")");
	    strcat(cadena_expr,cad_aux);
	    borrar(campos);
	  } 
	  else { /* se trata de una instancia */
	    //o de unqa variable global
	    lo = busca_obj(lobj,exp->ident);
	    if (exp->tipooperando==definido) {
	      campos = NULL;
	      while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
		exp=exp->sig;
		exp=exp->sig;
		if (campos==NULL)
		  campos = cr_lista_id(campos,exp->ident);
		else
		  campos = insertar(campos, exp->ident);
	      }
	      switch (exp->tipooperando) {
	      case lista_enteros: 
		sprintf(cad_aux,"tail_lent(");
		strcat(cadena_expr,cad_aux);
		break;
	      case lista_reales:
		sprintf(cad_aux,"tail_lreal(");
		strcat(cadena_expr,cad_aux);
		break;
	      }
              switch (lo->clvar) {
	      case instancia:
		sprintf(cad_aux,"&%si_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
		break;
	      case var:
		sprintf(cad_aux,"&%sv_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
		break;
	      default:
		fprintf(stderr, "ERROR del compilador Arlips.\n");
		fprintf(stderr, "Se esperaba un objeto lo con lo->clvar instancia o var,\n");
		fprintf(stderr, "Pero se recibi %d. (Ver en tds.h la enum clasobj\n", lo->clvar);
		fprintf(stderr, "Revisa en genera.c la funcin Genera_exp() en su caso tailsim.\n");
		fprintf(stderr, "Revisa la insercin de elementos en la tds\n");
		break;
	      }  
	      strcat(cadena_expr,cad_aux);
	      campaux = campos;
	      while ( campaux!=NULL) {
		sprintf(cad_aux,".%s",campaux->nombre);
		strcat(cadena_expr,cad_aux);
		campaux = campaux->sig;
	      }
	      sprintf(cad_aux,")");
	      strcat(cadena_expr,cad_aux);
	      borrar(campos);
	    }
	    else {
	      switch (exp->tipooperando) {
	      case lista_enteros:
		sprintf(cad_aux,"tail_lent(");
		strcat(cadena_expr,cad_aux);
		break;
	      case lista_reales:
		sprintf(cad_aux,"tail_lreal(");
		strcat(cadena_expr,cad_aux);
		break;
	      }
	      sprintf(cad_aux,"&%s%s)",prefijo, lo->nombre);
	      strcat(cadena_expr,cad_aux);
	    }
	  }
	  exp=exp->sig; /* pasamos el parentesis */
	  break;
	  
	case popheadsim :  /* todavia no sabemos si es una lista */
                                    /* de enteros o reales */
	  exp=exp->sig;
	  exp=exp->sig; /* pasamos el parentesis */
	  if (exp->tipooperador==existe) { /* si es una variable patron */
	    exp=exp->sig;
	    lo = busca_obj(lobj,exp->ident);
	    campos = NULL;
	    while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
	      exp=exp->sig;  /* ahora obtenemos el campo: las patern var siempre lo tienen */
	      exp=exp->sig;
	      if (campos==NULL)
		campos = cr_lista_id(campos,exp->ident);
	      else
		campos = insertar(campos, exp->ident);
	    }
	    switch (exp->tipooperando) {
	    case lista_enteros:
	      sprintf(cad_aux,"pop_head_lent(");
	      strcat(cadena_expr,cad_aux);
	      break;
	    case lista_reales: sprintf(cad_aux,"pop_head_lreal(");
	      strcat(cadena_expr,cad_aux);
	      break;
	    }
	    sprintf(cad_aux,"&%si_%s[%s]",prefijo,lo->instancia_de,lo->nombre);
	    strcat(cadena_expr,cad_aux);
	    campaux = campos;
	    while ( campaux!=NULL) {
	      sprintf(cad_aux,".%s",campaux->nombre);
	      strcat(cadena_expr,cad_aux);
	      campaux = campaux->sig;
	    }
	    sprintf(cad_aux,")");
	    strcat(cadena_expr,cad_aux);
	    borrar(campos);
	  }
	  else { /* se trata de una instancia */
	    //o de una variable global
	    lo = busca_obj(lobj,exp->ident);
	    if (exp->tipooperando==definido) {
	      campos = NULL;
	      while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
		exp=exp->sig;
		exp=exp->sig;
		if (campos==NULL)
		  campos = cr_lista_id(campos,exp->ident);
		else
		  campos = insertar(campos, exp->ident);
	      }
	      switch (exp->tipooperando) {
	      case lista_enteros:
		sprintf(cad_aux,"pop_head_lent(");
		strcat(cadena_expr,cad_aux);
		break;
	      case lista_reales:
		sprintf(cad_aux,"pop_head_lreal(");
		strcat(cadena_expr,cad_aux);
		break;
	      }
              switch (lo->clvar) {
	      case instancia:
		sprintf(cad_aux,"&%si_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
		break;
	      case var:
		sprintf(cad_aux,"&%sv_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
		break;
	      default:
		fprintf(stderr, "ERROR del compilador Arlips.\n");
		fprintf(stderr, "Se esperaba un objeto lo con lo->clvar instancia o var,\n");
		fprintf(stderr, "Pero se recibi %d. (Ver en tds.h la enum clasobj\n", lo->clvar);
		fprintf(stderr, "Revisa en genera.c la funcin Genera_exp() en su caso popheadsim.\n");
		fprintf(stderr, "Revisa la insercin de elementos en la tds\n");
		break;
	      }  
	      strcat(cadena_expr,cad_aux);
	      campaux = campos;
	      while ( campaux!=NULL) {
		sprintf(cad_aux,".%s",campaux->nombre);
		strcat(cadena_expr,cad_aux);
		campaux = campaux->sig;
	      }
	      sprintf(cad_aux,")");
	      strcat(cadena_expr,cad_aux);
	      borrar(campos);
	    }
	    else {
	      switch (exp->tipooperando) {
	      case lista_enteros: 
		sprintf(cad_aux,"pop_head_lent(");
		strcat(cadena_expr,cad_aux);
		break;
	      case lista_reales:
		sprintf(cad_aux,"pop_head_lreal(");
		strcat(cadena_expr,cad_aux);
		break;
	      }
	      sprintf(cad_aux,"&%s%s)",prefijo, lo->nombre);
	      strcat(cadena_expr,cad_aux);
	    }
	  }
	  exp=exp->sig; /* pasamos el parentesis */
	  break;
	  
	case poptailsim :  /* todavia no sabemos si es una lista */
	  /* de enteros o reales */
	  exp=exp->sig;
	  exp=exp->sig; /* pasamos el parentesis */
	  if (exp->tipooperador==existe) { /* si es una variable patron */
	    exp=exp->sig;
	    lo = busca_obj(lobj,exp->ident);
	    campos = NULL;
	    while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
	      exp=exp->sig;  /* ahora obtenemos el campo: las patern var siempre lo tienen */
	      exp=exp->sig;
	      if (campos==NULL) 
		campos = cr_lista_id(campos,exp->ident);
	      else
		campos = insertar(campos, exp->ident);
	    }
	    switch (exp->tipooperando) {
	    case lista_enteros:
	      sprintf(cad_aux,"pop_tail_lent(");
	      strcat(cadena_expr,cad_aux);
	      break;
	    case lista_reales:
	      sprintf(cad_aux,"pop_tail_lreal(");
	      strcat(cadena_expr,cad_aux);
	      break;
	    }
	    sprintf(cad_aux,"&%si_%s[%s]",prefijo,lo->instancia_de,lo->nombre);
	    strcat(cadena_expr,cad_aux);
	    campaux = campos;
	    while ( campaux!=NULL) {
	      sprintf(cad_aux,".%s",campaux->nombre);
	      strcat(cadena_expr,cad_aux);
	      campaux = campaux->sig;
	    }
	    sprintf(cad_aux,")");
	    strcat(cadena_expr,cad_aux);
	    borrar(campos);
	  } 
	  else { /* se trata de una instancia */
	    //o de una variable global
	    lo = busca_obj(lobj,exp->ident);
	    if (exp->tipooperando==definido) {
	      campos = NULL;
	      while (exp->sig!=NULL && !exp->sig->esoperando && exp->sig->tipooperador==punto) {
		exp=exp->sig;
		exp=exp->sig;
		if (campos==NULL)
		  campos = cr_lista_id(campos,exp->ident);
		else
		  campos = insertar(campos, exp->ident);
	      }
	      switch (exp->tipooperando) {
	      case lista_enteros:
		sprintf(cad_aux,"pop_tail_lent(");
		strcat(cadena_expr,cad_aux);
		break;
	      case lista_reales:
		sprintf(cad_aux,"pop_tail_lreal(");
		strcat(cadena_expr,cad_aux);
		break;
	      }
	      switch (lo->clvar) {
	      case instancia:
		sprintf(cad_aux,"&%si_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
		break;
	      case var:
		sprintf(cad_aux,"&%sv_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
		break;
	      default:
		fprintf(stderr, "ERROR del compilador Arlips.\n");
		fprintf(stderr, "Se esperaba un objeto lo con lo->clvar instancia o var,\n");
		fprintf(stderr, "Pero se recibi %d. (Ver en tds.h la enum clasobj\n", lo->clvar);
		fprintf(stderr, "Revisa en genera.c la funcin Genera_exp() en su caso poptailsim.\n");
		fprintf(stderr, "Revisa la insercin de elementos en la tds\n");
		break;
	      }  
	      strcat(cadena_expr,cad_aux);
	      campaux = campos;
	      while ( campaux!=NULL) {
		sprintf(cad_aux,".%s",campaux->nombre);
		strcat(cadena_expr,cad_aux);
		campaux = campaux->sig;
	      }
	      sprintf(cad_aux,")");
	      strcat(cadena_expr,cad_aux);
	      borrar(campos);
	    }
	    else {
	      switch (exp->tipooperando) {
	      case lista_enteros:
		sprintf(cad_aux,"pop_tail_lent(");
		strcat(cadena_expr,cad_aux);
		break;
	      case lista_reales:
		sprintf(cad_aux,"pop_tail_lreal(");
		strcat(cadena_expr,cad_aux);
		break;
	      }
	      sprintf(cad_aux,"&%s%s)",prefijo, lo->nombre);
	      strcat(cadena_expr,cad_aux);
	    }
	  }
	  exp=exp->sig; /* pasamos el parentesis */
	  break;
	  
	}
      else {
	if (exp->claseoperando==constante) {
	  if (exp->tipooperando==real) {
	    sprintf(cad_aux,"%f",exp->valor.real);
	    strcat(cadena_expr,cad_aux);
	  }
	  else
	    if (exp->tipooperando==entero || exp->tipooperando==subrango) {
	      sprintf(cad_aux,"%d",exp->valor.entero);
	      strcat(cadena_expr,cad_aux);
	    } 
	    else
	      if (exp->tipooperando==boolean)
                           if (exp->tipooperador==truesim) {
                                     sprintf(cad_aux,"TRUE");
                                     strcat(cadena_expr,cad_aux);
                           } else {
                                     sprintf(cad_aux,"FALSE");
                                     strcat(cadena_expr,cad_aux);
                           }
                      } else if (exp->claseoperando==instancia && !exp->escampo) {
                         lo = busca_obj(lobj,exp->ident);
                         if (exp->tipooperando==definido) {
                            sprintf(cad_aux,"%si_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
                            strcat(cadena_expr,cad_aux);
                         } else {
                            sprintf(cad_aux,"%s%s", prefijo,lo->nombre);
                            strcat(cadena_expr,cad_aux);
                         }
                      } else if (exp->claseoperando==paramt && !exp->escampo) {
                         lo = busca_obj(lobj,exp->ident);
                         if (exp->tipooperando==definido) {
                              if (lo->por_ref) {
                                  strcat(cadena_expr,"(*");
                              }
                               sprintf(cad_aux,"%s%s",prefijo,exp->ident);
                               strcat(cadena_expr,cad_aux);
                              if (lo->por_ref) {
                                  strcat(cadena_expr,")");
                              }
                         } else {
                              if (lo->por_ref) {
                                  strcat(cadena_expr,"*");
                              }
                              sprintf(cad_aux,"%s%s", prefijo,lo->nombre);
                              strcat(cadena_expr,cad_aux);
                         }
                      } else if (exp->claseoperando==var && !exp->escampo) {
                         lo = busca_obj(lobj,exp->ident);
                         if (exp->tipooperando==definido) {
                            sprintf(cad_aux,"%sv_%s[%d]",prefijo,lo->instancia_de,lo->cod.num_instancia);
                            strcat(cadena_expr,cad_aux);
                         } else {
                            sprintf(cad_aux,"%s%s", prefijo,lo->nombre);
                            strcat(cadena_expr,cad_aux);
                         }
                      } 
	else {
	  sprintf(cad_aux,"%s",exp->ident);
	  strcat(cadena_expr,cad_aux);

	  //Debemos averiguar cuantas comas requiere la lista de parmetros de la funcin.
	  //Para ello la buscamos y recorremos la lista de parmetros en la declaracin.
	  if (exp->claseoperando == funcion) {
	    funcion_en_tds = busca_en_funciones(exp->ident);
	    if (funcion_en_tds && funcion_en_tds->param) {
	      for (i = strlen(funcion_en_tds->param); i > 0; i--) { 
		if (funcion_en_tds->param[i] == ',') { //Contamos las comas, para saber el n de param en la declaracin
		  comas_en_lista_param_funcion++;
		}
	      }
	      
	      //Como Genera_exp hace un break cada vez que se encuentra con una coma
	      //debemos encargarnos nosotros de la generacin de toda la lista de argumentos de la funcin
	      //Para ello usamos recursin, siendo el nmero de llamadas que se espera hacer
	      //igual al nmero de comas que se esperan segn la declaracin que se hizo 
	      //en la funcin externa.
	      
	      for (; comas_en_lista_param_funcion && exp && exp->sig; comas_en_lista_param_funcion--) {
		exp = Genera_exp(exp->sig, cad_aux, lobj);
		strcat(cadena_expr, cad_aux);
		//Ni tan siquiera comprobamos que la expresin devuelta sea una coma
		//porque el funcionamiento de esta "chapuza" depende de que de Genera_exp
		//solo se pueda salir porque la expresin sea una coma o porque ya se haya 
		//generado toda ella. Si hubiese algn otro motivo para salir de Genera_exp,
		//no podramos detectarlo, no habra quien lo pudiera manejar y 
		//desacompasaramos las comas,
		strcat(cadena_expr, ", ");
	      }
	      
	      //Tras esto, si aun quedan comas en el contador, es que la llamada se ha realizado
	      //con menor nmero de parmetros que los formales. Deba haberse mostrado un error.
	      //Lo malo es que no podemos controlar cuando se han escrito ms parmetros de los
	      //que tocan.
	      if (!exp) {
		fprintf(stderr, "WARNING: %s requires more arguments\n", funcion_en_tds->nombre_ext);
		//Atencin: Vas a tener una violacin de segmento
	      }
	    }
	  }

	}
      }
    if (exp->indice_array != NULL) {
      strcat(cadena_expr, "[");
      Genera_exp(exp->indice_array, cad_aux, lobj);
      strcat(cadena_expr, cad_aux);
      strcat(cadena_expr, "]");
    }
    exp_ant = exp;
    exp=exp->sig;
  }

  //Lo que sigue (a excepcin del return) solo se debe ejecutar cuando la expresin a generar va a ser 
  //la orden de impresin de todo un array. Solo en esos casos la ltima
  //expresin procesada es de tipooperando == array. Ntese que esto se 
  //podra separar en una funcin independiente, pero para ello la expresin debiera
  //ser recordada, puesto que si todo va bien, se devolver un NULL
  if (exp_ant->tipooperando == array && exp_ant->indice_array == NULL) {
    strcat(cadena_expr, ", ");    
    elemento = busca_en_tds(lo->nombre);
    //Si la instancia es de tipo simple, es un array directamente.
    if (elemento->tipo == array) {
      tbase_array = elemento->tbase_array;
      tam_array = elemento->tam_lista;
    }
    else {
      //pero si es de otro tipo, significa que es definido y tiene campo. 
      //Hay que buscar el campo pertinente.
      l_campos = busca_campo(lo->campo, elemento->lc);
      tbase_array = l_campos->tbase_array;
      tam_array = l_campos->tam_lista;
    }
    
    switch(tbase_array) {
    case entero:
      strcat(cadena_expr, "entero, ");
      break;
    case real:
      strcat(cadena_expr, "real, ");
      break;
    case boolean:
      strcat(cadena_expr, "boolean, ");
      break;
    }
    sprintf(cad_aux, "%d", tam_array);
    strcat(cadena_expr, cad_aux);
  }


  return(exp);
}

/************************ GENERA_LISTA_CONDS_VP *************************/
void Genera_lista_conds_vp(lista_reglas,lista_num_conds,entvp,reglas_inv)
    LISTA* lista_reglas;
    LIS_NUM *lista_num_conds;
    REGTDS * entvp;
    LISTA ** reglas_inv;
{
  REGREGLA * entr;
  LISTA_CONDIC * ent_cond;       /* entrada de condicion */
  LISTA *lvpeaux;
  int cont;
  LISTA * reglas_aux;
  LISTA_OBJ *lo2;
  int nvp_igualclase; /* numero de variables patron de la condicion */
                      /* con la misma clase */
  int num_cond_gen;   /* numero de condiciones generadas por variable patron */
  int ya_gen;         
 
 while (lista_reglas!=NULL) {
        entr=busca_en_reglas(lista_reglas->nombre);
        /* ahora se localiza la condicion */
        ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
        if (!entr->tiene_copias) {
          if (ent_cond->es_orig) {

              /* primero contamos el numero de parametros de la condicion */
              /* cuya clase es la misma que la vp modificada */
              lvpeaux = ent_cond->param;
              nvp_igualclase=0;
              while (lvpeaux!=NULL) {
                 lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                 if (lo2->cuantificador==existe) {
                    if (strcmp(lo2->instancia_de,entvp->instancia_de)==0) 
                           nvp_igualclase++;
                 }
                 lvpeaux=lvpeaux->sig;
              }

              num_cond_gen = 0;
              do {
                 fprintf(fich_gen,"\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
                 lvpeaux = ent_cond->param;
                 cont=0;
                 ya_gen = FALSE;
                 while (lvpeaux!=NULL) {
                    lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                    if (lo2->cuantificador==existe) {
                      if (cont>0) fprintf(fich_gen,",");
                      if (strcmp(lo2->instancia_de,entvp->instancia_de)==0)
                          {
			   if (ya_gen==num_cond_gen) 
                                fprintf(fich_gen,"%s",entvp->nombre);
                           else 
			        fprintf(fich_gen,"-1");
                           ya_gen++;
                         } else fprintf(fich_gen,"-1");
                      cont++;
                    }
                    lvpeaux=lvpeaux->sig;
                 }
                 fprintf(fich_gen,");\n");
                 num_cond_gen++;               
              } while (num_cond_gen<nvp_igualclase);
          }
          reglas_aux=insertar_sindup(*reglas_inv,lista_reglas->nombre);
          if (reglas_aux!=NULL) {
             *reglas_inv = reglas_aux;
             /* solo se inserta cuando no estaba previamente en la lista */
             fprintf(fich_gen,"\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
          }
        }
        lista_reglas=lista_reglas->sig;
        lista_num_conds = lista_num_conds->sig;
  }
}

/************************ GENERA_CONDS_VP *************************/
void Genera_conds_vp(entvp,campos,reglas_inv)
    REGTDS * entvp;
    EXPRESION *campos;
    LISTA ** reglas_inv;
{
  REGREGLA * entr;
  LISTA_CONDIC * ent_cond;       /* entrada de condicion */
  LISTA_CAMPOS *lcamp, *lcamp_aux;
  LISTA *lista_reglas;
  LIS_NUM *lista_num_conds;
  LISTA *lvpeaux;
  EXPRESION *exp;
  int cont;
  LISTA * reglas_aux;
  LISTA_OBJ *lo2;
  char cad_aux[255];
  char cad_aux2[500];
  char cad_aux3[255];
  int instref;
  int i;
  EHTDS *auxtds;

  /* ahora generamos codigo para que si se */
  /* modifica una instancia que es referenciada */
  /* en cualquier regla se llame a la condicion */
  /* de activacion de dicha regla */
  cad_aux[0]='\x0';
  cad_aux2[0]='\x0';
  cad_aux3[0]='\x0';
  instref = 0;
  sprintf(cad_aux,"\t\tswitch(%s) {\n", entvp->nombre);
  strcat(cad_aux2,cad_aux);
  for(i=0;i<HASH_SIZE;i++)
     if (tds[i]!=NULL) {
        auxtds=tds[i];
        while(auxtds!=NULL) {
             if (auxtds->etds.clvar==instancia && auxtds->etds.tipo==definido &&
                 strcmp(entvp->instancia_de,auxtds->etds.instancia_de)==0) {

                 /* ahora se van recorriendo los campos */
                 exp = campos;
                 lcamp_aux = auxtds->etds.lc;
                 while (exp!=NULL) {
                       exp=exp->sig;
                       lcamp = busca_campo(exp->ident, lcamp_aux);
                       lcamp_aux = lcamp->lc;
                       exp=exp->sig;
                 }

                 lista_reglas= lcamp->reglas;
                 lista_num_conds = lcamp->num_conds;
                 cad_aux3[0]='\x0';
                 sprintf(cad_aux,"\t\t  case %d:\n",auxtds->etds.cod.num_instancia);
                 strcat(cad_aux3,cad_aux);
                 while (lista_reglas!=NULL) {
                       entr=busca_en_reglas(lista_reglas->nombre);
                       ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
                       instref++;
                       if (!entr->tiene_copias) {
                        if (ent_cond->es_orig) {
                          lvpeaux = ent_cond->param;
                          cont=0;
                          sprintf(cad_aux,"\t\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
                          strcat(cad_aux3,cad_aux);
                          while (lvpeaux!=NULL) {
                                lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                                if (lo2->cuantificador==existe) {
                                   if (cont>0) {
                                      sprintf(cad_aux,",");
                                      strcat(cad_aux3,cad_aux);
                                   }
                                   if (strcmp(lo2->instancia_de,entvp->instancia_de)==0) {
                                      sprintf(cad_aux,"%s", entvp->nombre);
                                      strcat(cad_aux3,cad_aux);
                                   } else {
                                      sprintf(cad_aux,"-1");
                                      strcat(cad_aux3,cad_aux);
                                   }
                                   cont++;
                                }
                                lvpeaux=lvpeaux->sig;
                          }
                          sprintf(cad_aux,");\n");
                          strcat(cad_aux3,cad_aux);
                        }
                        reglas_aux=insertar_sindup(*reglas_inv,lista_reglas->nombre);
                        if (reglas_aux!=NULL) {
                                *reglas_inv = reglas_aux;
                                /* solo se aade esta instruccion de activacion */
                                /* de llamada a la parte izquierda, si anteriormente */
                                /* no habia sido utilizada para la misma asignacion */
                                sprintf(cad_aux,"\t\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
                                strcat(cad_aux3,cad_aux);
                        }
                       }
                       lista_reglas=lista_reglas->sig;
                       lista_num_conds = lista_num_conds->sig;
                 }
                 sprintf(cad_aux,"\t\t\tbreak;\n");
                 strcat(cad_aux3,cad_aux);
                 if (lcamp->reglas!=NULL) strcat(cad_aux2,cad_aux3);
             }
             auxtds=auxtds->sig;
        }
     }
     sprintf(cad_aux,"\t\t}\n");
     strcat(cad_aux2,cad_aux);
          if (instref>0) fprintf(fich_gen,"%s",cad_aux2);

}

/************************ GENERA_ACCION_ASIG_VP **********************/
void Genera_accion_asig_vp(//vp, campos,tip_izq,expr,lobj,reglas_inv,num_asig)
			   char * vp,
			   EXPRESION * campos,
			   char tip_izq,
			   EXPRESION * expr,
			   LISTA_OBJ *lobj,
			   LISTA ** reglas_inv,
			   int num_asig) {

  REGTDS * entvp;
  EXPRESION *exp;
  REGTIPO *entclase;
  LISTA_CAMPOS *lcamp, *lcamp_aux;
  REGREGLA * entr;
  LISTA_CONDIC * ent_cond;       /* entrada de condicion */
  LISTA *lvpeaux;
  int cont,cont2,cont3,cont4;
  LISTA * reglas_asig;           /* reglas que son generadas por la asignacion */
  LISTA * reglas_aux;
  LISTA_OBJ *lo,*lo2;
  char cad_aux[255];
  
  
  entvp=busca_en_tds(vp);
  /* se almacena el valor anterior */
  if (tip_izq==real)
    fprintf(fich_gen,"\told_value_real=");
  else
    fprintf(fich_gen,"\told_value_ent=");
  fprintf(fich_gen,"%si_%s[%s]",prefijo,entvp->instancia_de,vp);
  exp = campos;
  while (exp!=NULL) {
    exp=exp->sig;
    fprintf(fich_gen,".%s",exp->ident);
    if (exp->indice_array != NULL) {
      Genera_exp(exp->indice_array, cad_aux, lobj);
      fprintf(fich_gen, "[%s]", cad_aux);
    }
    exp=exp->sig;
  }
  fprintf(fich_gen,";\n");

  fprintf(fich_gen,"\t%si_%s[%s]",prefijo,entvp->instancia_de,vp);
  exp = campos;
  while (exp!=NULL) {
    exp=exp->sig;
    fprintf(fich_gen,".%s",exp->ident);
    if (exp->indice_array != NULL) {
      Genera_exp(exp->indice_array, cad_aux, lobj);
      fprintf(fich_gen, "[%s]", cad_aux);
    }
    exp=exp->sig;
  }
  fprintf(fich_gen,"=");

  Genera_exp(expr,cad_aux,lobj);
  fprintf(fich_gen,"%s;\n",cad_aux);

  fprintf(fich_gen,"\tcambio_asig[%d]=",num_asig);
  if (tip_izq==real)
    fprintf(fich_gen,"\told_value_real");
  else
    fprintf(fich_gen,"\told_value_ent");
  fprintf(fich_gen," != ");
  fprintf(fich_gen,"%si_%s[%s]",prefijo,entvp->instancia_de,vp);
  exp = campos;
  while (exp!=NULL) {
    exp=exp->sig;
    fprintf(fich_gen,".%s",exp->ident);
    if (exp->indice_array != NULL) {
      Genera_exp(exp->indice_array, cad_aux, lobj);
      fprintf(fich_gen, "[%s]", cad_aux);
    }
    exp=exp->sig;
  }
  fprintf(fich_gen,";\n");
  
  /* ahora llamamos a las condiciones afectadas por esta */
  /* accion */
  entclase=busca_en_tipos(entvp->instancia_de);

  /* ahora se van recorriendo los campos */
  exp = campos;
  lcamp_aux = entclase->lc;
  while (exp!=NULL) {
     exp=exp->sig;
     lcamp = busca_campo(exp->ident, lcamp_aux);
     lcamp_aux = lcamp->lc;
     exp=exp->sig;
  }

  fprintf(fich_gen,"\tif (cambio_asig[%d]) {\n",num_asig);

  reglas_asig = NULL;

  Genera_lista_conds_vp(lcamp->reglas,lcamp->num_conds,entvp,&reglas_asig);
  Genera_conds_vp(entvp,campos,&reglas_asig);

  /* ahora se actualizan las reglas cuya parte izquierda ha de llamarse */
  /* debido a las posibles modificaciones realizadas en esta accion */
  while (reglas_asig!=NULL)
  {
   reglas_aux = insertar_sindup(*reglas_inv,reglas_asig->nombre);
   if (reglas_aux!=NULL) *reglas_inv = reglas_aux;
   reglas_asig= reglas_asig->sig;
  }
  fprintf(fich_gen,"\t}\n");

}

/************************ GENERA_ACCION_ASIG_INST **********************/
void Genera_accion_asig_inst(
			     char * inst,
			     EXPRESION * campos,
			     char tip_izq,
			     EXPRESION * expr,
			     LISTA_OBJ *lobj,
			     LISTA ** reglas_inv,
			     int num_asig) {
  
  REGTDS * entins;
  EXPRESION *exp;
  REGTIPO *entclase;
  LISTA_CAMPOS *lcamp, *lcamp_aux;
  LISTA *lista_reglas;
  LIS_NUM *lista_num_conds;
  REGREGLA * entr;
  LISTA_CONDIC * ent_cond;       /* entrada de condicion */
  LISTA *lvpeaux;
  int cont,cont2,cont3,cont4;
  LISTA * reglas_aux;
  LISTA * reglas_asig;
  char cad_aux[255];
  int instref;
  int i;
  EHTDS *auxtds;
  LISTA_OBJ *lo,*lo2;
  int *param;
  int num_param;
  
  entins=busca_en_tds(inst);
  
  /* se almacena el valor anterior */
  if (tip_izq==real)
    fprintf(fich_gen,"\told_value_real=");
  else
    fprintf(fich_gen,"\told_value_ent=");
  
  if (entins->tipo!=definido) {
    fprintf(fich_gen,"%s%s",prefijo,inst);
    if (entins->tipo == array) {
      Genera_exp(campos, cad_aux, lobj);
      fprintf(fich_gen, "[%s]", cad_aux);
    }
  }
       
  else {
    fprintf(fich_gen,"%si_%s[%d]",prefijo,entins->instancia_de,entins->cod.num_instancia);
    exp = campos;
    if (exp->indice_array != NULL) {
      Genera_exp(exp->indice_array, cad_aux, lobj); 
      fprintf(fich_gen, "[%s]", cad_aux);
    }
    while (exp!=NULL) {
      exp=exp->sig;
      fprintf(fich_gen,".%s",exp->ident);
      if (exp->indice_array != NULL) {
	Genera_exp(exp->indice_array, cad_aux, lobj); 
	fprintf(fich_gen, "[%s]", cad_aux);
      }
      exp=exp->sig;
    }
  }
  fprintf(fich_gen,";\n");
  
  if (entins->tipo!=definido) {
    fprintf(fich_gen,"\t%s%s",prefijo,inst);
    if (entins->tipo == array) {
      Genera_exp(campos, cad_aux, lobj);
      fprintf(fich_gen, "[%s]", cad_aux);
    }
    fprintf(fich_gen, " = ");
  }  
  else {
    fprintf(fich_gen,"\t%si_%s[%d]",prefijo,entins->instancia_de,entins->cod.num_instancia);
    exp = campos;
    if (exp->indice_array != NULL) {
      Genera_exp(exp->indice_array, cad_aux, lobj); 
      fprintf(fich_gen, "[%s]", cad_aux);
    }    
    while (exp!=NULL) {
      exp=exp->sig;
      fprintf(fich_gen,".%s",exp->ident);
      if (exp->indice_array != NULL) {
	Genera_exp(exp->indice_array, cad_aux, lobj); 
	fprintf(fich_gen, "[%s]", cad_aux);
      }
      exp=exp->sig;
    }
    fprintf(fich_gen,"=");
  }
  Genera_exp(expr,cad_aux,lobj);
  fprintf(fich_gen,"%s;\n",cad_aux);

  fprintf(fich_gen,"\tcambio_asig[%d]=",num_asig);
  if (tip_izq==real)
    fprintf(fich_gen,"\told_value_real");
  else
    fprintf(fich_gen,"\told_value_ent");
  fprintf(fich_gen," != ");

  if (entins->tipo!=definido) {
    fprintf(fich_gen,"%s%s",prefijo,inst);
    if (entins->tipo == array) {
      Genera_exp(campos, cad_aux, lobj);
      fprintf(fich_gen, "[%s]", cad_aux);
    }  
  }
  else {
    fprintf(fich_gen,"%si_%s[%d]",prefijo,entins->instancia_de,entins->cod.num_instancia);
    exp = campos;
    if (exp->indice_array != NULL) {
      Genera_exp(exp->indice_array, cad_aux, lobj); 
      fprintf(fich_gen, "[%s]", cad_aux);
    }
    while (exp!=NULL) {
      exp=exp->sig;
      fprintf(fich_gen,".%s",exp->ident);
      if (exp->indice_array != NULL) {
	Genera_exp(exp->indice_array, cad_aux, lobj); 
	fprintf(fich_gen, "[%s]", cad_aux);
      }
      exp=exp->sig;
    }
  }
  fprintf(fich_gen,";\n");
  
  fprintf(fich_gen,"\tif (cambio_asig[%d]) {\n",num_asig);
  
  reglas_asig = NULL;
  /* ahora llamamos a los procedimientos que se */
  /* corresponden con la parte izq de la regla */
  if (entins->tipo==definido) {
    entclase=busca_en_tipos(entins->instancia_de);
    /* ahora se van recorriendo los campos */
    exp = campos;
    lcamp_aux = entclase->lc;
    while (exp!=NULL) {
          exp=exp->sig;
          lcamp = busca_campo(exp->ident, lcamp_aux);
          lcamp_aux = lcamp->lc;
          exp=exp->sig;
     }
     lista_reglas= lcamp->reglas;
     lista_num_conds = lcamp->num_conds;
  } else {
     lista_reglas= entins->reglas;
     lista_num_conds = entins->num_conds;
  }

  while (lista_reglas!=NULL) {
        entr=busca_en_reglas(lista_reglas->nombre);
        ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
        if (!entr->tiene_copias) {
         if (ent_cond->es_orig) {
           fprintf(fich_gen,"\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
           lvpeaux = ent_cond->param;
           cont=0;
           while (lvpeaux!=NULL) {
                 lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                 if (lo2->cuantificador==existe) {
                    if (cont>0) fprintf(fich_gen,",");
                    if (strcmp(lo2->instancia_de,entins->instancia_de)==0) fprintf(fich_gen,"%d",entins->cod.num_instancia);
                                                            else fprintf(fich_gen,"-1");
                    cont++;
                 }
                 lvpeaux=lvpeaux->sig;
           }
           fprintf(fich_gen,");\n");
         }
         reglas_aux=insertar_sindup(reglas_asig,lista_reglas->nombre);
         if (reglas_aux!=NULL) {
               reglas_asig = reglas_aux;
               /* se genera la activacion de este logico si antes no se habia */
               /* generado ya */
               fprintf(fich_gen,"\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
         }
        }
        lista_reglas=lista_reglas->sig;
        lista_num_conds = lista_num_conds->sig;
  }

  /* ahora se busca si hay alguna regla que */
  /* use en su condicion de activacion la instancia */
  /* que estamos modificando */
  if (entins->tipo==definido) {
     exp = campos;
     lcamp_aux = entins->lc;
     while (exp!=NULL) {
           exp=exp->sig;
           lcamp = busca_campo(exp->ident, lcamp_aux);
           lcamp_aux = lcamp->lc;
           exp=exp->sig;
     }
     lista_reglas= lcamp->reglas;
     lista_num_conds = lcamp->num_conds;
     while (lista_reglas!=NULL) {
       entr=busca_en_reglas(lista_reglas->nombre);
       ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
       if (!entr->tiene_copias) {
        if (ent_cond->es_orig) {
           fprintf(fich_gen,"\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
           lvpeaux = ent_cond->param;
           cont=0;
           while (lvpeaux!=NULL) {
               lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
               if (lo2->cuantificador==existe) {
                   if (cont>0) fprintf(fich_gen,",");
                   if (strcmp(lo2->instancia_de,entins->instancia_de)==0) fprintf(fich_gen,"%d",entins->cod.num_instancia);
                                                                         else fprintf(fich_gen,"-1");
                   cont++;
               }
               lvpeaux=lvpeaux->sig;
           }
           fprintf(fich_gen,");\n");
        }
        reglas_aux=insertar_sindup(reglas_asig,lista_reglas->nombre);
        if (reglas_aux!=NULL) {
                reglas_asig = reglas_aux;
                fprintf(fich_gen,"\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
        }
       }
       lista_reglas=lista_reglas->sig;
       lista_num_conds = lista_num_conds->sig;
     }
  }
  while (reglas_asig!=NULL)
  {
      reglas_aux = insertar_sindup(*reglas_inv,reglas_asig->nombre);
      if (reglas_aux!=NULL) *reglas_inv = reglas_aux;
      reglas_asig= reglas_asig->sig;
  }
  fprintf(fich_gen,"\t}\n");
}


void Genera_accion_addlist(
			   char tipo_accion,
			   char * inst,  //Deber ser la lista que modificamos
			   EXPRESION * campos,
			   char tip_izq,  //Debera ser el tipo de la lista (lista_enteros o lista_reales)
			   EXPRESION * expr, //Debera ser el elemento o expresin a aadir
			   LISTA_OBJ *lobj,
			   LISTA ** reglas_inv,
			   int num_asig) {

  REGTDS * entins;
  EXPRESION *exp;
  REGTIPO *entclase;
  LISTA_CAMPOS *lcamp, *lcamp_aux;
  LISTA *lista_reglas;
  LIS_NUM *lista_num_conds;
  REGREGLA * entr;
  LISTA_CONDIC * ent_cond;       /* entrada de condicion */
  LISTA *lvpeaux;
  int cont,cont2,cont3,cont4;
  LISTA * reglas_aux;
  LISTA * reglas_asig;
  char cad_aux[255];
  int instref;
  int i;
  EHTDS *auxtds;
  LISTA_OBJ *lo,*lo2;
  int *param;
  int num_param;
  
  entins=busca_en_tds(inst);


  /* se almacena el valor anterior */
  //...

  //Imprimimos la llamada a la funcin...
  if (tipo_accion == oplista_addhead)
    fprintf(fich_gen, "\tadd_head_");
  else
    fprintf(fich_gen, "\tadd_tail_");


  switch(tip_izq) {
  case lista_enteros:
    fprintf(fich_gen, "lent(");
    break;
  case lista_reales:
    fprintf(fich_gen, "lreal(");
    break;
  default:
    fprintf(stderr, "ERROR en el compilador. La comprobacin de tipos de lista ha fallado\n");
    fprintf(stderr, "La funcin Genera_accion_addtail de genera.c recibi un tipo no tratado\n");
    fprintf(stderr, "cuando estaba esperando una lista. El tipo fue %d\n", tip_izq);
    fprintf(stderr, "Revisa genera.c -> Genera_accion_addtail(...)\n");
    fprintf(stderr, "Revisa sinsem.c -> accion(...)\n");
    break;  
  } 
  
  Genera_exp(expr,cad_aux,lobj);
  fprintf(fich_gen,"%s, &",cad_aux); 
  
  if (entins->tipo!=definido)
    fprintf(fich_gen,"%s%s",prefijo,inst);
  else {
    switch (entins->clvar) {
    case paternvar:
      fprintf(fich_gen,"%si_%s[%s]", prefijo, entins->instancia_de, inst);
      break;
    case instancia:    
      fprintf(fich_gen,"%si_%s[%d]", prefijo, entins->instancia_de, entins->cod.num_instancia);
      break;
    case var:
      fprintf(fich_gen,"%sv_%s[%d]", prefijo, entins->instancia_de, entins->cod.num_instancia);
      break;
    }
    exp = campos;
    while (exp!=NULL) {
      exp=exp->sig;
      fprintf(fich_gen,".%s",exp->ident);
      exp=exp->sig;
    }
  }
  
  fprintf(fich_gen, ");\n");

 
  //Ya se ha impreso la llamada a la funcin que aade el elemento a la lista\n
  //Se contina procesando las condiciones de reglas afectadas...

  reglas_asig = NULL;
  /* ahora llamamos a los procedimientos que se */
  /* corresponden con la parte izq de la regla */
  //CGM: Qu pasar con las variables globales?
  if (entins->tipo==definido) {
    entclase=busca_en_tipos(entins->instancia_de);
    /* ahora se van recorriendo los campos */
    exp = campos;
    lcamp_aux = entclase->lc;
    while (exp!=NULL) {
      exp=exp->sig;
      lcamp = busca_campo(exp->ident, lcamp_aux);
      lcamp_aux = lcamp->lc;
      exp=exp->sig;
    }
    lista_reglas= lcamp->reglas;
    lista_num_conds = lcamp->num_conds;
  }
  else {
    lista_reglas= entins->reglas;
    lista_num_conds = entins->num_conds;
  }


  while (lista_reglas!=NULL) {
    entr=busca_en_reglas(lista_reglas->nombre);
    ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
    if (!entr->tiene_copias) {
      if (ent_cond->es_orig) {
	fprintf(fich_gen,"\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
	lvpeaux = ent_cond->param;
	cont=0;
	while (lvpeaux!=NULL) {
	  lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
	  if (lo2->cuantificador==existe) {
	    if (cont>0)
	      fprintf(fich_gen,",");
	    if (strcmp(lo2->instancia_de,entins->instancia_de)==0)
	      //fprintf(fich_gen,"%d",entins->cod.num_instancia);
	      fprintf(fich_gen,"%s",entins->nombre);
	    else fprintf(fich_gen,"-1");
	    cont++;
	  }
	  lvpeaux=lvpeaux->sig;
	}
	fprintf(fich_gen,");\n");
      }
      
      reglas_aux=insertar_sindup(reglas_asig,lista_reglas->nombre);
      if (reglas_aux!=NULL) {
	reglas_asig = reglas_aux;
	/* se genera la activacion de este logico si antes no se habia */
	/* generado ya */
	fprintf(fich_gen,"\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
      }
    }
    lista_reglas=lista_reglas->sig;
    lista_num_conds = lista_num_conds->sig;
  }

  /* ahora se busca si hay alguna regla que */
  /* use en su condicion de activacion la instancia */
  /* que estamos modificando */

  if (entins->tipo==definido) {
    exp = campos;
    lcamp_aux = entins->lc;
    while (exp!=NULL) {
      exp=exp->sig;
      lcamp = busca_campo(exp->ident, lcamp_aux);
      lcamp_aux = lcamp->lc;
      exp=exp->sig;
    }
    lista_reglas= lcamp->reglas;
    lista_num_conds = lcamp->num_conds;

    while (lista_reglas!=NULL) {
      entr=busca_en_reglas(lista_reglas->nombre);
      ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
      
      if (!entr->tiene_copias) {
        if (ent_cond->es_orig) {
	  fprintf(fich_gen,"\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
	  lvpeaux = ent_cond->param;
	  cont=0;
	  while (lvpeaux!=NULL) {
	    lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
	    if (lo2->cuantificador==existe) {
	      if (cont>0)
		fprintf(fich_gen,",");
	      if (strcmp(lo2->instancia_de,entins->instancia_de)==0)
		fprintf(fich_gen,"%d",entins->cod.num_instancia);
	      else
		fprintf(fich_gen,"-1");
	      cont++;
	    }
	    lvpeaux=lvpeaux->sig;
	  }
	  fprintf(fich_gen,");\n");
        }
        reglas_aux=insertar_sindup(reglas_asig,lista_reglas->nombre);
        if (reglas_aux!=NULL) {
	  reglas_asig = reglas_aux;
	  fprintf(fich_gen,"\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
        }
      }
      lista_reglas=lista_reglas->sig;
      lista_num_conds = lista_num_conds->sig;
    }
  }

  while (reglas_asig!=NULL) {
    reglas_aux = insertar_sindup(*reglas_inv,reglas_asig->nombre);
    if (reglas_aux!=NULL)
      *reglas_inv = reglas_aux;
    reglas_asig= reglas_asig->sig;
  }
}//Genera_accion_addlist


/*********************** GENERA_ACCION_LLAM_FUNC *************************/
void Genera_accion_llam_func(funcion,argm,lobj,reglas_inv)
   char * funcion;
   EXPRESION * argm;
   LISTA_OBJ * lobj;
   LISTA ** reglas_inv;
{
  REGFUNC *func;
  EXPRESION *exp, *expaux, *campos_arg;
  MODIFICACIONES *mod_aux;
  LISTA_PARAM *lparam, *lp;
  LISTA_CAMPOS *lcamp, *lcamp_aux;
  REGTDS * entvp;
  REGTIPO *entclase;
  char cad_aux[255];
  int num_ent;
  int num_real;
  int i2;
  LISTA *lista_reglas;
  LIS_NUM *lista_num_conds;
  REGREGLA * entr;
  LISTA_CONDIC * ent_cond;       /* entrada de condicion */
  LISTA *lvpeaux;
  int cont,cont2,cont3,cont4;
  LISTA * reglas_aux;
  LISTA * reglas_asig;
  int instref;
  int i;
  EHTDS *auxtds;
  LISTA_OBJ *lo,*lo2;
  int *param;
  int num_param;


/* FALTA: considerar si los cambios producidos por las acciones de la */
/* funcion cambian realmente la BH */

   func = busca_en_funciones(funcion);

   if (func->tipo==interna) {
    /* primero se almacenan los valores que se van a modificar */
        mod_aux = func->modif;
        num_ent = 0;
        num_real = 0;
        while (mod_aux!=NULL) {
           if (mod_aux->clase_obj==instancia) {
              switch (mod_aux->tip_izq) {
	       case boolean:
               case subrango:
                case entero: fprintf(fich_gen,"\told_val_ent[%d]=",num_ent);
                              num_ent++;
                              break;
                 case real:  fprintf(fich_gen,"\told_val_real[%d]=",num_real);
                             num_real++;
                             break;
              }
              entvp=busca_en_tds(mod_aux->nombre);
              if (entvp->tipo==definido) {
                 fprintf(fich_gen,"%si_%s[%d]",prefijo,entvp->instancia_de,entvp->cod.num_instancia);
                 exp = mod_aux->campos;
                 while (exp!=NULL) {
                       exp=exp->sig;
                       fprintf(fich_gen,".%s",exp->ident);
                       exp=exp->sig;
                 }
              } else fprintf(fich_gen,"%s%s",prefijo,mod_aux->nombre);
              fprintf(fich_gen,";\n");
           } else if (mod_aux->clase_obj==paramt) {
                  /* si es un parametro por ref, depende de si */
                  /* lo pasado en los argumentos de la funcion */
                  /* es o no un objeto de la Base de Hechos */
                  /* se busca la posicion del parametro */
                  lparam = func->lp;
                  i2=0;
                  while(strcmp(lparam->nombre,mod_aux->nombre)) {
                          lparam=lparam->sig;
                          i2++;
                  }
                  expaux=busca_expresion(argm,i2);
                  /* si hay primero un operador existe porque es una patternvar, se pasa */
                  if (!expaux->esoperando && expaux->tipooperador==existe) expaux=expaux->sig;

                  entvp=busca_en_tds(expaux->ident);
                  if ((entvp->clvar == instancia) ||
                      (entvp->clvar == paternvar)) {
                      /* si es un objeto de la BH se utiliza para almacenar el valor anterior */
                      /* FALTA: y si tiene campos, y  ademas en la funcion es utilizado con campos */
                      switch (mod_aux->tip_izq) {
		            case boolean:
		           case subrango:
                             case entero: fprintf(fich_gen,"\told_val_ent[%d]=",num_ent);
                                          num_ent++;
                                          break;
                             case real:   fprintf(fich_gen,"\told_val_real[%d]=",num_real);
                                          num_real++;
                                          break;
                      }
                      if (entvp->tipo==definido) {
                         fprintf(fich_gen,"%si_%s[%d]",prefijo,entvp->instancia_de,entvp->cod.num_instancia);
                         /* primero se escriben los campos del argumento */
                         exp = expaux->sig;
                         while (exp!=NULL) {
                               exp=exp->sig;
                               fprintf(fich_gen,".%s",exp->ident);
                               exp=exp->sig;
                         }
                         /* ahora los campos de la asignacion */
                         exp = mod_aux->campos;
                         while (exp!=NULL) {
                               exp=exp->sig;
                               fprintf(fich_gen,".%s",exp->ident);
                               exp=exp->sig;
                         }
                      } else fprintf(fich_gen,"%s%s",prefijo,expaux->ident);
                      fprintf(fich_gen,";\n");
                  }
           }
           mod_aux=mod_aux->sig;
        }
   }

   /* Se escribe la funcion con sus argumentos */
   if (func->tipo==interna) fprintf(fich_gen,"\t%s%s(",prefijo,funcion);
                       else fprintf(fich_gen,"\t%s(",func->nombre_ext);

   lp = func->lp;
   exp = argm;
   while (exp!=NULL) {
         if (!exp->esoperando && exp->tipooperador==coma) {
            fprintf(fich_gen,", ");
            exp=exp->sig;
         }
         exp=Genera_exp(exp,cad_aux,lobj);
         if (func->tipo==interna && lp->por_ref) fprintf(fich_gen,"&");
         fprintf(fich_gen,"%s",cad_aux);
         if (func->tipo==interna) lp = lp->sig;
   }
   fprintf(fich_gen,");\n");

   /* ahora se llama a las condiciones afectadas */
   /* por aquellas acciones que dependen de los parametros */
   /* (las asignaciones a parametros) */
   if (func->tipo==interna) {

    /* primero se comprueban los valores que se han modificado */
     mod_aux = func->modif;
     num_ent = 0;
     num_real = 0;
        while (mod_aux!=NULL) {
           if (mod_aux->clase_obj==instancia) {
              switch (mod_aux->tip_izq) {
	        case boolean:
	       case subrango:
                 case entero: fprintf(fich_gen,"\tif (old_val_ent[%d]!=",num_ent);
                              num_ent++;
                              break;
                 case real:  fprintf(fich_gen,"\tif (old_val_real[%d]!=",num_real);
                             num_real++;
                             break;
              }
              entvp=busca_en_tds(mod_aux->nombre);
              if (entvp->tipo==definido) {
                 fprintf(fich_gen,"%si_%s[%d]",prefijo,entvp->instancia_de,entvp->cod.num_instancia);
                 exp = mod_aux->campos;
                 while (exp!=NULL) {
                       exp=exp->sig;
                       fprintf(fich_gen,".%s",exp->ident);
                       exp=exp->sig;
                 }
              } else fprintf(fich_gen,"%s%s",prefijo,mod_aux->nombre);
              fprintf(fich_gen,") {\n");

              reglas_asig = NULL;
              /* ahora llamamos a los procedimientos que se */
              /* corresponden con la parte izq de la regla */
              if (entvp->tipo==definido) {
                 /* se procesan ahora las condiciones de la clase */
                 entclase = busca_en_tipos(entvp->instancia_de);
                 /* ahora se van recorriendo los campos */
                 exp = mod_aux->campos;
                 lcamp_aux = entclase->lc;
                 while (exp!=NULL) {
                      exp=exp->sig;
                      lcamp = busca_campo(exp->ident, lcamp_aux);
                      lcamp_aux = lcamp->lc;
                      exp=exp->sig;
                 }
                 lista_reglas= lcamp->reglas;
                 lista_num_conds = lcamp->num_conds;
              } else {
                lista_reglas= entvp->reglas;
                lista_num_conds = entvp->num_conds;
              }

              while (lista_reglas!=NULL) {
               entr=busca_en_reglas(lista_reglas->nombre);
               ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
               if (!entr->tiene_copias) {
                 if (ent_cond->es_orig) {
                    fprintf(fich_gen,"\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
                    lvpeaux = ent_cond->param;
                    cont=0;
                    while (lvpeaux!=NULL) {
                       lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                       if (lo2->cuantificador==existe) {
                         if (cont>0) fprintf(fich_gen,",");
                         if (strcmp(lo2->instancia_de,entvp->instancia_de)==0) fprintf(fich_gen,"%d",entvp->cod.num_instancia);
                                                            else fprintf(fich_gen,"-1");
                         cont++;
                       }
                       lvpeaux=lvpeaux->sig;
                    }
                    fprintf(fich_gen,");\n");
                 }
                 reglas_aux=insertar_sindup(reglas_asig,lista_reglas->nombre);
                 if (reglas_aux!=NULL) {
                    reglas_asig = reglas_aux;
                    /* se genera la activacion de este logico si antes no se habia */
                    /* generado ya */
                    fprintf(fich_gen,"\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
                 }
               }
               lista_reglas=lista_reglas->sig;
               lista_num_conds = lista_num_conds->sig;
              }

              /* ahora se busca si hay alguna regla que */
              /* use en su condicion de activacion la instancia */
              /* que estamos modificando */
              if (entvp->tipo==definido) {
                 /* ahora se van recorriendo los campos */
                 exp = mod_aux->campos;
                 lcamp_aux = entvp->lc;
                 while (exp!=NULL) {
                      exp=exp->sig;
                      lcamp = busca_campo(exp->ident, lcamp_aux);
                      lcamp_aux = lcamp->lc;
                      exp=exp->sig;
                 }
                 lista_reglas= lcamp->reglas;
                 lista_num_conds = lcamp->num_conds;
                 while (lista_reglas!=NULL) {
                    entr=busca_en_reglas(lista_reglas->nombre);
                    ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
                    if (!entr->tiene_copias) {
                       if (ent_cond->es_orig) {
                          fprintf(fich_gen,"\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
                          lvpeaux = ent_cond->param;
                          cont=0;
                          while (lvpeaux!=NULL) {
                             lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                             if (lo2->cuantificador==existe) {
                                if (cont>0) fprintf(fich_gen,",");
                                if (strcmp(lo2->instancia_de,entvp->instancia_de)==0) fprintf(fich_gen,"%d",entvp->cod.num_instancia);
                                                                         else fprintf(fich_gen,"-1");
                                cont++;
                             }
                             lvpeaux=lvpeaux->sig;
                          }
                          fprintf(fich_gen,");\n");
                       }
                       reglas_aux=insertar_sindup(reglas_asig,lista_reglas->nombre);
                       if (reglas_aux!=NULL) {
                          reglas_asig = reglas_aux;
                          fprintf(fich_gen,"\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
                       }
                    }
                    lista_reglas=lista_reglas->sig;
                    lista_num_conds = lista_num_conds->sig;
                 }
              }
              while (reglas_asig!=NULL)
              {
                   reglas_aux = insertar_sindup(*reglas_inv,reglas_asig->nombre);
                   if (reglas_aux!=NULL) *reglas_inv = reglas_aux;
                   reglas_asig= reglas_asig->sig;
              }
              fprintf(fich_gen,"\t}\n");
           } else if (mod_aux->clase_obj==paramt) {
                  /* si es un parametro por ref, depende de si */
                  /* lo pasado en los argumentos de la funcion */
                  /* es o no un objeto de la Base de Hechos */
                  /* se busca la posicion del parametro */
                  lparam = func->lp;
                  i2=0;
                  while(strcmp(lparam->nombre,mod_aux->nombre)) {
                          lparam=lparam->sig;
                          i2++;
                  }
                  expaux=busca_expresion(argm,i2);
                  /* si hay primero un operador existe porque es una patternvar, se pasa */
                  if (!expaux->esoperando && expaux->tipooperador==existe) expaux=expaux->sig;
                  campos_arg = expaux->sig;

                  entvp=busca_en_tds(expaux->ident);
                  if ((entvp->clvar == instancia) ||
                      (entvp->clvar == paternvar)) {
                      /* si es un objeto de la BH se utiliza para almacenar el valor anterior */
                      /* FALTA: y si tiene campos, y  ademas en la funcion es utilizado con campos */
                      switch (mod_aux->tip_izq) {

		            case boolean:
                           case subrango:
                             case entero: fprintf(fich_gen,"\tif (old_val_ent[%d]!=",num_ent);
                                          num_ent++;
                                          break;
                             case real:   fprintf(fich_gen,"\tif (old_val_real[%d]!=",num_real);
                                          num_real++;
                                          break;
                      }
                      if (entvp->tipo==definido) {
                         if (entvp->clvar==instancia) fprintf(fich_gen,"%si_%s[%d]",prefijo,entvp->instancia_de,entvp->cod.num_instancia);
                                 else fprintf(fich_gen,"%si_%s[%s]",prefijo,entvp->instancia_de,entvp->nombre);
                         /* primero se escriben los campos del argumento */
                         exp = campos_arg;
                         while (exp!=NULL) {
                               exp=exp->sig;
                               fprintf(fich_gen,".%s",exp->ident);
                               exp=exp->sig;
                         }
                         /* ahora los campos de la asignacion */
                         exp = mod_aux->campos;
                         while (exp!=NULL) {
                               exp=exp->sig;
                               fprintf(fich_gen,".%s",exp->ident);
                               exp=exp->sig;
                         }
                      } else fprintf(fich_gen,"%s%s",prefijo,expaux->ident);
                      fprintf(fich_gen,") {\n");

                      if (entvp->clvar == instancia) {
              reglas_asig = NULL;
              /* ahora llamamos a los procedimientos que se */
              /* corresponden con la parte izq de la regla */
              if (entvp->tipo==definido) {
                 /* se procesan ahora las condiciones de la clase */
                 entclase = busca_en_tipos(entvp->instancia_de);
                 /* ahora se van recorriendo los campos */
                 lcamp_aux = entclase->lc;
                 expaux=campos_arg;
                 while (expaux!=NULL && (expaux->esoperando || expaux->tipooperador!=coma) ) {  /* primero se recorren los campos del */
                        /* parametro actual */
                        expaux=expaux->sig;
                        lcamp = busca_campo(expaux->ident, lcamp_aux);
                        lcamp_aux = lcamp->lc;
                        expaux=expaux->sig;
                 }
                 /* ahora se recorren los campos de la asignacion al parametro */
                 expaux = mod_aux->campos;
                 while (expaux!=NULL) {
                        expaux=expaux->sig;
                        lcamp = busca_campo(expaux->ident, lcamp_aux);
                        lcamp_aux = lcamp->lc;
                        expaux=expaux->sig;
                 }
                 lista_reglas= lcamp->reglas;
                 lista_num_conds = lcamp->num_conds;
              } else {
                lista_reglas= entvp->reglas;
                lista_num_conds = entvp->num_conds;
              }

              while (lista_reglas!=NULL) {
               entr=busca_en_reglas(lista_reglas->nombre);
               ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
               if (!entr->tiene_copias) {
                 if (ent_cond->es_orig) {
                    fprintf(fich_gen,"\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
                    lvpeaux = ent_cond->param;
                    cont=0;
                    while (lvpeaux!=NULL) {
                       lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                       if (lo2->cuantificador==existe) {
                         if (cont>0) fprintf(fich_gen,",");
                         if (strcmp(lo2->instancia_de,entvp->instancia_de)==0) fprintf(fich_gen,"%d",entvp->cod.num_instancia);
                                                            else fprintf(fich_gen,"-1");
                         cont++;
                       }
                       lvpeaux=lvpeaux->sig;
                    }
                    fprintf(fich_gen,");\n");
                 }
                 reglas_aux=insertar_sindup(reglas_asig,lista_reglas->nombre);
                 if (reglas_aux!=NULL) {
                    reglas_asig = reglas_aux;
                    /* se genera la activacion de este logico si antes no se habia */
                    /* generado ya */
                    fprintf(fich_gen,"\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
                 }
               }
               lista_reglas=lista_reglas->sig;
               lista_num_conds = lista_num_conds->sig;
              }

              /* ahora se busca si hay alguna regla que */
              /* use en su condicion de activacion la instancia */
              /* que estamos modificando */
              if (entvp->tipo==definido) {
                 /* ahora se van recorriendo los campos */
                 lcamp_aux = entvp->lc;
                 expaux=campos_arg;
                 while (expaux!=NULL && (expaux->esoperando || expaux->tipooperador!=coma) ) {  /* primero se recorren los campos del */
                        /* parametro actual */
                        expaux=expaux->sig;
                        lcamp = busca_campo(expaux->ident, lcamp_aux);
                        lcamp_aux = lcamp->lc;
                        expaux=expaux->sig;
                 }
                 /* ahora se recorren los campos de la asignacion al parametro */
                 expaux = mod_aux->campos;
                 while (expaux!=NULL) {
                        expaux=expaux->sig;
                        lcamp = busca_campo(expaux->ident, lcamp_aux);
                        lcamp_aux = lcamp->lc;
                        expaux=expaux->sig;
                 }
                 lista_reglas= lcamp->reglas;
                 lista_num_conds = lcamp->num_conds;
                 while (lista_reglas!=NULL) {
                      entr=busca_en_reglas(lista_reglas->nombre);
                      ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
                      if (!entr->tiene_copias) {
                         if (ent_cond->es_orig) {
                            fprintf(fich_gen,"\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
                            lvpeaux = ent_cond->param;
                            cont=0;
                            while (lvpeaux!=NULL) {
                               lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                               if (lo2->cuantificador==existe) {
                                  if (cont>0) fprintf(fich_gen,",");
                                  if (strcmp(lo2->instancia_de,entvp->instancia_de)==0) fprintf(fich_gen,"%d",entvp->cod.num_instancia);
                                                                           else fprintf(fich_gen,"-1");
                                  cont++;
                               }
                               lvpeaux=lvpeaux->sig;
                            }
                            fprintf(fich_gen,");\n");
                         }
                         reglas_aux=insertar_sindup(reglas_asig,lista_reglas->nombre);
                         if (reglas_aux!=NULL) {
                            reglas_asig = reglas_aux;
                            fprintf(fich_gen,"\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
                         }
                      }
                      lista_reglas=lista_reglas->sig;
                      lista_num_conds = lista_num_conds->sig;
                 }
               }
               while (reglas_asig!=NULL)
               {
                  reglas_aux = insertar_sindup(*reglas_inv,reglas_asig->nombre);
                  if (reglas_aux!=NULL) *reglas_inv = reglas_aux;
                  reglas_asig= reglas_asig->sig;
               }
              } else { /* es una pattern var */
                   /* se procesan ahora las condiciones de la clase */
                 entclase = busca_en_tipos(entvp->instancia_de);
                 /* ahora se van recorriendo los campos */
                 lcamp_aux = entclase->lc;
                 expaux=campos_arg;
                 while (expaux!=NULL && (expaux->esoperando || expaux->tipooperador!=coma) ) {  /* primero se recorren los campos del */
                        /* parametro actual */
                        expaux=expaux->sig;
                        lcamp = busca_campo(expaux->ident, lcamp_aux);
                        lcamp_aux = lcamp->lc;
                        expaux=expaux->sig;
                 }
                 /* ahora se recorren los campos de la asignacion al parametro */
                 expaux = mod_aux->campos;
                 while (expaux!=NULL) {
                        expaux=expaux->sig;
                        lcamp = busca_campo(expaux->ident, lcamp_aux);
                        lcamp_aux = lcamp->lc;
                        expaux=expaux->sig;
                 }
                 Genera_lista_conds_vp(lcamp->reglas,lcamp->num_conds,entvp,reglas_inv);
                 Genera_conds_vp(entvp,mod_aux->campos,reglas_inv);

              }
              fprintf(fich_gen,"\t}\n");

              }
           }
           mod_aux=mod_aux->sig;
     }
   }
}

/************************ GENERA_CABECERA_ARTIS **************************/
void Genera_cabecera_artis() {
  fprintf(fich_gen, "#ifdef MODULE\n");
  fprintf(fich_gen, "  #include <rtl_core.h>\n");
  fprintf(fich_gen, "  //Se desactiva la gestin de ficheros para los print de estructuras\n");
  fprintf(fich_gen, "  #define FILE void\n");
  fprintf(fich_gen, "  #define stdout NULL\n");
  if (log_) {
    fprintf(fich_gen, "  //Se desactivan las funciones de log para tiempo real estricto\n");
    fprintf(fich_gen, "  #define %slog_inicia_log()\n", prefijo);
    fprintf(fich_gen, "  #define %slog_end_log(r)\n", prefijo);
    fprintf(fich_gen, "  #define %slog_reset()\n", prefijo);
    fprintf(fich_gen, "  #define %slog_end_reset()\n", prefijo);
    fprintf(fich_gen, "  #define %slog_regla_mas_prio(r)\n", prefijo);
    fprintf(fich_gen, "  #define %slog_end_fire()\n", prefijo);
    fprintf(fich_gen, "  #define %slog_end_actions()\n", prefijo);
    fprintf(fich_gen, "  #define %slog_insregla(r)\n");
    fprintf(fich_gen, "  #define %slog_elimregla(r)\n");
    fprintf(fich_gen, "  #define %slog_facts()\n", prefijo);
  }
  fprintf(fich_gen, "#else\n");
  fprintf(fich_gen, "  #include <stdio.h>\n");
  if (log_) {
    fprintf(fich_gen, "  #include \"%s\"\n", fich_generados[cab_log]);
  }  
  fprintf(fich_gen, "#endif\n\n");    
}

/*************************** GENERA_INCLUDES *****************************/

void Genera_includes() {

  fprintf(fich_gen, "#include \"%s\"\n", fich_generados[fich_bh]);
  fprintf(fich_gen, "#include \"tipos_.h\"\n");
  fprintf(fich_gen, "#include \"%s\"\n",fich_generados[cab_agenda]); 

  if (artis_mode) {
    fprintf(fich_gen, "#include <userclass.h>\n"); 
    fprintf(fich_gen, "#include <artis/include/kdm.h>\n\n\n");
    Genera_cabecera_artis();
  }
  else {
    fprintf(fich_gen, "#include <stdio.h>\n");
    fprintf(fich_gen, "#include <stdlib.h>\n");    
    if (log_) { //Si se pide log con artis_mode, se encarga Genera_cabecera_artis()
      fprintf(fich_gen, "#include \"%s\"\n", fich_generados[cab_log]);
    }
  }
  fprintf(fich_gen,"\n");  
}


/*************************** GENERA_CONSTANTES *****************************/

void Genera_constantes()
{
  EHTDS *aux;
  int i;

  for(i=0;i<HASH_SIZE;i++)
     if(tds[i]!=NULL) {
         aux=tds[i];
         while(aux!=NULL) {
             if (aux->etds.clvar==constante) {
                fprintf(fich_gen,"#define %s%s",prefijo,aux->etds.nombre);
                switch (aux->etds.tipo) {
                          case entero  :  fprintf(fich_gen,"  %d",aux->etds.valor.entero);
                                          break;
                          case real    :  fprintf(fich_gen,"  %f",aux->etds.valor.real);
                                          break;
               }
               fprintf(fich_gen,"\n");
             }
        aux=aux->sig;
        }
     }
     fprintf(fich_gen,"\n");
}

//********************* DECLARACIN DE NOMBRES DE LISTAS ******************/
void declara_nombres_listas() {
  int i;
  EHTDS *l_ele; //lista de elementos de una entrada hash de la tds
  LISTA_CAMPOS *l_campos;

  fprintf(fich_gen,"/*-------------- INICIALIZACIN DE NOMBRES DE LISTAS ----------------*/\n");
  for (i = 0; i < HASH_SIZE; i++) {
    if (tds[i] != NULL) {
      l_ele = tds[i];
      while (l_ele != NULL) {
	if ((l_ele->etds.clvar == instancia) || 
	    (l_ele->etds.clvar == var)) {
	  l_campos = l_ele->etds.lc;
	  while (l_campos != NULL) {
	    if ((l_campos->tipo == lista_enteros) ||
		(l_campos->tipo == lista_reales)) {
	      fprintf(fich_gen, "char %slnombre_%s_%s[] = \"%s.%s\";\n", prefijo,
		      l_ele->etds.nombre, l_campos->nombre,
		      l_ele->etds.nombre, l_campos->nombre);
	    } //Si el campo es una lista
	    l_campos = l_campos->siguiente;
	  } //Mientras la instancia o la variable tenga slots (y por tanto sea de tipo definido)
	  if ((l_ele->etds.tipo == lista_enteros) ||
	      (l_ele->etds.tipo == lista_reales)) {
	    fprintf(fich_gen, "char %slnombre_%s[] = \"%s\";\n", 
		    prefijo, l_ele->etds.nombre, l_ele->etds.nombre);
	  } //Si la instancia o var era simple (no de una clase) y de tipo lista
	} //Si es una instancia o variable
	l_ele = l_ele->sig;
      } //Mientras el registro de la tds tenga elementos por visitar
    } //Si el registro de la tds tiene elementos
  } //Para cada entrada de la tds
  fprintf(fich_gen, "\n\n\n");
} //declara_nombre_listas
	    
	

		      


/*********************** GENERA_FUNCIONES_EXTERNAS *************************/

void Genera_funciones_externas()
{
  EHFUNC *aux;
  int i;
  REGFUNC *funcion;
  LISTA_PARAM *lparam;

  fprintf(fich_gen,"/*--------------------- FUNCIONES EXTERNAS --------------------*/\n");
  fprintf(fich_gen,"\n");
  for(i=0;i<HASH_SIZE;i++)
     if(funciones[i]!=NULL) {
         aux=funciones[i];
         while(aux!=NULL) {
          if (aux->efunc.tipo==externa) {
           fprintf(fich_gen,"extern ");
           funcion= &(aux->efunc);
           switch (funcion->tipo_dev) {
             case boolean:
             case entero: fprintf(fich_gen,"int ");
                         break;
             case real:  fprintf(fich_gen,"float ");
                         break;
             case vacio: fprintf(fich_gen,"void ");
           }
           fprintf(fich_gen,"%s(",funcion->nombre_ext);
           /* ahora se generan los parametros: si tiene */
           if (funcion->param!=NULL){
                fprintf(fich_gen,"%s", funcion->param);
           }
           fprintf(fich_gen,");\n");
          }
         aux=aux->sig;
         }
     }

}
/*************************** GENERA_CLASES *****************************/
void Genera_decl_clase(entclase)
 REGTIPO *entclase;
{
  LISTA_CAMPOS * l;
  REGTIPO * aux;
  int es_lista;
 /* esta funcion genera la declaracion para una clase, pero antes intenta */
 /* declarar recursivamente todas las clases que utiliza */

 if (!entclase->ya_declarada) {
    if (!entclase->declarando_actualmente) {
      /* esto sirve para controlar si hay estructuras recursivas en las */
      /* declaraciones */
      entclase->declarando_actualmente = TRUE;

      /* primero se recorren los campos para ver si hay alguno que tenga */
      /* un tipo definido */

      l=entclase->lc;
      while(l!=NULL){
           if (l->tipo==definido) {
              aux= busca_en_tipos(l->instancia_de);
              Genera_decl_clase(aux);
           }
           l=l->siguiente;
      }
      /* ahora se declara */
      fprintf(fich_gen,"typedef struct {\n");
      l=entclase->lc;
      /* ahora se generan sus campos */
      while(l!=NULL){
	es_lista= FALSE;
	fprintf(fich_gen,"\t");
	switch (l->tipo) {
	
	case entero:
	  fprintf(fich_gen,"int     ");
	  break;
	
	case subrango:
	  fprintf(fich_gen,"int     ");
	  break;
	
	case real:  
	  fprintf(fich_gen,"float   ");
	  break;
          
	case caracter:
	  fprintf(fich_gen,"caracter");
	  break;
	
	case boolean:
	  fprintf(fich_gen,"int     ");
	  break;
	
	case lista_enteros:
	  fprintf(fich_gen,"listaent(%s,%d);\n",l->nombre,l->tam_lista);
	  es_lista=TRUE;
	  break;
	
	case lista_reales:
	  fprintf(fich_gen,"listareal(%s,%d);\n",l->nombre,l->tam_lista);
	  es_lista=TRUE;
	  break;
	
	case definido:
	  fprintf(fich_gen,"%s",l->instancia_de);
	  break;

	case array:
	  switch (l->tbase_array) {
	  case entero:
	  case boolean:
	    fprintf(fich_gen, "int ");
	    break;

	  case real:
	    fprintf(fich_gen, "float ");
	    break;
	  }
	  fprintf(fich_gen, "%s[%d];\n", l->nombre, l->tam_lista);
	  es_lista = TRUE; //Los arrays, como las listas, ya se han acabado de declarar
	  break;

	}
	if (!es_lista) 
	  fprintf(fich_gen,"  %s;\n",l->nombre);


	l=l->siguiente;
      }
      fprintf(fich_gen,"} %s%s;\n",prefijo,entclase->nombre);
      entclase->ya_declarada=TRUE;
      entclase->declarando_actualmente = FALSE;
    } else
    {
      /* esta estructura estaba ya siendo declarada */
      /* lo que significa que estamos usando una estructura que nos usa */
      menerror(103);
    }
 }
}

void Genera_clases()
{
  EHTIPO *aux;
  REGTIPO *auxptro;
  LISTA_CAMPOS * l, * lcaux;
  int i;

  fprintf(fich_gen,"/*--------------- ESTRUCTURAS DE DATOS PARA LAS CLASES ----------------*/\n");
  fprintf(fich_gen,"\n");
  for(i=0;i<HASH_SIZE;i++)
     if(tipos[i]!=NULL) {
         aux=tipos[i];
         while(aux!=NULL) {
             Genera_decl_clase(&aux->etipo);
             aux=aux->sig;
         }
     }
     fprintf(fich_gen,"\n");
}

/*************************** GENERA_TABLAS_REGLAS *****************************/

void Genera_tablas_reglas() {
  /* este procedimiento crea las estructuras de datos que contienen para
     cada regla, la combinacion de instancias que pueden estar activando dicha
     regla */
  
  EHREGLA *aux;
  LISTA_OBJ *lo;
  LISTA *lvpeaux;
  REGTIPO *auxptro;
  int i;
  
  fprintf(fich_gen,"/*--------------- ESTRUCTURAS DE DATOS PARA LAS REGLAS ----------------*/\n");
  fprintf(fich_gen,"\n");
  for(i=0;i<HASH_SIZE;i++)
    if(reglas[i]!=NULL) {
      aux=reglas[i];
      while(aux!=NULL) {
	if (!aux->eregla.tiene_copias)
	  {
	    fprintf(fich_gen,"int %sR%d", prefijo,aux->eregla.num_regla);
	    lvpeaux = aux->eregla.param;
	    while (lvpeaux!=NULL) {
	      lo=busca_obj(aux->eregla.lobj_cond,lvpeaux->nombre);
	      if (lo->cuantificador==existe) {
		auxptro = busca_en_tipos(lo->instancia_de);
		fprintf(fich_gen,"[%d]",auxptro->num_instancias);
	      }
	      lvpeaux=lvpeaux->sig;
	    }
	    fprintf(fich_gen,";     /* regla %s */ \n",aux->eregla.nombre);
	  }
	aux=aux->sig;
      }
      fprintf(fich_gen,"\n");
    }
} //Genera_tablas_reglas


/********************* GENERA_ESTRUCTURAS_CONDICIONES **********************/

void Genera_estructuras_condiciones() {
  /* este procedimiento crea las estructuras de datos para cada condicion de
     cada regla */

  EHREGLA *aux;
  LISTA_OBJ *lo;
  LISTA *lvpeaux;
  REGTIPO *auxptro;
  char nombre_condicion_regla[100];  /* almacena la cadena que identifica las estructuras */
  int i;
  LISTA_CONDIC  * lc;
  int tam_cambios;                  /* sirve para almacenar el tamao del vector */
                                    /* que almacena los cambios */
  int num_dim;                      /* numero de dimensiones del array de activaciones */
  
  fprintf(fich_gen,"/*------------- ESTRUCTURAS DE DATOS PARA LAS CONDICIONES -------------*/\n");
  fprintf(fich_gen,"\n");
  for(i=0;i<HASH_SIZE;i++)
    if(reglas[i]!=NULL) {
      aux=reglas[i];
      while(aux!=NULL) {
	if (!aux->eregla.tiene_copias) {
	  lc=aux->eregla.lcond;
	  while (lc!=NULL) {
	    sprintf(nombre_condicion_regla,"_cond%d_%s%s",lc->num_condic,prefijo,aux->eregla.nombre);
	    fprintf(fich_gen,"/* CONDICION: %d REGLA: %s",lc->num_condic,aux->eregla.nombre);
	    if (lc->es_orig) {
	      fprintf(fich_gen,"*/ \n\n");
	      if (lc->num_vpe==0) { /* condicion simple */
		fprintf(fich_gen,"int activ%s;\n",nombre_condicion_regla);
	      }
	      else {
		fprintf(fich_gen,"int activ%s",nombre_condicion_regla);
		lvpeaux = lc->param;
		tam_cambios = 0;
		num_dim = 0;
		while (lvpeaux!=NULL) {
		  lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		  if (lo->cuantificador==existe) {
		    auxptro = busca_en_tipos(lo->instancia_de);
		    fprintf(fich_gen,"[%d]",auxptro->num_instancias);
		    if (tam_cambios==0) 
		      tam_cambios=auxptro->num_instancias;
		    else  
		      tam_cambios=tam_cambios * auxptro->num_instancias;
		    num_dim++;
		  }
		  lvpeaux=lvpeaux->sig;
		}
		fprintf(fich_gen,";\n");
		/* ahora se generan la estructura donde se anotan los cambios */
		/* para que estos no esten repetidos */
		fprintf(fich_gen,"int camb%s",nombre_condicion_regla);
		lvpeaux = lc->param;
		num_dim = 0;
		while (lvpeaux!=NULL) {
		  lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		  if (lo->cuantificador==existe) {
		    auxptro = busca_en_tipos(lo->instancia_de);
		    fprintf(fich_gen,"[%d]",auxptro->num_instancias);
		    num_dim++;
		  }
		  lvpeaux=lvpeaux->sig;
		}
		fprintf(fich_gen,";\n");
		/* ahora se genera el vector de cambios */
		fprintf(fich_gen,"struct {\n");
		fprintf(fich_gen,"\tint  tipo_cambio;\n");
		fprintf(fich_gen,"\tint  instancias[%d];\n",num_dim);
		fprintf(fich_gen,"} cambios%s[%d];\n",nombre_condicion_regla,tam_cambios);
		fprintf(fich_gen,"int num_cambios%s;\n",nombre_condicion_regla);
	      }
	    }
	    else {/* si no es la original */
	      fprintf(fich_gen," (copia de condicion %d regla %s) */\n\n",lc->cond_orig,lc->regla_orig);
	      if (lc->num_vpe>0) { /* condicion con vpe */
		fprintf(fich_gen,"int num_cambios%s;\n",nombre_condicion_regla);
	      }
	    }
	    lc=lc->sig;
	    fprintf(fich_gen,"\n");
	  }
	}
	aux=aux->sig;
      }
    }
  fprintf(fich_gen,"\n");
} //Genera_estructuras_condiciones()



void busca_cond_copias_de(numc, nomregla)
   int numc;
   char *nomregla;
{
/* este procedimiento busca copias de la condicion pasada como parametro */
/* y para cada copia saca una actualizacion de los cambios */
  EHREGLA *aux;
  int i;
  LISTA_CONDIC  * lc;

  for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
             if (!aux->eregla.tiene_copias) {
                lc = aux->eregla.lcond;
                while (lc!=NULL) {
                    if (!lc->es_orig) { /* si es una copia */
                       if (!strcmp(lc->regla_orig,nomregla) && (lc->cond_orig==numc))
                           fprintf(fich_gen,"\tnum_cambios_cond%d_%s%s=num_cambios_cond%d_%s%s;\n",lc->num_condic,prefijo,aux->eregla.nombre,numc,prefijo,nomregla);
                    }
                    lc= lc->sig;
                 }
             }
             aux=aux->sig;
         }
     }
}

int tiene_copias_cond(numc, nomregla)
   int numc;
   char *nomregla;
{
/* este procedimiento devuelve TRUE si encuentra copias de la condicion */
/* pasada como parametro */
  EHREGLA *aux;
  int i;
  LISTA_CONDIC  * lc;

  for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
             if (!aux->eregla.tiene_copias) {
                lc = aux->eregla.lcond;
                while (lc!=NULL) {
                    if (!lc->es_orig) { /* si es una copia */
                       if (!strcmp(lc->regla_orig,nomregla) && (lc->cond_orig==numc))
                           return(TRUE);
                    }
                    lc= lc->sig;
                 }
             }
             aux=aux->sig;
         }
     }
  return(FALSE);
}


/**************************** GENERA CONDICIONES ****************************/
void Genera_condiciones()
/* este procedimiento crea una funcion por cada condicion de
  cada regla */
{
  EHREGLA *aux;
  LISTA_OBJ *lo;
  LISTA *lvpeaux;
  REGTIPO *auxptro;
  char nombre_condicion_regla[100];  /* almacena la cadena que identifica las estructuras */
  char cad_condic[255];
  int i;
  LISTA_CONDIC  * lc;
  int num_dim;                      /* numero de dimensiones del array de activaciones */
  int n;
  int cont;
  int cont_pt;
  int num_instancias_pt[10];        /* se admiten hasta 10 paratodos en una condicion */

  fprintf(fich_gen,"/*---------------------------  CONDICIONES -----------------------*/\n");
  fprintf(fich_gen,"\n");
  for(i=0;i<HASH_SIZE;i++)
    if(reglas[i]!=NULL) {
      aux=reglas[i];
      while(aux!=NULL) {
	if (!aux->eregla.tiene_copias) {
	  lc=aux->eregla.lcond;
	  while (lc!=NULL) {
	    if (lc->es_orig) {   /* si no es una copia */
	      sprintf(nombre_condicion_regla,"cond%d_%s%s",lc->num_condic,prefijo,aux->eregla.nombre);
	      fprintf(fich_gen,"/* CONDICION %d %s */ \n\n",lc->num_condic,aux->eregla.nombre);
	      fprintf(fich_gen,"void %s(",nombre_condicion_regla);
	      
	      lvpeaux = lc->param;
	      num_dim=0;
	      while (lvpeaux!=NULL) {
		lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		if (lo->cuantificador==existe) {
		  if (num_dim>0)
		    fprintf(fich_gen,",");
		  fprintf(fich_gen,"%s",lvpeaux->nombre);
		  num_dim++;
		}
		lvpeaux=lvpeaux->sig;
	      }
	      fprintf(fich_gen,")\n");
	      
	      /* se declaran los parametros de la funcion-condicion */
	      lvpeaux = lc->param;
	      while (lvpeaux!=NULL) {
		lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		if (lo->cuantificador==existe)
		  fprintf(fich_gen,"\tint %s;\n",lvpeaux->nombre);
		lvpeaux=lvpeaux->sig;
	      }
	      fprintf(fich_gen,"{\n");
	      
	      /* ahora se crean los iteradores para las vpu */
	      lvpeaux = lc->param;
	      while (lvpeaux!=NULL) {
		lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		if (lo->cuantificador==paratodo)
		  fprintf(fich_gen,"\tint %s;\n",lvpeaux->nombre);
		lvpeaux=lvpeaux->sig;
	      }
	      if (lc->num_vpu>0)
		fprintf(fich_gen,"\tint cumple_paratodo[%d];\n",lc->num_vpu);
	      /* se declaran los limites superiores e inferiores si */
	      /* hay vpe */
	      if (lc->num_vpe>0) {
		fprintf(fich_gen,"\tint lim_inf[%d];\n",lc->num_vpe);
		fprintf(fich_gen,"\tint lim_sup[%d];\n",lc->num_vpe);
		if (tiene_copias_cond(lc->num_condic,aux->eregla.nombre)) 
		  fprintf(fich_gen,"\tint nc;\n");
		fprintf(fich_gen,"\n");
		n=0;
		lvpeaux = lc->param;
		while (lvpeaux!=NULL) {
		  lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		  if (lo->cuantificador==existe) {
		    auxptro=busca_en_tipos(lo->instancia_de);
		    fprintf(fich_gen,"\tif (%s==-1) {\n",lvpeaux->nombre);
		    fprintf(fich_gen,"\t\tlim_inf[%d]=0;\n",n);
		    fprintf(fich_gen,"\t\tlim_sup[%d]=%d;\n",n,auxptro->num_instancias);
		    fprintf(fich_gen,"\t} else {\n");
		    fprintf(fich_gen,"\t\tlim_inf[%d]=%s;\n",n,lvpeaux->nombre);
		    fprintf(fich_gen,"\t\tlim_sup[%d]=%s+1;\n",n,lvpeaux->nombre);
		    fprintf(fich_gen,"\t}\n");
		    n++;
		  }
		  lvpeaux=lvpeaux->sig;
		}
	      }
	      
	      /* ahora se crean los bucles para las vpe si hay mas de 1 */
	      if (lc->num_vpe>0) {
		lvpeaux = lc->param;
		n=0;
		while (lvpeaux!=NULL) {
		  lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		  if (lo->cuantificador==existe) {
		    fprintf(fich_gen,"\tfor(%s=lim_inf[%d]; %s<lim_sup[%d];%s++) {\n",lvpeaux->nombre,n,lvpeaux->nombre,n,lvpeaux->nombre);
		    n++;
		  }
		  lvpeaux=lvpeaux->sig;
		}
	      }
	      
	      /* ahora se crean los bucles para las vpu */
	      lvpeaux = lc->param;
	      cont_pt = 0;
	      while (lvpeaux!=NULL) {
		lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		if (lo->cuantificador==paratodo) {
		  fprintf(fich_gen,"\tcumple_paratodo[%d]=0;\n",cont_pt);
		  auxptro = busca_en_tipos(lo->instancia_de);
		  num_instancias_pt[cont_pt]=auxptro->num_instancias;
		  cont_pt++;
		  fprintf(fich_gen,"\tfor(%s=0;%s<%d;%s++){\n",lvpeaux->nombre,lvpeaux->nombre,auxptro->num_instancias,lvpeaux->nombre);
		}
		lvpeaux=lvpeaux->sig;
	      }
	      
	      /* ahora se genera el if principal */
	      cad_condic[0]='\0';
	      fprintf(fich_gen,"\tif (");

	      
	      Genera_exp(lc->lado_izq,cad_condic,lc->lobj_cond);
	      fprintf(fich_gen,"%s",cad_condic);
	      switch (lc->oprel) {
	      case igual : 
		fprintf(fich_gen," == ");
		break;
	      case distinto:
		fprintf(fich_gen," != ");
		break;
	      case menor:
		fprintf(fich_gen," < ");
		break;
	      case mayor:
		fprintf(fich_gen," > ");
		break;
	      case mayorig:
		fprintf(fich_gen," >= ");
		break;
	      case menorig:
		fprintf(fich_gen," <= ");
		break;
	      }
	      cad_condic[0]='\0';
	      if (lc->lado_der!=NULL)     /* si es una expresion booleana, no tiene lado derecho de la expresion */
		Genera_exp(lc->lado_der,cad_condic,lc->lobj_cond);
	      fprintf(fich_gen,"%s) {\n",cad_condic);
	      
	      if (cont_pt>0) {  /* si hay algun para todo */
		for (cont=0;cont<cont_pt; cont++) 
		  fprintf(fich_gen,"\t\tcumple_paratodo[%d]++;\n",cont);
		fprintf(fich_gen,"\t\t\tif (");
		for (cont=0;cont<cont_pt; cont++) {
		  if (cont>2)
		    fprintf(fich_gen," && ");
		  fprintf(fich_gen,"cumple_paratodo[%d]==%d",cont,num_instancias_pt[cont]);
		}
		fprintf(fich_gen,") {\n");
	      }


	      //El tratamiento del not con variables patrn existenciales es errneo.
	      //Lo aqu realizado hace que la regla se dispare con las instancias que
	      //no satisfacen una condicin, pero un "no existe cond" equivale a un 
	      //"para todo, no cond". Es posible implementar el no existe como un para 
	      //todo no, pero ello implica modificar esta seccin, y la generacin
	      //de segmentos unin, entre otros (LIz_...). Adems, se genera otro problema;
	      //Las variables patrn cuantificadas existencialmente pueden ser usadas en el 
	      //lado derecho de las reglas, con lo que o se controla esta situacin, o se
	      //producir un absurdo (con que instancia se instancia un para todo no...)
	      if (lc->num_vpe==0) {
		if (!lc->esta_negada) {
		  fprintf(fich_gen,"\t\tactiv_%s=TRUE;\n",nombre_condicion_regla);
		  if (cont_pt>0)
		    fprintf(fich_gen,"\t\t}\n");
		  fprintf(fich_gen,"\t} else {\n");
		  fprintf(fich_gen,"\t\tactiv_%s=FALSE;\n",nombre_condicion_regla);
		}
		else {
		  fprintf(fich_gen,"\t\tactiv_%s=FALSE;\n",nombre_condicion_regla);
		  if (cont_pt>0)
		    fprintf(fich_gen,"\t\t}\n");
		  fprintf(fich_gen,"\t} else {\n");
		  fprintf(fich_gen,"\t\tactiv_%s=TRUE;\n",nombre_condicion_regla);
		}
		if (cont_pt>0)
		  fprintf(fich_gen,"\t\treturn;\n");
	      }
	      else {
		fprintf(fich_gen,"\t\tactiv_%s",nombre_condicion_regla);
		lvpeaux = lc->param;
		while (lvpeaux!=NULL) {
		  lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		  if (lo->cuantificador==existe)
		    fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		  lvpeaux=lvpeaux->sig;
		}
		if (lc->esta_negada) 
		  fprintf(fich_gen, "=CELDA_VACIA;\n");
		else
		  fprintf(fich_gen,"=CELDA_OCUPADA;\n");

		fprintf(fich_gen,"\t\t/* ahora se anotan los cambios */\n");
		/**/
		if (tiene_copias_cond(lc->num_condic,aux->eregla.nombre)) {
		  fprintf(fich_gen,"\t\tif (camb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }
		  
		  if (lc->esta_negada)
		    fprintf(fich_gen, "==activacion) {\n");
		  else
		    fprintf(fich_gen,"==desactivacion) {\n");
		  
		  fprintf(fich_gen,"\t\t\tcamb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }
		  if (lc->esta_negada)
		    fprintf(fich_gen, "=desactivacion;\n");
		  else
		    fprintf(fich_gen,"=activacion;\n");
		  fprintf(fich_gen,"\t\t\tnc=0;\n");
		  fprintf(fich_gen,"\t\t\twhile (nc<num_cambios_%s) {\n",nombre_condicion_regla);
		  if (lc->esta_negada)
		    fprintf(fich_gen,"\t\t\t\tif (cambios_%s[nc].tipo_cambio==activacion)\n",nombre_condicion_regla);
		  else
		    fprintf(fich_gen,"\t\t\t\tif (cambios_%s[nc].tipo_cambio==desactivacion)\n",nombre_condicion_regla);
		  
		  lvpeaux = lc->param;
		  n=0;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe) {
		      fprintf(fich_gen,"\t\t\t\tif (cambios_%s[nc].instancias[%d]==%s) \n",
			      nombre_condicion_regla,n,lvpeaux->nombre);
		      n++;
		    }
		    lvpeaux=lvpeaux->sig;
		  }
		  if (lc->esta_negada)
		    fprintf(fich_gen, "\t\t\t\t\t{ cambios_%s[nc].tipo_cambio=desactivacion;\n",nombre_condicion_regla);
		  else
		    fprintf(fich_gen,"\t\t\t\t\t{ cambios_%s[nc].tipo_cambio=activacion;\n",nombre_condicion_regla);
		  fprintf(fich_gen,"\t\t\t\t\tbreak; }\n");
		  fprintf(fich_gen,"\t\t\t\tnc++;\n");
		  fprintf(fich_gen,"\t\t\t}\n");
		  fprintf(fich_gen,"\t\t}\n");
		  fprintf(fich_gen,"\t\telse if (camb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }
		  if (lc->esta_negada)
		    fprintf(fich_gen, "==CELDA_OCUPADA) {\n");
		  else
		    fprintf(fich_gen,"==CELDA_VACIA) {\n");
		  fprintf(fich_gen,"\t\t\tcamb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }
		  
		  if (lc->esta_negada) {
		    fprintf(fich_gen, "=desactivacion;\n");
		    fprintf(fich_gen,"\t\t\tcambios_%s[num_cambios_%s].tipo_cambio=desactivacion;\n",
			    nombre_condicion_regla, nombre_condicion_regla);
		  }
		  else {
		    fprintf(fich_gen, "=activacion;\n");
		    fprintf(fich_gen, "\t\t\tcambios_%s[num_cambios_%s].tipo_cambio=activacion;\n",
			    nombre_condicion_regla, nombre_condicion_regla);
		  }
		  
		  lvpeaux = lc->param;
		  n=0;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe) {
		      fprintf(fich_gen,"\t\t\tcambios_%s[num_cambios_%s].instancias[%d]=%s;\n",
			      nombre_condicion_regla,nombre_condicion_regla,n,lvpeaux->nombre);
		      n++;
		    }
		    lvpeaux=lvpeaux->sig;
		  }
		  fprintf(fich_gen,"\t\t\tnum_cambios_%s++;\n",nombre_condicion_regla);
		  fprintf(fich_gen,"\t\t}\n");
		}
		else {
		  fprintf(fich_gen,"\t\tif (camb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }
		  
		  if (lc->esta_negada) 
		    fprintf(fich_gen,"!=desactivacion) {\n");
		  else
		    fprintf(fich_gen,"!=activacion) {\n");
		  fprintf(fich_gen,"\t\t\tcamb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }

		  if (lc->esta_negada) {
		    fprintf(fich_gen,"=desactivacion;\n");		  
		    fprintf(fich_gen,"\t\t\tcambios_%s[num_cambios_%s].tipo_cambio=desactivacion;\n",
			    nombre_condicion_regla,nombre_condicion_regla);  
		  }
		  else {
		    fprintf(fich_gen,"=activacion;\n");		  
		    fprintf(fich_gen,"\t\t\tcambios_%s[num_cambios_%s].tipo_cambio=activacion;\n",
			    nombre_condicion_regla,nombre_condicion_regla);
		  }
		  
		  lvpeaux = lc->param;
		  n=0;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe) {
		      fprintf(fich_gen,"\t\t\tcambios_%s[num_cambios_%s].instancias[%d]=%s;\n",
			      nombre_condicion_regla,nombre_condicion_regla,n,lvpeaux->nombre);
		      n++;
		    }
		    lvpeaux=lvpeaux->sig;
		  }
		  fprintf(fich_gen,"\t\t\tnum_cambios_%s++;\n",nombre_condicion_regla);
		  fprintf(fich_gen,"\t\t}\n");
		}
		/**/
		
		if (cont_pt>0) fprintf(fich_gen,"\t\t}\n");
		fprintf(fich_gen,"\t} else {\n");
		fprintf(fich_gen,"\t\tif (");
		fprintf(fich_gen,"activ_%s",nombre_condicion_regla);
		lvpeaux = lc->param;
		while (lvpeaux!=NULL) {
		  lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		  if (lo->cuantificador==existe)
		    fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		  lvpeaux=lvpeaux->sig;
		}

		if (lc->esta_negada) 
		  fprintf(fich_gen, "!=CELDA_OCUPADA) {\n");
		else
		  fprintf(fich_gen,"!=CELDA_VACIA) {\n");
		fprintf(fich_gen,"\t\t\tactiv_%s",nombre_condicion_regla);
		lvpeaux = lc->param;
		while (lvpeaux!=NULL) {
		  lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		  if (lo->cuantificador==existe)
		    fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		  lvpeaux=lvpeaux->sig;
		}
		if (lc->esta_negada)
		  fprintf(fich_gen, "=CELDA_OCUPADA;\n");
		else
		  fprintf(fich_gen,"=CELDA_VACIA;\n");
		fprintf(fich_gen,"\t\t\t/* ahora se anotan los cambios */\n");
		if (tiene_copias_cond(lc->num_condic,aux->eregla.nombre)) {
		  fprintf(fich_gen,"\t\t\tif (camb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }
		  if (lc->esta_negada)
		    fprintf(fich_gen,"==desactivacion) {\n");
		  else
		    fprintf(fich_gen,"==activacion) {\n");
		  fprintf(fich_gen,"\t\t\t\tcamb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }
		  if (lc->esta_negada)
		    fprintf(fich_gen, "=activacion;\n");
		  else
		    fprintf(fich_gen,"=desactivacion;\n");
		  fprintf(fich_gen,"\t\t\t\tnc=0;\n");
		  fprintf(fich_gen,"\t\t\t\twhile (nc<num_cambios_%s) {\n",nombre_condicion_regla);
		  if (lc->esta_negada)
		    fprintf(fich_gen,"\t\t\t\t\tif (cambios_%s[nc].tipo_cambio==desactivacion)\n",nombre_condicion_regla);
		  else
		    fprintf(fich_gen,"\t\t\t\t\tif (cambios_%s[nc].tipo_cambio==activacion)\n",nombre_condicion_regla);
		  
		  lvpeaux = lc->param;
		  n=0;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe) {
		      fprintf(fich_gen,"\t\t\t\t\tif (cambios_%s[nc].instancias[%d]==%s) \n",nombre_condicion_regla,n,lvpeaux->nombre);
		      n++;
		    }
		    lvpeaux=lvpeaux->sig;
		  }

		  if (lc->esta_negada)
		    fprintf(fich_gen,"\t\t\t\t\t\t{ cambios_%s[nc].tipo_cambio=activacion;\n",nombre_condicion_regla);
		  else
		    fprintf(fich_gen,"\t\t\t\t\t\t{ cambios_%s[nc].tipo_cambio=desactivacion;\n",nombre_condicion_regla);
		  fprintf(fich_gen,"\t\t\t\t\t\tbreak; }\n");
		  fprintf(fich_gen,"\t\t\t\t\tnc++;\n");
		  fprintf(fich_gen,"\t\t\t\t}\n");
		  fprintf(fich_gen,"\t\t\t}\n");
		  fprintf(fich_gen,"\t\t\telse if (camb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }

		  if (lc->esta_negada)
		    fprintf(fich_gen,"==CELDA_OCUPADA) {\n"); 
		  else
		    fprintf(fich_gen,"==CELDA_VACIA) {\n");
		  fprintf(fich_gen,"\t\t\t\tcamb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }

		  if (lc->esta_negada) {
		     fprintf(fich_gen,"=activacion;\n");
		     fprintf(fich_gen,"\t\t\t\tcambios_%s[num_cambios_%s].tipo_cambio=activacion;\n",
			     nombre_condicion_regla,nombre_condicion_regla);
		  }
		  else {
		    fprintf(fich_gen,"=desactivacion;\n");
		    fprintf(fich_gen,"\t\t\t\tcambios_%s[num_cambios_%s].tipo_cambio=desactivacion;\n",
			    nombre_condicion_regla,nombre_condicion_regla);
		  }
		  
		  lvpeaux = lc->param;
		  n=0;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe) {
		      fprintf(fich_gen,"\t\t\t\tcambios_%s[num_cambios_%s].instancias[%d]=%s;\n",
			      nombre_condicion_regla,nombre_condicion_regla,n,lvpeaux->nombre);
		      n++;
		    }
		    lvpeaux=lvpeaux->sig;
		  }
		  fprintf(fich_gen,"\t\t\t\tnum_cambios_%s++;\n",nombre_condicion_regla);
		  fprintf(fich_gen,"\t\t\t}\n");
		}
		else {
		  fprintf(fich_gen,"\t\t\tif (camb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }

		  if (lc->esta_negada)
		    fprintf(fich_gen,"!=activacion) {\n");
		  else
		    fprintf(fich_gen,"!=desactivacion) {\n");
		  fprintf(fich_gen,"\t\t\t\tcamb_%s",nombre_condicion_regla);
		  lvpeaux = lc->param;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe)
		      fprintf(fich_gen,"[%s]",lvpeaux->nombre);
		    lvpeaux=lvpeaux->sig;
		  }

		  if (lc->esta_negada) {
		    fprintf(fich_gen,"=activacion;\n");
		    fprintf(fich_gen,"\t\t\t\tcambios_%s[num_cambios_%s].tipo_cambio=activacion;\n",
			    nombre_condicion_regla,nombre_condicion_regla);
		  }
		  else {
		    fprintf(fich_gen,"=desactivacion;\n");
		    fprintf(fich_gen,"\t\t\t\tcambios_%s[num_cambios_%s].tipo_cambio=desactivacion;\n",
			    nombre_condicion_regla,nombre_condicion_regla);
		  }
		  lvpeaux = lc->param;
		  n=0;
		  while (lvpeaux!=NULL) {
		    lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
		    if (lo->cuantificador==existe) {
		      fprintf(fich_gen,"\t\t\t\tcambios_%s[num_cambios_%s].instancias[%d]=%s;\n",
			      nombre_condicion_regla,nombre_condicion_regla,n,lvpeaux->nombre);
		      n++;
		    }
		    lvpeaux=lvpeaux->sig;
		  }
		  fprintf(fich_gen,"\t\t\t\tnum_cambios_%s++;\n",nombre_condicion_regla);
		  fprintf(fich_gen,"\t\t\t}\n");
		}
		fprintf(fich_gen,"\t\t}\n");
		if (cont_pt>0) fprintf(fich_gen,"\t\tbreak;\n");
	      }
	      
	      fprintf(fich_gen,"\t}\n");
	      /* ahora se cierran los bucles para las vpu */
	      for(n=0;n<lc->num_vpu;n++)
		fprintf(fich_gen,"\t}\n");
	      
	      if (lc->num_vpe>0)
		for(n=0;n<lc->num_vpe;n++)
		  fprintf(fich_gen,"\t}\n");
	      
	      /* ahora se buscan condiciones que sean copias de esta */
	      /* para actualizar su num_cambios */
	      if (lc->num_vpe>0) {
		busca_cond_copias_de(lc->num_condic,aux->eregla.nombre);
	      }
	      
	      fprintf(fich_gen,"}\n");
	    }
	    else {
	      fprintf(fich_gen,"/* CONDICION %d %s",lc->num_condic,aux->eregla.nombre);
	      fprintf(fich_gen," (copia de condicion %d regla %s) */\n",lc->cond_orig,lc->regla_orig);
	    }
	    lc=lc->sig;
	    fprintf(fich_gen,"\n");
	  }
	}
	aux=aux->sig;
      }
    }
  fprintf(fich_gen,"\n");
  
}
/*************************** GENERA_FUNCIONES_TABLAS *****************************/

void Genera_funciones_tablas()
{
  EHREGLA *aux;
  LISTA_OBJ *lo;
  LISTA *lvpeaux;
  int i;
  int dim;
  int cont,cont2,cont3,cont4;
  REGTIPO *entclase;

   /* Generamos la funcion esta_activada que devuelve TRUE si la regla pasada
  como parametro, con la instanciacion pasada como parametro esta activada */
 fprintf(fich_gen,"/*------------- FUNCIONES PARA LAS ESTRUCTURAS DE LAS REGLAS -------------*/\n");
 fprintf(fich_gen,"\n");
 fprintf(fich_gen,"int %sesta_activada(r)\n",prefijo);
 fprintf(fich_gen,"\t %sREGAGENDA r;\n",prefijo);
 fprintf(fich_gen,"{\n");
 fprintf(fich_gen,"\tswitch (r.num_regla) {\n");
 for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
           if (!aux->eregla.tiene_copias) {
             dim=0;
             fprintf(fich_gen,"\t\tcase %d:  ", aux->eregla.num_regla);
             fprintf(fich_gen,"return (%sR%d", prefijo,aux->eregla.num_regla);
             lvpeaux = aux->eregla.param;
             while (lvpeaux!=NULL) {
                     lo=busca_obj(aux->eregla.lobj_cond,lvpeaux->nombre);
                     if (lo->cuantificador==existe) {
                           fprintf(fich_gen,"[r.instancias[%d]]",dim);
                           dim++;
                      }
                      lvpeaux=lvpeaux->sig;
             }
             fprintf(fich_gen,");\n");
           }
           aux=aux->sig;
         }
    }
    fprintf(fich_gen,"\t}\n");
    fprintf(fich_gen,"}\n\n");

 /* Generamos la funcion marca_activada que marca la regla pasada
  como parametro, con la instanciacion pasada como parametro como activada */
 fprintf(fich_gen,"void %smarca_activada(r,entagen)\n",prefijo);
 fprintf(fich_gen,"\t %sREGAGENDA r;\n",prefijo);
 fprintf(fich_gen,"\t int entagen;\n");
 fprintf(fich_gen,"{\n");
 fprintf(fich_gen,"\tswitch (r.num_regla) {\n");
 for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
           if (!aux->eregla.tiene_copias) {
             dim=0;
             fprintf(fich_gen,"\t\tcase %d:  ", aux->eregla.num_regla);
             fprintf(fich_gen,"%sR%d", prefijo,aux->eregla.num_regla);
             lvpeaux = aux->eregla.param;
             while (lvpeaux!=NULL) {
                     lo=busca_obj(aux->eregla.lobj_cond,lvpeaux->nombre);
                     if (lo->cuantificador==existe) {
                           fprintf(fich_gen,"[r.instancias[%d]]",dim);
                           dim++;
                      }
                      lvpeaux=lvpeaux->sig;
             }
             fprintf(fich_gen,"=entagen;\n");
             if (debug) {
                 fprintf(fich_gen,"\t\tprintf(\">Activacion: %s ",aux->eregla.nombre);
                 lvpeaux = aux->eregla.param;
                 while (lvpeaux!=NULL) {
                     lo=busca_obj(aux->eregla.lobj_cond,lvpeaux->nombre);
                     if (lo->cuantificador==existe) {
                           fprintf(fich_gen," %s: ",lo->nombre);
                           /* se que lo siguiente puede parecer raro ...*/
                           /* pero no he encontrado otra forma de que sacara */
                           /* %d en el fichero generado (solo en UNIX no funciona*/
                           fprintf(fich_gen,"%s","%d");
                      }
                      lvpeaux=lvpeaux->sig;
                 }
                 fprintf(fich_gen," \\n\"");
                 lvpeaux = aux->eregla.param;
                 dim=0;
                 while (lvpeaux!=NULL) {
                     lo=busca_obj(aux->eregla.lobj_cond,lvpeaux->nombre);
                     if (lo->cuantificador==existe) {
                           fprintf(fich_gen,",r.instancias[%d]",dim);
                           dim++;
                      }
                      lvpeaux=lvpeaux->sig;
                 }
                 fprintf(fich_gen,");\n");
             }
             fprintf(fich_gen,"\t\t\t break;\n");
           }
           aux=aux->sig;
         }
    }
    fprintf(fich_gen,"\t}\n");
    fprintf(fich_gen,"}\n\n");

     /* Generamos la funcion marca_desactivada que marca la regla pasada
  como parametro, con la instanciacion pasada como parametro como desactivada */
 fprintf(fich_gen,"int %smarca_desactivada(r)\n",prefijo);
 fprintf(fich_gen,"\t %sREGAGENDA r;\n",prefijo);
 fprintf(fich_gen,"{\n");
 fprintf(fich_gen,"\tint entagen;\n");
 fprintf(fich_gen,"\n");
 fprintf(fich_gen,"\tswitch (r.num_regla) {\n");
 for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
           if (!aux->eregla.tiene_copias) {
             dim=0;
             fprintf(fich_gen,"\t\tcase %d:  ", aux->eregla.num_regla);

             fprintf(fich_gen,"entagen=");
             fprintf(fich_gen,"%sR%d",prefijo, aux->eregla.num_regla);
             lvpeaux = aux->eregla.param;
             while (lvpeaux!=NULL) {
                     lo=busca_obj(aux->eregla.lobj_cond,lvpeaux->nombre);
                     if (lo->cuantificador==existe) {
                           fprintf(fich_gen,"[r.instancias[%d]]",dim);
                           dim++;
                      }
                      lvpeaux=lvpeaux->sig;
             }
             fprintf(fich_gen,";\n");
             fprintf(fich_gen,"\t\t\t");
             dim=0;
             fprintf(fich_gen,"%sR%d", prefijo,aux->eregla.num_regla);
             lvpeaux = aux->eregla.param;
             while (lvpeaux!=NULL) {
                     lo=busca_obj(aux->eregla.lobj_cond,lvpeaux->nombre);
                     if (lo->cuantificador==existe) {
                           fprintf(fich_gen,"[r.instancias[%d]]",dim);
                           dim++;
                      }
                      lvpeaux=lvpeaux->sig;
             }
             fprintf(fich_gen,"=0;\n");
             if (debug) {
                 fprintf(fich_gen,"\t\tprintf(\">Desactivacion: %s ",aux->eregla.nombre);
                 lvpeaux = aux->eregla.param;
                 while (lvpeaux!=NULL) {
                     lo=busca_obj(aux->eregla.lobj_cond,lvpeaux->nombre);
                     if (lo->cuantificador==existe) {
                           fprintf(fich_gen," %s: ",lo->nombre);
                           fprintf(fich_gen,"%s","%d");
                      }
                      lvpeaux=lvpeaux->sig;
                 }
                 fprintf(fich_gen," \\n\"");
                 lvpeaux = aux->eregla.param;
                 dim=0;
                 while (lvpeaux!=NULL) {
                     lo=busca_obj(aux->eregla.lobj_cond,lvpeaux->nombre);
                     if (lo->cuantificador==existe) {
                           fprintf(fich_gen,",r.instancias[%d]",dim);
                           dim++;
                      }
                      lvpeaux=lvpeaux->sig;
                 }
                 fprintf(fich_gen,");\n");
             }
             fprintf(fich_gen,"\t\t\t break;\n");
           }
           aux=aux->sig;
         }
    }
    fprintf(fich_gen,"\t}\n");
    fprintf(fich_gen,"\treturn(entagen);\n");
    fprintf(fich_gen,"}\n\n");

  /* Generamos la funcion elim_agenda que elimina todas las ocurrencias de una
  regla de la agenda */
 fprintf(fich_gen,"void %selim_agenda(num_regla)\n",prefijo);
 fprintf(fich_gen,"\t int num_regla;\n");
 fprintf(fich_gen,"{\n");
 if (num_maxvp >0) {
    fprintf(fich_gen,"\tint iter[%d];\n",num_maxvp);
 }
 fprintf(fich_gen,"\t%sREGAGENDA r;\n",prefijo);
 fprintf(fich_gen,"\n");
 fprintf(fich_gen,"\tswitch (num_regla) {\n");
 for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
           if (!aux->eregla.tiene_copias) {
             dim=0;
             fprintf(fich_gen,"\t\tcase %d:  \n", aux->eregla.num_regla);
             cont=0;
             lvpeaux = aux->eregla.param;
             while (lvpeaux!=NULL) {
                   for(cont2=0; cont2<cont; cont2++) fprintf(fich_gen,"\t");
                   lo=busca_obj(aux->eregla.lobj_cond,lvpeaux->nombre);
                   if (lo->cuantificador==existe) {
                            entclase=busca_en_tipos(lo->instancia_de);
                            fprintf(fich_gen,"\t\tfor(iter[%d]=0;iter[%d]<%d; iter[%d]++)\n",cont,cont,entclase->num_instancias,cont);
                            cont++;
                   }
                   lvpeaux=lvpeaux->sig;
             }
             if (cont>0) fprintf(fich_gen,"\t\t {\n");
             fprintf(fich_gen,"\t\t     r.num_regla=%d;\n",aux->eregla.num_regla);
             for (cont3=0; cont3<aux->eregla.num_vpe; cont3++)
                    fprintf(fich_gen,"\t\t     r.instancias[%d]=iter[%d];\n",cont3,cont3);
             fprintf(fich_gen,"\t\t\t if (%sesta_activada(r)) %selimregla_agenda(r);\n",prefijo,prefijo);
             if (cont>0) fprintf(fich_gen,"\t\t }\n");
             fprintf(fich_gen,"\t\t\t break;\n");
           }
           aux=aux->sig;
         }
    }
    fprintf(fich_gen,"\t}\n");
    fprintf(fich_gen,"}\n\n");
}

/*************************    DECLARA_INSTANCIAS    *************************/
void Declara_instancias()
{
  EHTIPO *aux;
  EHTDS *auxtds;
  int i;
  int es_lista;
  
  fprintf(fich_gen,"/*----------------- DECLARACION DE INSTANCIAS --------------------*/\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"/* Instancias de tipo definido */\n");
  for(i=0;i<HASH_SIZE;i++)
    if(tipos[i]!=NULL) {
      aux=tipos[i];
      while(aux!=NULL) {
	if (aux->etipo.num_instancias>0) {
	  fprintf(fich_gen,"%s%s ",prefijo,aux->etipo.nombre);
	  fprintf(fich_gen,"%si_%s[%d];\n",prefijo,aux->etipo.nombre,aux->etipo.num_instancias);
	}
	aux=aux->sig;
      }
    }
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"/* Instancias de tipo simple */\n");
  for(i=0;i<HASH_SIZE;i++)
    if(tds[i]!=NULL) {
      auxtds=tds[i];
      while(auxtds!=NULL) {
	if (auxtds->etds.clvar==instancia && auxtds->etds.tipo!=definido) {
	  es_lista = FALSE;
	  switch(auxtds->etds.tipo) {
	  case entero  :  fprintf(fich_gen,"int     ");
	    break;
	  case subrango:  fprintf(fich_gen,"int     ");
	    break;
	  case real    :  fprintf(fich_gen,"float   ");
	    break;
	  case caracter:  fprintf(fich_gen,"caracter");
	    break;
	  case boolean :  fprintf(fich_gen,"int     ");
	    break;
	  case lista_enteros : 
	    fprintf(fich_gen,"listaent(%s%s,%d);\n", 
		    prefijo, auxtds->etds.nombre, 
		    auxtds->etds.tam_lista);
	    es_lista=TRUE;
	    break;
	  case lista_reales  :  
	    fprintf(fich_gen,"listareal(%s%s,%d);\n",
		    prefijo, auxtds->etds.nombre,
		    auxtds->etds.tam_lista);
	    es_lista=TRUE;
	    break;
	    
	  case array:
	    es_lista = TRUE;
	    switch (auxtds->etds.tbase_array) {
	    case entero:
	    case boolean:
	      fprintf(fich_gen, "int ");
	      break;
	      
	    case real:
	      fprintf(fich_gen, "float ");
	      break;
	      
	    default:
	      fprintf(stderr, "sinsem.c acepto un tipo para el que genera.c no puede generar cdigo\n");
	      fprintf(stderr, "revisa en sinsem.c la funcin r_facts (menerror(126,...))\n");
	      fprintf(stderr, "revisa en genera.c la funcin Declara_instancias para las simples\n");
	      break;
	    }
	    fprintf(fich_gen,"  %s%s[%d];\n",prefijo,auxtds->etds.nombre, auxtds->etds.tam_lista);
	  }
	  if (!es_lista) fprintf(fich_gen,"  %s%s;\n",prefijo,auxtds->etds.nombre);
	}
	auxtds=auxtds->sig;
      }
    }
  fprintf(fich_gen,"\n");
}

/*************************    DECLARA_VARIABLES    *************************/
void Declara_variables()
{
  EHTIPO *aux;
  EHTDS *auxtds;
  int i;
  int es_lista;
  
  fprintf(fich_gen,"/*----------------- DECLARACION DE VARIABLES --------------------*/\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"/* Variables de tipo definido */\n");
  for(i=0;i<HASH_SIZE;i++)
    if(tipos[i]!=NULL) {
      aux=tipos[i];
      while(aux!=NULL) {
	if (aux->etipo.num_vars>0) {
	  fprintf(fich_gen,"%s%s ",prefijo,aux->etipo.nombre);
	  fprintf(fich_gen,"%sv_%s[%d];\n",prefijo,aux->etipo.nombre,aux->etipo.num_vars);
	}
	aux=aux->sig;
      }
    }
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"/* Variables de tipo simple */\n");
  for(i=0;i<HASH_SIZE;i++)
    if(tds[i]!=NULL) {
      auxtds=tds[i];
      while(auxtds!=NULL) {
	if (auxtds->etds.clvar==var && auxtds->etds.tipo!=definido) {
	  es_lista = FALSE;
	  switch(auxtds->etds.tipo) {
	  case entero  :  fprintf(fich_gen,"int     ");
	    break;
	  case subrango:  fprintf(fich_gen,"int     ");
	    break;
	  case real    :  fprintf(fich_gen,"float   ");
	    break;
	  case caracter:  fprintf(fich_gen,"caracter");
	    break;
	  case boolean :  fprintf(fich_gen,"int     ");
	    break;
	  case lista_enteros :
	    fprintf(fich_gen,"listaent(%s%s,%d);\n", prefijo, auxtds->etds.nombre,auxtds->etds.tam_lista);
	    es_lista=TRUE;
	    break;
	  case lista_reales  :  
	    fprintf(fich_gen,"listareal(%s%s,%d);\n", prefijo, auxtds->etds.nombre,auxtds->etds.tam_lista);
	    es_lista=TRUE;
	    break;
	  case array:
	    switch (auxtds->etds.tbase_array) {
	    case boolean:
	    case entero:
	      fprintf(fich_gen, "int ");
	      break;
	    case real:
	      fprintf(fich_gen, "float ");
	      break;
	    }
	    fprintf(fich_gen, "%s%s[%d];\n", prefijo, auxtds->etds.nombre, auxtds->etds.tam_lista);
	    es_lista = TRUE;
	  }
	  if (!es_lista) fprintf(fich_gen,"  %s%s;\n",prefijo,auxtds->etds.nombre);
	}
	auxtds=auxtds->sig;
      }
    }
  fprintf(fich_gen,"\n");
}

/********************** INICIALIZA_VARIABLES  ************************/
void Inicializa_variables() {
  EHTDS *aux;
  int i, j;
  LISTA_CAMPOS * l;
  LIS_NUM * lnum;
  
  fprintf(fich_gen,"/*----------------- INICIALIZACION DE VARIABLES -------------------*/\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"void %sInicializa_variables()\n",prefijo);
  fprintf(fich_gen,"{\n");
  for(i=0;i<HASH_SIZE;i++)
    if(tds[i]!=NULL) {
      aux=tds[i];
      while(aux!=NULL) {
	if (aux->etds.clvar==var) {
	  fprintf(fich_gen,"\t/* inicializando variable %s */\n",aux->etds.nombre );
	  /* FALTA: los campos que son de tipo definido */
	  l=aux->etds.lc;
	  while(l!=NULL) {
	    if (l->tipo==lista_enteros || l->tipo==lista_reales) {
	      /* si es un tipo lista aunque no este inicializado se */
	      /* inicializa */
	      switch (l->tipo) {
	      case lista_enteros :  
		fprintf(fich_gen, "\tinicializa_lent(&%sv_%s[%d].%s, %d, %slnombre_%s_%s, %sv_%s[%d].%s_elementos);\n",
			prefijo, aux->etds.instancia_de, aux->etds.cod.num_instancia, l->nombre,
			l->tam_lista,
			prefijo, aux->etds.nombre, l->nombre,
			prefijo, aux->etds.instancia_de, aux->etds.cod.num_instancia, l->nombre);
		break;
	      case lista_reales  :  
		fprintf(fich_gen, "\tinicializa_lreal(&%sv_%s[%d].%s, %d, %slnombre_%s_%s, %sv_%s[%d].%s_elementos);\n",
			prefijo, aux->etds.instancia_de, aux->etds.cod.num_instancia, l->nombre,
			l->tam_lista,
			prefijo, aux->etds.nombre, l->nombre,
			prefijo, aux->etds.instancia_de, aux->etds.cod.num_instancia, l->nombre);
		break;
	      }
	    }
	    if (l->inicializado) {
	      if (l->tipo!=lista_enteros && l->tipo!=lista_reales && l->tipo != array) { /* si no es de tipo lista */
		fprintf(fich_gen,"\t%sv_%s[%d].%s=", prefijo, aux->etds.instancia_de, 
			aux->etds.cod.num_instancia, l->nombre);
		switch (l->tipo) {
		case entero  :  
		  fprintf(fich_gen,"%d;\n",l->valor.entero);
		  break;
		case subrango:
		  fprintf(fich_gen,"%d;\n",l->valor.entero);
		  break;
		case real    :  
		  fprintf(fich_gen,"%f;\n",l->valor.real);
		  break;
		case caracter:  /* FALTA */
		  break;
		case boolean :  
		  if (l->valor.entero==0)
		    fprintf(fich_gen,"FALSE;\n");
		  else
		    fprintf(fich_gen,"TRUE;\n");
		  break;
		}
	      }
	      else {
		if (l->tipo==lista_enteros || l->tipo==lista_reales || l->tipo == array) {
		  j = 0;
		  lnum = l->ld->prim;
		  while (lnum!=NULL) {
		    switch (l->tipo) {
		    case lista_enteros :  
		      fprintf(fich_gen, "\tadd_tail_lent(%.0f, &%sv_%s[%d].%s);\n",
			      lnum->num.real, 
			      prefijo, aux->etds.instancia_de, aux->etds.cod.num_instancia, l->nombre);
		      break;
		      
		    case lista_reales  :  
		      fprintf(fich_gen, "\tadd_tail_lreal(%f, &%sv_%s[%d].%s);\n",
			      lnum->num.real, 
			      prefijo, aux->etds.instancia_de, aux->etds.cod.num_instancia, l->nombre);
		      break;
		      
		    case array:
		      fprintf(fich_gen, "\t%sv_%s[%d].%s[%d] = ",
			      prefijo, aux->etds.instancia_de, aux->etds.cod.num_instancia, l->nombre, j);
		      switch (l->tbase_array) {
		      case boolean:
		      case entero:
			fprintf(fich_gen, "%.0f;\n", lnum->num.real);
			break;
		      case real:
			fprintf(fich_gen, "%f;\n", lnum->num.real);
			break;
		      }
		      j++;
		      break;
		    }
		    lnum = lnum->sig;
		  }
		  fprintf(fich_gen,"\n");
		}
	      }
	    }
	    l=l->siguiente;
	  }
	  if (aux->etds.tipo!=definido) {
	    if (aux->etds.tipo==lista_enteros || aux->etds.tipo==lista_reales) {     
	      /* si es un tipo lista aunque no este inicializado se */
	      /* inicializa */
	      switch (aux->etds.tipo) {
	      case lista_enteros :  
		fprintf(fich_gen, "\tinicializa_lent(&%s%s, %d, %slnombre_%s, %s%s_elementos);\n",
			prefijo, aux->etds.nombre, 
			aux->etds.tam_lista, 
			prefijo, aux->etds.nombre,
			prefijo, aux->etds.nombre);
		break;
	      case lista_reales  :  
		fprintf(fich_gen, "\tinicializa_lreal(&%s%s, %d, %slnombre_%s, %s%s_elementos);\n",
			prefijo, aux->etds.nombre,
			aux->etds.tam_lista,
			prefijo, aux->etds.nombre,
			prefijo, aux->etds.nombre);
		break;
	      }
	    }
	    if (aux->etds.inicializado) {
	      if (aux->etds.tipo==lista_enteros || aux->etds.tipo==lista_reales || aux->etds.tipo == array) {     
		j = 0;
		lnum = aux->etds.ld->prim;
		while (lnum!=NULL) {
		  switch (aux->etds.tipo) {
		  case lista_enteros :  
		    fprintf(fich_gen,"\tadd_tail_lent(%.0f, &%s%s);\n",
			    lnum->num.real, prefijo, aux->etds.nombre);
		    break;
		  case lista_reales  :  
		    fprintf(fich_gen,"\tadd_tail_lreal(%f, &%s%s);\n",
			    lnum->num.real, prefijo, aux->etds.nombre);
		    break;
		    
		  case array:
		    fprintf(fich_gen, "\t%s%s[%d] = ", prefijo, aux->etds.nombre, j);
		    switch (aux->etds.tbase_array) {
		    case boolean:
		    case entero:
		      fprintf(fich_gen, "%.0f;\n", lnum->num.real);
		      break;
		    case real:
		      fprintf(fich_gen, "%f;\n", lnum->num.real);
		      break;
		    }
		    j++;
		  }
		  lnum = lnum->sig;
		}
		fprintf(fich_gen,"\n");
	      }
	      else {   /* si no es una lista */
		fprintf(fich_gen,"\t%s%s=",prefijo,aux->etds.nombre);
		switch (aux->etds.tipo) {
		case entero  :  
		  fprintf(fich_gen,"%d;\n",aux->etds.valor.entero);
		  break;
		case subrango:
		  fprintf(fich_gen,"%d;\n",aux->etds.valor.entero);
		  break;
		case real    :  
		  fprintf(fich_gen,"%f;\n",aux->etds.valor.real);
		  break;
		case caracter:  /* FALTA */
		  break;
		case boolean : 
		  if (aux->etds.valor.entero==0)
		    fprintf(fich_gen,"FALSE;\n");
		  else
		    fprintf(fich_gen,"TRUE;\n");
		  break;
		}
	      }
	    }
	  }
	}
	aux=aux->sig;
      }
    }
  fprintf(fich_gen,"}\n\n");
}

/********************** INICIALIZA_INSTANCIAS ************************/

void Inicializa_instancias() {
  EHTDS *aux;
  int i, j;
  LISTA_CAMPOS * l;
  LIS_NUM * lnum;
  
  fprintf(fich_gen,"/*----------------- INICIALIZACION DE INSTANCIAS -------------------*/\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"void %sInicializa_instancias()\n",prefijo);
  fprintf(fich_gen,"{\n");
  for(i=0;i<HASH_SIZE;i++)
    if(tds[i]!=NULL) {
      aux=tds[i];
      while(aux!=NULL) {
	if (aux->etds.clvar==instancia) {
	  fprintf(fich_gen,"\t/* inicializando instancia %s */\n",aux->etds.nombre );
	  /* FALTA: los campos que son de tipo definido */
	  l=aux->etds.lc;
	  while(l!=NULL) {
	    if (l->tipo==lista_enteros || l->tipo==lista_reales) {
	      /* si es un tipo lista aunque no este inicializado se */
	      /* inicializa */
	      switch (l->tipo) {
	      case lista_enteros : 			
		fprintf(fich_gen,"\tinicializa_lent(&%si_%s[%d].%s, %d, %slnombre_%s_%s, %si_%s[%d].%s_elementos);\n",
			prefijo,aux->etds.instancia_de,aux->etds.cod.num_instancia,l->nombre,
			l->tam_lista, 
			prefijo, aux->etds.nombre, l->nombre,
			prefijo,aux->etds.instancia_de,aux->etds.cod.num_instancia,l->nombre);
		break;
	      case lista_reales  :
		fprintf(fich_gen,"\tinicializa_lreal(&%si_%s[%d].%s, %d, %slnombre_%s_%s, %si_%s[%d].%s_elementos);\n",
			prefijo,aux->etds.instancia_de,aux->etds.cod.num_instancia,l->nombre,
			l->tam_lista, 
			prefijo, aux->etds.nombre, l->nombre,
			prefijo,aux->etds.instancia_de,aux->etds.cod.num_instancia,l->nombre);
		break;
	      }
	    }
	    if (l->inicializado) {
	      if (l->tipo!=lista_enteros && l->tipo!=lista_reales && l->tipo != array) { /* si no es de tipo lista */
		fprintf(fich_gen,"\t%si_%s[%d].%s=",prefijo,aux->etds.instancia_de,aux->etds.cod.num_instancia,l->nombre);
		switch (l->tipo) {
		case entero  :
		  fprintf(fich_gen,"%d;\n",l->valor.entero);
		  break;
		case subrango:
		  fprintf(fich_gen,"%d;\n",l->valor.entero);
		  break;
		case real    :
		  fprintf(fich_gen,"%f;\n",l->valor.real);
		  break;
		case caracter:  /* FALTA */
		  break;
		case boolean :  
		  if (l->valor.entero==0) fprintf(fich_gen,"FALSE;\n");
		  else fprintf(fich_gen,"TRUE;\n");
		  break;
		}
	      } 
	      else {
		if (l->tipo==lista_enteros || l->tipo==lista_reales || l->tipo == array) {
		  lnum = l->ld->prim;
		  j = 0;
		  while (lnum!=NULL) {
		    switch (l->tipo) {
		    case lista_enteros :
		      fprintf(fich_gen,"\tadd_tail_lent(%.0f, &%si_%s[%d].%s);\n",
			      lnum->num.real, prefijo,aux->etds.instancia_de,aux->etds.cod.num_instancia,
			      l->nombre);
		      break;

		    case lista_reales  :  
		      fprintf(fich_gen,"\tadd_tail_lreal(%f, &%si_%s[%d].%s);\n",
			      lnum->num.real, prefijo,aux->etds.instancia_de,aux->etds.cod.num_instancia,
			      l->nombre);
		      break;

		    case array:
		      fprintf(fich_gen, "\t%si_%s[%d].%s[%d] = ", 
			      prefijo, aux->etds.instancia_de, aux->etds.cod.num_instancia, l->nombre, j);
		      switch (l->tbase_array) {

		      case entero:
		      case boolean:
			fprintf(fich_gen, "%.0f;\n", lnum->num.real);
			break;

		      case real:
			fprintf(fich_gen, "%f;\n", lnum->num.real);
			break;
		      }
		      j++;
		    }
		    lnum = lnum->sig;
		  }
		  fprintf(fich_gen,"\n");
		}
	      }
	    }
	    l=l->siguiente;
	  }
	  if (aux->etds.tipo!=definido) {
	    if (aux->etds.tipo==lista_enteros || aux->etds.tipo==lista_reales) {
	      /* si es un tipo lista aunque no este inicializado se inicializa */
	      switch (aux->etds.tipo) {	       
	      case lista_enteros :  
		fprintf(fich_gen,"\tinicializa_lent(&%s%s, %d, %slnombre_%s, %s%s_elementos);\n",
			prefijo, aux->etds.nombre, 
			aux->etds.tam_lista,
			prefijo, aux->etds.nombre,
			prefijo, aux->etds.nombre);
		break;
	      case lista_reales  :  
		fprintf(fich_gen,"\tinicializa_lreal(&%s%s, %d, %slnombre_%s, %s%s_elementos);\n",
			prefijo, aux->etds.nombre, 
			aux->etds.tam_lista,
			prefijo, aux->etds.nombre,
			prefijo, aux->etds.nombre);
		break;
	      }
	    }


	    if (aux->etds.tipo==lista_enteros || aux->etds.tipo==lista_reales || aux->etds.tipo == array) {
	      if (aux->etds.ld == NULL) {
		lnum = NULL;
	      }
	      else {
		lnum = aux->etds.ld->prim;
	      }
	      
	      j = 0;
	      while (lnum!=NULL) {
		switch (aux->etds.tipo) {
		case lista_enteros :
		  fprintf(fich_gen,"\tadd_tail_lent(%.0f, &%s%s);\n",
			  lnum->num.real,prefijo,aux->etds.nombre);
		  break;
		case lista_reales  :  
		  fprintf(fich_gen,"\tadd_tail_lreal(%f, &%s%s);\n",
			  lnum->num.real,prefijo,aux->etds.nombre);
		  break;

		case array:
		  fprintf(fich_gen, "\t%s%s[%d] = ", 
			  prefijo, aux->etds.nombre, j);
		  switch (aux->etds.tbase_array) {
		  case entero:
		  case boolean:
		    fprintf(fich_gen, "%.0f;\n", lnum->num.real);
		    break;
		  case real:
		    fprintf(fich_gen, "%f;\n", lnum->num.real);
		    break;
		  }
		  j++;
		}
		lnum = lnum->sig;
	      }
	      fprintf(fich_gen,"\n");
	    } 
	    else {   /* si no es una lista */
	      if (aux->etds.inicializado) {
		fprintf(fich_gen,"\t%s%s=",prefijo,aux->etds.nombre);
		switch (aux->etds.tipo) {
		case entero  :  fprintf(fich_gen,"%d;\n",aux->etds.valor.entero);
		  break;
		case subrango:  fprintf(fich_gen,"%d;\n",aux->etds.valor.entero);
		  break;
		case real    :  fprintf(fich_gen,"%f;\n",aux->etds.valor.real);
		  break;
		case caracter:  /* FALTA */
		  break;
		case boolean :
		  if (aux->etds.valor.entero==0) fprintf(fich_gen,"FALSE;\n");
		  else fprintf(fich_gen,"TRUE;\n");
		  break;
		}
	      }
	    }
	  }
	}
	aux=aux->sig;
      }
    }
  fprintf(fich_gen,"}\n\n");
}


/**************************** LEER_DEL_KDM ******************************/
void Genera_lectura_del_kdm() {
  int i;
  EHTDS *reg_tds;
  LISTA_CAMPOS *campo;
  REGTIPO *clase;

  fprintf(fich_gen, "/*----------------- LECTURA DE INSTANCIAS KDM -------------------*/\n");
  fprintf(fich_gen, "/*Esta funcin lee todos los datos de las instancias Arlips declaradas\n");
  fprintf(fich_gen, "en el KDM en tiempo de ejecucin, para la actualizacin de los valores\n");
  fprintf(fich_gen, "con los que debe trabajar Arlips*/\n\n");
  fprintf(fich_gen, "void %sLee_kdm() {\n", prefijo);
  
  for (i=0; i<HASH_SIZE; i++) {
    if (tds[i] != NULL) {
      reg_tds = tds[i];
      while (reg_tds != NULL) {
	if ((reg_tds->etds.clvar == instancia) &&
	    (reg_tds->etds.from_kdm)) {	    	  
	  fprintf(fich_gen, "\n\t //Actualizando la instancia %s\n", reg_tds->etds.nombre);
	  campo = reg_tds->etds.lc;
	  while (campo != NULL) {
	    //1. Se usan como condicin de una regla.
	    //2. Se usan como accin de una regla.
	    //3. Existe una variable patrn de la clase de la instancia.
	    if (campo->campo_buscado) {
	      //El campo en cuestin de la instancia se usa por alguna regla,
	      //o bien existe una var patrn que usa ese campo.
	      fprintf(fich_gen, "\t %si_%s[%d].%s = ", prefijo, reg_tds->etds.instancia_de, 
		      reg_tds->etds.cod.num_instancia, campo->nombre);
	      switch (campo->tipo) {
	      case entero:
		fprintf(fich_gen, "*(int*)");
		break;
	      case real:
		fprintf(fich_gen, "*(float*)");
		break;
	      default:
		fprintf(stderr, "COMPILADOR INCOMPLETO!!!\n");
		fprintf(stderr, "Se ha detectado una instancia del KDM que no es de un tipo esperado.\n");
		fprintf(stderr, "Revisar la funcin Genera_lectura_del_kdm del fichero genera.c.\n");
		fprintf(stderr, "Revisar el manejo de los tipos en la TDS.\n");
		fprintf(stderr, "Revisar los tipos admitidos por el parser del KDM.\n");
		fprintf(stderr, "El tipo no contemplado adecuadamente es el %d.\n", campo->tipo);
		break;
	      }
	      fprintf(fich_gen, "%s(%s->%s);\n", f_GetKDM, reg_tds->etds.nombre, campo->nombre);
	    } //end si el campo se usa en reglas
	    else fprintf(fich_gen, "\t//Ninguna regla usa el campo %s. No se actualizar...\n", 
			 campo->nombre);
	    campo = campo->siguiente;
	  } // while de todos los campos de una instancia kdm
	} // if instancia del kdm
	reg_tds = reg_tds->sig;
      } //while de un elemento de la hash de la TDS
    } //if entrada de la hash tiene algn elemento
  } //for de entradas hash
  fprintf (fich_gen, "}\n\n");
} //Genera_lectura_del_kdm()


/**************************** ESCRIBIR_EN_KDM ******************************/
void Genera_escritura_en_kdm() {
  int i;
  EHTDS *reg_tds;
  LISTA_CAMPOS *campo;
  
  fprintf(fich_gen, "/*----------------- ESCRITURA DE INSTANCIAS EN KDM -------------------*/\n");
  fprintf(fich_gen, "/*Esta funcin escribe en el KDM todos los datos importados del KDM que\n"); 
  fprintf(fich_gen, "HAYAN SIDO USADOS por alguna regla de este sistema de prodccin Arlips.\n");
  fprintf(fich_gen, "Actualiza algunos valores del KDM*/\n");
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "void %sEscribe_kdm() {\n", prefijo);
  
  for (i=0; i<HASH_SIZE; i++) {
    if (tds[i] != NULL) {
      reg_tds = tds[i];
      while (reg_tds != NULL) {
	if ((reg_tds->etds.clvar == instancia) &&
	    (reg_tds->etds.from_kdm)) {	    
	  campo = reg_tds->etds.lc;
	  fprintf(fich_gen, "\n\t //Escribiendo la instancia %s en el KDM\n", reg_tds->etds.nombre);
	  while (campo != NULL) {
	    if (campo->campo_buscado) { 
	      //DECISIN: Debe escribir campos que no cambian en el sp?
	      //Todos los campos mencionados por alguna regla, o afectados por 
	      //alguna variable patrn se escriben.

	      //El campo en cuestin se usa por alguna regla

	      //Imprime KDM_Set_Value(inst_kdm->campo, PREFi_arlClass[x].slot);
	      fprintf(fich_gen, "\t %s(%s->%s,  %si_%s[%d].%s);\n",    
		      f_SetKDM, reg_tds->etds.nombre, campo->nombre, 
		      prefijo, reg_tds->etds.instancia_de, reg_tds->etds.cod.num_instancia, campo->nombre);
	    } //end si el campo se usa en reglas
	    else fprintf(fich_gen, "\t//Ninguna regla usa el campo %s. No se actualizar...\n", 
			 campo->nombre);
	    campo = campo->siguiente;
	  } // while de todos los campos de una instancia kdm
	} // if instancia del kdm
	reg_tds = reg_tds->sig;
      } //while de un elemento de la hash de la TDS
    } //if entrada de la hash tiene algn elemento
  } //for de entradas hash
  fprintf (fich_gen, "}\n\n");
} //Genera_escritura_en_kdm
			


/********************** GENERA_FUNCION_PRINCIPAL ************************/

void Genera_funcion_principal()
{
  if (!nomain) {
    fprintf(fich_gen, "void %s(int %slimit, int %ssteps) {\n", nom_sp, prefijo, prefijo);
    fprintf(fich_gen, "  int %spaso = 0;\n", prefijo);
    fprintf(fich_gen, "  int %siteracion = 0;\n", prefijo);
    fprintf(fich_gen, "  int %slong_cadena = 20;\n", prefijo);
    fprintf(fich_gen, "  char %scadena[%slong_cadena];\n", prefijo, prefijo);
  }
  else {
    fprintf(fich_gen, "void %s() {\n", nom_sp);
  }

  if (log_) {
    if (artis_mode) {
      fprintf(fich_gen, "#ifndef MODULE\n");
    }
    fprintf(fich_gen, "  if (!%slog_inicia_log()) return;\n", prefijo);
    if (artis_mode) {
      fprintf(fich_gen, "#endif\n");
    }
  }

  fprintf(fich_gen, "  %sInicializa_estructuras();\n", prefijo);
  fprintf(fich_gen, "  %sInicializa_agenda();\n", prefijo);
  fprintf(fich_gen, "  %sInicializa_instancias();\n", prefijo);
  fprintf(fich_gen, "  %sInicializa_variables();\n", prefijo);
  if (artis_mode) 
    fprintf(fich_gen,"  %sLee_kdm();\n", prefijo);
  fprintf(fich_gen,"  %sReset();\n",prefijo);
  
  fprintf(fich_gen, "  while (!%sagenda_vacia && !%shalt) {\n", prefijo, prefijo);
  if (!nomain) {
    fprintf(fich_gen, "    if (%slimit) {\n", prefijo);
    fprintf(fich_gen, "      if (%siteracion == %slimit) {\n", prefijo, prefijo);
    if (log_) {
      fprintf(fich_gen, "        %slog_end_log(%selimit);\n", prefijo, prefijo);
    }
    fprintf(fich_gen, "        exit(1);\n");
    fprintf(fich_gen, "      }\n");
    fprintf(fich_gen, "      else {\n");
    fprintf(fich_gen, "        %siteracion++;\n", prefijo);
    fprintf(fich_gen, "      }\n");    
    fprintf(fich_gen, "    }\n\n");
    
    fprintf(fich_gen, "    if (%ssteps) {\n", prefijo);
    fprintf(fich_gen, "      if (%spaso < %ssteps) {\n", prefijo);
    fprintf(fich_gen, "        %spaso++;\n", prefijo);
    fprintf(fich_gen, "      }\n");
    fprintf(fich_gen, "      else {\n");
    fprintf(fich_gen, "        fprintf(stdout, \"[q]uit, [Intro] or ####>\");\n");
    fprintf(fich_gen, "        fgets(%scadena, %slong_cadena, stdin);\n", prefijo);
    fprintf(fich_gen, "        switch(%scadena[0]) {\n", prefijo);
    fprintf(fich_gen, "        case 'q':\n");
    fprintf(fich_gen, "        case 'Q':\n");
    if (log_) 
      fprintf (fich_gen, "          %slog_end_log(%sequit);\n", prefijo, prefijo);
    fprintf(fich_gen, "          exit(1);\n");
    fprintf(fich_gen, "          break;\n");
    fprintf(fich_gen, "        case '0': case '1': case '2': case '3': case '4':\n");
    fprintf(fich_gen, "        case '5': case '6': case '7': case '8': case '9':\n");
    fprintf(fich_gen, "          sscanf(%scadena, \"%s\", &%ssteps);\n", prefijo, "%d", prefijo);
    fprintf(fich_gen, "        default:\n");
    fprintf(fich_gen, "          %spaso = 1;\n", prefijo);
    fprintf(fich_gen, "          break;\n");
    fprintf(fich_gen, "        }\n");
    fprintf(fich_gen, "      }\n");
    fprintf(fich_gen, "    }\n");
  }
  
  fprintf(fich_gen, "    %sStep();\n" ,prefijo);
  if (log_) { 
    fprintf(fich_gen, "    if (!%sagenda_vacia)\n", prefijo);
    fprintf(fich_gen, "      %slog_end_fire();\n", prefijo);
  }

  fprintf(fich_gen, "  }\n");

  if (artis_mode)
    fprintf(fich_gen, "  %sEscribe_kdm();\n", prefijo);
  if (log_) {
    fprintf(fich_gen, "  if (%shalt) \n", prefijo);
    fprintf(fich_gen, "    %slog_end_log(%sehalt);\n", prefijo, prefijo);
    fprintf(fich_gen, "  else\n");
    fprintf(fich_gen, "    %slog_end_log(%seempty);\n", prefijo, prefijo);
  }
  fprintf(fich_gen,"}\n");
  fprintf(fich_gen,"\n");
}

/********************** GENERA_FUNCION_MAIN ************************/

void Genera_funcion_main()
{

  //Recordar que todo esto no se genera cuando se funciona en modo ARTIS
  fprintf(fich_gen, "int main(int argc, char *argv[]) {\n",nom_sp);  
  fprintf(fich_gen, "  int %ssteps = 0;\n", prefijo);
  fprintf(fich_gen, "  int %slimit = 0;\n", prefijo);
  fprintf(fich_gen, "  int i;\n\n");

  fprintf(fich_gen, "  for (i = 1; i < argc; i++) {\n");
  fprintf(fich_gen, "    if (!%slimit && !strcmp(argv[i], \"-limit\")) {\n", prefijo);
  fprintf(fich_gen, "      i++;\n");
  fprintf(fich_gen, "      %slimit = atoi(argv[i]);\n", prefijo);
  fprintf(fich_gen, "      if (%slimit <= 0) {\n", prefijo);
  fprintf(fich_gen, "        fprintf(stderr, \"Uso incorrecto. Tras -limit se esperaba un entero positivo\\n\");\n");
  fprintf(fich_gen, "        exit(-1);\n");
  fprintf(fich_gen, "      }\n");
  fprintf(fich_gen, "    }\n");
  fprintf(fich_gen, "    else if (!%ssteps && !strcmp(argv[i], \"-steps\")) {\n", prefijo);
  fprintf(fich_gen, "      i++;\n");
  fprintf(fich_gen, "      %ssteps = atoi(argv[i]);\n", prefijo);
  fprintf(fich_gen, "      if (%ssteps <= 0) {\n", prefijo);
  fprintf(fich_gen, "        fprintf(stderr, \"Uso incorrecto. A -steps debe seguirle un valor entero positivo\\n\");\n");
  fprintf(fich_gen, "        exit(-1);\n");
  fprintf(fich_gen, "      }\n");
  fprintf(fich_gen, "    }\n");
  fprintf(fich_gen, "    else {\n");
  fprintf(fich_gen, "      fprintf(stderr, \"%s %s\\n\", argv[0], argv[i]);\n", "%s", "%s");
  fprintf(fich_gen, "      fprintf(stderr, \"Uso:\\n\");\n");
  fprintf(fich_gen, "      fprintf(stderr, \"%s [-limit l] [-steps s]\\n\", argv[0]);\n", "%s");
  fprintf(fich_gen, "      fprintf(stderr, \"-limit l establece en l el mximo nmero de reglas a disparadar.\\n\");\n");
  fprintf(fich_gen, "      fprintf(stderr, \"-steps s establece en s el nmero de reglas a ejecutar antes de la sguiente pausa:\\n\");\n");
  fprintf(fich_gen, "      fprintf(stderr, \"  Tras la ejecucin de s pasos el sistema de produccin espera:\\n\");\n");
  fprintf(fich_gen, "      fprintf(stderr, \"  q fuerza la salida inmediata-\\n\");\n");
  fprintf(fich_gen, "      fprintf(stderr, \"  un entero positivo sobreescribe el nmero de reglas a ejecutar entre pausas.\\n\");\n");
  fprintf(fich_gen, "      fprintf(stderr, \"  cualquier otra tecla continua la ejecucin del sistema.\\n\");\n");
  fprintf(fich_gen, "      exit(-1);\n");
  fprintf(fich_gen, "    }\n");
  fprintf(fich_gen, "  }\n");

  fprintf(fich_gen, "  %s(%slimit, %ssteps);\n",nom_sp, prefijo, prefijo);
  fprintf(fich_gen, "}\n");
}

/************************** GENERA_REGLAS_IZQ ****************************/

void Genera_reglas_izq()
{
  EHREGLA *aux;
  int i;
  LISTA_CONDIC *lc;
  int cond_simples;     /* numero de condiciones simples */
  int num_cond;         /* numero de condiciones */
  char nombre_condicion_regla[100];  /* almacena la cadena que identifica las estructuras */
  int n,pos,num_vp;
  LISTA *param_regla;
  LISTA *param_cond;
  LISTA *param_aux;
  LISTA_OBJ * lo, *lo_aux;
  REGTIPO *auxptro;
  LISTA_CONDIC *resto_cond;
  char cad_exp[255];

  fprintf(fich_gen,"/*--------------- PARTES IZQUIERDA DE LAS REGLAS ----------------*/\n");

  for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
           if (!aux->eregla.tiene_copias) {
             fprintf(fich_gen,"void Iz_%s%s()\n",prefijo,aux->eregla.nombre);
             fprintf(fich_gen,"{\n");
             fprintf(fich_gen,"\tint ncamb;\n");
             fprintf(fich_gen,"\t%sREGAGENDA r;\n",prefijo);
             /* ahora se declaran todas las vpe de la regla */
             param_regla = aux->eregla.param;
             while (param_regla!=NULL) {
                      fprintf(fich_gen,"\tint  %s;\n",param_regla->nombre);
                      param_regla=param_regla->sig;
             }
             fprintf(fich_gen,"\n");

             lc = aux->eregla.lcond;
             num_cond= 0;
             cond_simples=0;
             /* primero se generan las condiciones simples */
             while (lc != NULL) {
                if (lc->num_vpe==0) cond_simples++;
                num_cond++;
                lc=lc->sig;
             }
             if (cond_simples>0) {
                fprintf(fich_gen,"\tif (");
                lc = aux->eregla.lcond;
                n = 0;
                while (lc != NULL) {
                   if (lc->num_vpe==0) {
                     if (n>0) fprintf(fich_gen," && ");
                     if (lc->es_orig)
                          sprintf(nombre_condicion_regla,"cond%d_%s%s",lc->num_condic,prefijo,aux->eregla.nombre);
                     else
                          sprintf(nombre_condicion_regla,"cond%d_%s%s",lc->cond_orig,prefijo,lc->regla_orig);
                     fprintf(fich_gen,"(activ_%s)",nombre_condicion_regla);
                     n++;
                   }
                   lc=lc->sig;
                }
                fprintf(fich_gen,") {\n");
                if (num_cond==cond_simples) {
                   fprintf(fich_gen,"\t\tr.num_regla = %d;\n",aux->eregla.num_regla);
                   fprintf(fich_gen,"\t\tr.prioridad = ");
                   if (aux->eregla.expr_prio!=NULL) {
                     Genera_exp(aux->eregla.expr_prio,cad_exp,aux->eregla.lobj_cond);
                     fprintf(fich_gen,"%s;\n",cad_exp);
                     if (!es_constante(aux->eregla.expr_prio)) { /* comprobacion dinamica de rango */
                        fprintf(fich_gen,"\t\tif (r.prioridad>10 || r.prioridad<-10) {\n");
			/* EVR FEB05 */
                        if (!noprints)  fprintf(fich_gen,"\t\t\tprintf(\"ERROR: Prioridad fuera de rango\\n\");\n");
			fprintf(fich_gen,"\t\t\texit(-1);\n");
                        fprintf(fich_gen,"\t\t}\n");
                     }
                   } else fprintf(fich_gen,"0;\n"); /* prioridad por defecto */
                   fprintf(fich_gen,"\t\t%sinsregla_agenda(r);\n",prefijo);

                }
             }

             /* ahora se comprueban los cambios del resto de condiciones */
             lc = aux->eregla.lcond;
             n = 0;
             while (lc != NULL) {
                 if (lc->num_vpe>0) {
                   if (lc->es_orig)
                        sprintf(nombre_condicion_regla,"cond%d_%s%s",lc->num_condic,prefijo,aux->eregla.nombre);
                   else
                        sprintf(nombre_condicion_regla,"cond%d_%s%s",lc->cond_orig,prefijo,lc->regla_orig);
                   fprintf(fich_gen,"\t\tif (num_cambios_cond%d_%s%s>0) {\n",lc->num_condic,prefijo,aux->eregla.nombre);
                   fprintf(fich_gen,"\t\t\tncamb=0;\n");
                   fprintf(fich_gen,"\t\t\tr.num_regla = %d;\n",aux->eregla.num_regla);
                   fprintf(fich_gen,"\t\t\tr.prioridad = ");
                   if (aux->eregla.expr_prio!=NULL) {
                      Genera_exp(aux->eregla.expr_prio,cad_exp,aux->eregla.lobj_cond);
                      fprintf(fich_gen,"%s;\n",cad_exp);
                     if (!es_constante(aux->eregla.expr_prio)) {  /* comprobacion dinamica de rango */
                        fprintf(fich_gen,"\t\t\tif (r.prioridad>10 || r.prioridad<-10) {\n");
                        /* EVR FEB05 */
			if (!noprints) fprintf(fich_gen,"\t\t\t\tprintf(\"ERROR: Prioridad fuera de rango\\n\");\n");
                        fprintf(fich_gen,"\t\t\t\texit(-1);\n");
                        fprintf(fich_gen,"\t\t\t}\n");
                     }
                   } else fprintf(fich_gen,"0;\n"); /* prioridad por defecto */
                   fprintf(fich_gen,"\t\t\twhile (ncamb<num_cambios_cond%d_%s%s) {\n",lc->num_condic,prefijo,aux->eregla.nombre);

                   param_cond = lc->param;
                   n = 0;
                   while (param_cond!=NULL) {
                       lo=busca_obj(lc->lobj_cond,param_cond->nombre);
                       if (lo->cuantificador==existe) {
                          fprintf(fich_gen,"\t\t\t    %s=cambios_%s[ncamb].instancias[%d];\n",param_cond->nombre,nombre_condicion_regla,n);
                          /* hay que ver que posicion ocupa esta vp en la lista de vpe de la regla */
                          param_regla= aux->eregla.param;
                          pos=0;
                          while (strcmp(param_regla->nombre,param_cond->nombre)) {
                                lo_aux = busca_obj(aux->eregla.lobj_cond,param_regla->nombre);
                                if (lo_aux->cuantificador==existe) pos++;
                                /* punto de comprobacion */
                                if (param_regla==NULL) {
                                     fprintf(fich_gen,"ERROR GRAVE en Genera_reglas_izq\n");
                                     exit(-1);
                                }
                                param_regla=param_regla->sig;
                          }
                          fprintf(fich_gen,"\t\t\t    r.instancias[%d]=%s;\n",pos,param_cond->nombre);
                          n++;
                       }
                       param_cond=param_cond->sig;
                   }

                   if (lc->es_orig) {
                     fprintf(fich_gen,"\t\t\t    camb_%s",nombre_condicion_regla);
                     param_cond = lc->param;
                     while (param_cond!=NULL) {
                       lo=busca_obj(lc->lobj_cond,param_cond->nombre);
                       if (lo->cuantificador==existe) {
                          fprintf(fich_gen,"[%s]",param_cond->nombre);
                       }
                       param_cond=param_cond->sig;
                     }
                     fprintf(fich_gen,"=CELDA_VACIA;\n");
                   }
                   fprintf(fich_gen,"\t\t\t    if (cambios_%s[ncamb].tipo_cambio==desactivacion) {\n",nombre_condicion_regla);

                   param_regla = aux->eregla.param;
                   while (param_regla!=NULL) {
                       lo=busca_obj(aux->eregla.lobj_cond,param_regla->nombre);
                       if (lo->cuantificador==existe) {
                           lo_aux = busca_obj(lc->lobj_cond,param_regla->nombre);
                           if (lo_aux==NULL) {  /* no esta en la condicion */
                              auxptro = busca_en_tipos(lo->instancia_de);
                              fprintf(fich_gen,"\t\t\t\tfor(%s=0;%s<%d; %s++)\n",param_regla->nombre,param_regla->nombre,auxptro->num_instancias,param_regla->nombre);
                           }
                       }
                       param_regla=param_regla->sig;
                   }

                   fprintf(fich_gen,"\t\t\t\tif (%sR%d",prefijo,aux->eregla.num_regla);
                   param_regla = aux->eregla.param;
                   while (param_regla!=NULL) {
                       lo=busca_obj(aux->eregla.lobj_cond,param_regla->nombre);
                       if (lo->cuantificador==existe) {
                              fprintf(fich_gen,"[%s]",param_regla->nombre);
                       }
                       param_regla=param_regla->sig;
                   }
                   fprintf(fich_gen,">0) {\n");

                   n=0;
                   param_regla = aux->eregla.param;
                   while (param_regla!=NULL) {
                       lo=busca_obj(aux->eregla.lobj_cond,param_regla->nombre);
                       if (lo->cuantificador==existe) {
                           lo_aux = busca_obj(lc->lobj_cond,param_regla->nombre);
                           if (lo_aux==NULL) {  /* no esta en la condicion */
                              fprintf(fich_gen,"\t\t\t\t\tr.instancias[%d]=%s;\n",n,param_regla->nombre);
                           }
                           n++;
                       }
                       param_regla=param_regla->sig;
                   }
                   fprintf(fich_gen,"\t\t\t\t\t%selimregla_agenda(r);\n",prefijo);
                   fprintf(fich_gen,"\t\t\t\t}\n");

                   fprintf(fich_gen,"\t\t\t    } else {  /* el cambio es una activacion */\n");

                   param_cond = lc->param;
                   resto_cond = aux->eregla.lcond;
                   while (resto_cond!=NULL) {   /* se recorren el resto de condiciones */
                         if (lc->num_condic!=resto_cond->num_condic && resto_cond->num_vpe>0) { /* si no es la actual y no es simple*/
                             /* si esta condicion tiene las mismas vp que la actual */
                             /* entonces se verifica primero que esta activada      */
                             param_aux = resto_cond->param;
                             num_vp = 0;
                             while (param_aux!=NULL) {
                                 if (busca_lista(param_cond,param_aux->nombre)) num_vp++;
                                 param_aux=param_aux->sig;
                             }
                             if (num_vp==resto_cond->num_vpe) { /* si todas estan ya en la actual */
                                if (resto_cond->es_orig)
                                      fprintf(fich_gen,"\t\t\t\tif (activ_cond%d_%s%s",resto_cond->num_condic,prefijo,aux->eregla.nombre);
                                else
                                      fprintf(fich_gen,"\t\t\t\tif (activ_cond%d_%s%s",resto_cond->cond_orig,prefijo,resto_cond->regla_orig);
                                param_aux = resto_cond->param;
                                num_vp = 0;
                                while (param_aux!=NULL) {
                                   lo=busca_obj(resto_cond->lobj_cond,param_aux->nombre);
                                   if (lo->cuantificador==existe) fprintf(fich_gen,"[%s]",param_aux->nombre);
                                   param_aux=param_aux->sig;
                                }
                                fprintf(fich_gen,")\n");
                             }

                         }
                         resto_cond = resto_cond->sig;
                   }

                   /* ahora se generan los bucles para el resto de vp */
                   /* se recorren las vp de la regla y se cojen aquellas */
                   /* que no esten en la condicion actual */
                   param_regla = aux->eregla.param;
                   while (param_regla!=NULL) {
                       lo=busca_obj(aux->eregla.lobj_cond,param_regla->nombre);
                       if (lo->cuantificador==existe) {
                           lo_aux = busca_obj(lc->lobj_cond,param_regla->nombre);
                           if (lo_aux==NULL) {  /* no esta en la condicion */
                              auxptro = busca_en_tipos(lo->instancia_de);
                              fprintf(fich_gen,"\t\t\t\tfor(%s=0;%s<%d; %s++)\n",param_regla->nombre,param_regla->nombre,auxptro->num_instancias,param_regla->nombre);
                           }
                       }
                       param_regla=param_regla->sig;
                   }

                   /* se vuelven a recorrer las que no tienen vp coincidentes */
                   resto_cond = aux->eregla.lcond;
                   while (resto_cond!=NULL) {   /* se recorren el resto de condiciones */
                         if (lc->num_condic!=resto_cond->num_condic) { /* si no es la actual */
                             /* si esta condicion tiene las mismas vp que la actual */
                             /* entonces se verifica primero que esta activada      */
                             param_aux = resto_cond->param;
                             num_vp = 0;
                             while (param_aux!=NULL) {
                                 if (busca_lista(param_cond,param_aux->nombre)) num_vp++;
                                 param_aux=param_aux->sig;
                             }
                             if (num_vp!=resto_cond->num_vpe) { /* si no estan todas en la actual */
                                if (resto_cond->es_orig)
                                      fprintf(fich_gen,"\t\t\t\tif (activ_cond%d_%s%s",resto_cond->num_condic,prefijo,aux->eregla.nombre);
                                else
                                      fprintf(fich_gen,"\t\t\t\tif (activ_cond%d_%s%s",resto_cond->cond_orig,prefijo,resto_cond->regla_orig);
                                param_aux = resto_cond->param;
                                num_vp = 0;
                                while (param_aux!=NULL) {
                                   lo=busca_obj(resto_cond->lobj_cond,param_aux->nombre);
                                   if (lo->cuantificador==existe) fprintf(fich_gen,"[%s]",param_aux->nombre);
                                   param_aux=param_aux->sig;
                                }
                                fprintf(fich_gen,")\n");
                             }

                         }
                         resto_cond = resto_cond->sig;
                   }
                   fprintf(fich_gen,"\t\t\t\tif (%sR%d",prefijo,aux->eregla.num_regla);
                   param_regla = aux->eregla.param;
                   while (param_regla!=NULL) {
                       lo=busca_obj(aux->eregla.lobj_cond,param_regla->nombre);
                       if (lo->cuantificador==existe) {
                              fprintf(fich_gen,"[%s]",param_regla->nombre);
                       }
                       param_regla=param_regla->sig;
                   }
                   fprintf(fich_gen,"==0) {\n");

                   n=0;
                   param_regla = aux->eregla.param;
                   while (param_regla!=NULL) {
                       lo=busca_obj(aux->eregla.lobj_cond,param_regla->nombre);
                       if (lo->cuantificador==existe) {
                           lo_aux = busca_obj(lc->lobj_cond,param_regla->nombre);
                           if (lo_aux==NULL) {  /* no esta en la condicion */
                              fprintf(fich_gen,"\t\t\t\t\tr.instancias[%d]=%s;\n",n,param_regla->nombre);
                           }
                           n++;
                       }
                       param_regla=param_regla->sig;
                   }
                   fprintf(fich_gen,"\t\t\t\t\t%sinsregla_agenda(r);\n",prefijo);
                   fprintf(fich_gen,"\t\t\t\t}\n");
                   fprintf(fich_gen,"\t\t\t    }\n");
                   fprintf(fich_gen,"\t\t\t    ncamb++;\n");
                   fprintf(fich_gen,"\t\t\t}\n");
                   fprintf(fich_gen,"\t\t\tnum_cambios_cond%d_%s%s=0;\n",lc->num_condic,prefijo,aux->eregla.nombre);
                   fprintf(fich_gen,"\t\t}\n");
                 }
                 lc=lc->sig;
             }
             if (cond_simples>0) {
                fprintf(fich_gen,"\t} else {\n");
                fprintf(fich_gen,"\t\t%selim_agenda(%d);\n",prefijo,aux->eregla.num_regla);
		/*                lc = aux->eregla.lcond;
		   while (lc != NULL) {
                   if (lc->num_vpe>0) {
                     sprintf(nombre_condicion_regla,"cond%d_%s",lc->num_condic,aux->eregla.nombre);
                     fprintf(fich_gen,"\t\tnum_cambios_%s=0;\n",nombre_condicion_regla);
                    }
                    lc=lc->sig;
                }
                */
                fprintf(fich_gen,"\t}\n");
             }
             fprintf(fich_gen,"\tLIz_%s%s=FALSE;\n",prefijo,aux->eregla.nombre);
             fprintf(fich_gen,"}\n");
           }
         fprintf(fich_gen,"\n");
         aux=aux->sig;
        }
     }
}

/************************ GENERA_VARIABLES_REGLAS_IZQ ********************/
void Genera_variables_reglas_izq()
{
  EHREGLA *aux;
  int i;

  fprintf(fich_gen,"/* Variables auxiliares para las partes izquierda de las reglas */\n");
  for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
           if (!aux->eregla.tiene_copias) {
             fprintf(fich_gen,"int LIz_%s%s=FALSE;\n",prefijo,aux->eregla.nombre);
           }
         aux=aux->sig;
        }
     }
}

/*************************** GENERA_CAB_AGENDA *****************************/
void Genera_cab_agenda()
{
  EHREGLA *aux;
  LISTA_OBJ *lo;
  int i;
  int num_vpe;


  fprintf(fich_gen, "#ifndef _AGENDA_%s_H_\n", nom_sp);
  fprintf(fich_gen, "#define _AGENDA_%s_H_\n\n", nom_sp);

  for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
             num_vpe=0;
             lo=aux->eregla.lobj_acc;
             while (lo!=NULL) {
                   if (lo->cuantificador==existe) num_vpe++;
                   lo=lo->sig;
             }
             if (num_vpe>num_maxvp) num_maxvp = num_vpe;
             aux=aux->sig;
         }
     }

  fprintf(fich_gen,"enum tipo_estrategia{profundidad=0, anchura};\n");
  fprintf(fich_gen,"typedef struct {\n");
  fprintf(fich_gen,"\tint num_regla;\n");
  fprintf(fich_gen,"\tchar prioridad;\n");
  if (num_maxvp>0) fprintf(fich_gen,"\tint instancias[%d];\n",num_maxvp);
  fprintf(fich_gen,"\tint ptro_sig, ptro_ant;\n");
  fprintf(fich_gen,"} %sREGAGENDA;\n",prefijo);
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"typedef struct {\n");
  fprintf(fich_gen,"\tint ptro_ini, ptro_fin;\n");
  fprintf(fich_gen,"} LISTA_REG;\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"extern int %sagenda_vacia;\n",prefijo);
  fprintf(fich_gen,"extern char %sestrategia;\n",prefijo);
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"void %sinsregla_agenda();\n",prefijo);
  fprintf(fich_gen,"void %selimregla_agenda();\n",prefijo);
  fprintf(fich_gen,"%sREGAGENDA %sobten_regla_mas_prio();\n",prefijo,prefijo);
  fprintf(fich_gen,"void %sInicializa_agenda();\n",prefijo);

  fprintf(fich_gen, "\n#endif //_AGENDA_%s_H_\n\n", nom_sp);
}

/*************************** GENERA_FICH_AGENDA *****************************/
void Genera_fich_agenda() {


  if(artis_mode) {
    Genera_cabecera_artis();
  }
    
  fprintf(fich_gen, "#include \"tipos_.h\"\n");
  fprintf(fich_gen, "#include \"%s\"\n",fich_generados[cab_agenda]);
  if (!artis_mode) {
    fprintf(fich_gen, "#include <stdlib.h>\n");
    fprintf(fich_gen, "#include <stdio.h>\n");
  }
  if (log_)
    fprintf(fich_gen, "#include \"%s\"\n", fich_generados[cab_log]);
  
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"#define %sTAM_AGENDA 100\n",prefijo);

  fprintf(fich_gen,"int %spos_libre=0;\n",prefijo);
  fprintf(fich_gen,"int %sagenda_vacia=FALSE;\n",prefijo);
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"%sREGAGENDA %sAGENDA[%sTAM_AGENDA];\n",prefijo,prefijo,prefijo);
  fprintf(fich_gen,"LISTA_REG %sTABLA_PRIO[21];\n",prefijo);
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"extern void %smarca_activada();\n",prefijo);
  fprintf(fich_gen,"extern int %smarca_desactivada();\n",prefijo);
  fprintf(fich_gen,"\n");
}

void Genera_funciones_agenda()
{
 /* Generamos la funcion obten_regla_mas_prio que obtiene de la agenda la
  instancia de la regla con mayor prioridad */

 fprintf(fich_gen,"%sREGAGENDA %sobten_regla_mas_prio()\n",prefijo,prefijo);
 fprintf(fich_gen,"{\n");
 fprintf(fich_gen,"\tint i;\n");
 fprintf(fich_gen,"\tint pos_regla;\n");
 fprintf(fich_gen,"\n");
 fprintf(fich_gen,"\ti=20;\n");
 fprintf(fich_gen,"\twhile(!%sTABLA_PRIO[i].ptro_ini && i>=0) i--;\n",prefijo);
 fprintf(fich_gen,"\tif (i>=0) {\n");
 fprintf(fich_gen,"\t\tpos_regla=%sTABLA_PRIO[i].ptro_ini;\n",prefijo);
 if (log_) {
   fprintf(fich_gen,"\t\t%slog_regla_mas_prio(%sAGENDA[pos_regla]);\n", prefijo, prefijo);
 }
 fprintf(fich_gen,"\t\treturn %sAGENDA[pos_regla];\n",prefijo);
 fprintf(fich_gen,"\t} else %sagenda_vacia = TRUE;\n",prefijo);
 fprintf(fich_gen,"}\n");
 fprintf(fich_gen,"\n");

 /* Generamos la funcion insregla_agenda que inserta una regla en la agenda */
 fprintf(fich_gen,"void %sinsregla_agenda(r)\n",prefijo);
 fprintf(fich_gen,"\t %sREGAGENDA r;\n",prefijo);
 fprintf(fich_gen,"{\n");
 fprintf(fich_gen,"\tint pos_regla;\n");
 fprintf(fich_gen,"\tint pos_ini;\n");
 fprintf(fich_gen,"\tint pos_fin;\n");
 fprintf(fich_gen,"\n");

 if (log_)
   fprintf(fich_gen, "\t%slog_insregla(r);\n", prefijo);
 fprintf(fich_gen,"\tpos_regla=%spos_libre;\n",prefijo);
 /* aqui si pos_libre fuera mayor que el tamao de la agenda se podria
  ejecutar un HALT -> esto significaria que se necesita una agenda de mayor
  tamao */

 /*EVR FEB05 */
 fprintf(fich_gen,"\tif (%spos_libre>=%sTAM_AGENDA) {\n",prefijo,prefijo);
 if (artis_mode) {
   fprintf(fich_gen, "#ifdef MODULE\n");
   fprintf(fich_gen, "  rtl_printf(\"Error: La agenda se ha desbordado\\n\");\n");
   fprintf(fich_gen, "#else\n");
   fprintf(fich_gen, "  printf(\"Error: La agenda se ha desbordado\\n\");\n");
   fprintf(fich_gen, "  exit(1);\n");
   fprintf(fich_gen, "#endif\n");
   fprintf(fich_gen, "}\n");
   /*
     CGM 15/06/05:
     Qu hacer cuando se ha desbordado la agenda si estamos funcionando en 
     modo tarea crtica peridica de artis? Una espera activa?
     Si no hacemos nada, la continuacin de la ejecucin llevar a...?
     Sera conveniente aumentar el tamao de la agenda?
     Calcular el tamao de forma que se ajuste al programa?
   */
 }
 else{ 
   if (!noprints)  
     fprintf(fich_gen,"\tprintf(\"Error: La agenda se ha desbordado\\n\");\n");
   fprintf(fich_gen,"\t    exit(1); }\n");
 }
 fprintf(fich_gen,"\t%spos_libre=%sAGENDA[%spos_libre].ptro_sig;\n",prefijo,prefijo,prefijo);
 fprintf(fich_gen,"\tif (%sestrategia==profundidad) {\n",prefijo);
 fprintf(fich_gen,"\t\tpos_ini=%sTABLA_PRIO[r.prioridad+10].ptro_ini;\n",prefijo);
 fprintf(fich_gen,"\t\t%sTABLA_PRIO[r.prioridad+10].ptro_ini=pos_regla;\n",prefijo);
 fprintf(fich_gen,"\t\tif (%sTABLA_PRIO[r.prioridad+10].ptro_fin==0) %sTABLA_PRIO[r.prioridad+10].ptro_fin=pos_regla;\n",prefijo,prefijo);
 fprintf(fich_gen,"\t\t\t    else %sAGENDA[pos_ini].ptro_ant=pos_regla;\n",prefijo);
 fprintf(fich_gen,"\t\t%sAGENDA[pos_regla] = r;\n",prefijo);
 fprintf(fich_gen,"\t\t%sAGENDA[pos_regla].ptro_sig=pos_ini;\n",prefijo);
 fprintf(fich_gen,"\t\t%sAGENDA[pos_regla].ptro_ant=0;\n",prefijo);
 fprintf(fich_gen,"\t} else if (%sestrategia==anchura) {\n",prefijo);
 fprintf(fich_gen,"\t\tpos_fin=%sTABLA_PRIO[r.prioridad+10].ptro_fin;\n",prefijo);
 fprintf(fich_gen,"\t\t%sTABLA_PRIO[r.prioridad+10].ptro_fin=pos_regla;\n",prefijo);
 fprintf(fich_gen,"\t\tif (%sTABLA_PRIO[r.prioridad+10].ptro_ini==0) %sTABLA_PRIO[r.prioridad+10].ptro_ini=pos_regla;\n",
	 prefijo,prefijo);
 fprintf(fich_gen,"\t\t\t    else %sAGENDA[pos_fin].ptro_sig=pos_regla;\n",prefijo);
 fprintf(fich_gen,"\t\t%sAGENDA[pos_regla] = r;\n",prefijo);
 fprintf(fich_gen,"\t\t%sAGENDA[pos_regla].ptro_sig=0;\n",prefijo);
 fprintf(fich_gen,"\t\t%sAGENDA[pos_regla].ptro_ant=pos_fin;\n",prefijo);
 fprintf(fich_gen,"\t}\n");
 fprintf(fich_gen,"\t%smarca_activada(r,pos_regla);\n",prefijo);
 fprintf(fich_gen,"}\n\n");

 /* Generamos la funcion elimregla_agenda que elimina una regla de la agenda */
 fprintf(fich_gen,"void %selimregla_agenda(r)\n",prefijo);
 fprintf(fich_gen,"\t %sREGAGENDA r;\n",prefijo);
 fprintf(fich_gen,"{\n");
 fprintf(fich_gen,"\tint pos_regla;\n");
 fprintf(fich_gen,"\tint pos_ant,pos_sig;\n");
 fprintf(fich_gen,"\tchar priori;\n");
 fprintf(fich_gen,"\n");

 if (log_)
   fprintf(fich_gen, "\t%slog_elimregla(r);\n", prefijo);
 fprintf(fich_gen,"\tpos_regla = %smarca_desactivada(r);\n",prefijo);
 fprintf(fich_gen,"\tpos_ant=%sAGENDA[pos_regla].ptro_ant;\n",prefijo);
 fprintf(fich_gen,"\tpos_sig=%sAGENDA[pos_regla].ptro_sig;\n",prefijo);
 fprintf(fich_gen,"\tpriori=%sAGENDA[pos_regla].prioridad+10;\n",prefijo);
 fprintf(fich_gen,"\tif (pos_ant!=0) { %sAGENDA[pos_ant].ptro_sig=pos_sig; }\n",prefijo);
 fprintf(fich_gen,"\t\t else { %sTABLA_PRIO[priori].ptro_ini=pos_sig; }\n",prefijo);
 fprintf(fich_gen,"\tif (pos_sig!=0) { %sAGENDA[pos_sig].ptro_ant=pos_ant; }\n",prefijo);
 fprintf(fich_gen,"\t\t else { %sTABLA_PRIO[priori].ptro_fin=pos_ant; }\n",prefijo);
 fprintf(fich_gen,"\t%sAGENDA[pos_regla].ptro_sig=%spos_libre;\n",prefijo,prefijo);
 fprintf(fich_gen,"\t%spos_libre=pos_regla;\n",prefijo);
 fprintf(fich_gen,"}\n\n");

 /* Generamos la funcion Inicializa_agenda, que inicializa las estructuras */
 /* de la agenda */
    fprintf(fich_gen,"void %sInicializa_agenda()\n",prefijo);
    fprintf(fich_gen,"{\n");
    fprintf(fich_gen,"\tint iter1;\n");
    fprintf(fich_gen,"\n");

    //Se inicializa la variable agenda_vacia.
    //En caso de no hacerlo aqu, no se podr tener un comportamiento
    //peridico en el SP, pues una vez puesta a TRUE, nunca se
    //actualizar.
    fprintf(fich_gen, "%sagenda_vacia = FALSE;\n", prefijo);


    /* ahora se inicializa la variable pos_libre */
    fprintf(fich_gen,"\t%spos_libre=1;\n",prefijo);

    /* ahora se inicializa la agenda */
    fprintf(fich_gen,"\tfor(iter1=1; iter1<%sTAM_AGENDA; iter1++)\n",prefijo);
    fprintf(fich_gen,"\t\t %sAGENDA[iter1].ptro_sig=iter1+1;\n",prefijo);

    /* ahora se inicializa la tabla de prioridades */
    fprintf(fich_gen,"\tfor(iter1=0; iter1<21; iter1++) {\n");
    fprintf(fich_gen,"\t\t %sTABLA_PRIO[iter1].ptro_ini=0;\n",prefijo);
    fprintf(fich_gen,"\t\t %sTABLA_PRIO[iter1].ptro_fin=0;\n",prefijo);
    fprintf(fich_gen,"\t}\n");
    fprintf(fich_gen,"}\n\n");
}

void Genera_cab_log() {
  int i;
  EHTIPO *tipo_tds;

  fich_gen = abrir_fichero_escritura(fich_generados[cab_log]);
  fprintf(fich_gen, "#ifndef _LOG_%s_H_\n", nom_sp);
  fprintf(fich_gen, "#define _LOG_%s_H_\n\n", nom_sp);
  fprintf(fich_gen, "#include \"%s\"\n\n", fich_generados[cab_agenda]);

  fprintf(fich_gen, "typedef enum %skind_termination {%selimit, %sequit, %sehalt, %seempty} %skind_termination;\n",
	  prefijo, prefijo, prefijo, prefijo, prefijo, prefijo);
  fprintf(fich_gen, "void %slog_insregla(%sREGAGENDA regla);\n", prefijo, prefijo);
  fprintf(fich_gen, "void %slog_elimregla(%sREGAGENDA regla);\n", prefijo, prefijo);
  fprintf(fich_gen, "void %slog_regla_mas_prio(%sREGAGENDA regla);\n", prefijo, prefijo);
  fprintf(fich_gen, "void %slog_end_actions();\n", prefijo);
  fprintf(fich_gen, "void %slog_end_fire();\n", prefijo);
  fprintf(fich_gen, "void %slog_reset();\n", prefijo);
  fprintf(fich_gen, "void %slog_end_reset();\n", prefijo);
  fprintf(fich_gen, "int %slog_inicia_log();\n", prefijo);
  fprintf(fich_gen, "void %slog_end_log(%skind_termination t);\n", prefijo, prefijo);
  fprintf(fich_gen, "void %slog_facts();\n", prefijo);
  for (i = 0; i < HASH_SIZE; i++) {
    tipo_tds = tipos[i];
    while (tipo_tds) {
      fprintf(fich_gen, "void %slog_class_%s();\n", prefijo, tipo_tds->etipo.nombre);
      tipo_tds = tipo_tds->sig;
    }
  }
    
     
  fprintf(fich_gen, "\n\n#endif //_LOG_%s_H_\n\n", nom_sp);
  cierra_fichero(fich_gen);
}

/*************************** GENERA_FICH_LOG ******************************/


//Funcin auxiliar para loguear un valor dado el nombre del cdigo C, su tipo y tamao
//nombre_c es el nombre de la instancia o slot mediante el cual se puede obneter su valor
//tipo representa el tipo de dicha instancia o slot
//tbase_array, cuando el tipo es array, representa el tipo base del array
//tam, cuando el tipo es array, representa el tamao del array
static void Genera_log_valor(char *nombre_c, tipoid tipo, tipoid tbase_array, int tam) {
  switch (tipo) {
  case entero:
  case subrango:
  case boolean:
    fprintf(fich_gen, "  fprintf(fd_log, \"%%d\", %s);\n", nombre_c);
    break;
  case real:
    fprintf(fich_gen, "  fprintf(fd_log, \"%%f\", %s);\n", nombre_c);
    break;
  case caracter: //Aun no los hay, pero debera ser...
    fprintf(fich_gen, "  fprintf(fd_log, \"%%c\", %s);\n", nombre_c);
    break;
  case cadena: //No las hay como tipo declarado por el usuario, pero debera ser...
    fprintf(fich_gen, "  fprintf(fd_log, \"%%s\", %s);\n", nombre_c);
    break;
  case lista_enteros:
    fprintf(fich_gen, "  print_lent(fd_log, (LISTA_GENERICA_ENTEROS *)&%s);\n", nombre_c);
    break;
  case lista_reales:
    fprintf(fich_gen, "  print_lreal(fd_log, (LISTA_GENERICA_REALES *)&%s);\n", nombre_c);
    break;
  case array:
    fprintf(fich_gen, "  print_array(fd_log, %s, ", nombre_c);
    switch (tbase_array) {
    case entero:
      fprintf(fich_gen, "entero");
      break;
    case real:
      fprintf(fich_gen, "real");
      break;
    case boolean:
      fprintf(fich_gen, "boolean");
      break;
    default:
      fprintf(stderr, "Error en el compilador: No se esperaba el tipo del array %s%s durante la generacin\n");
      fprintf(stderr, "de las funciones del log. Revisa genera.c Genera_fich_log\n");
      break;
    }
    fprintf(fich_gen, ", %d);\n", tam);
    break;
  default:
    fprintf(stderr, "Error en el compilador: No se esperaba el tipo de la instancia %s%s durante la generacin\n");
    fprintf(stderr, "de las funciones del log. Revisa genera.c Genera_fich_log\n");
    break; 
  }
}


void Genera_fich_log() {
  EHREGLA *regla_tds;
  EHTIPO *tipo_tds; //Para hacer el log de las clases de usuario
  EHTDS *hecho_tds; //hecho en la tds para loguear su estado
  LISTA_CAMPOS *slot_tds; //Para loguear los distintos slots de una instancia.
  char nombre_c[3*long_ident]; //Nombre en cdigo generado de la instancia (con slot) de la que se desea loguear su valor
  int i, j;
  int total_reglas = 0; //Almacena el nmero de reglas que se han declarado
  

  fich_gen = abrir_fichero_escritura(fich_generados[fich_log]);
  
  //Generacin de los includes necesarios:
  fprintf(fich_gen, "#include \"%s\"\n", fich_generados[fich_bh]);
  fprintf(fich_gen, "#include <stdio.h>\n");
  fprintf(fich_gen, "#include <time.h>\n");
  fprintf(fich_gen, "#include \"%s\"\n", fich_generados[cab_log]);


  //Obtencin del nmero total de reglas declaradas y generacin de 
  //nombres de las mismas:
  fprintf(fich_gen, "\n\n");
  for (i = 0; i < HASH_SIZE; i++) {
    regla_tds = reglas[i];
    while (regla_tds != NULL) {
      if (!regla_tds->eregla.tiene_copias) {
	fprintf(fich_gen, "static char nombre_regla_%d[] = \"%s\";\n", 
		regla_tds->eregla.num_regla, regla_tds->eregla.nombre);
	if (total_reglas < regla_tds->eregla.num_regla) {
	  total_reglas = regla_tds->eregla.num_regla;
	}
      }
      regla_tds = regla_tds->sig;
    }
  }
  fprintf(fich_gen, "static char *nombre_reglas[%d];\n", total_reglas + 1);


  //Generacin de las variables propias para el log:
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "static FILE *fd_log;\n");
  fprintf(fich_gen, "static char fmode_log[] = \"w\";\n");


  //Inicializacin del log...
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "int %slog_inicia_log() {\n", prefijo);
  fprintf(fich_gen, "  if ((fd_log = fopen(\"%s.xml\", fmode_log)) == NULL) {\n", nom_sp);
  fprintf(fich_gen, "    fprintf(stderr, \"Cannot open log file %s.xml\\n\");\n", nom_sp);
  fprintf(fich_gen, "    fprintf(stderr, \"%s aborted!\\n\");\n", nom_sp);
  fprintf(fich_gen, "    return 0;\n");
  fprintf(fich_gen, "  }\n");
  fprintf(fich_gen, "  //Se cambia el modo de apertura para permitir llamadas iterativas desde Artis\n");
  fprintf(fich_gen, "  fmode_log[0] = 'a';\n");
  for (i = 1; i <= total_reglas; i++) {
    fprintf(fich_gen, "  nombre_reglas[%d] = nombre_regla_%d;\n", i, i);
  }
  fprintf(fich_gen, "  fprintf(fd_log, \"<arlips_log RBS=\\\"%s\\\" timestamp=\\\"%%f\\\">\\n\", ((double)clock())/CLOCKS_PER_SEC);\n", nom_sp);
  fprintf(fich_gen, "  return 1;\n");
  fprintf(fich_gen, "}\n");

  //Finalizacin del log...
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "void %slog_end_log(%skind_termination end) {\n", 
	  prefijo, prefijo);
  fprintf(fich_gen, "  char *message;\n");
  fprintf(fich_gen, "  switch (end) {\n");
  fprintf(fich_gen, "  case %selimit:\n", prefijo);
  fprintf(fich_gen, "    message = \"limit\";\n");
  fprintf(fich_gen, "    break;\n");
  fprintf(fich_gen, "  case %sequit:\n", prefijo);
  fprintf(fich_gen, "    message = \"quit\";\n");
  fprintf(fich_gen, "    break;\n");
  fprintf(fich_gen, "  case %sehalt:\n", prefijo);
  fprintf(fich_gen, "    message = \"halt\";\n");
  fprintf(fich_gen, "    break;\n");
  fprintf(fich_gen, "  case %seempty:\n", prefijo);
  fprintf(fich_gen, "    message = \"empty\";\n");
  fprintf(fich_gen, "    break;\n");
  fprintf(fich_gen, "  }\n");
  fprintf(fich_gen, "  fprintf(fd_log, \"  <end>%s</end>\\n\", message);\n", 
	  "%s");
  fprintf(fich_gen, "  fprintf(fd_log, \"</arlips_log>\\n\\n\");\n");
  fprintf(fich_gen, "  fclose(fd_log);\n");
  fprintf(fich_gen, "}\n");
	  

  //log reset
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "void %slog_reset() {\n", prefijo);
  fprintf(fich_gen, "  fprintf(fd_log, \"  <reset>\\n\");\n");
  fprintf(fich_gen, "  fprintf(fd_log, \"    <actions>\\n\");\n");
  fprintf(fich_gen, "}\n");

  //log end reset
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "void %slog_end_reset() {\n", prefijo);
  fprintf(fich_gen, "  %slog_end_actions();\n", prefijo);
  fprintf(fich_gen, "  fprintf(fd_log, \"  </reset>\\n\");\n");
  fprintf(fich_gen, "}\n");

  //log de la regla ms prioritaria
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "void %slog_regla_mas_prio(%sREGAGENDA r) { \n", prefijo, prefijo);
  fprintf(fich_gen, "  fprintf(fd_log, \"  <fired>\\n\");\n");
  fprintf(fich_gen, "  fprintf(fd_log, \"    <rule>%s</rule>\\n\", nombre_reglas[r.num_regla]);\n", "%s");
  fprintf(fich_gen, "  fprintf(fd_log, \"    <priority>%s</priority>\\n\", r.prioridad);\n", "%d");
  fprintf(fich_gen, "  fprintf(fd_log, \"    <actions>\\n\");\n");
  fprintf(fich_gen, "}\n");

  //log fin del disparo de una regla
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "void %slog_end_fire() {\n", prefijo);
  fprintf(fich_gen, "  %slog_end_actions();\n");
  fprintf(fich_gen, "  fprintf(fd_log, \"  </fired>\\n\");\n");
  fprintf(fich_gen, "}\n");

  //log finaliza la lista de acciones
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "void %slog_end_actions() {\n");
  fprintf(fich_gen, "  fprintf(fd_log, \"    </actions>\\n\");\n");
  fprintf(fich_gen, "}\n");

  //log de la insercin de una regla en la agenda
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "void %slog_insregla(%sREGAGENDA r) {\n");
  fprintf(fich_gen, "  fprintf(fd_log, \"      <activation>\\n\");\n");
  fprintf(fich_gen, "  fprintf(fd_log, \"        <rule>%s</rule>\\n\", nombre_reglas[r.num_regla]);\n", "%s");
  fprintf(fich_gen, "  fprintf(fd_log, \"        <priority>%s</priority>\\n\", r.prioridad);\n", "%d");
  fprintf(fich_gen, "  fprintf(fd_log, \"      </activation>\\n\");\n");
  fprintf(fich_gen, "}\n");

  //log de la eliminacin de una regla en la agenda
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "void %slog_elimregla(%sREGAGENDA r) {\n", prefijo, prefijo);
  fprintf(fich_gen, "  fprintf(fd_log, \"      <deactivation>\\n\");\n");
  fprintf(fich_gen, "  fprintf(fd_log, \"        <rule>%s</rule>\\n\", nombre_reglas[r.num_regla]);\n", "%s");
  fprintf(fich_gen, "  fprintf(fd_log, \"      </deactivation>\\n\");\n");
  fprintf(fich_gen, "}\n");

  

  //log del estado de la base de hechos bajo demanda
  fprintf(fich_gen, "\n\n");
  fprintf(fich_gen, "void %slog_facts() {\n", prefijo);
  fprintf(fich_gen, "  fprintf(fd_log, \"      <working_memory>\\n\");\n");
  //Empezamos por loguear clases
  for (i = 0; i < HASH_SIZE; i++) { //Llamamos a las funciones que loguean cada una de las clases.
    tipo_tds = tipos[i];
    while (tipo_tds) {
      fprintf(fich_gen, "  %slog_class_%s();\n", prefijo, tipo_tds->etipo.nombre);
      tipo_tds = tipo_tds->sig;
    }
  }
  //Seguimos con las instancias de tipo simple
  fprintf(fich_gen, "  fprintf(fd_log, \"        <class id=\\\"system\\\">\\n\");\n");
  fprintf(fich_gen, "  fprintf(fd_log, \"          <instances>\\n\");\n");
  for (i = 0; i < HASH_SIZE; i++) {
    hecho_tds = tds[i];
    while (hecho_tds) {
      if (hecho_tds->etds.tipo != definido && hecho_tds->etds.clvar == instancia) {
	//Del kdm solo pueden venir instancias de clase, por lo tanto
	//no hace falta comprobar si las instancias simples son del 
	//kdm y se usan.
	fprintf(fich_gen, "  fprintf(fd_log, \"            <%s>\");\n", hecho_tds->etds.nombre);
	strcpy(nombre_c, prefijo);
	strcat(nombre_c, hecho_tds->etds.nombre);
	Genera_log_valor(nombre_c, hecho_tds->etds.tipo, hecho_tds->etds.tbase_array, hecho_tds->etds.tam_lista);
	fprintf(fich_gen, "  fprintf(fd_log, \"</%s>\\n\");\n", hecho_tds->etds.nombre);
      }
      hecho_tds = hecho_tds->sig;
    }
  }
  fprintf(fich_gen, "  fprintf(fd_log, \"          </instances>\\n\");\n");
  //Y ahora repetimos un procedimiento similar en busca de variables...  
  fprintf(fich_gen, "  fprintf(fd_log, \"          <variables>\\n\");\n");
  for (i = 0; i < HASH_SIZE; i++) {
    hecho_tds = tds[i];
    while (hecho_tds) {
      if (hecho_tds->etds.tipo != definido && hecho_tds->etds.clvar == var) {
	//Del kdm solo pueden venir instancias de clase, por lo tanto
	//no hace falta comprobar si las instancias simples son del 
	//kdm y se usan.
	fprintf(fich_gen, "  fprintf(fd_log, \"            <%s>\");\n", hecho_tds->etds.nombre);
	strcpy(nombre_c, prefijo);
	strcat(nombre_c, hecho_tds->etds.nombre);
	Genera_log_valor(nombre_c, hecho_tds->etds.tipo, hecho_tds->etds.tbase_array, hecho_tds->etds.tam_lista);
	fprintf(fich_gen, "  fprintf(fd_log, \"</%s>\\n\");\n", hecho_tds->etds.nombre);
      }
      hecho_tds = hecho_tds->sig;
    }
  }
  fprintf(fich_gen, "  fprintf(fd_log, \"          </variables>\\n\");\n");
  fprintf(fich_gen, "  fprintf(fd_log, \"        </class>\\n\");\n");  
  fprintf(fich_gen, "  fprintf(fd_log, \"      </working_memory>\\n\");\n");
  fprintf(fich_gen, "}\n\n");

  //Generamos una funcin por clase para loguear los hechos de esa clase
  for (i = 0; i < HASH_SIZE; i++) {
    tipo_tds = tipos[i];
    while (tipo_tds) {
      //Para cada clase generamos una funcin que loguea todas sus instancias
      fprintf(fich_gen, "void %slog_class_%s() {\n", prefijo, tipo_tds->etipo.nombre);
      fprintf(fich_gen, "  fprintf(fd_log, \"        <class id=\\\"%s\\\">\\n\");\n", tipo_tds->etipo.nombre);
      fprintf(fich_gen, "  fprintf(fd_log, \"          <instances>\\n\");\n");
      for (j = 0; j < HASH_SIZE; j++) {
	hecho_tds = tds[j];
	while (hecho_tds) {
	  if (hecho_tds->etds.tipo == definido && hecho_tds->etds.clvar == instancia && 
	      !strcmp(hecho_tds->etds.instancia_de, tipo_tds->etipo.nombre)) {
	    fprintf(fich_gen, "  fprintf(fd_log, \"            <%s>\\n\");\n", hecho_tds->etds.nombre);
	    slot_tds = tipo_tds->etipo.lc;
	    while (slot_tds) {
	      if (!hecho_tds->etds.from_kdm || slot_tds->campo_buscado) {
		fprintf(fich_gen, "  fprintf(fd_log, \"              <%s>\");\n", slot_tds->nombre);
		sprintf(nombre_c, "%si_%s[%d].%s", prefijo, hecho_tds->etds.instancia_de, 
			hecho_tds->etds.cod.num_instancia, slot_tds->nombre);
		Genera_log_valor(nombre_c, slot_tds->tipo, slot_tds->tbase_array, slot_tds->tam_lista);
		fprintf(fich_gen, "  fprintf(fd_log, \"              </%s>\\n\");\n", slot_tds->nombre);
	      }
	      slot_tds = slot_tds->siguiente;	      
	    }
	    fprintf(fich_gen, "  fprintf(fd_log, \"            </%s>\\n\");\n", hecho_tds->etds.nombre);
	  }
	  hecho_tds = hecho_tds->sig;
	}
      }
      fprintf(fich_gen, "  fprintf(fd_log, \"          </instances>\\n\");\n");
      //Y tambin sus variables globales
      fprintf(fich_gen, "  fprintf(fd_log, \"          <variables>\\n\");\n");      
      for (j = 0; j < HASH_SIZE; j++) {
	hecho_tds = tds[j];
	while (hecho_tds) {
	  if (hecho_tds->etds.tipo == definido && hecho_tds->etds.clvar == var && 
	      !strcmp(hecho_tds->etds.instancia_de, tipo_tds->etipo.nombre)) {
	    fprintf(fich_gen, "  fprintf(fd_log, \"            <%s>\\n\");\n", hecho_tds->etds.nombre);
	    slot_tds = tipo_tds->etipo.lc;
	    while (slot_tds) {
	      if (slot_tds->campo_buscado) {
		fprintf(fich_gen, "  fprintf(fd_log, \"              <%s>\");\n", slot_tds->nombre);
		sprintf(nombre_c, "%sv_%s[%d].%s", prefijo, hecho_tds->etds.instancia_de, 
			hecho_tds->etds.cod.num_instancia, slot_tds->nombre);
		Genera_log_valor(nombre_c, slot_tds->tipo, slot_tds->tbase_array, slot_tds->tam_lista);
		fprintf(fich_gen, "  fprintf(fd_log, \"              </%s>\\n\");\n", slot_tds->nombre);
	      }
	      slot_tds = slot_tds->siguiente;	      
	    }
	    fprintf(fich_gen, "  fprintf(fd_log, \"            </%s>\\n\");\n", hecho_tds->etds.nombre);
	  }
	  hecho_tds = hecho_tds->sig;
	}
      }
      //Y ya hemos acabado con todas las instancias y variables de esta clase.  
      fprintf(fich_gen, "  fprintf(fd_log, \"          </variables>\\n\");\n");
      fprintf(fich_gen, "  fprintf(fd_log, \"        </class>\\n\");\n");
      fprintf(fich_gen, "}\n\n");
      tipo_tds = tipo_tds->sig;
    }
  }	  
  cierra_fichero(fich_gen);
}


/*************************** GENERA_CAB_TIPOS *****************************/
void Genera_cab_tipos() {
  fprintf(fich_gen, "#ifndef _TIPOS_ARL_H_\n");
  fprintf(fich_gen, "#define _TIPOS_ARL_H_\n\n");

  if (artis_mode) {
    Genera_cabecera_artis();
  }
  else {
    fprintf(fich_gen, "#include <stdio.h>\n");
  }
  
     fprintf(fich_gen, "#define TRUE 1\n");
     fprintf(fich_gen, "#define FALSE 0\n");
     fprintf(fich_gen, "#define CELDA_VACIA 0\n");
     fprintf(fich_gen, "#define CELDA_OCUPADA 1\n");
     
     fprintf(fich_gen, "/* ahora se generan las macros que sirven para declarar listas */\n");
     fprintf(fich_gen, "/* de enteros y reales */\n");

     fprintf(fich_gen, "#define listaent(x,taml)  int x ## _elementos[taml];LISTA_GENERICA_ENTEROS x\n");
     fprintf(fich_gen, "#define listareal(x,taml) float x ## _elementos[taml];LISTA_GENERICA_REALES x\n");
     fprintf(fich_gen, "\n");
     /* ahora se define el tipo generico que manejaran las funciones */
     /* que operan con listas */
     fprintf(fich_gen, "typedef struct {\n");
     fprintf(fich_gen, "\tchar *nombre;\n");
     fprintf(fich_gen, "\tint cabeza;\n");
     fprintf(fich_gen, "\tint cola;\n");
     fprintf(fich_gen, "\tint tam;\n");
     fprintf(fich_gen, "\tint num_elem;\n");
     fprintf(fich_gen, "\tint *lista;\n");
     fprintf(fich_gen, "} LISTA_GENERICA_ENTEROS;\n");
     fprintf(fich_gen, "\n");

     fprintf(fich_gen, "typedef struct {\n");
     fprintf(fich_gen, "\tchar *nombre;\n"); 
     fprintf(fich_gen, "\tint cabeza;\n");
     fprintf(fich_gen, "\tint cola;\n");
     fprintf(fich_gen, "\tint tam;\n");
     fprintf(fich_gen, "\tint num_elem;\n");
     fprintf(fich_gen, "\tfloat *lista;\n");
     fprintf(fich_gen, "} LISTA_GENERICA_REALES;\n");
     fprintf(fich_gen, "\n");

     //Se generan los tipos posibles de un array
     fprintf(fich_gen, "//enumerado de los tipos admitidos para la impresin de arrays\n");
     fprintf(fich_gen, "typedef enum enum_tipos {boolean, entero, real} enum_tipos;\n\n");
     

     /* ahora se generan los prototipos de las funciones */
     fprintf(fich_gen, "void inicializa_lent(LISTA_GENERICA_ENTEROS *lent, const int tamlis, char *nombre, int *elementos);\n");
     fprintf(fich_gen, "void inicializa_lreal(LISTA_GENERICA_REALES *lreal, const int tamlis, char *nombre, float *elementos);\n");
     fprintf(fich_gen, "void add_head_lent(const int num, LISTA_GENERICA_ENTEROS *lent);\n");
     fprintf(fich_gen, "void add_head_lreal(const float num, LISTA_GENERICA_REALES *lreal);\n");
     fprintf(fich_gen, "void add_tail_lent(const int num, LISTA_GENERICA_ENTEROS *lent);\n");
     fprintf(fich_gen, "void add_tail_lreal(const float num, LISTA_GENERICA_REALES *lreal);\n");
     fprintf(fich_gen, "int head_lent(const LISTA_GENERICA_ENTEROS *lent);\n");
     fprintf(fich_gen, "int tail_lent(const LISTA_GENERICA_ENTEROS *lent);\n");
     fprintf(fich_gen, "float head_lreal(const LISTA_GENERICA_REALES *lreal);\n");
     fprintf(fich_gen, "float tail_lreal(const LISTA_GENERICA_REALES *lreal);\n");
     fprintf(fich_gen, "int pop_head_lent(LISTA_GENERICA_ENTEROS *lent);\n");
     fprintf(fich_gen, "int pop_tail_lent(LISTA_GENERICA_ENTEROS *lent);\n");
     fprintf(fich_gen, "float pop_head_lreal(LISTA_GENERICA_REALES *lreal);\n");
     fprintf(fich_gen, "float pop_tail_lreal(LISTA_GENERICA_REALES *lreal);\n");
     fprintf(fich_gen, "void print_lent(FILE *fd, const LISTA_GENERICA_ENTEROS *lent);\n");
     fprintf(fich_gen, "void print_lreal(FILE *fd, const LISTA_GENERICA_REALES *lreal);\n");

     fprintf(fich_gen, "void print_array(FILE *fd, void *array, const enum_tipos tipo, const int tam);\n");

     fprintf(fich_gen, "\n#endif //_TIPOS_ARL_H_\n\n");
}

/*************************** GENERA_FICH_TIPOS *****************************/


void Genera_fich_tipos() {
  fprintf(fich_gen, "#include \"tipos_.h\"\n");
  if (!noprints) {
    if (artis_mode) {
      Genera_cabecera_artis();
    }
    else {
      fprintf(fich_gen,"#include <stdio.h>\n");
    }
  }      
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\n");


  if (!noprints) { 
    //Se genera el tipo de datos necesario para avisar
    //cuando se producen errores en listas.
    fprintf(fich_gen, "typedef enum {head, tail, pophead, poptail, addhead, addtail} funcion_lista;\n\n\n"); 
    
    
    
    //Se genera la funcin que imprime cualquier clase de array
    fprintf(fich_gen, "void print_array(FILE *fd, void *array, const enum_tipos tipo, const int tam) {\n");
    fprintf(fich_gen, "  int i;\n");
    fprintf(fich_gen, "  int *int_array;\n");
    fprintf(fich_gen, "  float *float_array;\n\n");
 
    if (artis_mode) {
      fprintf(fich_gen, "#ifdef MODULE\n");
      fprintf(fich_gen, "  rtl_printf(\"(\");\n");
      fprintf(fich_gen, "#else\n");
      fprintf(fich_gen, "  fprintf(fd, \"(\");\n");
      fprintf(fich_gen, "#endif\n\n");
    }
    else {
      fprintf(fich_gen, "  fprintf(fd, \"(\");\n\n");
    }

    fprintf(fich_gen, "  switch (tipo) {\n");
    fprintf(fich_gen, "  case boolean:\n");
    fprintf(fich_gen, "  case entero:\n");
    fprintf(fich_gen, "    int_array = (int *)array;\n");
    fprintf(fich_gen, "    for (i = 0; i < tam; i++) {\n");
    if (artis_mode) {
      fprintf(fich_gen, "#ifdef MODULE\n");
      fprintf(fich_gen, "      rtl_printf(\" %s \", int_array[i]);\n", "%d");
      fprintf(fich_gen, "#else\n");
      fprintf(fich_gen, "      fprintf(fd, \" %s \", int_array[i]);\n", "%d");
      fprintf(fich_gen, "#endif\n\n");
    }
    else {
      fprintf(fich_gen, "      fprintf(fd, \" %s \", int_array[i]);\n", "%d");
    }
    fprintf(fich_gen, "    }\n");
    fprintf(fich_gen, "    break;\n");
    
    fprintf(fich_gen, "  case real:\n");
    fprintf(fich_gen, "    float_array = (float *)array;\n");
    fprintf(fich_gen, "    for (i = 0; i < tam; i++) {\n");
    if (artis_mode) {
      fprintf(fich_gen, "#ifdef MODULE\n");
      fprintf(fich_gen, "      rtl_printf(\" %s \", float_array[i]);\n", "%f");
      fprintf(fich_gen, "#else\n");
      fprintf(fich_gen, "      fprintf(fd, \" %s \", float_array[i]);\n", "%f");
      fprintf(fich_gen, "#endif\n\n");
    }
    else {
      fprintf(fich_gen, "      fprintf(fd, \" %s \", float_array[i]);\n", "%f");
    }
    fprintf(fich_gen, "    }\n");
    fprintf(fich_gen, "    break;\n");
    fprintf(fich_gen, "  }\n");

    if (artis_mode) {
      fprintf(fich_gen, "#ifdef MODULE\n");
      fprintf(fich_gen, "  rtl_printf(\")\");\n");
      fprintf(fich_gen, "#else\n");
      fprintf(fich_gen, "  fprintf(fd, \")\");\n");
      fprintf(fich_gen, "#endif\n\n");
    }
    else {
      fprintf(fich_gen, "  fprintf(fd, \")\");\n");
    }
    fprintf(fich_gen, "}\n\n\n");
    

    
    //Se genera la funcin error_lista que se encarga de avisar
    //cuando se producen errores en listas, y que depende de su tipo asociado.
    fprintf(fich_gen, "void error_lista(const funcion_lista f, const char *l) {\n");
    fprintf(fich_gen, "  char *msg;\n\n");
    
    fprintf(fich_gen, "  switch (f) {\n");
    fprintf(fich_gen, "  case head:\n");
    fprintf(fich_gen, "    msg = \"Arlips cannot execute head on the empty list\";\n");
    fprintf(fich_gen, "    break;\n\n");
    
    fprintf(fich_gen, "  case tail:\n");
    fprintf(fich_gen, "    msg = \"Arlips cannot execute tail on the empty list\";\n");
    fprintf(fich_gen, "    break;\n\n");

    fprintf(fich_gen, "  case pophead:\n");
    fprintf(fich_gen, "  case poptail:\n");			   
    fprintf(fich_gen, "    msg = \"Arlips cannot pop an element from the empty list\";\n");
    fprintf(fich_gen, "    break;\n\n");
    
    fprintf(fich_gen, "  case addhead:\n");
    fprintf(fich_gen, "  case addtail:\n");			   
    fprintf(fich_gen, "    msg = \"Arlips cannot add an element on the full list\";\n");
    fprintf(fich_gen, "    break;\n");
    fprintf(fich_gen, "  } //del switch (f)\n\n");
    
    if (artis_mode) {
      fprintf(fich_gen, "#ifdef MODULE\n");
      fprintf(fich_gen, "  rtl_printf(\"ERROR: %s %s\\n\", msg, l);\n", "%s", "%s");
      fprintf(fich_gen, "#else\n");
      fprintf(fich_gen, "  fprintf(stderr, \"ERROR: %s %s\\n\", msg, l);\n", "%s", "%s");
      fprintf(fich_gen, "#endif\n\n");
    }
    else
      fprintf(fich_gen, "  fprintf(stderr, \"ERROR: %s %s\\n\", msg, l);\n", "%s", "%s");
    fprintf(fich_gen, "} //de error_lista()\n\n\n");
  }
      
    
  
  /* se genera la funcion inicializa_lent que se encarga */
  /* de inicializar una lista de enteros */
  fprintf(fich_gen, "void inicializa_lent(LISTA_GENERICA_ENTEROS *lent, const int tamlis, char *nombre, int *comienzo) {\n");
  fprintf(fich_gen, "\tlent->nombre = nombre;\n");
  fprintf(fich_gen, "\tlent->cabeza=-1;\n");
  fprintf(fich_gen, "\tlent->cola =0;\n");
  fprintf(fich_gen, "\tlent->tam = tamlis;\n");
  fprintf(fich_gen, "\tlent->num_elem = 0;\n");
  fprintf(fich_gen, "\tlent->lista = comienzo;\n");
  fprintf(fich_gen, "}\n");
  fprintf(fich_gen, "\n");
  
  /* se genera la funcion inicializa_real que se encarga */
  /* de inicializar una lista de reales */
  
  fprintf(fich_gen, "void inicializa_lreal(LISTA_GENERICA_REALES *lreal, const int tamlis, char *nombre, float *comienzo) {\n");
  fprintf(fich_gen, "\tlreal->nombre = nombre;\n");
  fprintf(fich_gen, "\tlreal->cabeza=-1;\n");
  fprintf(fich_gen, "\tlreal->cola =0;\n");
  fprintf(fich_gen, "\tlreal->tam = tamlis;\n");
  fprintf(fich_gen, "\tlreal->num_elem = 0;\n");
  fprintf(fich_gen, "\tlreal->lista = comienzo;\n");
  fprintf(fich_gen, "}\n");
  fprintf(fich_gen, "\n");

  //Se genera la funcin que imprime una lista de enteros
  if (!noprints) {
    fprintf(fich_gen, "void print_lent(FILE *fd, const LISTA_GENERICA_ENTEROS *l) {\n");
    fprintf(fich_gen, "  int i;\n");
    if (artis_mode) {
      fprintf(fich_gen, "#ifdef MODULE\n");
      fprintf(fich_gen, "  rtl_printf(\"(\");\n");
      fprintf(fich_gen, "  for(i = 0; i < l->num_elem; i++) {\n");
      fprintf(fich_gen, "    rtl_printf(\" %s \", (*l).lista[((*l).cabeza+i)%(*l).tam]);\n",
	      "%d");
      fprintf(fich_gen, "  }\n");
      fprintf(fich_gen, "  rtl_printf(\")\");\n");  
      fprintf(fich_gen, "#else\n");
      fprintf(fich_gen, "  fprintf(fd, \"(\");\n");
      fprintf(fich_gen, "  for(i = 0; i < l->num_elem; i++) {\n");
      fprintf(fich_gen,
	      "    fprintf(fd, \" %s \", (*l).lista[((*l).cabeza+i)%(*l).tam]);\n",
	      "%d");
      fprintf(fich_gen, "  }\n");
      fprintf(fich_gen, "  fprintf(fd, \")\");\n");
      fprintf(fich_gen, "#endif\n");
    } //si artis_mode
    else {
      fprintf(fich_gen, "  fprintf(fd, \"(\");\n");
      fprintf(fich_gen, "  for(i = 0; i < l->num_elem; i++) {\n");
      fprintf(fich_gen,
	      "    fprintf(fd, \" %s \", (*l).lista[((*l).cabeza+i)%(*l).tam]);\n",
	      "%d");
      fprintf(fich_gen, "  }\n");
      fprintf(fich_gen, "  fprintf(fd, \")\");\n");
    } //else artis_mode (si modo normal)
    fprintf(fich_gen, "}\n\n\n");
  } //Si hay que generar los prints de listas


  //Se genera la funcin que imprime una lista de reales
  if (!noprints) {
    fprintf(fich_gen, "void print_lreal(FILE *fd, const LISTA_GENERICA_REALES *l) {\n");
    fprintf(fich_gen, "  int i;\n");
    if (artis_mode) {
      fprintf(fich_gen, "#ifdef MODULE\n");
      fprintf(fich_gen, "  rtl_printf(\"(\");\n");
      fprintf(fich_gen, "  for(i = 0; i < l->num_elem; i++) {\n");
      fprintf(fich_gen, "    rtl_printf(\" %s \", (*l).lista[((*l).cabeza+i)%(*l).tam]);\n",
	      "%f");
      fprintf(fich_gen, "  }\n");
      fprintf(fich_gen, "  rtl_printf(\")\");\n");  
      fprintf(fich_gen, "#else\n");
      fprintf(fich_gen, "  fprintf(fd, \"(\");\n");
      fprintf(fich_gen, "  for(i = 0; i < l->num_elem; i++) {\n");
      fprintf(fich_gen,
	      "    fprintf(fd, \" %s \", (*l).lista[((*l).cabeza+i)%(*l).tam]);\n",
	      "%f");
      fprintf(fich_gen, "  }\n");
      fprintf(fich_gen, "  fprintf(fd, \")\");\n");
      fprintf(fich_gen, "#endif\n");
    } //si artis_mode
    else {
      fprintf(fich_gen, "  fprintf(fd, \"(\");\n");
      fprintf(fich_gen, "  for(i = 0; i < l->num_elem; i++) {\n");
      fprintf(fich_gen,
	      "    fprintf(fd, \" %s \", (*l).lista[((*l).cabeza+i)%(*l).tam]);\n",
	      "%f");
      fprintf(fich_gen, "  }\n");
      fprintf(fich_gen, "  fprintf(fd, \")\");\n");
    } //else artis_mode (si modo normal)
    fprintf(fich_gen, "}\n\n\n");
  } //Si hay que generar los prints de listas




  /* se genera la funcion head_lent que devuelve la cabeza de una lista */
  /* entera */
  fprintf(fich_gen,"int head_lent(const LISTA_GENERICA_ENTEROS *lent) {\n");
  fprintf(fich_gen, "\tint devol;\n");
  fprintf(fich_gen, "\tint cab;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lent).cabeza==-1) {   /* LISTA VACIA */\n");
  fprintf(fich_gen, "\t\t//lista_vacia=TRUE;\n");
  if (!noprints)
    fprintf(fich_gen, "\t\terror_lista(head, (*lent).nombre);\n");
  fprintf(fich_gen, "\t\treturn(0);\n");
  fprintf(fich_gen, "\t}\n");
  fprintf(fich_gen, "\tdevol = (*lent).lista[(*lent).cabeza];\n");
  fprintf(fich_gen, "\treturn (devol);\n");
  fprintf(fich_gen, "}\n");
  fprintf(fich_gen, "\n");
  
  /* se genera la funcion tail_lent que devuelve la cola de una lista */
  /* entera */
  fprintf(fich_gen,"int tail_lent(const LISTA_GENERICA_ENTEROS *lent) {\n");
  fprintf(fich_gen,"\tint devol;\n");
  fprintf(fich_gen,"\tint col;\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"\tif ((*lent).cabeza==-1) {   /* LISTA VACIA */\n");
  fprintf(fich_gen,"\t\t//lista_vacia=TRUE;\n");
  if (!noprints)
    fprintf(fich_gen, "\t\terror_lista(tail, (*lent).nombre);\n");
  fprintf(fich_gen,"\t\treturn(0);\n");
  fprintf(fich_gen,"\t}\n");
  fprintf(fich_gen,"\tcol= (*lent).cola;\n");
  fprintf(fich_gen,"\tif (col==0) devol = (*lent).lista[(*lent).tam-1];\n");
  fprintf(fich_gen,"\t else devol = (*lent).lista[(*lent).cola-1];\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"\treturn (devol);\n");
  fprintf(fich_gen,"}\n");
  fprintf(fich_gen,"\n");
  
  /* se genera la funcion pop_head_lent que extrae y devuelve la cabeza de una lista */
  /* entera */
  fprintf(fich_gen,"int pop_head_lent(LISTA_GENERICA_ENTEROS *lent) {\n");
  fprintf(fich_gen,"\tint devol;\n");
  fprintf(fich_gen,"\tint cab;\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"\tif ((*lent).cabeza==-1) {   /* LISTA VACIA */\n");
  fprintf(fich_gen,"\t\t//lista_vacia=TRUE;\n");
  if (!noprints)
    fprintf(fich_gen, "\t\terror_lista(pophead, (*lent).nombre);\n");
  fprintf(fich_gen,"\t\treturn(0);\n");
  fprintf(fich_gen,"\t}\n");
  fprintf(fich_gen,"\tdevol = (*lent).lista[(*lent).cabeza];\n");
  fprintf(fich_gen,"\tcab= (*lent).cabeza;\n");
  fprintf(fich_gen,"\tif ((*lent).cabeza==(*lent).tam-1) (*lent).cabeza=0;\n");
  fprintf(fich_gen,"\t\telse   (*lent).cabeza++;\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"\tif ((*lent).cabeza==(*lent).cola) {\n");
  fprintf(fich_gen,"\t\t//lista_vacia = TRUE;\n");
  fprintf(fich_gen,"\t\t(*lent).cabeza=-1;\n");
  fprintf(fich_gen,"\t\t(*lent).cola=0;\n");
  fprintf(fich_gen,"\t}\n");
  fprintf(fich_gen,"\t(*lent).num_elem--;\n");
  fprintf(fich_gen,"\treturn (devol);\n");
  fprintf(fich_gen,"}\n");
  fprintf(fich_gen,"\n");
  
  /* se genera la funcion tail_lent que extrae y devuelve la cola de una lista */
  /* entera */
  fprintf(fich_gen,"int pop_tail_lent(LISTA_GENERICA_ENTEROS *lent) {\n");
  fprintf(fich_gen,"\tint devol;\n");
  fprintf(fich_gen,"\tint col;\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"\tif ((*lent).cabeza==-1) {   /* LISTA VACIA */\n");
  fprintf(fich_gen,"\t\t//lista_vacia=TRUE;\n");
  if (!noprints)
    fprintf(fich_gen, "\t\terror_lista(poptail, (*lent).nombre);\n");
  fprintf(fich_gen,"\t\treturn(0);\n");
  fprintf(fich_gen,"\t}\n");
  fprintf(fich_gen,"\tcol= (*lent).cola;\n");
  fprintf(fich_gen,"\tif (col==0) devol = (*lent).lista[(*lent).tam-1];\n");
  fprintf(fich_gen,"\t else devol = (*lent).lista[(*lent).cola-1];\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"\tif ((*lent).cola==0) (*lent).cola=(*lent).tam-1;\n");
  fprintf(fich_gen,"\telse   (*lent).cola--;\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"\tif ((*lent).cabeza==(*lent).cola) {\n");
  fprintf(fich_gen,"\t\t//lista_vacia = TRUE;\n");
  fprintf(fich_gen,"\t\t(*lent).cabeza=-1;\n");
  fprintf(fich_gen,"\t\t(*lent).cola=0;\n");
  fprintf(fich_gen,"\t}\n");
  fprintf(fich_gen,"\t(*lent).num_elem--;\n");
  fprintf(fich_gen,"\treturn (devol);\n");
  fprintf(fich_gen,"}\n");
  fprintf(fich_gen,"\n");
  
  
  /* se genera la funcion add_head_lent que aade un elemento a la */
  /* cabeza de una lista entera */
  fprintf(fich_gen,"void add_head_lent(const int num, LISTA_GENERICA_ENTEROS *lent) {\n");
  fprintf(fich_gen, "\tint cab;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lent).cabeza==(*lent).cola) {    /* si esta llena */\n");
  fprintf(fich_gen, "\t\t//lista_llena=TRUE;\n");
  if (!noprints) 
    fprintf(fich_gen, "\t\terror_lista(addhead, (*lent).nombre);\n");
  fprintf(fich_gen, "\t\treturn;\n");
  fprintf(fich_gen, "\t}\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tcab = (*lent).cabeza;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lent).cabeza<=0) (*lent).cabeza=(*lent).tam-1;\n");
  fprintf(fich_gen, "\t\telse  (*lent).cabeza--;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\t(*lent).lista[(*lent).cabeza] = num;\n");
  fprintf(fich_gen, "\t(*lent).num_elem++;\n");
  fprintf(fich_gen, "}\n");
  fprintf(fich_gen, "\n");
  

  fprintf(fich_gen,"void add_tail_lent(const int num, LISTA_GENERICA_ENTEROS *lent) {\n");
  fprintf(fich_gen, "\tint col;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lent).cabeza==(*lent).cola) {\n");
  fprintf(fich_gen, "\t\t//lista_llena=TRUE;\n");
  if (!noprints)
    fprintf(fich_gen, "\t\terror_lista(addtail, (*lent).nombre);\n");
  fprintf(fich_gen, "\t\treturn;\n");
  fprintf(fich_gen, "\t}\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lent).cabeza==-1) /* lista vacia */\n");
  fprintf(fich_gen, "\t{\n");
  fprintf(fich_gen, "\t\t(*lent).cabeza = (*lent).tam-1;\n");
  fprintf(fich_gen, "\t\t(*lent).lista[(*lent).cabeza]=num;\n");
  fprintf(fich_gen, "\t} else {\n");
  fprintf(fich_gen, "\t\tcol= (*lent).cola;\n");
  fprintf(fich_gen, "\t\tif ((*lent).cola==(*lent).tam-1)  (*lent).cola=0;\n");
  fprintf(fich_gen, "\t\t\telse (*lent).cola++;\n");
  fprintf(fich_gen, "\t\t(*lent).lista[col] = num;\n");
  fprintf(fich_gen, "\t}\n");
  fprintf(fich_gen, "\t(*lent).num_elem++;\n");
  fprintf(fich_gen, "}\n");
  fprintf(fich_gen, "\n");
  
  
  /* se genera la funcion head_lreal que devuelve la cabeza de una lista */
  /* real */
  fprintf(fich_gen,"float head_lreal(const LISTA_GENERICA_REALES *lreal) {\n");
  fprintf(fich_gen,"\tfloat devol;\n");
  fprintf(fich_gen,"\tint cab;\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"\tif ((*lreal).cabeza==-1) {   /* LISTA VACIA */\n");
  fprintf(fich_gen,"\t\t//lista_vacia=TRUE;\n");
  if (!noprints)
    fprintf(fich_gen, "\t\terror_lista(head, (*lreal).nombre);\n");
  fprintf(fich_gen,"\t\treturn(0.0);\n");
  fprintf(fich_gen,"\t}\n");
  fprintf(fich_gen,"\tdevol = (*lreal).lista[(*lreal).cabeza];\n");
  fprintf(fich_gen,"\treturn (devol);\n");
  fprintf(fich_gen,"}\n");
  fprintf(fich_gen,"\n");
  
  /* se genera la funcion tail_lreal que devuelve la cola de una lista */
  /* real */
  fprintf(fich_gen,"float tail_lreal(const LISTA_GENERICA_REALES *lreal) {\n");
  fprintf(fich_gen, "\tfloat devol;\n");
  fprintf(fich_gen, "\tint col;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lreal).cabeza==-1) {   /* LISTA VACIA */\n");
  fprintf(fich_gen, "\t\t//lista_vacia=TRUE;\n");
  if (!noprints)
    fprintf(fich_gen, "\t\terror_lista(tail, (*lreal).nombre);\n");
  fprintf(fich_gen,"\t\treturn(0.0);\n");
  fprintf(fich_gen,"\t}\n");
  fprintf(fich_gen,"\tcol= (*lreal).cola;\n");
  fprintf(fich_gen,"\tif (col==0) devol = (*lreal).lista[(*lreal).tam-1];\n");
  fprintf(fich_gen,"\t       else devol = (*lreal).lista[(*lreal).cola-1];\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"\treturn (devol);\n");
  fprintf(fich_gen,"}\n");
  fprintf(fich_gen,"\n");
  
  /* se genera la funcion pop_head_lreal que extrae y devuelve */
  /* el elemento que ocupa la posicion inicial de una lista real */
  fprintf(fich_gen,"float pop_head_lreal(LISTA_GENERICA_REALES *lreal) {\n");
  fprintf(fich_gen, "\tfloat devol;\n");
  fprintf(fich_gen, "\tint cab;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lreal).cabeza==-1) {   /* LISTA VACIA */\n");
  fprintf(fich_gen, "\t\t//lista_vacia=TRUE;\n");
  if(!noprints)
    fprintf(fich_gen, "\t\terror_lista(pophead, (*lreal).nombre);\n");
  fprintf(fich_gen, "\t\treturn(0.0);\n");
  fprintf(fich_gen, "\t}\n");
  fprintf(fich_gen, "\tdevol = (*lreal).lista[(*lreal).cabeza];\n");
  fprintf(fich_gen, "\tcab= (*lreal).cabeza;\n");
  fprintf(fich_gen, "\tif ((*lreal).cabeza==(*lreal).tam-1) (*lreal).cabeza=0;\n");
  fprintf(fich_gen, "\t\telse   (*lreal).cabeza++;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lreal).cabeza==(*lreal).cola) {\n");
  fprintf(fich_gen, "\t\t//lista_vacia = TRUE;\n");
  fprintf(fich_gen, "\t\t(*lreal).cabeza=-1;\n");
  fprintf(fich_gen, "\t\t(*lreal).cola=0;\n");
  fprintf(fich_gen, "\t}\n");
  fprintf(fich_gen, "\t(*lreal).num_elem--;\n");
  fprintf(fich_gen, "\treturn (devol);\n");
  fprintf(fich_gen, "}\n");
  fprintf(fich_gen, "\n");
  
  /* se genera la funcion pop_tail_lreal que extrae y devuelve el */
  /* elemento que ocupa la posicion final de una lista real */
  fprintf(fich_gen,"float pop_tail_lreal(LISTA_GENERICA_REALES *lreal) {\n");
  fprintf(fich_gen, "\tfloat devol;\n");
  fprintf(fich_gen, "\tint col;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lreal).cabeza==-1) {   /* LISTA VACIA */\n");
  fprintf(fich_gen, "\t\t//lista_vacia=TRUE;\n");
  if (!noprints)
    fprintf(fich_gen, "\t\terror_lista(poptail, (*lreal).nombre);\n");
  fprintf(fich_gen, "\t\treturn(0.0);\n");
  fprintf(fich_gen, "\t}\n");
  fprintf(fich_gen, "\tcol= (*lreal).cola;\n");
  fprintf(fich_gen, "\tif (col==0) devol = (*lreal).lista[(*lreal).tam-1];\n");
  fprintf(fich_gen, "\t       else devol = (*lreal).lista[(*lreal).cola-1];\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lreal).cola==0) (*lreal).cola=(*lreal).tam-1;\n");
  fprintf(fich_gen, "\telse   (*lreal).cola--;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lreal).cabeza==(*lreal).cola) {\n");
  fprintf(fich_gen, "\t\t//lista_vacia = TRUE;\n");
  fprintf(fich_gen, "\t\t(*lreal).cabeza=-1;\n");
  fprintf(fich_gen, "\t\t(*lreal).cola=0;\n");
  fprintf(fich_gen, "\t}\n");
  fprintf(fich_gen, "\t(*lreal).num_elem--;\n");
  fprintf(fich_gen, "\treturn (devol);\n");
  fprintf(fich_gen, "}\n");
  fprintf(fich_gen, "\n");
  
  /* se genera la funcion add_head_lreal que aade un elemento a la */
  /* cabeza de una lista entera */
  fprintf(fich_gen,"void add_head_lreal(const float num, LISTA_GENERICA_REALES *lreal) {\n");
  fprintf(fich_gen, "\tint cab;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lreal).cabeza==(*lreal).cola) {    /* si esta llena */\n");
  fprintf(fich_gen, "\t\t//lista_llena=TRUE;\n");
  fprintf(fich_gen, "\t\terror_lista(addhead, (*lreal).nombre);\n");
  fprintf(fich_gen, "\t\treturn;\n");
  fprintf(fich_gen, "\t}\n");
  fprintf(fich_gen, "\n");
  
  fprintf(fich_gen,"\tcab = (*lreal).cabeza;\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"\tif ((*lreal).cabeza<=0) (*lreal).cabeza=(*lreal).tam-1;\n");
  fprintf(fich_gen,"\t\telse  (*lreal).cabeza--;\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"\t(*lreal).lista[(*lreal).cabeza] = num;\n");
  fprintf(fich_gen,"\t(*lreal).num_elem++;\n");
  fprintf(fich_gen,"}\n");
  fprintf(fich_gen,"\n");



  fprintf(fich_gen,"void add_tail_lreal(const float num, LISTA_GENERICA_REALES *lreal) {\n");
  fprintf(fich_gen, "\tint col;\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lreal).cabeza==(*lreal).cola) {\n");
  fprintf(fich_gen, "\t\t//lista_llena=TRUE;\n");
  if (!noprints)
    fprintf(fich_gen, "\t\terror_lista(addtail, (*lreal).nombre);\n");
  fprintf(fich_gen, "\t\treturn;\n");
  fprintf(fich_gen, "\t}\n");
  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\tif ((*lreal).cabeza==-1) /* lista vacia */\n");
  fprintf(fich_gen, "\t{\n");
  fprintf(fich_gen, "\t\t(*lreal).cabeza = (*lreal).tam-1;\n");
  fprintf(fich_gen, "\t\t(*lreal).lista[(*lreal).cabeza]=num;\n");
  fprintf(fich_gen, "\t} else {\n");
  fprintf(fich_gen, "\t\tcol= (*lreal).cola;\n");
  fprintf(fich_gen, "\t\tif ((*lreal).cola==(*lreal).tam-1)  (*lreal).cola=0;\n");
  fprintf(fich_gen, "\t\t\telse (*lreal).cola++;\n");
  fprintf(fich_gen, "\t\t(*lreal).lista[col] = num;\n");
  fprintf(fich_gen, "\t}\n");
  fprintf(fich_gen, "\t(*lreal).num_elem++;\n");
  fprintf(fich_gen, "}\n");
  
  /* se genera la funcion size_lent que devuelve el tamao de una lista */
  /* entera */
  fprintf(fich_gen,"int size_lent(const LISTA_GENERICA_ENTEROS *lent) {\n");
  fprintf(fich_gen,"\treturn (lent->num_elem);\n");
  fprintf(fich_gen,"}\n");
  fprintf(fich_gen,"\n");
  
  /* se genera la funcion size_lreal que devuelve el tamao de una lista */
  /* entera */
  fprintf(fich_gen,"int size_lreal(const LISTA_GENERICA_REALES *lreal) {\n");
  fprintf(fich_gen,"\treturn (lreal->num_elem);\n");
  fprintf(fich_gen,"}\n");
  fprintf(fich_gen,"\n");
  
  fprintf(fich_gen,"\n");
} //de Genera_fich_tipos()

/********************** GENERA_INICIALIZACION_DATOS ************************/

void Genera_Inicializacion_datos()
{
 /* se inicializan las estructuras de las reglas que contienen las
  instanciaciones activadas */
  EHREGLA *aux;
  LISTA_OBJ *lo;
  LISTA *lvpeaux;
  REGTIPO * auxptro;
  int i;
  int dim;
  int cont;
  char nombre_condicion_regla[100];  /* almacena la cadena que identifica las estructuras */
  LISTA_CONDIC  * lc;

 fprintf(fich_gen,"void %sInicializa_estructuras()\n",prefijo);
 fprintf(fich_gen,"{\n");
 /* declaramos el maximo de iteradores */
 fprintf(fich_gen,"\tint ");
 for(i=1;i<=num_maxvp-1;i++) fprintf(fich_gen,"iter%d, ",i);
 fprintf(fich_gen,"iter%d;\n", num_maxvp);

 /* ahora creamos los bucles que recorreran las estructuras */
 for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
           if (!aux->eregla.tiene_copias) {
             fprintf(fich_gen,"/*** REGLA: %s ***/ \n\n",aux->eregla.nombre);
             dim=1;
             lvpeaux = aux->eregla.param;
             while (lvpeaux!=NULL) {
                     lo=busca_obj(aux->eregla.lobj_cond,lvpeaux->nombre);
                     if (lo->cuantificador==existe) {
                           for(cont=0; cont<dim; cont++) fprintf(fich_gen,"\t");
                           auxptro = busca_en_tipos(lo->instancia_de);
                           fprintf(fich_gen,"for(iter%d=0;iter%d<%d; iter%d++)\n",dim,dim,auxptro->num_instancias,dim);
                           dim++;
                      }
                      lvpeaux=lvpeaux->sig;
             }
             for(cont=0; cont<dim; cont++) fprintf(fich_gen,"\t");
             fprintf(fich_gen,"%sR%d",prefijo,aux->eregla.num_regla);
             for(cont=1; cont<dim; cont++) fprintf(fich_gen,"[iter%d]",cont);
             fprintf(fich_gen,"=0;\n");
             fprintf(fich_gen,"\n");

             /* ahora se inicializan las estructuras de las condiciones */
             lc=aux->eregla.lcond;
             while (lc!=NULL) {
               sprintf(nombre_condicion_regla,"_cond%d_%s%s",lc->num_condic,prefijo,aux->eregla.nombre);
               fprintf(fich_gen,"  /* CONDICION: %d REGLA: %s */ \n\n",lc->num_condic,aux->eregla.nombre);
               if (lc->es_orig) {
                 if (lc->num_vpe==0) { /* condicion simple */
                   fprintf(fich_gen,"\tactiv%s=FALSE;\n",nombre_condicion_regla);
                 } else {
                   lvpeaux = lc->param;
                   dim = 1;
                   while (lvpeaux!=NULL) {
                      lo=busca_obj(lc->lobj_cond,lvpeaux->nombre);
                      if (lo->cuantificador==existe) {
                           for(cont=0; cont<dim; cont++) fprintf(fich_gen,"\t");
                           auxptro = busca_en_tipos(lo->instancia_de);
                           fprintf(fich_gen,"for(iter%d=0;iter%d<%d; iter%d++)\n",dim,dim,auxptro->num_instancias,dim);
                           dim++;
                      }
                      lvpeaux=lvpeaux->sig;
                   }
                   for(cont=0; cont<dim; cont++) fprintf(fich_gen,"\t");
                   fprintf(fich_gen,"{  activ%s",nombre_condicion_regla);
                   for(cont=1; cont<dim; cont++) fprintf(fich_gen,"[iter%d]",cont);
                   fprintf(fich_gen,"=CELDA_VACIA;\n");
                   for(cont=0; cont<dim; cont++) fprintf(fich_gen,"\t");
                   fprintf(fich_gen,"   camb%s",nombre_condicion_regla);
                   for(cont=1; cont<dim; cont++) fprintf(fich_gen,"[iter%d]",cont);
                   fprintf(fich_gen,"=CELDA_VACIA;}\n");
                   fprintf(fich_gen,"\n");
                   fprintf(fich_gen,"\tnum_cambios%s=0;\n",nombre_condicion_regla);
                 }
               } else {
                 if (lc->num_vpe>0)
                    fprintf(fich_gen,"\tnum_cambios%s=0;\n",nombre_condicion_regla);
               }
               lc=lc->sig;
               fprintf(fich_gen,"\n");
             }
           }
           aux=aux->sig;
         }
    }
    fprintf(fich_gen,"}\n");
    fprintf(fich_gen,"\n");

}

/************************** GENERA_FUNCION_RESET ****************************/

void Genera_funcion_reset() {
  /* la funcion reset se ejecuta al principio para a partir de los valores
     iniciales de las instancias cargar la agenda */
  EHTDS *aux;
  int i;
  LISTA_CAMPOS * l;
  LISTA_CAMPOS * lins;
  LIS_NUM * lnum;
  LISTA *lista_reglas;
  LISTA *reglas_invol;
  LIS_NUM *lista_num_conds;
  REGREGLA *entr;
  LISTA_CONDIC * ent_cond;       /* entrada de condicion */
  LISTA *lvpeaux;
  int cont;
  LISTA_OBJ *lo2;
  REGTIPO *entip;
  LISTA * reglas_aux;
  int nvp_igualclase; /* numero de variables patron de la condicion */
                      /* con la misma clase */
  int num_cond_gen;   /* numero de condiciones generadas por variable patron */
  int ya_gen;         
  
  
  fprintf(fich_gen,"/*----------------- FUNCION RESET -------------------*/\n");
  fprintf(fich_gen,"void %sReset()\n",prefijo);
  fprintf(fich_gen,"{\n");
  
  if (log_)
    fprintf(fich_gen, "\t%slog_reset();\n\n", prefijo);

  reglas_invol = NULL;
  for(i=0;i<HASH_SIZE;i++)
    if(tds[i]!=NULL) {
      aux=tds[i];
      while(aux!=NULL) {
	if (aux->etds.clvar==instancia) {
	  if (aux->etds.tipo==definido) {
	    entip = busca_en_tipos(aux->etds.instancia_de);
	    lins = aux->etds.lc;
	    l=entip->lc;
	    /* FALTA: que se llame a las condiciones afectadas */
	    /* por las inicializaciones de los campos que a su */
	    /* vez tienen campos */
	    while(l!=NULL) {
	      if (lins->inicializado) {
		lista_reglas = l->reglas;
		lista_num_conds = l->num_conds;
		while (lista_reglas!=NULL) {
		  entr=busca_en_reglas(lista_reglas->nombre);
		  /* ahora se localiza la condicion */
		  ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
		  if (!entr->tiene_copias) {
		    if (ent_cond->es_orig) {
		      
		      /* primero contamos el numero de parametros de la condicion */
		      /* cuya clase es la misma que la vp modificada */
		      lvpeaux = ent_cond->param;
		      nvp_igualclase=0;
		      while (lvpeaux!=NULL) {
			lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
			if (lo2->cuantificador==existe) {
			  if (strcmp(lo2->instancia_de,aux->etds.instancia_de)==0) 
			    nvp_igualclase++;
			}
			lvpeaux=lvpeaux->sig;
		      }
		      num_cond_gen = 0;
		      do {
			fprintf(fich_gen,"\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
			lvpeaux = ent_cond->param;
			cont=0;
			ya_gen = 0;
			while (lvpeaux!=NULL) {
			  lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
			  if (lo2->cuantificador==existe) {
			    if (cont>0) fprintf(fich_gen,",");
			    if (strcmp(lo2->instancia_de,aux->etds.instancia_de)==0) {
			      if (ya_gen==num_cond_gen) 
				fprintf(fich_gen,"%d",aux->etds.cod.num_instancia);
			      else   
				fprintf(fich_gen,"-1");
			      ya_gen++;
			    } 
			    else fprintf(fich_gen,"-1");
			    cont++;
			  }
			  lvpeaux=lvpeaux->sig;
			}
			fprintf(fich_gen,");\n");
			num_cond_gen++;               
		      } while (num_cond_gen<nvp_igualclase);
		    }              
		    reglas_aux = insertar_sindup(reglas_invol,lista_reglas->nombre);
		    if (reglas_aux!=NULL) reglas_invol = reglas_aux;
		  }
		  lista_reglas=lista_reglas->sig;
		  lista_num_conds = lista_num_conds->sig;
		}
		
		lista_reglas = lins->reglas;
		lista_num_conds = lins->num_conds;
		while (lista_reglas!=NULL) {
		  entr=busca_en_reglas(lista_reglas->nombre);
		  /* ahora se localiza la condicion */
		  ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
		  if (!entr->tiene_copias) {
		    if (ent_cond->es_orig) {
		      lvpeaux = ent_cond->param;
		      cont=0;
		      fprintf(fich_gen,"\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
		      while (lvpeaux!=NULL) {
			lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
			if (cont>0) fprintf(fich_gen,",");
			if (strcmp(lo2->instancia_de,aux->etds.instancia_de)==0) fprintf(fich_gen,"%d",aux->etds.cod.num_instancia);
			else fprintf(fich_gen,"-1");
			cont++;
			lvpeaux=lvpeaux->sig;
		      }
		      fprintf(fich_gen,");\n");
		    }
		    reglas_aux = insertar_sindup(reglas_invol,lista_reglas->nombre);
		    if (reglas_aux!=NULL) reglas_invol = reglas_aux;
		  }
		  lista_reglas=lista_reglas->sig;
		  lista_num_conds = lista_num_conds->sig;
		}
		
	      }
	      l=l->siguiente;
	      lins = lins->siguiente;
	    }
	  } 
	  else {
	    lista_reglas = aux->etds.reglas;
	    lista_num_conds = aux->etds.num_conds;
	    while (lista_reglas!=NULL) {
	      entr=busca_en_reglas(lista_reglas->nombre);
	      /* ahora se localiza la condicion */
	      ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
	      if (!entr->tiene_copias) {
		if (ent_cond->es_orig) {
		  lvpeaux = ent_cond->param;
		  cont=0;
		  fprintf(fich_gen,"\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
		  while (lvpeaux!=NULL) {
		    lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
		    if (cont>0) fprintf(fich_gen,",");
		    if (strcmp(lo2->instancia_de,aux->etds.instancia_de)==0) fprintf(fich_gen,"%d",aux->etds.cod.num_instancia);
		    else fprintf(fich_gen,"-1");
		    cont++;
		    lvpeaux=lvpeaux->sig;
		  }
		  fprintf(fich_gen,");\n");
		}
		reglas_aux = insertar_sindup(reglas_invol,lista_reglas->nombre);
		if (reglas_aux!=NULL) reglas_invol = reglas_aux;
	      }
	      lista_reglas=lista_reglas->sig;
	      lista_num_conds = lista_num_conds->sig;
	    }
	  }
	}
	aux=aux->sig;
      }
    }
  /* ahora se imprimen todas las reglas de las que */
  /* alguna condicion haya sido afectada por una inicializacion */
  lvpeaux= reglas_invol;
  while(lvpeaux!=NULL) {
    fprintf(fich_gen,"\tIz_%s%s();\n",prefijo,lvpeaux->nombre);
    lvpeaux=lvpeaux->sig;
  }
  borrar(reglas_invol);

  if (log_)
    fprintf(fich_gen, "\n\t%slog_end_reset();\n", prefijo);
  
  fprintf(fich_gen,"}\n\n");
}

/************************** GENERA_FUNCION_RUN ****************************/

void Genera_funcion_run()
{
  EHREGLA *aux;
  LISTA_OBJ *lo;
  LISTA *lvpeaux;
  int i;
  int dim;
  REGREGLA *regla;

 fprintf(fich_gen,"void %sStep()\n",prefijo);
 fprintf(fich_gen,"{\n");
 /* Declaracion de variables */
 fprintf(fich_gen,"\t %sREGAGENDA r;\n",prefijo);
 fprintf(fich_gen,"\n");
 fprintf(fich_gen,"\tr = %sobten_regla_mas_prio();\n",prefijo);
 fprintf(fich_gen,"\tif (!%sagenda_vacia) {\n",prefijo);
 fprintf(fich_gen,"\t\t%selimregla_agenda(r);\n",prefijo);
 fprintf(fich_gen,"\t\tswitch (r.num_regla) {\n");
 for(i=0;i<HASH_SIZE;i++)
     if(reglas[i]!=NULL) {
         aux=reglas[i];
         while(aux!=NULL) {
           if (!aux->eregla.tiene_copias) {
             regla=&(aux->eregla);
             fprintf(fich_gen,"\t\t\tcase %d:  ", regla->num_regla);
             while (regla->copia_de!=NULL) regla=regla->copia_de;
             dim=0;
             fprintf(fich_gen,"De_%s%s(", prefijo,regla->nombre);
             lvpeaux = regla->param;
             while (lvpeaux!=NULL) {
                     lo=busca_obj(regla->lobj_cond,lvpeaux->nombre);
                     if (lo->cuantificador==existe) {
                       if (dim>0) fprintf(fich_gen,",");
                       fprintf(fich_gen,"r.instancias[%d]",dim);
                       dim++;
                      }
                      lvpeaux=lvpeaux->sig;
             }

             fprintf(fich_gen,");\n");
             fprintf(fich_gen,"\t\t\t\t break;\n");
           }
           aux=aux->sig;
         }
    }
    fprintf(fich_gen,"\t\t}\n");
    fprintf(fich_gen,"\t}\n");
    fprintf(fich_gen,"}\n\n");

}

/****** Busca_op_lista ******/
void Busca_op_lista(exp,reglas_inv)
  EXPRESION *exp;
  LISTA **reglas_inv;
{
  /* esta funcion se encarga de buscar operaciones con listas pop_head y  */
  /* pop_tail en una expresion de una accion, para poder llamar a las     */
  /* condiciones afectadas */
  REGTDS *entvp;
  REGTIPO * entclase;
  LISTA *campos,*camp_aux;
  LISTA_CAMPOS *lcamp,*lcamp_aux;
  LISTA * lista_reglas;
  LIS_NUM *lista_num_conds;
  REGREGLA *entr;
  LISTA_CONDIC * ent_cond;       /* entrada de condicion */
  LISTA *lvpeaux;
  int cont,cont2,cont3,cont4;
  LISTA * reglas_aux;
  LISTA * reglas_asig;
  char cad_aux[255];
  int instref;
  int i;
  EHTDS *auxtds;
  LISTA_OBJ *lo,*lo2;
  int *param;
  int num_param;
  EXPRESION *camp_exp;

  while (exp!=NULL) {

   if (!exp->esoperando && (exp->tipooperador==popheadsim || exp->tipooperador == poptailsim)) {
     /* pasamos el parentesis */
     exp=exp->sig;
     exp=exp->sig;

     if (!exp->esoperando && exp->tipooperador==existe) exp=exp->sig;

     entvp = busca_en_tds(exp->ident);
     campos = NULL;
     camp_exp = NULL;
     if (entvp->tipo==definido) {

         while (exp->sig!=NULL) {
          if(!exp->sig->esoperando && exp->sig->tipooperador==punto) {
              exp=exp->sig;
              /* se crea una expresion con los campos */
              camp_exp=inserta_lista_exp(camp_exp,0,0,0,0,0,(char *)NULL,(ENTREAL)0);

              exp=exp->sig;
              camp_exp=inserta_lista_exp(camp_exp,0,0,0,0,0,exp->ident,(ENTREAL)0);
              if (campos==NULL) campos = cr_lista_id(campos,exp->ident);
                           else campos = insertar(campos, exp->ident);
          } else break;
	 }    
     }

     if (entvp->clvar ==instancia ) /* si es una instancia */
     {
       reglas_asig = NULL;
       if (entvp->tipo==definido) {
          entclase=busca_en_tipos(entvp->instancia_de);
          /* ahora se van recorriendo los campos */
          camp_aux = campos;
          lcamp_aux = entclase->lc;
          while (camp_aux!=NULL) {
             lcamp = busca_campo(camp_aux->nombre, lcamp_aux);
             lcamp_aux = lcamp->lc;
             camp_aux=camp_aux->sig;
          }
          lista_reglas= lcamp->reglas;
          lista_num_conds = lcamp->num_conds;
       } else {
          lista_reglas= entvp->reglas;
          lista_num_conds = entvp->num_conds;
       }

       while (lista_reglas!=NULL) {
          entr=busca_en_reglas(lista_reglas->nombre);
          ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
          if (!entr->tiene_copias) {
            if (ent_cond->es_orig) {
             fprintf(fich_gen,"\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
             lvpeaux = ent_cond->param;
             cont=0;
             while (lvpeaux!=NULL) {
                 lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                 if (lo2->cuantificador==existe) {
                    if (cont>0) fprintf(fich_gen,",");
                    if (strcmp(lo2->instancia_de,entvp->instancia_de)==0) fprintf(fich_gen,"%d",entvp->cod.num_instancia);
                                                            else fprintf(fich_gen,"-1");
                    cont++;
                 }
                 lvpeaux=lvpeaux->sig;
           }
           fprintf(fich_gen,");\n");
         }
         reglas_aux=insertar_sindup(reglas_asig,lista_reglas->nombre);
         if (reglas_aux!=NULL) {
               reglas_asig = reglas_aux;
               /* se genera la activacion de este logico si antes no se habia */
               /* generado ya */
               fprintf(fich_gen,"\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
         }
        }
        lista_reglas=lista_reglas->sig;
        lista_num_conds = lista_num_conds->sig;
       }

       if (entvp->tipo==definido) {
          /* ahora se van recorriendo los campos */
          camp_aux = campos;
          lcamp_aux = entvp->lc;
          while (camp_aux!=NULL) {
             lcamp = busca_campo(camp_aux->nombre, lcamp_aux);
             lcamp_aux = lcamp->lc;
             camp_aux=camp_aux->sig;
          }
          lista_reglas= lcamp->reglas;
          lista_num_conds = lcamp->num_conds;
          while (lista_reglas!=NULL) {
            entr=busca_en_reglas(lista_reglas->nombre);
            ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
            if (!entr->tiene_copias) {
              if (ent_cond->es_orig) {
               fprintf(fich_gen,"\t\tcond%d_%s%s(",lista_num_conds->num.entero,prefijo,lista_reglas->nombre);
               lvpeaux = ent_cond->param;
               cont=0;
               while (lvpeaux!=NULL) {
                 lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                 if (lo2->cuantificador==existe) {
                    if (cont>0) fprintf(fich_gen,",");
                    if (strcmp(lo2->instancia_de,entvp->instancia_de)==0) fprintf(fich_gen,"%d",entvp->cod.num_instancia);
                                                            else fprintf(fich_gen,"-1");
                    cont++;
                 }
                 lvpeaux=lvpeaux->sig;
             }
             fprintf(fich_gen,");\n");
           }
           reglas_aux=insertar_sindup(reglas_asig,lista_reglas->nombre);
           if (reglas_aux!=NULL) {
               reglas_asig = reglas_aux;
               /* se genera la activacion de este logico si antes no se habia */
               /* generado ya */
               fprintf(fich_gen,"\t\tLIz_%s%s=TRUE;\n",prefijo,lista_reglas->nombre);
           }
          }
          lista_reglas=lista_reglas->sig;
          lista_num_conds = lista_num_conds->sig;
         }
       }


       while (reglas_asig!=NULL)
       {
         reglas_aux = insertar_sindup(*reglas_inv,reglas_asig->nombre);
         if (reglas_aux!=NULL) *reglas_inv = reglas_aux;
         reglas_asig= reglas_asig->sig;
       }

     } else if (entvp->clvar == paternvar ) /* si es una variable patron */
       {
         reglas_asig = NULL;
         entclase=busca_en_tipos(entvp->instancia_de);
         /* ahora se van recorriendo los campos */
         camp_aux = campos;
         lcamp_aux = entclase->lc;
         while (camp_aux!=NULL) {
             lcamp = busca_campo(camp_aux->nombre, lcamp_aux);
             lcamp_aux = lcamp->lc;
             camp_aux=camp_aux->sig;
         }
         lista_reglas= lcamp->reglas;
         lista_num_conds = lcamp->num_conds;

         Genera_lista_conds_vp(lista_reglas,lista_num_conds,entvp,&reglas_asig);
	 Genera_conds_vp(entvp,camp_exp,&reglas_asig); 

         while (reglas_asig!=NULL)
         {
           reglas_aux = insertar_sindup(*reglas_inv,reglas_asig->nombre);
           if (reglas_aux!=NULL) *reglas_inv = reglas_aux;
           reglas_asig= reglas_asig->sig;
         }

       }
   }  
   exp=exp->sig;
  }
}
/************************** GENERA_REGLAS_DER ****************************/

void Genera_reglas_der() {
  EHREGLA *aux;
  LISTA_OBJ *lo,*lo2;
  REGTIPO *entclase;
  REGTDS * entvp;
  REGREGLA * entr;
  LISTA_ACCIONES *la;
  EXPRESION *exp;
  EXPRESION *expaux;
  LISTA *lista_reglas;
  LIS_NUM *lnum;
  LISTA *lvpeaux;
  LISTA *campos, *campaux;
  LISTA_CAMPOS *lcamp, *lcamp_aux;
  int i;
  int cont,cont2,cont3,cont4;
  int num_obj;
  int masdeuno;
  REGREGLA *regla;
  EHTDS *auxtds;
  int i2;
  char tipo_expr;
  char cad_entrecom[255];
  char cad_variables[255];
  char cad_aux[255];
  int nueva_exp;
  LISTA_CONDIC * ent_cond;       /* entrada de condicion */
  LISTA * reglas_invol;          /* reglas que deben llamarse al final de la PD de la regla */
  LISTA * reglas_aux;
  int instref;
  REGFUNC *func;
  LISTA_PARAM *lp,*lparam;
  LISTA_ACCIONES *laaux;
  int num_asig;
  
  int num_max_enteros;  /* numero maximo de enteros y reales que se cambian */
  int num_max_reales;   /* en la funcion */
  int num_ent;
  int num_real;
  MODIFICACIONES * mod_aux;
  
  fprintf(fich_gen,"/*--------------- PARTES DERECHA DE LAS REGLAS ----------------*/\n");
  num_maxvp=0;
  for(i=0;i<HASH_SIZE;i++)
    if(reglas[i]!=NULL) {
      aux=reglas[i];
      while(aux!=NULL) {
	if (!aux->eregla.tiene_copias) {
	  regla= &(aux->eregla);
	  if (regla->copia_de!=NULL) {
	    while (regla->copia_de!=NULL) regla=regla->copia_de;
	  }
	  if (!regla->ya_generado) {
	    /* se inicializan las reglas involucradas */
	    reglas_invol = NULL;
	    
	    regla->ya_generado = TRUE;
	    fprintf(fich_gen,"void De_%s%s(",prefijo,regla->nombre);
	    /* ahora se genera el perfil de la funcion */
	    lvpeaux = regla->param;
	    masdeuno = FALSE;
	    while (lvpeaux!=NULL) {
	      lo=busca_obj(regla->lobj_cond,lvpeaux->nombre);
	      if (lo->cuantificador==existe) {
		if (masdeuno)
		  fprintf(fich_gen,",");
		fprintf(fich_gen,"%s",lo->nombre);
		masdeuno = TRUE;
	      }
	      lvpeaux=lvpeaux->sig;
	    }
	    fprintf(fich_gen,")\n");
	    
	    /* aqui se generan la declaracion de los parametros */
	    lvpeaux = regla->param;
	    while (lvpeaux!=NULL) {
	      lo=busca_obj(regla->lobj_cond,lvpeaux->nombre);
	      if (lo->cuantificador==existe) {
		fprintf(fich_gen,"\tint %s;\n",lo->nombre);
	      }
	      lvpeaux=lvpeaux->sig;
	    }
	    fprintf(fich_gen,"{\n");
	    /* ahora contamos las asignaciones para declarar una array */
	    /* de booleanos que contenga si se ha producido o no cambio */
	    la=regla->lacc;
	    num_asig= 0;
	    num_max_enteros = 0;
	    num_max_reales = 0;
	    while (la!=NULL) {
	      if (la->tipo_accion==asig_patter || la->tipo_accion==asig_instan)
		num_asig++;
	      /* si es una funcion tambien han de considerarse las modificaciones */
	      if (la->tipo_accion==llamada_subprg) {
		/* ahora se procesan las modificaciones para ver cuantos */
		/* enteros y reales hay que guardar para almacenar los cambios */
		num_ent = 0;
		num_real = 0;
		func = busca_en_funciones(la->ident);
		if (func->tipo==interna) {
		  mod_aux = func->modif;
		  while (mod_aux!=NULL) {
		    if (mod_aux->clase_obj==instancia) {
		      switch (mod_aux->tip_izq) {
			//CGM 14/ene/2004: el caso de declaracion de un booleano no estaba contemplado
			//y no se declaraba un vector que luego se usaba (ejemplos\avanza.arl)
			//Esto soluciono ese problema, quedando por considerar el switch de abajo
		      case boolean: 
		      case subrango:
		      case entero:
			num_ent++;
			break;
		      case real:
			num_real++;
			break;			      
		      }
		    } 
		    else
		      if (mod_aux->clase_obj==paramt) {
			/* si es un parametro por ref, depende de si */
			/* lo pasado en los argumentos de la funcion */
			/* es o no un objeto de la Base de Hechos */
			
			/* se busca la posicion del parametro */
			lparam = func->lp;
			i2=0;
			while(strcmp(lparam->nombre,mod_aux->nombre)) {
			  lparam=lparam->sig;
			  i2++;
			}

			expaux=busca_expresion(la->expr,i2);
			if (!expaux->esoperando && expaux->tipooperador==existe) 
			  expaux=expaux->sig;
			entvp=busca_en_tds(expaux->ident);
			if ((entvp->clvar == instancia) ||
			    (entvp->clvar == paternvar)) {
			  /* si es un objeto de la BH se cuenta */				
			  switch (mod_aux->tip_izq) {
			  case boolean:
			  case subrango:
			  case entero:
			    num_ent++;
			    break;
			  
			  case real:
			    num_real++;
			    break;
			  }
			}
		      }
		    mod_aux=mod_aux->sig;
		  }
		}
		/* se controla el numero maximo de enteros y de reales */
		/* que se van a necesitar al llamar a una funcion */
		if (num_ent>num_max_enteros) 
		  num_max_enteros = num_ent;
		if (num_real>num_max_reales)
		  num_max_reales = num_real;
	      }
	      la=la->sig;
	    }
	    if (num_asig > 0) {
	      fprintf(fich_gen,"\tint   cambio_asig[%d];\n",num_asig);
	      fprintf(fich_gen,"\tint   old_value_ent;\n");
	      fprintf(fich_gen,"\tfloat old_value_real;\n");
	      fprintf(fich_gen,"\tint   iter;\n");
	    }
	    if (num_max_enteros>0)
	      fprintf(fich_gen,"\tint   old_val_ent[%d];\n",num_max_enteros);
	    if (num_max_reales>0)
	      fprintf(fich_gen,"\tint   old_val_real[%d];\n",num_max_reales);

	    fprintf(fich_gen,"\n");

	    if (debug) {
	      fprintf(fich_gen,"\tprintf(\">Ejecucion: %s ",regla->nombre);
	      if (regla->param!=NULL) {
		lvpeaux = regla->param;
		while (lvpeaux!=NULL) {
		  lo=busca_obj(regla->lobj_cond,lvpeaux->nombre);
		  if (lo->cuantificador==existe) {
		    fprintf(fich_gen," %s: ",lo->nombre);
		    fprintf(fich_gen,"%s","%d");
		  }
		  lvpeaux=lvpeaux->sig;
		}
		fprintf(fich_gen,"\\n\"");
		lvpeaux = regla->param;
		while (lvpeaux!=NULL) {
		  lo=busca_obj(regla->lobj_cond,lvpeaux->nombre);
		  if (lo->cuantificador==existe) {
		    fprintf(fich_gen,",%s",lo->nombre);
		  }
		  lvpeaux=lvpeaux->sig;
		}
		fprintf(fich_gen,");\n");
	      }
	      else
		fprintf(fich_gen,"\\n\");\n");
	      fprintf(fich_gen,"\n");
	    }
	    if (num_asig> 0 )
	      fprintf(fich_gen,"\tfor(iter=0;iter<%d; iter++) cambio_asig[iter]=FALSE;\n\n",num_asig);
	    
	    la=regla->lacc;
	    num_asig= 0;
	    while (la!=NULL) {
	      cad_aux[0]='\0';
	      switch (la->tipo_accion) {
		
	      case oplista_addhead:
	      case oplista_addtail:
		Genera_accion_addlist(la->tipo_accion, la->ident, la->campos, 
				       la->tip_izq, la->expr, regla->lobj_acc, 
				       &reglas_invol, num_asig);
		break;
		
	      case asig_patter:
		Genera_accion_asig_vp(la->ident,la->campos,la->tip_izq,la->expr,regla->lobj_acc,&reglas_invol,num_asig);
		num_asig++;
		break;
	
	      case asig_instan:
		Genera_accion_asig_inst(la->ident,la->campos,la->tip_izq,la->expr,regla->lobj_acc,&reglas_invol,num_asig);
		num_asig++;
		break;
		
	      case asig_var:
		entvp=busca_en_tds(la->ident);
		if (entvp->tipo==definido) {
		  fprintf(fich_gen,"\t%sv_%s[%d]",prefijo,entvp->instancia_de,entvp->cod.num_instancia);
		  exp = la->campos;
		  while (exp!=NULL) {
		    exp=exp->sig;
		    fprintf(fich_gen,".%s",exp->ident);
		    if (exp->indice_array != NULL) {
		      Genera_exp(exp->indice_array, cad_aux, regla->lobj_acc);
		      fprintf(fich_gen, "[%s]", cad_aux);
		    }
		    exp=exp->sig;
		  }
		} 
		else {
		  fprintf(fich_gen,"\t%s%s",prefijo,la->ident);
		  if (entvp->tipo == array) {
		    Genera_exp(la->campos, cad_aux, regla->lobj_acc);
		    fprintf(fich_gen, "[%s]", cad_aux);
		  }
		}
		  
		fprintf(fich_gen,"=");
		Genera_exp(la->expr,cad_aux,regla->lobj_acc);
		fprintf(fich_gen,"%s;\n",cad_aux);
		break;
		
	      case llamada_subprg:
		Genera_accion_llam_func(la->ident,la->expr,regla->lobj_acc,&reglas_invol);
		break;
		
	      case log_facts:
		if (log_) {
		  fprintf(fich_gen, "\t%slog_facts();\n", prefijo);
		}
		break;
		
	      case imprimir:
		if (!noprints) { 
		  i2=0;
		  exp = la->expr;
		  while (exp!=NULL) {
		    cad_entrecom[0]='\0';
		    cad_variables[0]='\0';
		    tipo_expr = la->tip_lista_expr[i2];
		    switch (tipo_expr) {
		    case real:
		    case entero:
		    case subrango:
		    case cadena:
		      //Para los tipos simples de c existe una parte propia al tipo
		      //y una parte comn a todos ellos. Pero adems se soporta la 
		      //impresin de tipos propios de arlips. Por eso el doble switch.
		      switch (tipo_expr) {
			//Parte propia de cada tipo simple de c.
		      case real:
			strcat(cad_entrecom,"\%f");
			exp=Genera_exp(exp,cad_aux,regla->lobj_acc);
			strcat(cad_variables,", ");
			strcat(cad_variables,cad_aux);
			break;
		      case entero:
		      case subrango:
			strcat(cad_entrecom,"\%d");
			exp=Genera_exp(exp,cad_aux,regla->lobj_acc);
			strcat(cad_variables,", ");
			strcat(cad_variables,cad_aux);
			break;
		      case cadena:
			strcat(cad_entrecom,exp->cad);
			exp=exp->sig;
			break;
		      } //switch de un tipo simple en c (y de la parte propia del tipo)
		      
		      //Parte comn a todos los tipos simples de c:
		      if (artis_mode) {
			fprintf(fich_gen, "#ifdef MODULE\n");
			fprintf(fich_gen, "\trtl_printf(");
			fprintf(fich_gen, "\"%s\"", cad_entrecom);
			fprintf(fich_gen, "%s);\n", cad_variables);
			fprintf(fich_gen, "#else\n");
			fprintf(fich_gen, "\tprintf(");
			fprintf(fich_gen, "\"%s\"", cad_entrecom);
			fprintf(fich_gen, "%s);\n", cad_variables);
			fprintf(fich_gen, "#endif\n\n");
		      } //si artis_mode
		      else {			 
			fprintf(fich_gen, "\tprintf(\"");
			fprintf(fich_gen, "%s\"",cad_entrecom);
			fprintf(fich_gen, "%s);\n", cad_variables);
		      } //else artis_mode
		      break;
		      
		      //Se han tratado los tipos simples.
		      //Para los tipos definidos en arlips se necesitan
		      //funciones propias que hagan la impresin
		    case lista_enteros:
		      exp = Genera_exp(exp, cad_aux, regla->lobj_acc);
		      fprintf(fich_gen, "\tprint_lent(stdout, &%s);\n", cad_aux); 
		      break;
		    case lista_reales:
		      exp = Genera_exp(exp, cad_aux, regla->lobj_acc);
		      fprintf(fich_gen, "\tprint_lreal(stdout, &%s);\n", cad_aux); 
		      break;
		    case array:
		      exp = Genera_exp(exp, cad_aux, regla->lobj_acc);
		      fprintf(fich_gen, "\tprint_array(stdout, %s);\n", cad_aux);
		      break;
		    } //switch tipo de expresion a imprimir
		    
		    //CGM: Para qu se hacen estas comprobaciones
		    //antes de pasar al siguiente elemento?
		    //Pueden no cumplirse estas condiciones?
		    if (exp!=NULL)
		      if (!exp->esoperando && exp->tipooperador==coma)
			exp=exp->sig;
		    i2++;
		  } //Mientras haya expresiones
		} //Si hay que imprimir
		break;
		
	      case parada:
		fprintf(fich_gen,"\t%shalt=TRUE;\n",prefijo);
	      }
	      
	      if (la->expr!=NULL) {
		Busca_op_lista(la->expr,&reglas_invol);
	      } 
	      la=la->sig;
	      fprintf(fich_gen,"\n");
	    }
	    
	    /* ahora se imprimen todas las reglas de las que */
	    /* alguna condicion haya sido afectada por la PD de la regla */
	    lvpeaux= reglas_invol;
            while(lvpeaux!=NULL) {
	      fprintf(fich_gen,"\tif (LIz_%s%s) Iz_%s%s();\n",prefijo,lvpeaux->nombre,prefijo,lvpeaux->nombre);
	      lvpeaux=lvpeaux->sig;
	    }
	    borrar(reglas_invol);
	    fprintf(fich_gen,"\n}\n");
	  }
	}
	aux=aux->sig;
      }
    }
}

/**************************** GENERA FUNCIONES ****************************/
void Genera_funciones()
{
  EHFUNC *aux;
  int i;
  LISTA_ACCIONES *la;
  char cad_entrecom[255];
  char cad_variables[255];
  char cad_aux[255];
  int nueva_exp;
  int i2;
  EXPRESION *exp;
  char tipo_expr;
  LISTA_OBJ *lo,*lo2;
  REGTIPO *entclase;
  REGTDS * entvp;
  REGFUNC *funcion;
  LISTA *campos, *campaux;
  LISTA_CAMPOS *lcamp_aux, *lcamp;
  int num_obj;
  LISTA_PARAM *lparam;
  char cad_porref[2];
  int *param;
  int num_param;
  LISTA *lista_reglas;
  LIS_NUM *lista_num_conds;
  REGREGLA *entr;
  LISTA_CONDIC *ent_cond;
  LISTA *lvpeaux;
  int cont;

  cad_porref[0]='\x0';
  fprintf(fich_gen,"/*--------------------- FUNCIONES INTERNAS --------------------*/\n");
  fprintf(fich_gen,"\n");
  for(i=0;i<HASH_SIZE;i++)
     if(funciones[i]!=NULL) {
         aux=funciones[i];
         while(aux!=NULL) {
          if (aux->efunc.tipo==interna) {
           funcion= &(aux->efunc);
           declarando_funcion = TRUE;
           entfuncion = funcion;
           switch (funcion->tipo_dev) {
             case boolean:
             case entero: fprintf(fich_gen,"int ");
                         break;

             case real:  fprintf(fich_gen,"float ");
                         break;
             case vacio: fprintf(fich_gen,"void ");
           }
           fprintf(fich_gen,"%s%s(",prefijo,funcion->nombre);
           lparam = funcion->lp;
           while (lparam!=NULL) {
              fprintf(fich_gen,"%s",lparam->nombre);
              lparam=lparam->sig;
              if (lparam!=NULL) fprintf(fich_gen,",");
           }
           fprintf(fich_gen,")\n");
           lparam = funcion->lp;
           while (lparam!=NULL) {
              cad_porref[0]='\x0';
              if (lparam->por_ref) strcat(cad_porref,"*");
              switch(lparam->ptro->tipo) {
                case definido: lo = busca_obj(funcion->lobj_func,lparam->nombre);
                               fprintf(fich_gen,"\t%s%s  %s %s;\n",prefijo,lo->instancia_de,cad_porref,lparam->nombre);
                               break;
                case entero:

                case boolean: fprintf(fich_gen,"\tint   %s %s;\n",cad_porref,lparam->nombre);
                              break;
                case real:    fprintf(fich_gen,"\tfloat %s %s;\n",cad_porref,lparam->nombre);
              }
              lparam=lparam->sig;
           }
           fprintf(fich_gen,"{\n");
           la=funcion->lacc;
           funcion->cond_ac = NULL;
           while (la!=NULL) {
                   cad_aux[0]='\0';
                   switch (la->tipo_accion) {

                   case imprimir:
		     if(!noprints) {
                                    fprintf(fich_gen,"\tprintf(\"");
                                    cad_entrecom[0]='\0';
                                    cad_variables[0]='\0';
                                    i2=0;
                                    exp = la->expr;
                                    while (exp!=NULL) {
                                         tipo_expr = la->tip_lista_expr[i2];
                                         if (tipo_expr==real) {
                                                 strcat(cad_entrecom,"\%f");
                                                 exp=Genera_exp(exp,cad_aux,funcion->lobj_func);
                                                 strcat(cad_variables,", ");
                                                 strcat(cad_variables,cad_aux);
                                              } else if (tipo_expr==entero || tipo_expr == subrango) {
                                                        strcat(cad_entrecom,"\%d");
                                                        exp=Genera_exp(exp,cad_aux,funcion->lobj_func);
                                                        strcat(cad_variables,", ");
                                                        strcat(cad_variables,cad_aux);
                                                   } else if (tipo_expr==cadena) {
                                                             strcat(cad_entrecom,exp->cad);
                                                             exp=exp->sig;
						   }
					  if (exp!=NULL)
                                               if (!exp->esoperando && exp->tipooperador==coma)
                                                               exp=exp->sig;
                                          i2++;
                                    }
                                    fprintf(fich_gen,"%s\"",cad_entrecom);
                                    if (i2==1 && cad_variables[0]!='\0') fprintf(fich_gen,","); /* si solo hay una variable */
                                    fprintf(fich_gen,"%s);\n", cad_variables);
		     }	    
		     break;

                     case retorno_subprg:   fprintf(fich_gen,"\treturn(");
                                            Genera_exp(la->expr,cad_aux,funcion->lobj_func);
                                            fprintf(fich_gen,"%s);\n",cad_aux);
                                            break;
                     case asig_instan:
                                    entvp=busca_en_tds(la->ident);
                                    if (entvp->tipo!=definido) fprintf(fich_gen,"\t%s%s=",prefijo,la->ident);
                                    else {
                                          fprintf(fich_gen,"\t%si_%s[%d]",prefijo,entvp->instancia_de,entvp->cod.num_instancia);
                                          exp = la->campos;
                                          while (exp!=NULL) {
                                                exp=exp->sig;
                                                fprintf(fich_gen,".%s",exp->ident);
                                                exp=exp->sig;
                                           }
                                           fprintf(fich_gen,"=");
                                    }

                                    Genera_exp(la->expr,cad_aux,funcion->lobj_func);
                                    fprintf(fich_gen,"%s;\n",cad_aux);

                                    /* ahora acumulamos los procedimientos que se */
                                    /* corresponden con la parte izq de la regla */
                                    if (entvp->tipo==definido) {
                                       entclase=busca_en_tipos(entvp->instancia_de);
                                       /* ahora se van recorriendo los campos */
                                       exp = la->campos;
                                       lcamp_aux = entclase->lc;
                                       while (exp!=NULL) {
                                             exp=exp->sig;
                                             lcamp = busca_campo(exp->ident, lcamp_aux);
                                             lcamp_aux = lcamp->lc;
                                             exp=exp->sig;
                                       }
                                       lista_reglas= lcamp->reglas;
                                       lista_num_conds = lcamp->num_conds;
                                    } else {
                                       lista_reglas= entvp->reglas;
                                       lista_num_conds = entvp->num_conds;
                                    }

                                    while (lista_reglas!=NULL) {
                                          entr=busca_en_reglas(lista_reglas->nombre);
                                          ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
                                          if (!entr->tiene_copias) {
                                             lvpeaux = ent_cond->param;
                                             cont=0;
                                             while (lvpeaux!=NULL) {
                                                   lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                                                   if (lo2->cuantificador==existe) {
                                                      cont++;
                                                   }
                                                   lvpeaux=lvpeaux->sig;
                                             }
                                             num_param = cont;
                                             param = (int *) malloc(sizeof(int)*num_param);
                                             cont = 0;
                                             lvpeaux = ent_cond->param;
                                            while (lvpeaux!=NULL) {
                                                   lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                                                   if (lo2->cuantificador==existe) {
                                                      if (strcmp(lo2->instancia_de,entvp->instancia_de)==0) param[cont]=entvp->cod.num_instancia;
                                                            else param[cont]=-1;
                                                      cont++;
                                                   }
                                                   lvpeaux=lvpeaux->sig;
                                             }
                                             inserta_condac(&funcion->cond_ac,lista_num_conds->num.entero,entr->num_regla,param,num_param);
                                          }
                                          lista_reglas=lista_reglas->sig;
                                          lista_num_conds = lista_num_conds->sig;
                                    }

                                    /* ahora se busca si hay alguna regla que */
                                    /* use en su condicion de activacion la instancia */
                                    /* que estamos modificando */
                                    if (entvp->tipo==definido) {
                                       exp = la->campos;
                                       lcamp_aux = entvp->lc;
                                       while (exp!=NULL) {
                                             exp=exp->sig;
                                             lcamp = busca_campo(exp->ident, lcamp_aux);
                                             lcamp_aux = lcamp->lc;
                                             exp=exp->sig;
                                       }
                                       lista_reglas= lcamp->reglas;
                                       lista_num_conds = lcamp->num_conds;
                                       while (lista_reglas!=NULL) {
                                             entr=busca_en_reglas(lista_reglas->nombre);
                                             ent_cond = busca_en_cond(entr->lcond,lista_num_conds->num.entero);
                                             if (!entr->tiene_copias) {
                                                lvpeaux = ent_cond->param;
                                                cont=0;
                                                while (lvpeaux!=NULL) {
                                                      lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                                                      if (lo2->cuantificador==existe) {
                                                        cont++;
                                                      }
                                                      lvpeaux=lvpeaux->sig;
                                                }
                                                num_param = cont;
                                                param = (int *) malloc(sizeof(int)*num_param);
                                                lvpeaux = ent_cond->param;
                                                cont=0;
                                                while (lvpeaux!=NULL) {
                                                      lo2= busca_obj(ent_cond->lobj_cond,lvpeaux->nombre);
                                                      if (lo2->cuantificador==existe) {
                                                        if (strcmp(lo2->instancia_de,entvp->instancia_de)==0)
                                                           param[cont]=entvp->cod.num_instancia;
                                                        else
                                                           param[cont]=-1;
                                                        cont++;
                                                      }
                                                      lvpeaux=lvpeaux->sig;
                                                }
                                                inserta_condac(&funcion->cond_ac,lista_num_conds->num.entero,entr->num_regla,param,num_param);
                                             }
                                             lista_reglas=lista_reglas->sig;
                                             lista_num_conds = lista_num_conds->sig;
                                       }
                                    }

                                    break;

                     case asig_paramt:
                                    lparam=busca_param(funcion->lp,la->ident);
                                    cad_porref[0]='\x0';
                                    if (lparam->por_ref) strcat(cad_porref,"*");
                                    entvp=busca_en_tds(la->ident);
                                    if (entvp->tipo!=definido) fprintf(fich_gen,"\t%s%s=",cad_porref,la->ident);
                                    else {
                                          fprintf(fich_gen,"\t");
                                          if (lparam->por_ref) fprintf(fich_gen,"(");
                                          fprintf(fich_gen,"%s%s",cad_porref,la->ident);
                                          if (lparam->por_ref) fprintf(fich_gen,")");
                                          exp = la->campos;
                                          while (exp!=NULL) {
                                                exp=exp->sig;
                                                fprintf(fich_gen,".%s",exp->ident);
                                                exp=exp->sig;
                                           }
                                           fprintf(fich_gen,"=");
                                    }
                                    Genera_exp(la->expr,cad_aux,funcion->lobj_func);
                                    fprintf(fich_gen,"%s;\n",cad_aux);
                                    break;

                   case asig_var:
                                 entvp=busca_en_tds(la->ident);
                                 if (entvp->tipo==definido) {
                                    fprintf(fich_gen,"\t%sv_%s[%d]",prefijo,entvp->instancia_de,entvp->cod.num_instancia);
                                    exp = la->campos;
                                    while (exp!=NULL) {
                                       exp=exp->sig;
                                       fprintf(fich_gen,".%s",exp->ident);
                                       exp=exp->sig;
                                    }
                                 } else
                                    fprintf(fich_gen,"\t%s%s",prefijo,la->ident);
                                 fprintf(fich_gen,"=");
                                 Genera_exp(la->expr,cad_aux,funcion->lobj_func);
                                 fprintf(fich_gen,"%s;\n",cad_aux);

                                 break;

                     case parada:  fprintf(fich_gen,"\t%shalt=TRUE;\n",prefijo);
                   }
                   la=la->sig;
             }
           fprintf(fich_gen,"}\n");
           declarando_funcion = FALSE;
          }
         aux=aux->sig;
         }
     }
}

void Genera_make_artis(char *nom_sp, int log_) {
  int i;

  //Generando fichero para el modo de KS opcional
  fprintf(fich_gen, "%s.o :\t%s_temp.o agenda_%s.o tipos_.o", nom_sp, nom_sp, nom_sp);
  if (log_) {
    fprintf(fich_gen, " log_%s.o", nom_sp);
  }
  
  //Ficheros externos...
  for (i = 0; i < num_fich_ext; i++) {
    fprintf(fich_gen," %s",fich_ext[i]);
  }

  fprintf(fich_gen, "\n");
  fprintf(fich_gen, "\t$(LD) $(LDFLAGS) -r -o %s.o -static %s_temp.o agenda_%s.o tipos_.o", 
	  nom_sp, nom_sp, nom_sp);
  if (log_)
    fprintf(fich_gen, " log_%s.o", nom_sp);
  
  for (i = 0; i < num_fich_ext; i++) {
    //Los ficheros externos se aaden a la orden de linkado y se libera la memoria reservada
    //para sus nombres.
    fprintf(fich_gen, " %s", fich_ext[i]);
    //    free(fich_ext[num_fich_ext]);
  }
  fprintf(fich_gen, "\n\n");
  
  fprintf(fich_gen, "%s_temp.o : \t%s.c\n", nom_sp, nom_sp);
  fprintf(fich_gen, "\t${CC} ${CFLAGS} -I/usr/local -I../ -c %s.c -o %s_temp.o\n\n",
	  nom_sp, nom_sp);

  fprintf(fich_gen, "agenda_%s.o :\tagenda_%s.c\n", nom_sp, nom_sp);
  fprintf(fich_gen, "\t${CC} -c ${CFLAGS} agenda_%s.c\n\n", nom_sp);

  fprintf(fich_gen, "tipos_.o :\ttipos_.c\n");
  fprintf(fich_gen, "\t${CC} -c ${CFLAGS} tipos_.c\n\n");

  if (log_) {
    fprintf(fich_gen, "log_%s.o : \tlog_%s.c\n", nom_sp, nom_sp);
    fprintf(fich_gen, "\t${CC} -c ${CFLAGS} log_%s.c\n\n", nom_sp);
  }    

  fprintf(fich_gen, "clean :\n");
  fprintf(fich_gen, "\trm -f %s.o %s_temp.o agenda_%s.o tipos_.o log_%s.o *~ *#\n", nom_sp, nom_sp, nom_sp, nom_sp);

}

void Genera_make_rt_artis(char *nom_sp) {
  //Generando el fichero para el make en modo de tarea crtica
  FILE *fd_origen;
  FILE *fd_rt;
  char rtfile[90];
  int c;
  extern char *output_path;
  
  strcpy(rtfile, "rt_");
  strcat(rtfile, nom_sp);
  strcat(rtfile, ".mak");
  if ((fd_rt = abrir_fichero_escritura(rtfile)) == NULL) {
    fprintf(stderr, "Error al tratar de crear el fichero para el make en modo rt\n");
    fprintf(stderr, "No se crear un fichero para el make en modo rt\n");
  }
  else {
    fprintf(fd_rt, "include /usr/rtlinux/rtl.mk\n\n");

    Genera_make_artis(nom_sp, 0); //No se puede generar el rt con log...
    fclose(fd_rt);
  }
}

void Genera_make(nom_sp)
   char *nom_sp;
{
   int cont;
  /* en esta funcion se genera el fichero Makefile para facilitar la compilacion del fichero */
  /* generado por Arlips */
   fprintf(fich_gen, "CC = gcc");
   if (debug) 
     fprintf(fich_gen, " -g");
   fprintf(fich_gen, "\n");
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"%s : %s.c agenda_%s.o tipos_.o",nom_sp, nom_sp, nom_sp);
  if (log_)
    fprintf(fich_gen, " log_%s.o", nom_sp);
  /* ahora se ponen los ficheros declarados como externos */
  cont = num_fich_ext;
  while(cont>0) {
      cont--;
      fprintf(fich_gen," %s",fich_ext[cont]);
  }
  fprintf(fich_gen,"\n");
  if (!nomain) {
    if (log_) {
      fprintf(fich_gen,"\t${CC} %s.c agenda_%s.o tipos_.o log_%s.o -o %s", nom_sp, nom_sp, nom_sp, nom_sp);
    }
    else {
      fprintf(fich_gen,"\t${CC} %s.c agenda_%s.o tipos_.o -o %s",nom_sp,nom_sp,nom_sp);
    }
  }
  else
         fprintf(fich_gen,"\t${CC} -c %s.c",nom_sp);
  /* ahora se ponen los ficheros declarados como externos */
  while(num_fich_ext>0) {
      num_fich_ext--;
      fprintf(fich_gen," %s",fich_ext[num_fich_ext]);
      /* en este ultimo recorrido tambien se libera la memoria */
      //free(fich_ext[num_fich_ext]);
  }
  fprintf(fich_gen,"\n");

  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"agenda_%s.o : agenda_%s.c\n",nom_sp,nom_sp);
  fprintf(fich_gen,"\t${CC} -c agenda_%s.c\n",nom_sp);
  fprintf(fich_gen,"\n");
  fprintf(fich_gen,"tipos_.o : tipos_.c\n");
  fprintf(fich_gen,"\t${CC} -c tipos_.c\n");
  fprintf(fich_gen,"\n");
  if (log_) {
    fprintf(fich_gen,"log_%s.o : log_%s.c\n",nom_sp,nom_sp);
    fprintf(fich_gen,"\t${CC} -c log_%s.c\n",nom_sp);
    fprintf(fich_gen,"\n");  
  }

  fprintf(fich_gen, "clean :\n");
  fprintf(fich_gen, "\trm -f %s agenda_%s.o tipos_.o log_%s.o *~ *#\n", nom_sp, nom_sp, nom_sp);
}

void libera_fich_ext() {
  for (num_fich_ext--; num_fich_ext >= 0; num_fich_ext--) {
    free(fich_ext[num_fich_ext]);
  }
  num_fich_ext = 0;
}
 
void Genera_fuente(//fich_dest, estr_control,sp,verb)
   char * fich_dest,
   char estr_control,
   char * sp,
   int verb)
{
  
  strcpy(fich_generados[fich_bh], fich_dest);
  fich_generados[fich_bh] [strlen(fich_dest) - 1] = 'h';

  strncpy(prefijo, sp, 7);
  strcat(prefijo, "_");
  prefijo[8] = '\0';

   strcpy(nom_sp,sp);
   verbos = verb;
   /* ahora se construyen los nombres del fichero agenda */
   strcpy(fich_generados[cab_agenda],"agenda_");
   strcat(fich_generados[cab_agenda],nom_sp);

   /* los nombres en MS-DOS estan limitados a 8 caracteres 
   if (strlen(fich_generados[cab_agenda])>8) fich_generados[cab_agenda][8]='\0'; */

   strcat(fich_generados[cab_agenda],".h");

   strcpy(fich_generados[fich_agenda],"agenda_");
   strcat(fich_generados[fich_agenda],nom_sp);

   /* los nombres en MS-DOS estan limitados a 8 caracteres 
   if (strlen(fich_generados[fich_agenda])>8) fich_generados[fich_agenda][8]='\0'; */
   strcat(fich_generados[fich_agenda],".c");


   //Se generan los nombres para los ficheros de log
   
   if (log_) {
     strcpy(fich_generados[cab_log], "log_");
     strcat(fich_generados[cab_log], nom_sp);
     strcpy(fich_generados[fich_log], fich_generados[cab_log]);
     strcat(fich_generados[cab_log], ".h");
     strcat(fich_generados[fich_log], ".c");
   }  

   /* Se genera el fichero AGENDA_.H */
   if (verbos) printf("Generando fichero con la implementacion de la agenda (%s)\n",fich_generados[fich_agenda]);
   fich_gen = abrir_fichero_escritura(fich_generados[cab_agenda]);
   Genera_cab_agenda();
   cierra_fichero(fich_gen);

   /* Se genera el fichero AGENDA_.C */
   fich_gen = abrir_fichero_escritura(fich_generados[fich_agenda]);
   Genera_fich_agenda();
   Genera_funciones_agenda();
   cierra_fichero(fich_gen);



   /* Se genera el fichero TIPOS_.H */
   if (verbos) printf("Generando el fichero con la implementacion de los tipos (%s)\n",fich_generados[fich_tipos]);
   fich_gen = abrir_fichero_escritura(fich_generados[cab_tipos]);
   Genera_cab_tipos();
   cierra_fichero(fich_gen);

   /* Se genera el fichero TIPOS_.C */
   fich_gen = abrir_fichero_escritura(fich_generados[fich_tipos]);
   Genera_fich_tipos();
   cierra_fichero(fich_gen);

   if (!nomake) {
      /* ahora se genera el fichero Makefile */
      if (nomain) {
          strcpy(fich_make,nom_sp);
          strcat(fich_make,".mak");
      }
      if (verbos) printf("Generando el fichero Makefile (%s)\n",fich_make);
      fich_gen = abrir_fichero_escritura(fich_make);
      if (artis_mode) {
	Genera_make_artis(nom_sp, log_);
	cierra_fichero(fich_gen);
	Genera_make_rt_artis(nom_sp);
      }
      else {
	Genera_make(nom_sp);
      }
      if (!artis_mode)
	cierra_fichero(fich_gen);
   }

   //Liberamos la memoria reservada para los ficheros extern declarados
   //pues una vez generado el Makefile, no es necesaria.
   libera_fich_ext();

   //Generando la base de hechos...
   fich_gen = abrir_fichero_escritura(fich_generados[fich_bh]);
   fprintf(fich_gen, "#include \"%s\"\n\n", fich_generados[cab_tipos]);
   Genera_clases();
   Declara_instancias();
   Declara_variables();
   cierra_fichero(fich_gen);
     

   /* ahora se genera el fichero pasado como parametros */
   if (verbos) printf("Generando el fichero principal (%s)\n",fich_dest);
   fich_gen = abrir_fichero_escritura(fich_dest);
   Genera_includes();
   Genera_constantes();
   declara_nombres_listas();
   Genera_funciones_externas();

   /* FALTA:  buscar un sitio mas apropiado para generar esta variable */
   fprintf(fich_gen,"int %shalt=FALSE;\n",prefijo);
   fprintf(fich_gen,"char %sestrategia=",prefijo);
   if (estr_control==profundidad) fprintf(fich_gen,"profundidad;\n");
                         else     fprintf(fich_gen,"anchura;\n");
   fprintf(fich_gen,"enum tipos_cambio{activacion=1, desactivacion};\n");
   fprintf(fich_gen,"\n");

   Genera_variables_reglas_izq();
   //Se mueve al .h
   //Genera_clases();
   Genera_tablas_reglas();
   Genera_estructuras_condiciones();
   Genera_funciones_tablas();
   Genera_Inicializacion_datos();
   //Las declaraciones se mueven a un fichero .h
   //Declara_instancias();
   //Declara_variables();
   Inicializa_instancias();
   Inicializa_variables();
   Genera_funciones();
   Genera_condiciones();
   Genera_reglas_izq();
   Genera_reglas_der();
   if (artis_mode) {
     Genera_lectura_del_kdm();
     Genera_escritura_en_kdm();
   }
   Genera_funcion_reset();
   Genera_funcion_run();
   Genera_funcion_principal();
   if (!nomain) Genera_funcion_main();
   cierra_fichero(fich_gen);

   //Se genera el fichero para log
   if (log_) {
     if (verbos) {
       printf("Generando fichero con la implementacin del log\n");
     }
     Genera_cab_log();
     Genera_fich_log();
   }
}
